/**
	$Id: ExperimentExplorer.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2006 Johan Enell, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web;

import java.awt.Color;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import javax.servlet.jsp.PageContext;

import net.sf.basedb.util.StaticCache;
import net.sf.basedb.util.Values;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.taglib.table.TableColumn;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicPositionQuery;
import net.sf.basedb.core.DynamicReporterQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.VirtualTable;
import net.sf.basedb.core.ItemContext.SortDirection;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.query.Aggregations;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlQuery;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.formatter.Formatter;

/**
	A class for storing settings and doing work related to the experiment 
	explorer tool in the web interface. The JSP pages that uses this class
	are located in the <code>www/views/experiments/explorer</code> directory
	and it's subdirectory.
	<p>
	This class is used two things:
	<ul>
	<li>Storing the settings
	<li>Creating the queries that retrieves the data according to the settings
	</ul>
	
	<p>
	Most of the settings are stored in regular {@link ItemContext}: objects. This
	means that they will be saved to the database when the user logs out and
	loaded again the next time they are needed. Each experiment explorer uses
	two <code>ItemContext</code> objects. One for storing the settings for the
	reporter search page and one for the view page. The reporter search settings
	are stored in the context <code>SessionControl.getCurrentContext(Item.REPORTER,
	"explorer.id")</code> and the view page settings are stored in 
	<code>SessionControl.getCurrentContext(Item.SPOTDATA, "explorer.id")</code> where
	"id" is replaced by the bioassayset's id.


	@author nicklas
	@version 2.0
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
*/
public class ExperimentExplorer
{

	/**
		Special value to use in the {@link #setPositionIndex(int)} method 
		to show all positions.
	*/
	public static final int SPOT_ALL = -1;
	
	/**
		Special value to use in the {@link #setPositionIndex(int)} method 
		to calculate the average for each bioassay.
	*/
	public static final int SPOT_AVG = -2;
	
	/**
	 	Predefined colors for annotation groups
	*/
	public static final String[] COLORS = 
		{"#ffcc00", "#00ffcc", "#00b3ff", "#ff00b3", "#00e090", "#ff0033", "#a8a8a8", "#33ff00"};
	
	/**
		The default context for the reporter search page.
	*/
	private static final ItemContext defaultReporterContext = 
		Base.createDefaultContext("@externalId", "externalId,symbol");
	
	/**
		Stores default context for the spot data list on the view page. 
		Different defaults for each raw data type since they can have different
		number of channels.
	*/
	private static final Map<RawDataType, ItemContext> defaultSpotContexts = 
		new HashMap<RawDataType, ItemContext>();

	static
	{
		for (RawDataType rdt : RawDataTypes.getRawDataTypes())
		{
			// Default visible columns are: position, ch1, ch2, ...
			StringBuilder columns = new StringBuilder("bioassay,position");
			for (int ch = 1; ch <= rdt.getChannels(); ++ch)
			{
				columns.append(",ch").append(ch);
			}
			defaultSpotContexts.put(rdt, 
				Base.createDefaultContext("POSITION", columns.toString()));
		}
	}
	
	/**
		Get the existing experiment explorer for a bioassayset or create a new
		one if none exist yet. The returned object is cached in the {@link
		SessionControl}:s session settings.
		@param bioAssaySet The bioassay set to get an experiment explorer for
		@return An <code>ExperimentExplorer</code> object
	*/
	public static ExperimentExplorer getExplorer(BioAssaySet bioAssaySet)
	{
		String explorerSetting = "explorer." + bioAssaySet.getId();
		SessionControl sc = bioAssaySet.getSessionControl();
		ExperimentExplorer explorer = (ExperimentExplorer)sc.getSessionSetting(explorerSetting);
		if (explorer == null)
		{
			explorer = new ExperimentExplorer(bioAssaySet);
			sc.setSessionSetting(explorerSetting, explorer);
		}
		return explorer;
	}
	
	/**
		The bioassay set we are exploring.
	*/
	private final BioAssaySet bioAssaySet;
	
	private final Experiment experiment;
	
	
	/**
		The snapshot manager that loads annotation values.
	*/
	private SnapshotManager snapshotManager;
	
	/**
		Maps a bioassays cube column number to it's id for fast
		loading by the {@link #getBioAssay(DbControl, short)} method.
	*/
	private final Map<Short, Integer> bioAssays;

	private int[] posCache;

	/**
		The cache of reporter ids matching the 
		current reporter filter.
	*/
	private ReporterPositions[] cache;
	
	/**
		The index of the current reporter on the view page.
		@see #getReporterIndex()
	*/
	private int reporterIndex = 0;
	/**
		The index of the curren position on the view page.
		@see #getPositionIndex()
	*/
	private int positionIndex = 0;
	
	/**
		Number of reporter that match the reporter search.
	*/
	private int matchingReporters = -1;
	
	/**
		The currently selected annotation types on the view page.
	*/
	private Set<Integer> annotationTypeIds;
	
	/**
		Caches annotation summaries.
	*/
	private Map<Integer, AnnotationSummary> summaryCache;
	
	/**
		Create a new experiment explorer.
	*/
	private ExperimentExplorer(BioAssaySet bioAssaySet)
	{
		this.bioAssaySet = bioAssaySet;
		this.experiment = bioAssaySet.getExperiment();
		this.bioAssays = new HashMap<Short, Integer>();
		this.snapshotManager = new SnapshotManager();
		this.summaryCache = new HashMap<Integer, AnnotationSummary>();
		for (BioAssay ba : bioAssaySet.getBioAssays().list(bioAssaySet.getDbControl()))
		{
			bioAssays.put(ba.getDataCubeColumnNo(), ba.getId());
		}
	}

	/**
		Get the bioassay set that is currently beeing explored.
		@param dc The DbControl used for access the database
		@return The <code>BioAssaySet</code> object
		@see BioAssaySet#getById(DbControl, int)
	*/
	public BioAssaySet getBioAssaySet(DbControl dc)
	{
		return BioAssaySet.getById(dc, bioAssaySet.getId());
	}	

	/**
		Get the experiment that is currently beeing explored.
		@param dc The DbControl used for access the database
		@return The <code>Experiment</code> object
	*/
	public Experiment getExperiment(DbControl dc)
	{
		return Experiment.getById(dc, experiment.getId());
	}	

	/**
		Get the snapshot manager that is used to load annotation
		values.
		@since 2.14
	*/
	public SnapshotManager getSnapshotManager()
	{
		return snapshotManager;
	}
	
	/**
		Get the number of bioassays in the bioassayset.
	*/
	public int getNumBioAssays()
	{
		return bioAssays.size();
	}
	
	/**
		Get the bioassay with the specified column number in the
		data cube. This method is useful since spot queries (see 
		{@link #getSpotQuery(DbControl, int, int, List)})
		can only return the column number, but we usually want to display
		the name, etc.
		
		@param dc The DbControl object used to get the bioassay
		@param columnNo The colum number
		@return The bioassay with the specified column number or null
			if no bioassay is found
	*/
	public BioAssay getBioAssay(DbControl dc, short columnNo)
	{
		Integer baId = bioAssays.get(columnNo);
		return baId == null ? null : BioAssay.getById(dc, baId);
	}

	/**
		The number of reporter id'd to cache. For performance reasons
		the id's of the reporters matching the reporter search are cached.
		This means we can step forwards and backwards more quickly on the
		view page since we don't have to execute the reporter search each 
		time. This method should return the same value as {@link #getMatchingReporters(DbControl)}
		unless the cache hasn't been initialized.
		@return The size of the cache, or 0 if the cache has not
			been initialized
	*/
	public int getCacheSize()
	{
		return cache == null ? 0 : cache.length;
	}
	
	/**
		Get the total number of unique reporters found in the bioassayset.
		@see BioAssaySet#getNumReporters()
	*/
	public int getNumReporters()
	{
		return bioAssaySet.getNumReporters();
	}
	
	/**
		Get the total number of spots in the bioassayset.
	 	@see BioAssaySet#getNumSpots()
	*/
	public int getNumSpots()
	{
		return bioAssaySet.getNumSpots();
	}
	
	/**
		Get the subcontext name used for the {@link ItemContext} objects
		related to this explorer. The name is created like this:
		<code>explorer.id</code> where id is replaced with the id of
		the bioassayset we are exploring.
		
		@return The subcontext name
	*/
	public String getSubContext()
	{
		return "explorer." + bioAssaySet.getId();
	}
	
	/**
		Get the default spot context that is used to initialise the actual context.
		@return An <code>ItemContext</code> object
	*/
	public ItemContext getDefaultSpotContext()
	{
		return defaultSpotContexts.get(bioAssaySet.getRawDataType());
	}

	/**
		Get the default reporter context that is used to initialise the actual context.
		@return An <code>ItemContext</code> object
	*/
	public ItemContext getDefaultReporterContext()
	{
		return defaultReporterContext;
	}
	
	/**
		Get and update the current settings for the spot context. The spot context is used
		on the view page for the list with spot data.
		@param sc The SessionControl of the logged in user
		@param pageContext The {@link PageContext} containing the
			updated settings, or null to not update the settings
		@return An <code>ItemContext</code> object
		@see Base#getAndSetCurrentContext(SessionControl, Item, PageContext, ItemContext)
	 */
	public ItemContext getAndSetSpotContext(SessionControl sc, PageContext pageContext)
	{
		ItemContext cc = Base.getAndSetCurrentContext(sc, Item.SPOTDATA, getSubContext(), pageContext, getDefaultSpotContext());
		return cc;
	}
	
	/**
		Get and update the current settings for the reporter context. The reporter context
		is used on the reporter search page.
		@param sc The SessionControl of the logged in user
		@param pageContext The {@link PageContext} containing the
			updated settings, or null to not update the settings
		@return An <code>ItemContext</code> object
		@see Base#getAndSetCurrentContext(SessionControl, Item, PageContext, ItemContext)
	*/
	public ItemContext getAndSetReporterContext(SessionControl sc, PageContext pageContext)
	{
		ItemContext cc = Base.getAndSetCurrentContext(sc, Item.REPORTER, getSubContext(), pageContext, getDefaultReporterContext());
		if (pageContext != null)
		{
			// Settings may have changed, clear cache
			matchingReporters = -1;
			cache = null;
		}
		return cc;
	}
	
	/**
		Get the selected annotation types.
		@return A set with the id's of the selected annotation types
		@since 2.14
	*/
	public Set<Integer> getAnnotationTypeIds()
	{
		if (annotationTypeIds == null) annotationTypeIds = new HashSet<Integer>();
		return annotationTypeIds;
	}
	
	/**
		Update the selected annotation types.
		@param ids A collection containing the id's of all annotation types
			that should currently be selected, all other annotation types are
			automatically deselected
		@since 2.14
	*/
	public void setAnnotationTypeIds(Collection<Integer> ids)
	{
		Set<Integer> current = getAnnotationTypeIds();
		current.clear();
		for (Integer id : ids)
		{
			current.add(id);
			summaryCache.remove(id);
		}
		// The snapshot manager also has an in-memory cache
		snapshotManager = new SnapshotManager();
	}
	
	/**
		Get the annotation types that are of interest in this experiment.
		Ie. the experimental factors of the experiment the bioassayset we
		are exploring belongs to. Optionally, you can also get annotations
		for bioassays.

		@param dc The DbControl for database access
		@param includeBioAssayAnnotations If bioassay annotation should be included
			even if they are not experimental factors
		@return A list of annotation types, null may be returned if no annotation types
			exists
		@see net.sf.basedb.core.Experiment#getExperimentalFactors()
	*/
	public List<AnnotationType> getAnnotationTypes(DbControl dc, boolean includeBioAssayAnnotations)
	{
		final ItemQuery<AnnotationType> query = AnnotationType.getQuery(null);
		query.join(Hql.leftJoin("experiments", Item.EXPERIMENT.getAlias()));

		// Restriction for experimental factors
		Restriction ef = Restrictions.eq(
			Hql.alias(Item.EXPERIMENT.getAlias()), 
			Hql.entity(getBioAssaySet(dc).getExperiment())
		);
		// Restrictions for BioAssay annotation types
		if (includeBioAssayAnnotations)
		{
			query.join(Hql.innerJoin("itemTypes", "itemType"));
			Restriction baa = Restrictions.eq(
				Hql.alias("itemType"),
				Expressions.integer(Item.BIOASSAY.getValue())
			);
			ef = Restrictions.or(ef, baa);
		}
		query.restrict(ef);
		query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		query.order(Orders.asc(Hql.property("name")));
		query.setDistinct(true); // Needed because of the left join
		return query.list(dc);
	}
	
	/**
		Get an AnnotationSummary object which can be used to calculate means
		and other values for each annotation group. When the annotation summary
		object is create each bioassay is placed in one annotation group
		based on which annotatations it have. Each annotation group is unique and
		is the only one with the same combination of annotation values. More
		than one bioassay may be placed in the same group if they are annotated
		with the same values.
		
		@param dc The DbControl for database access
		@param annotationType The annotation type, mustn't be null
		@return An <code>AnnotationSummary</code> object
		@see AnnotationGroup#getAnnotationValues()
	*/
	public AnnotationSummary getAnnotationSummary(DbControl dc, AnnotationType annotationType)
	{
		AnnotationSummary summary = summaryCache.get(annotationType.getId());
		if (summary == null)
		{
			summary = new AnnotationSummary(this, dc, annotationType);
			summaryCache.put(annotationType.getId(), summary);
		}
		else
		{
			summary.reset();
		}
		return summary;
	}
	
	/**
	  	Get the number of reporters matching the current reporter
		filter.
		@param dc The DbControl for database access
	  	@return an <code>int</code> of matching reporters 
	*/
	public int getMatchingReporters(DbControl dc)
	{
		if (matchingReporters == -1) initReporterCache(dc);
		return matchingReporters;
	}
	
	/**
		Get the index of the current reporter. It should be a value between
		0 and {@link #getMatchingReporters(DbControl)}-1.
		@see #getReporter(DbControl, int)
	*/
	public int getReporterIndex()
	{
		return reporterIndex;
	}
	
	/**
		Set the index of the currently viewed reporter. If the index is below
		zero or greater than or equal to {@link #getMatchingReporters(DbControl)} it is set
		to the nearest index.
		@param reporterIndex The new index
		@see #getReporter(DbControl, int)
	*/
	public void setReporterIndex(int reporterIndex)
	{
		if (reporterIndex >= matchingReporters) reporterIndex = matchingReporters - 1;
		if (reporterIndex < 0) reporterIndex = 0;
		this.reporterIndex = reporterIndex;
	}
	
	/**
		Get the reporter for the specified index. The index should be between 
		0 and the number of matching reporters minus 1. The returned reporter 
		information is taken from the cloned table if that is present in the
		current experiment.
		
		@param dc The DbControl to use for database access
		@param index The index
		@return The reporter or null if no reporter is found at the specified index
		@see #getMatchingReporters(DbControl)
	*/
	public ReporterData getReporter(DbControl dc, int index)
	{
		if (cache == null) initReporterCache(dc);
		int reporterId = index >= 0 && index < cache.length ? cache[index].id : 0;
		
		return reporterId == 0 ? null : Reporter.getById(dc, reporterId, getExperiment(dc));
	}

	/**
		Get the reporter id for the specified index. The index should be
		between 0 and the number of matching reporters minus 1. This method
		assumed that the reporter/position cache has already been initialized.
		@param index The index
		@return The internal id of the reporter, or 0 if the cache is not
			initialized or if the index is outside the allowed range
		@since 3.0
		@see #getMatchingReporters(DbControl)
	*/
	public int getReporterId(int index)
	{
		if (cache == null) return 0;
		return index >= 0 && index < cache.length ? cache[index].id : 0;
	}
	
	/**
		Get the number of unique positions the reporter of the specifed
		index is present in among all the bioassays in the bioassay set.
		The index should be between 0 and the number of matching reporters
		minus 1.
		
		@param dc The DbControl to use for database access
		@param index The reporter index
		@return The number of positions, or 0 if the index is outside 
			the valid range
		@see #getReporter(DbControl, int)
		@see #getMatchingReporters(DbControl)
		@since 2.11
	*/
	public int getNumPositions(DbControl dc, int index)
	{
		if (cache == null) initReporterCache(dc);
		return index >= 0 && index < cache.length ? cache[index].pos.length : 0;
	}
	
	/**
		Get the index of the current position. It should be a value between
		0 and {@link #getNumPositions(DbControl, int)}-1, or one
		of the {@link #SPOT_ALL} or {@link #SPOT_AVG} special values (negative).
		@return The current index
		@see #getPosition(DbControl, int, int)
	*/
	public int getPositionIndex()
	{
		return positionIndex;
	}
	
	/**
		Set the index of the curren position.
		@param positionIndex The new index
		@see #getPositionIndex()
		@see #getPosition(DbControl, int, int)
	*/
	public void setPositionIndex(int positionIndex)
	{
		this.positionIndex = positionIndex;
	}
	
	/**
		Get the position number of the position specified by the
		given reporter and position index. For a given reporter
		the positions are sorted in ascending order. The reporter
		index should be a value between 0 and the number of matching
		reporter minus 1. The position index should be a value between
		0 and the number of positions for the given reporter minus 1.
		
		@param reporterIndex The reporter index
		@param positionIndex The position index
		@return The position number or -1 if no position can be found
			with the given indexes
		@see #getMatchingReporters(DbControl)
		@see #getNumPositions(DbControl, int)
		@since 2.11
	*/
	public int getPosition(DbControl dc, int reporterIndex, int positionIndex)
	{
		if (cache == null) initReporterCache(dc);
		int[] positions = null;
		if (reporterIndex >= 0 && reporterIndex < cache.length)
		{
			positions = cache[reporterIndex].pos;
		}
		int position = -1;
		if (positions != null)
		{
			if (positionIndex >= 0 && positionIndex < positions.length)
			{
				position = positions[positionIndex];
			}
		}
		return position;
	}
	
	/**
		Get a query returning the reporter data used by the reporter search page.
		Note that the query is {@link DynamicReporterQuery} and not a 
		{@link net.sf.basedb.core.DataQuery}. This means that we don't get 
		{@link ReporterData} objects as a result and that we must specify the
		properties to select (see {@link ItemContext#configureQuery(DbControl, SqlQuery, List)}).
		<p>
		The reporter id is always included as the first selected property so it
		doesn't have to be in the selection list.
		<p>
		
		NOTE! The query is restricted using information in the reporter/position cache
		to only return the exactly the reporter information that is displayed on the
		current page. Thus, some things do not work in the same way as
		most other queries used when listing items.
		
		<ul>
		<li>The query doesn't return a total count. Use {@link #getMatchingReporters(DbControl)}
			instead.
		<li>Paging options do not work (eg. {@link Query#setFirstResult(int)} and {@link Query#setMaxResults(int)})
		</ul>
		
		@param dc The DbControl for database access
		@param selectionList The properties that the query should select
		@return A query
		@see ItemContext#configureQuery(DbControl, SqlQuery, List)
	*/
	public DynamicReporterQuery getReporterQuery(DbControl dc, List<String> selectionList)
	{
		// Make sure the reporter/position cache is up to date
		if (cache == null) initReporterCache(dc);

		Experiment exp = getExperiment(dc);
		ItemContext cc = dc.getSessionControl().getCurrentContext(Item.REPORTER, getSubContext(), null);
		if (exp.getVirtualDb().hasClonedReporters())
		{
			cc.setPropertyInspector(exp.getVirtualDb().getQueryPropertyInspector());
		}
		else
		{
			cc.setPropertyInspector(null);
		}

		// Get the reporter id:s to show on the current page
		List<Integer> reporterIds = new ArrayList<Integer>(cc.getRowsPerPage()+1);
		int start = cc.getRowsPerPage() * cc.getPage();
		int end = start + cc.getRowsPerPage();
		for (int index = start; index < end && index < cache.length; ++index)
		{
			int reporterId = cache[index].id;
			if (reporterId > 0) reporterIds.add(reporterId);
		}
		
		// Create reporter query
		DynamicReporterQuery reporterQuery = Reporter.getDynamicQuery(exp);
		// Guard against no hits
		if (reporterIds.size() == 0) reporterIds.add(-1);
		reporterQuery.restrict(Restrictions.in(Dynamic.reporter("id"), new InExpression(reporterIds)));
		
		reporterQuery.select(Selects.expression(Dynamic.reporter("id"), "id", true));
		cc.configureQuery(dc, (SqlQuery)reporterQuery, selectionList);
		
		// Do not return the total count. Do not set paging options.
		reporterQuery.setReturnTotalCount(false);
		reporterQuery.setFirstResult(0);
		reporterQuery.setMaxResults(-1);
		
		return reporterQuery;
	}
	
	/**
		Get a query returning spot data for a given reporter and position. This query is used
		by the view page to list the spot data for each bioassay.
		<p>
		The column number is always included as the first selected property. If the
		{@link #SPOT_AVG} is specified for <code>positionIndex</code> a count of the
		number of positions is included as the second selected property. Otherwise, the 
		position number is included as the second selected property.
		<p>
		Then, all channel intensities are selected.
		
		<p> 
		See {@link ItemContext#getDynamicExpression(DbControl, SqlQuery, String)} for more information 
		about how to specify additional selection properties.
		

		@param dc The DbControl for database access
		@param reporterIndex The index of the reporter
		@param positionIndex The index of the position, or one the special
			values {@link #SPOT_ALL} or {@link #SPOT_AVG}
		@param selectionList The data that the query should select
		@return A query
		@since 2.11
	*/
	public DynamicSpotQuery getSpotQuery(DbControl dc, int reporterIndex, int positionIndex, List<TableColumn> selectionList)
	{
		if (cache == null) initReporterCache(dc);
		int[] positions = null;
		if (reporterIndex >= 0 && reporterIndex < cache.length)
		{
			positions = cache[reporterIndex].pos;
		}
		
		BioAssaySet bioAssaySet = getBioAssaySet(dc);
		int channels = bioAssaySet.getRawDataType().getChannels();
		DynamicSpotQuery spotQuery = bioAssaySet.getSpotData();

		// Always select the column (ie. bioassay)
		spotQuery.select(Dynamic.select(VirtualColumn.COLUMN));
		
		// Restricting on the data-cube column seems to give better
		// performance -- probably because the primary key index
		// can be used to lookup rows
		spotQuery.restrict(Restrictions.in(Dynamic.column(VirtualColumn.COLUMN), new InExpression(bioAssays.keySet())));
		
		if (positionIndex == SPOT_ALL)
		{
			// Select the position number
			spotQuery.select(Dynamic.select(VirtualColumn.POSITION));
			
			// Restrict by the positions known to have the given reporter
			spotQuery.restrict(Restrictions.in(Dynamic.column(VirtualColumn.POSITION), new InExpression(positions)));
		}
		else if (positionIndex == SPOT_AVG)
		{
			// Select the number of spots we are averaging over
			spotQuery.select(
				Selects.expression(
					Aggregations.count(Dynamic.column(VirtualColumn.COLUMN), false), 
					"cnt"
				)
			);
			
			// Create group per bioassay
			spotQuery.group(Dynamic.column(VirtualColumn.COLUMN));
			
			// Restrict by the positions known to have the given reporter
			spotQuery.restrict(Restrictions.in(Dynamic.column(VirtualColumn.POSITION), new InExpression(positions)));
		}
		else
		{
			// Select the position
			spotQuery.select(Dynamic.select(VirtualColumn.POSITION));

			// Get spots for one specific position
			int position = -1;
			if (positions != null)
			{
				if (positionIndex >= 0 && positionIndex < positions.length)
				{
					position = positions[positionIndex];
				}
			}
			spotQuery.restrict(
				Restrictions.eq(
					Dynamic.column(VirtualColumn.POSITION), 
					Expressions.integer(position)
				)
			);
		}
		
		// Select channel intensities
		Formula.AverageMethod avgMethod = bioAssaySet.getIntensityTransform().getAverageMethod();
		for (int ch = 1; ch <= channels; ++ch)
		{
			Expression select = Dynamic.column(VirtualColumn.channelRaw(ch));
			if (positionIndex == SPOT_AVG)
			{
				select = avgMethod.getAverageExpression(select);
			}
			spotQuery.select(Selects.expression(select, "ch" + ch));
		}
		
		// Set additional selected columns
		int i = 0;
		for (TableColumn tc : selectionList)
		{
			// Skip channel intensities since they are already selected
			if (!tc.getProperty().startsWith("%"))
			{
				Expression select = ItemContext.getDynamicExpression(dc, spotQuery, tc.getProperty());
				if (select != null)
				{
					i++;
					if (positionIndex == SPOT_AVG)
					{
						select = tc.getAverageMethod().getAverageExpression(select);
					}
					spotQuery.select(Selects.expression(select, "e" + i));
				}
			}
		}
		
		// Add sort order information
		ItemContext cc = dc.getSessionControl().getCurrentContext(Item.SPOTDATA, getSubContext());
		String sortProperties = cc.getSortProperty();
		if (sortProperties != null && sortProperties.length() > 0)
		{
			String[] sp = sortProperties.split(",");
			for (String sortProperty : sp)
			{
				Expression sortby = ItemContext.getDynamicExpression(dc, spotQuery, sortProperty);
				if (positionIndex == SPOT_AVG)
				{
					// Find TableColumn to know which average method to use
					if (!sortProperty.startsWith("%"))
					{
						boolean found = false;					
						for (TableColumn tc : selectionList)
						{
							if (sortProperty.equals(tc.getProperty()))
							{
								sortby = tc.getAverageMethod().getAverageExpression(sortby);
								found = true;
								break;
							}
						}
						if (!found) sortby = null;
					}				
					else
					{
						sortby = avgMethod.getAverageExpression(sortby);
					}
				}
				if (sortby != null)
				{
					spotQuery.order(SortDirection.ASC == cc.getSortDirection() ? Orders.asc(sortby) : Orders.desc(sortby));
				}
			}
		}	
		return spotQuery;
	}
	
	private void initReporterCache(DbControl dc)
	{
		BioAssaySet bas = getBioAssaySet(dc);
		
		if (posCache == null)
		{
			int id = bas.getId();
			StaticCache staticCache = Application.getStaticCache();
			String cacheKey = "explorer/" + (id % 1000) + "/poslist-" + id + ".ser";
			List<Integer> posList = staticCache.load(cacheKey, 100);
			if (posList == null)
			{
				Select spotPos = Dynamic.select(VirtualTable.SPOT, VirtualColumn.POSITION);
				DynamicSpotQuery posQuery = bas.getSpotData();
				posQuery.select(spotPos);
				posQuery.order(Orders.asc(Expressions.selected(spotPos)));
				posQuery.setDistinct(true);
				DynamicResultIterator it = posQuery.iterate(dc);
				posList = new ArrayList<Integer>(bas.getNumReporters());
				try
				{
					while (it.hasNext())
					{
						posList.add(it.next().getInt(1));
					}
					staticCache.store(cacheKey, posList, 100);
				}
				catch (SQLException ex)
				{
					throw new BaseException(ex);
				}
				finally
				{
					it.close();
				}
			}
			posCache = new int[posList.size()];
			for (int i = 0; i < posCache.length; ++i)
			{
				posCache[i] = posList.get(i);
			}
		}
		
		DynamicPositionQuery reporterQuery = bas.getPositionData();
		Select position = Dynamic.select(VirtualTable.POSITION, VirtualColumn.POSITION);
		reporterQuery.joinReporters(JoinType.INNER);
		reporterQuery.select(Dynamic.select(VirtualTable.POSITION, VirtualColumn.REPORTER_ID));
		reporterQuery.select(position);
		ItemContext cc = dc.getSessionControl().getCurrentContext(Item.REPORTER, getSubContext(), null);
		if (getExperiment(dc).getVirtualDb().hasClonedReporters())
		{
			cc.setPropertyInspector(getExperiment(dc).getVirtualDb().getQueryPropertyInspector());
		}
		else
		{
			cc.setPropertyInspector(null);
		}
		cc.configureQuery(dc, (SqlQuery)reporterQuery, null);
		reporterQuery.order(Orders.asc(Expressions.selected(position)));
		reporterQuery.setDistinct(true);
		reporterQuery.setMaxResults(-1);
		reporterQuery.setFirstResult(-1);
		reporterQuery.setReturnTotalCount(false);
		
		Map<Integer, List<Integer>> map = new LinkedHashMap<Integer, List<Integer>>();
		DynamicResultIterator it = reporterQuery.iterate(dc);
		try
		{
			while (it.hasNext())
			{
				SqlResult result = it.next();
				int id = result.getInt(1);
				int pos = result.getInt(2);
				if (Arrays.binarySearch(posCache, pos) >= 0)
				{
					List<Integer> list = map.get(id);
					if (list == null)
					{
						list = new ArrayList<Integer>();
						map.put(id, list);
					}
					list.add(pos);
				}
			}
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			it.close();
		}
		
		matchingReporters = map.size();
		cache = new ReporterPositions[matchingReporters];
		int index = 0;
		for (Map.Entry<Integer, List<Integer>> entry : map.entrySet())
		{
			int id = entry.getKey();
			List<Integer> list = entry.getValue();
			int[] pos = new int[list.size()];
			for (int i = 0; i < pos.length; ++i)
			{
				pos[i] = list.get(i);
			}
			cache[index] = new ReporterPositions(id, pos);
			++index;
		}
	}
	
	/**
		Cache entry for a reporter with positions. Contains the
		reporter ID and an array with the position numbers it is
		present on in the bioassay set.
	*/
	public static class ReporterPositions
	{
		final int id;
		final int[] pos;
		ReporterPositions(int id, int[] pos)
		{
			this.id = id;
			this.pos = pos;
		}
	}
	
	/**
		Creates a list of (numerical) values that can be used as the
		right-hand operand to the IN operator. Eg. foo IN (v1, v2, ...)
	*/
	private static class InExpression
		implements Expression
	{

		private final String in;
		
		InExpression(int... values)
		{
			StringBuilder sb = new StringBuilder();
			for (int v : values)
			{
				if (sb.length() > 0) sb.append(",");
				sb.append(v);
			}
			in = sb.toString();
		}
		
		InExpression(Collection<? extends Number> values)
		{
			StringBuilder sb = new StringBuilder();
			for (Number v : values)
			{
				if (sb.length() > 0) sb.append(",");
				sb.append(v);
			}
			in = sb.toString();
		}
		
		@Override
		public String toQl(Query query, DbControl dc) 
			throws BaseException
		{
			return in;
		}

		@Override
		public Collection<? extends QueryElement> getChildren()
		{
			return null;
		}
		
	}
	
	/**
		An annotation summary object keeps track of intensities and other 
		statistical information based on the annotation group.
	
		@author nicklas
		@version 2.0
		@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
	 */
	public static class AnnotationSummary
	{
	
		private final ExperimentExplorer explorer;
		private final Map<Short, AnnotationGroup> assayGroups;
		private final Set<AnnotationGroup> groups;
		private final AnnotationType annotationType;
		
		@SuppressWarnings({ "unchecked", "rawtypes" })
		private AnnotationSummary(ExperimentExplorer explorer, DbControl dc, AnnotationType annotationType)
		{
			this.explorer = explorer;
			this.assayGroups = new HashMap<Short, AnnotationGroup>();
			this.groups = new TreeSet<AnnotationGroup>();
			this.annotationType = annotationType;
			
			Formatter formatter = FormatterFactory.getTypeFormatter(dc.getSessionControl(), annotationType.getValueType());
			if (annotationType.supportUnits()) formatter = annotationType.getDefaultUnit().getFormatter(formatter);
			BioAssaySet bioAssaySet = explorer.getBioAssaySet(dc);
			int channels = bioAssaySet.getRawDataType().getChannels();
			Map<Set<?>, AnnotationGroup> temp = new HashMap<Set<?>, AnnotationGroup>();
			int groupId = 0;
			for (BioAssay ba : bioAssaySet.getBioAssays().list(dc))
			{
				Set<?> annotationValues = BioAssaySetUtil.getAnnotationValues(dc, explorer.getSnapshotManager(), ba, annotationType);
				AnnotationGroup group = temp.get(annotationValues);
				if (group == null)
				{
					group = new AnnotationGroup(++groupId, annotationValues, channels, formatter);
					temp.put(annotationValues, group);
					groups.add(group);
				}
				assayGroups.put(ba.getDataCubeColumnNo(), group);
			}
			setColorsAuto();
			temp.clear();
		}
		
		/**
			Get the annotation type this summary is based on.
			@since 2.14
		*/
		public AnnotationType getAnnotationType()
		{
			return annotationType;
		}
		
		/**
			Get the annotation group for the bioassay with the 
			give data cube column.
			
			@param column The data cube column of an bioassay
			@return An AnnotationGroup object, or null if none is found
		*/
		public AnnotationGroup getAnnotationGroup(short column)
		{
			return assayGroups.get(column);
		}

		/**
			Get all annotation groups that was found among the bioassays.
			@return The annotations groups.
		*/
		public Set<AnnotationGroup> getAnnotationGroups()
		{
			return groups;
		}
		
		/**
			Reset the collected statistics in all annotation groups.
			@since 2.14
		*/
		public void reset()
		{
			for (AnnotationGroup grp : groups)
			{
				grp.reset();
			}
		}
		
		/**
			Arranges the colors of the groups in the order they
			are returned by the {@link #getAnnotationGroups()} set.
			@since 2.14
		*/
		public void setColorsAuto()
		{
			int i = 0;
			for (AnnotationGroup grp : groups)
			{
				grp.setColor(COLORS[i % COLORS.length]);
				++i;
			}
		}
	}
	
	/**
		An annotation group contains statistical information about the spot
		values falling into the group.
	
		@author nicklas
		@version 2.0
		@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
	 */
	public static class AnnotationGroup
		implements Comparable<AnnotationGroup>
	{
		private final int id;
		private final Set<?> annotationValues;
		private final String title;
		private final float[] sum;
		private final int[] count;
		private final Map<String, Float> statistics;
		private String color;
		
		private <T> AnnotationGroup(int id, Set<T> annotationValues, int channels, Formatter<T> formatter)
		{
			this.id = id;
			this.annotationValues = annotationValues;
			this.title = Values.getString(annotationValues, ", ", true, formatter);
			this.sum = new float[channels+1]; // Channels are 1-based
			this.count = new int[channels+1];
			this.statistics = new HashMap<String, Float>();
			setColor(COLORS[(id - 1) % COLORS.length]);
		}
		
		/**
			The ID of the group. The ID is only useful within the objects 
			AnnotationSummary objects lifetime. The first group that is
			created is given an ID of 1, the second 2, and so on.
			@return The ID
		*/
		public int getId()
		{
			return id;
		}
		
		/**
			Get the title of this annotation group. The title is created
			by concatenating all annotation values.
			@return The title
		*/
		public String getTitle()
		{
			return title;
		}
		
		/**
			Get all annotation values in this annotation group.
			@return A set containing all annotation values
		*/
		public Set<?> getAnnotationValues()
		{
			return annotationValues;
		}

		/**
			Get a color code for HTML that is as distinct as possible from
			the color of the other groups. The colors may repeat if there are many
			groups.
			@return The color string
		*/
		public String getColor()
		{
			return color;
		}
		
		/**
			Get the AWT color object that corresponds to the HTML color string.
			The color string must have been set with a seven-character 
			string with 2 hex characters for every color component, eg. 
			'#rrggbb'.
			@since 2.14
		*/
		public Color getAwtColor()
		{
			int r = Integer.parseInt(color.substring(1, 3), 16);
			int g = Integer.parseInt(color.substring(3, 5), 16);
			int b = Integer.parseInt(color.substring(5), 16);
			return new Color(r, g, b);
		}
		
		/**
			Sets the color for this annotation group.
			@param color A HTML color code
			@since 2.14
		*/
		public void setColor(String color)
		{
			this.color = color;
		}
		
		/**
			Add an intensity value to this group.
			@param channel The number of the channel, starting with 1
			@param intensity The spot's intensity
		*/
		public void addValue(int channel, float intensity)
		{
			sum[channel] += intensity;
			count[channel]++;
		}
		
		/**
			Get the sum of intensities for the specified channel.
			@param channel The number of the channel, starting with 1
			@return The sum of the intensities
			@see #addValue(int, float)
		*/
		public float getSum(int channel)
		{
			return sum[channel];
		}
		
		/**
		 	Get the number of intensity values added to the specified channel.
			@param channel The number of the channel, starting with 1
			@return The number of values
			@see #addValue(int, float)
		*/
		public int getCount(int channel)
		{
			return count[channel];
		}
		
		/**
			Get the mean of the intensity values for the given channel.
			@param channel he number of the channel, starting with 1
			@return The mean of the values, or NaN if no values has been added
			@see #addValue(int, float)
		*/
		public float getMean(int channel)
		{
			int theCount = count[channel];
			float theMean = theCount > 0 ? sum[channel] / theCount : Float.NaN;
			return theMean;
		}
		
		/**
			Add another statisitcal measurement to this group. All values
			with the same key are added together and can be retrieved with 
			the {@link #getStatistics(String)} method.
			
			@param key The key for the measurment
			@param value The value to add
		 */
		public void addStatistics(String key, float value)
		{
			Float current = statistics.get(key);
			if (current == null)
			{
				current = value;
			}
			else
			{
				current += value;
			}
			statistics.put(key, current);
		}
		
		/**
			Get the sum of all values added with the {@link #addStatistics(String, float)}
			method with the given key.
			@param key The measurement key
			@return The sum of the added values or Float.NaN if no value is found
		*/
		public float getStatistics(String key)
		{
			Float value = statistics.get(key);
			return value == null ? Float.NaN : value.floatValue();
		}
		
		/**
			Resets all collected statistics in this group.
			@since 2.14
		*/
		public void reset()
		{
			for (int i = 0; i < sum.length; ++i)
			{
				sum[i] = 0.0f;
				count[i] = 0;
			}
			statistics.clear();
		}

		@Override
		public int compareTo(AnnotationGroup o)
		{
			int result = this.title.compareTo(o.title);
			if (result == 0) result = this.id - o.id;
			return result;
		}
		
	}
}
