/**
	$Id: LoadServletsProcessor.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.servlet.ServletContext;

import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.extensions.manager.ExtensionsFile;
import net.sf.basedb.util.extensions.manager.ExtensionsFileProcessor;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;
import net.sf.basedb.util.extensions.manager.ExtensionsFile.WriteableExtensionsFile;
import net.sf.basedb.util.extensions.manager.ProcessResults;

/**
	Processor implementation that define servlets based on what it can
	find in the META-INF/servlets.xml file. The processor only works with
	JAR files.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (ons, 09 apr. 2014) $
*/
public class LoadServletsProcessor 
	implements ExtensionsFileProcessor 
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(LoadServletsProcessor.class);

	private final ServletContext context;
	private final ProcessResults results;
	
	private int numFiles;
	private int numError;
	private int numServlets;
	
	public LoadServletsProcessor(ServletContext context, ProcessResults results)
	{
		this.context = context;
		this.results = results;
	}
	
	/*
		From the ExtensionsFileProcessor interface
		------------------------------------------
	*/
	@Override
	public void begin(ExtensionsManager manager, int numFiles) 
	{
		this.numFiles = 0;
		this.numError = 0;
		this.numServlets = 0;
	}

	@Override
	public void processFile(ExtensionsManager manager, WriteableExtensionsFile wFile) 
	{
		ExtensionsFile xtFile = wFile.getExtensionsFile();
		if (!xtFile.isJar()) 
		{
			log.info("File is not a JAR file (skipping): " + xtFile);
			return;
		}

		log.info("Loading servlets from file: " +xtFile);
		InputStream in = null;
		ZipFile zipFile = null;
		int numServlets = 0;
		try
		{
			zipFile = new ZipFile(xtFile.getFile());
			ZipEntry zipEntry = zipFile.getEntry("META-INF/servlets.xml");
			if (zipEntry == null)
			{
				log.info("META-INF/servlets.xml not found in file: " + xtFile);
				return;
			}
			
			int num = 0;
			in = zipFile.getInputStream(zipEntry);			
			ClassLoader jarLoader = xtFile.getClassLoader();
			String jarFile = xtFile.getName();
			wFile.open();
			ServletLoader servletLoader = new ServletLoader();
			servletLoader.loadXmlFile(in, jarFile, jarLoader, true);
			for (ServletWrapper wrapper : servletLoader.getServlets())
			{
				wrapper.setServletContext(context);
				wFile.registerObject(new ServletWrapperKey(jarFile, wrapper.getServletName()), wrapper);
				num++;
			}
			numFiles++;
			if (num > 0 && results != null)
			{
				results.addMessage(xtFile, num + " servlet(s) loaded successfully.");
			}
			log.info("Loaded " + num + " servlets from file: " + xtFile);
		}
		catch (Throwable ex)
		{
			wFile.setError(true);
			numError++;
			if (results != null)
			{
				results.addErrorMessage(xtFile, "Failed to load servlets: " + ex.getClass().getName() + "; " + ex.getMessage());
			}
			log.error("Failed to load servlets from file: " + xtFile, ex);
		}
		finally
		{
			wFile.close();
			FileUtil.close(in);
			try
			{
				if (zipFile != null) zipFile.close();
			}
			catch (IOException ex)
			{}
		}
	}

	@Override
	public void done(ExtensionsManager manager) 
	{}

	@Override
	public void done(ExtensionsManager manager, Throwable t)
	{}
	// -----------------------------------------
	
	/**
		Get the number of files that was successfully processed.
	*/
	public int getNumFiles()
	{
		return numFiles;
	}
	
	/**
		Get the number of files that was had an error.
	*/
	public int getNumError()
	{
		return numError;
	}

	/**
		Get the total number of servlets found during the last processing.
	*/
	public int getNumServlets()
	{
		return numServlets;
	}
	
}
