/**
	$Id: ServletLoader.java 6880 2015-04-21 06:26:57Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;
import java.util.List;

import javax.servlet.Servlet;

import net.sf.basedb.util.ClassUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XmlUtil2;
import net.sf.basedb.util.extensions.xml.XmlLoader;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.Namespace;

/**
	Loads servlet definitions from an XML file. The file format is a 
	simplified version of the format allowed for <code>web.xml</code>
	files in Tomcat. We have kept only the &lt;servlet&gt; tag
	and it's subtags.

	@author nicklas
	@version 2.7
	@base.modified $Date: 2015-04-21 08:26:57 +0200 (tis, 21 apr. 2015) $
 */
public class ServletLoader
{

	/**
		The URL pointing to the servlets.xsd schema.
	*/
	private static final String schemaFileURL = 
		XmlLoader.class.getResource("/net/sf/basedb/core/xsd/servlets.xsd").toString();
	
	/**
		The name of the servlets namespace.
	*/
	private static final String namespace = "http://base.thep.lu.se/servlets.xsd";
	
	// The last document that was validated
	private Document validatedDom;
	
	// All loaded servlets
	private List<ServletWrapper> servlets;
	
	/**
		Create a new XML loader instance.
	*/
	public ServletLoader()
	{
		servlets = new LinkedList<ServletWrapper>();
	}

	
	/**
		Load a servlet definition XML file. This method may be called 
		multiple times with different XML files. The loading may also be
		done by a two-step process where the first step validates the file
		and the second step loads the extensions. See {@link 
		#validateXmlFile(InputStream, String)} and {@link 
		#loadLastValidatedFile(ClassLoader, boolean)}.
	
		@param xmlFile An input stream to read the XML data from
		@param filename The original filename the stream is coming from, or null
			if not known. This value is only used when generating error messages
		@param classLoader The classloader to use when loading classes that are
			named in the XML file, or null to use the default classloader (=the same
			class loader that loaded the BASE core classes)
		@param clear TRUE to clear all already loaded servlets before loading 
			the servlets in this file
		@throws IOException If there is an error reading the XML file
		@throws ClassNotFoundException If a servlet class named in the XML file can't
			be found
		@throws NoSuchMethodException If a servlet class doesn't implement a
			public, no-argument constructor
		@throws IllegalAccessException If a servlet class constructor isn't 
			public
		@see #validateXmlFile(InputStream, String)
		@see #loadLastValidatedFile(ClassLoader, boolean)
	*/
	public void loadXmlFile(InputStream xmlFile, String filename,
			ClassLoader classLoader, boolean clear)
		throws IOException, ClassNotFoundException, NoSuchMethodException,
			IllegalAccessException
	{
		validateXmlFile(xmlFile, filename);
		loadLastValidatedFile(classLoader, clear);
	}

	/**
		Validate an XML file against the servlet definition schema. If the
		file is valid you may continue to load the servlet definitions. See
		{@link #loadLastValidatedFile(ClassLoader, boolean)}.
		
		@param xmlFile An input stream to read the XML data from
		@param filename The original filename the stream is coming from, or null
			if not known. This value is only used when generating error messages
		@throws IOException If there is an error reading the XML file
	*/
	public void validateXmlFile(InputStream xmlFile, String filename)
		throws IOException
	{
		validatedDom = null;
		validatedDom = loadDocument(xmlFile, filename);
	}
	
	/**
		Continue loading servlet definitions from the last validated XML file. This is
		the second step in a two-step process and requires that 
		{@link #validateXmlFile(InputStream, String)} has been successfully called 
		first.
	
		@param classLoader The classloader to use when loading classes that are
			named in the XML file, or null to use the default classloader (=the same
			class loader that loaded the BASE core classes)
		@param clear TRUE to clear all already loaded servlets before loading 
			the servlets in this file

		@throws ClassNotFoundException If a servlet class named in the XML file can't
			be found
		@throws NoSuchMethodException If a servlet class doesn't implement a
			public, no-argument constructor
		@throws IllegalAccessException If a servlet class constructor isn't 
			public
	*/
	public void loadLastValidatedFile(ClassLoader classLoader, boolean clear)
		throws ClassNotFoundException, IllegalAccessException, NoSuchMethodException
	{
		if (validatedDom == null) 
		{
			throw new NullPointerException("No file has been validated");
		}
		if (clear)
		{
			servlets.clear();
		}
		loadServlets(validatedDom, classLoader);
		validatedDom = null;
	}	
	
	/**
		Checks if an XML file has passed validation in the
		{@link #validateXmlFile(InputStream, String)} method. If so,
		the {@link #loadLastValidatedFile(ClassLoader, boolean)} can be called
		to continue loading the servlet definitions.
		<p>
		Note that once the file has been loaded this flag is reset to
		FALSE.
		
		@return TRUE if a file has been validated, FALSE otherwise
	*/
	public boolean hasValidFile()
	{
		return validatedDom != null;
	}
	
	/**
		Get a list with all loaded servlet definitions.
	*/
	public List<ServletWrapper> getServlets()
	{
		return servlets;
	}
	
	/**
		Load and validate the XML file.
		@return A JDOM document object
	*/
	protected Document loadDocument(InputStream xmlFile, String filename)
		throws IOException
	{
		return XmlUtil2.getSchemaValidatedXML(xmlFile, filename, namespace, schemaFileURL);
	}
	
	/**
		Load the servlet definitions from an XML document.
	*/
	protected int loadServlets(Document dom, ClassLoader classLoader)
		throws ClassNotFoundException, NoSuchMethodException, IllegalAccessException
	{
		int numLoaded = 0;
		List<ServletWrapper> temp = new LinkedList<ServletWrapper>();
		/*
			Format of an servlet definition is:
			<servlet>
				<servlet-name>...</servlet-name>
				<servlet-class>...</servlet-class>
				<init-param>
					<param-name>...</param-name>
					<param-value>...</param-value>
				</init-param>
			</servlet>
		*/
		Element root = dom.getRootElement();
		Namespace ns = root.getNamespace();
		List<Element> servletTags = root.getChildren("servlet", ns);
		for (Element servletTag : servletTags)
		{
			// Servlet name (required)
			String servletName = Values.getStringOrNull(servletTag.getChildText("servlet-name", ns));
			if (servletName == null) throw new NullPointerException("servlet-name");
			
			// Servlet class name (required)
			String className = Values.getStringOrNull(servletTag.getChildText("servlet-class", ns));
			if (className == null) throw new NullPointerException("servlet-class[servlet-name=" + servletName + "]");
			
			// Load the servlet class and make sure it implements the Servlet interface
			Class<? extends Servlet> servletClass = loadServletClass(className, classLoader);
			ServletWrapper wrapper = new ServletWrapper(servletClass, servletName);
			
			// Set initialisation parameters
			List<Element> parameters = servletTag.getChildren("init-param", ns);
			for (Element param : parameters)
			{
				String parameterName = Values.getStringOrNull(param.getChildText("param-name", ns));
				String parameterValue = Values.getStringOrNull(param.getChildText("param-value", ns));
				wrapper.setParameter(parameterName, parameterValue);
			}
			numLoaded++;
			temp.add(wrapper);
		}
		servlets.addAll(temp);
		return numLoaded;
	}
	
	/**
		Load the servlet class. The class must implement the {@link Servlet}
		interface and have a public, no-argument contstructor
		
		@param className The name of the servlet class
		@param classLoader The classloader to use, or null to use
			the BASE core classloader
		@return An initialised factory
	*/
	protected Class<? extends Servlet> loadServletClass(String className, ClassLoader classLoader)
		throws ClassNotFoundException, NoSuchMethodException, IllegalAccessException
	{
		return (Class<? extends Servlet>)ClassUtil.checkAndLoadClass(classLoader, className, true, Servlet.class);
	}
	
}
