package net.sf.basedb.clients.web.extensions.fileviewer;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreEnabled;

/**
	This class is used for holding information about a file that is beeing displayed
	in the GUI in some context. A file item is always available by the {@link #getFile()}
	method. The {@link #getType()} can be used to determine what other information that is
	possible available.

	@author nicklas
	@since 3.16
*/
public class FileViewerContext 
{

	/**
		Creates a file-listing or single-file context without
		any other related items.
		@see Type#FILE
	*/
	public static FileViewerContext file()
	{
		return new FileViewerContext(Type.FILE, null);
	}

	/**
		Creates a data-file listing context where files belong
		to a parent item. The parent item is a {@link FileStoreEnabled}
		item and the context also contains a {@link FileSetMember} and
		{@link DataFileType}.
		@see Type#DATA_FILE
	*/
	public static FileViewerContext dataFile(BasicItem parentItem)
	{
		return new FileViewerContext(Type.DATA_FILE, parentItem);
	}
	
	/**
		Creates an any-to-any listing context where files belong
		to a parent item. The parent item can be any {@link BasicItem}
		item and the context also contains an {@link AnyToAny} link.
		@see Type#ANY_TO_ANY
	*/
	public static FileViewerContext anyToAny(BasicItem parentItem)
	{
		return new FileViewerContext(Type.ANY_TO_ANY, parentItem);
	}
	
	private final Type type;
	private final BasicItem parentItem;
	private File file;
	private AnyToAny any;
	private FileSetMember member;
	private DataFileType fileType;
	
	private FileViewerContext(Type type, BasicItem parentItem)
	{
		this.type = type;
		this.parentItem = parentItem;
	}
	
	/**
		Get the context type that the file is being displayed in.
	*/
	public Type getType()
	{
		return type;
	}

	/**
		If the context has a parent item this method returns it.
	*/
	public BasicItem getParentItem()
	{
		return parentItem;
	}

	/**
		Get the current file.
	*/
	public File getFile()
	{
		return file;
	}
	
	/**
		Get the current any-to-any link pointing to the file.
		This should always be available if the type is
		{@link Type#ANY_TO_ANY} or null otherwise.
	*/
	public AnyToAny getAnyToAny()
	{
		return any;
	}
	
	/**
		Get the current data file type. 
		This should always be available if the type is
		{@link Type#DATA_FILE} or null otherwise.
	*/
	public DataFileType getFileType()
	{
		return fileType;
	}
	
	/**
		Get the current file set member information for the file.
		This should always be available if the type is
		{@link Type#DATA_FILE} or null otherwise.
	*/
	public FileSetMember getMember()
	{
		return member;
	}
	
	/**
		Set the current file.
	*/
	public void setCurrentFile(File file)
	{
		this.file = file;
	}
	
	/**
		Sets the current data file.
	*/
	public void setCurrentDataFile(File file, DataFileType fileType, FileSetMember member)
	{
		this.file = file;
		this.fileType = fileType;
		this.member = member;
	}
	
	/**
		Sets the current any-to-any linked file.
	*/
	public void setCurrentAnyToAny(File file, AnyToAny any)
	{
		this.file = file;
		this.any = any;
	}
	
	public enum Type
	{
		/**
			The context is a regular file listing context. There is no
			parent item or other related items.
		*/
		FILE,
		
		/**
			The context is a list with any-to-any links from a parent item.
			A parent item should be available as well as the any-to-any item.
		 */
		ANY_TO_ANY,
		
		/**
			The context is a list with data-files that are attached to a
			parent item. A parent item should be available as well as 
			the data file type and file set membership information.
		*/
		DATA_FILE;
	}
	
}
