package net.sf.basedb.clients.web.extensions.list;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.formatter.LinkedItemFormatter;
import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.util.AnnotationUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.formatter.CollectionFormatter;
import net.sf.basedb.util.formatter.NameableFormatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	A column action implementation that finds a parent item of a given subtype and
	then get an annotation value from it.
	
	NOTE! Preliminary implementation. Need caching, lazy-loading, snapashot manager, 
	support for bioplate, support for "name", etc. 
	
	@author nicklas
	@since 3.16
*/
public abstract class ParentItemColumn
	extends AbstractListColumnBean<Annotatable, Object>
{
	
	/**
		For now, we assume that 'expr' is /ITEMTYPE/subtype-id/#annotationtype-id or
		/ITEMTYPE/subtype-id/property
	*/
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static ParentItemColumn create(DbControl dc, int index, String expr, ParentItemHelper helper)
	{
		ParentItemColumn col = null;
		try
		{
			String[] tmp = expr.split("/");
			Item parentType = Item.valueOf(tmp[1]);
			ItemSubtype subtype = ItemSubtype.getById(dc, Values.getInt(tmp[2]));
			Filter<Annotatable> parentFilter = new SubtypeFilter(subtype);
			if (tmp[3].startsWith("#"))
			{
				AnnotationType at = AnnotationType.getById(dc, Values.getInt(tmp[3].substring(1)));
				col = new ParentItemAnnotationColumn(index, expr, parentFilter, helper, at);
				col.setTitle(at.getName() + " [A]");
				col.setTooltip(subtype.getName() + " › "+at.getName());
				col.setValueType(at.getValueType());
				col.setClazz("parentitemcol");
				col.setExportFormatter(FormatterFactory.getTypeFormatter(dc.getSessionControl(), at.getValueType()));
				if (!helper.lazy) col.setFormatter(new CollectionFormatter(col.getExportFormatter()));
			}
			else
			{
				String path = tmp[3];
				PropertyPath property = null;
				if (!".".equals(path)) // special case '.' is equal to the "parent" item
				{
					Metadata m = Metadata.getInstance(parentType.getItemClass());
					property = m.getPropertyPath(path, false);
					col = new ParentItemPropertyColumn(index, expr, parentFilter, helper, property);
					col.setTitle(subtype.getName()+"."+pathToTitle(path));
					col.setTooltip(subtype.getName() + " › "+path);
				}
				else
				{
					col = new ParentItemPropertyColumn(index, expr, parentFilter, helper, null);
					col.setTitle(subtype.getName());
					col.setTooltip(subtype.getName());
					col.setFilterProperty("/"+tmp[1]+"/"+tmp[2]+"/name");
				}
				col.setClazz("parentitemcol");
				if (property == null || property.getHibernateType().isEntityType())
				{
					col.setExportFormatter(new NameableFormatter());
					if (col.getFilterProperty()==null) 
					{
						col.setFilterProperty(col.getProperty()+".name");
					}
					col.setValueType(Type.STRING);
					if (!helper.lazy) col.setFormatter(new CollectionFormatter(new LinkedItemFormatter()));
				}
				else
				{
					Type colType = Type.fromHibernateType(property.getHibernateType());
					if (colType != null)
					{
						col.setValueType(colType);
						col.setExportFormatter(FormatterFactory.getTypeFormatter(dc.getSessionControl(), colType));
					}
					else
					{
						// Last fallback -- if we get here the Type.fromHibernateType() should be update with special cases
						col.setValueType(Type.STRING);
						col.setExportFormatter(new ToStringFormatter<>());
					}
					if (!helper.lazy) col.setFormatter(new CollectionFormatter(col.getExportFormatter()));
				}
			}
		}
		catch (RuntimeException ex)
		{}
		
		return col;
	}

	private static Map<String, String> pathTitles = new HashMap<>();
	
	static
	{
		// See also add_parent_columns.js
		pathTitles.put("externalId", "External ID");
		pathTitles.put("bioWell.bioPlate", "Bioplate");
		pathTitles.put("entryDate", "Registered");
		pathTitles.put("remainingQuantity", "Remaining quantity");
		pathTitles.put("creationEvent.entryDate", "Registered");
		pathTitles.put("creationEvent.protocol", "Protocol");
		pathTitles.put("creationEvent.hardware", "Hardware");
		pathTitles.put("remainingQuantity", "Remaining quantity");
	}
	
	/**
		Convert some paths to a more user-friendly title instead.
	*/
	private static String pathToTitle(String path)
	{
		String title = pathTitles.get(path);
		return title == null ? path.substring(0, 1).toUpperCase() + path.substring(1) : title;
	}
	
	private final int index;
	private final Filter<Annotatable> filter;
	private final ParentItemHelper helper;
	
	ParentItemColumn(int index, String id, Filter<Annotatable> filter, ParentItemHelper helper)
	{
		this.index = index;
		this.filter = filter;
		this.helper = helper;
		setId(id);
		setProperty(id);
		setExportable(true);
		setFilterable(true);
	}

	/**
		We finalize this implementation to make sure that the helper 
		implementation always get a chance to re-cycle transactions.
		Subclasses should implement {@link #getValue(ParentItemHelper, Annotatable)}
		to return the requested value from the item and use the
		DbControl from {@link ParentItemHelper#dc}.
	*/
	@Override
	public final Object getValue(DbControl dc, Annotatable item) 
	{
		if (helper.lazy) return getProxyTag(item);
		item = helper.recycle(dc, item);
		return getValue(helper, item);
	}

	/**
		Generate a proxy tag that is returned instead of actual data. A
		JavaScript in 'lazy-parent-items.js' will scan the table and
		inject the actual data in batches.
	*/
	private String getProxyTag(Annotatable item)
	{
		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		sb.append(" class=\"lazy-parent not-loaded\"");
		sb.append(" id=\"lazy-parent-").append(index).append("-").append(item.getId()).append("\"");
		sb.append(" data-item-id=\"").append(item.getId()).append("\"");
		sb.append(" data-index=\"").append(index).append("\"");
		sb.append("></div>");
		return sb.toString();
	}
	
	/**
		We finalize this implementation to make sure that the helper 
		implementation always get a chance to re-cycle transactions.
		Subclasses should implement {@link #getValue(ParentItemHelper, Annotatable)}
		to return the requested value from the item and use the
		DbControl from {@link ParentItemHelper#dc}.
	*/
	@Override
	public final Object getExportValue(DbControl dc, Annotatable item) 
	{
		item = helper.recycle(dc, item);
		return getValue(helper, item);
	}

	
	/**
		Alternate implementation for loading data.
	*/
	protected abstract Object getValue(ParentItemHelper helper, Annotatable item);
	
	/**
		Get all parent items to the requested item that mathces the specified filter.
	*/
	protected Set<Annotatable> getParentItems(Annotatable item)
	{
		return AnnotationUtil.getAllAnnotatableParentItems(helper.dc, item, filter, helper.cache);
	}
	
}