package net.sf.basedb.clients.web.extensions.list;

import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.TransactionalAction;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.util.AnnotationUtil;

/**
	Helper class for storing request-scoped objects that may
	be useful to re-use between multiple columns.

	@author nicklas
	@since 3.16
*/
class ParentItemHelper 
{

	private static final int MAX_ITEMS_IN_TRANSACTION = 200;
	
	final SessionControl sc;
	final boolean lazy;
	// Use the snapshot manager for efficient loading of annotations
	final SnapshotManager manager;
	// Use the cache to improve performance of AnnotationUtil.getAllAnnotatableParents()
	final AnnotationUtil.Cache cache;
	private final Set<Integer> items;

	DbControl dc;
	
	ParentItemHelper(SessionControl sc, boolean lazy) 
	{
		this.sc = sc;
		this.lazy = lazy;
		this.manager = new SnapshotManager();
		this.cache = AnnotationUtil.createCache();
		this.items = new HashSet<>();
	}
	
	/**
		Re-cycle transactions if needed and re-load the given item
		in the new transaction.
	*/
	Annotatable recycle(DbControl dcExt, Annotatable item)
	{
		if (dc == null) dc = dcExt;

		int itemId = item.getId();
		if (items.add(itemId) && items.size() > MAX_ITEMS_IN_TRANSACTION)
		{
			// For every MAX_ITEMS_IN_TRANSACTION items, we create a new DbControl
			// in order to let garbage collection clean up alread processed items.
			// Otherwise we risk that the "entire" database is loaded into
			// a single DbControl
			if (dc != dcExt) 
			{
				// Close the old DbControl but only if it is one that we created
				dc.close();
			}
			else
			{
				// This makes sure that our last DbControl is closed when the
				// external transaction is completed
				dcExt.addTransactionalAction(new CloseMyDbControl(this));
			}
			
			dc = sc.newDbControl();
			cache.clear();
			manager.clear();
			items.clear();
			items.add(itemId);
		}
		
		// We reload the item in the current DbControl
		return item.getType().getById(dc, itemId);
	}
	
	// Close and cleanup
	void close()
	{
		if (dc != null) dc.close();
		manager.clear();
		cache.clear();
		items.clear();
	}
	
	static class CloseMyDbControl
		implements TransactionalAction
	{
		
		private final ParentItemHelper helper;
		public CloseMyDbControl(ParentItemHelper helper)
		{
			this.helper = helper;
		}
		
		@Override
		public void onRollback() 
		{
			helper.close();
		}
		
		@Override
		public void onBeforeCommit() 
		{
			helper.close();
		}
		
		@Override
		public void onAfterCommit() 
		{}
		
	}
	
	
}
