/**
	$Id $

	Copyright (C) 2014 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions.skin;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import javax.servlet.ServletContext;

import net.sf.basedb.core.Application;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.events.EventHandler;
import net.sf.basedb.util.extensions.events.EventType;
import net.sf.basedb.util.extensions.events.ExtensionPointEventFilter;
import net.sf.basedb.util.extensions.manager.ExtensionsManager;

/**
	Utility class for working with image remapping.
	
	@author nicklas
	@since 3.4
*/
public class ImageRemapperUtil 
{

	private static ServletContext servletContext;
	private static ImageRemapper currentImageRemapper;
	private static transient boolean newMapperIsNeeded;
	
	/**
		Initialized at server startup.
	*/
	public static synchronized void init(ServletContext context)
	{
		if (servletContext != null) return;

		servletContext = context;
		newMapperIsNeeded = true;
		
		// All events related to the global-skin extension point will trigger a remap
		ExtensionsManager manager = Application.getExtensionsManager();
		manager.getRegistry().registerEventHandler(
			new EventHandler() 
			{
				@Override
				public void handleEvent(EventType event, ExtensionPoint<?> extensionPoint, Extension<?> extension) 
				{
					newMapperIsNeeded = true;
				}
			}, 
			new ExtensionPointEventFilter("net.sf.basedb.clients.web.global-skin"), null);
	}
	
	/**
		Get a new image remapper if it is needed, otherwise null.
		Once the remapping is complete, use {@link #setCurrentMapper(ImageRemapper)}
		to change the current image remapper.
	*/
	public static ImageRemapper getNewImageRemapperIfNeeded(ServletContext context)
	{
		return newMapperIsNeeded ? new ImageRemapper() : null;
	}

	
	/**
		Get the current image remapper. Can be null if remapping is not needed.
	*/
	public static ImageRemapper getCurrentMapper()
	{
		return currentImageRemapper;
	}
	
	/**
		Changes the current image remapper. The remapping should
		have been complete before calling this method since the image
		remapper is not thread-safe for writing. If the new remapper
		size is 0 the current remapper is set to null.
	*/
	public static synchronized void setCurrentMapper(ImageRemapper remapper)
	{
		currentImageRemapper = remapper != null && remapper.size() > 0 ? remapper : null;
		newMapperIsNeeded = false;
	}

	/**
		Scan the given url for images (png and gif) and remap them. The url 
		can be a path that is relative the root of the server or
		root of the context. It should always start with a /.
		
		The remapping is done so that the given path is searched
		recursively and the subpaths found are remapped the full
		path. For example, if url is /extensions/example.jar/remap
		and this folder contains a subfolder with two images (images/baselogo.png and images/cancel.png)
		the following remap is done:
		
		/images/baselogo.png --&gt; /extensions/example.jar/remap/images/baselogo.png
		/images/cancel.png --&gt; /extensions/example.jar/remap/images/cancel.png
		
	*/
	public static void remapAllTo(ImageRemapper mapper, String url)
	{
		// Remove root context if present since 
		// ServerContext.getResourePaths use context-relative paths
		if (url.startsWith(servletContext.getContextPath()))
		{
			url = url.substring(servletContext.getContextPath().length());
		}
		
		List<String> pathsToCheck = new ArrayList<String>();
		pathsToCheck.add(url);

		int i = 0;
		while (pathsToCheck.size() > 0 && i < 100) // 100 is just a safety
		{
			i++;
			String pathToCheck = pathsToCheck.remove(pathsToCheck.size()-1);
			Set<String> paths = servletContext.getResourcePaths(pathToCheck);
			if (paths != null)
			{
				for (String p : paths)
				{
					if (p.endsWith(".png") || p.endsWith(".gif"))
					{
						mapper.map(p.substring(url.length()), p);
					}
					else if (p.endsWith("/"))
					{
						pathsToCheck.add(p);
					}
				}
			}
		}
	}
	
}
