/*
	$Id: AnnotationSet.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.mapping.PersistentClass;

import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.snapshot.AnnotationSetSnapshot;
import net.sf.basedb.core.snapshot.AnnotationSnapshot;
import net.sf.basedb.core.snapshot.AnnotationTypeFilter;
import net.sf.basedb.core.snapshot.SnapshotManager;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.util.AnnotationUtil;
import net.sf.basedb.util.filter.Filter;

/**
	An annotation set is the container for all annotations on an item.
	An item can have one annotation of each annotation type that is defined
	for it.
	<p>
	An item can also inherit annotation from it's parents. It is possible
	to inherit all annotations or only specific ones. In the first case this means
	that if the parent gets new annotations those are automatically inherited. 
	In the second case the inherticane is lost if the annotation is deleted. If a
	new annotation of the same annotation type is creted later, it is NOT automatically
	inherited again.

	@author Nicklas
	@version 2.0
	@see AnnotationType
	@see Annotation
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
*/
public class AnnotationSet
	extends BasicItem
	implements Transactional
{
	/**
		The type of item represented by this class.
		@see Item#ANNOTATIONSET
		@see #getType()
	*/
	public static final Item TYPE = Item.ANNOTATIONSET;
	
	/**
		This filter gives everybody read permission to annotations.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new annotation set for the specified item.
		@see AnnotatedItem#getAnnotationSet()
	*/
	static AnnotationSet getNew(DbControl dc, Annotatable item)
		throws BaseException
	{	
		AnnotationSet as = dc.newItem(AnnotationSet.class, item);		
		as.getData().setItemType(item.getType().getValue());		
		return as;
	}
	/**
		Get an <code>AnnotationSet</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>AnnotationSet</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static AnnotationSet getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		AnnotationSet as = dc.loadItem(AnnotationSet.class, id);
		if (as == null) throw new ItemNotFoundException("AnnotationSet[id="+id+"]");
		return as;
	}	
	/**
		Get a {@link ItemQuery} object configured to retrieve <code>AnnotationSet</code> 
		items. This query may return items which the logged in user doesn't have 
		read permission to. At the moment there is no way to solve this problem.
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<AnnotationSet> getQuery()
	{
		return new ItemQuery<AnnotationSet>(AnnotationSet.class, RUNTIME_FILTER);
	}

	/**
		Load the id of the item that this annotation set belongs to.
		@since 2.14
	*/
	static int loadItemId(org.hibernate.Session session, Item itemType, int annotationSetId)
	{
		String hql = "SELECT item.id"+
			" FROM "+itemType.getDataClass().getName()+" item"+
			" WHERE item.annotationSet = :annotationSet";
		org.hibernate.query.Query<Integer> query = HibernateUtil.createQuery(session, hql, Integer.class);
		query.setParameter("annotationSet", annotationSetId, TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query);
	}
	
	/**
		Create a new annotation snapshot for the given annotation set.
		NOTE! This method will always create a new snapshot from the 
		database. The snapshot is not saved in the cache.
		<p>
		It is recommended that snapshots are cached as files in the static cache
		since they are much quiker to load. Use {@link SnapshotManager#getSnapshot(DbControl, int)} 
		is instead of this method.
		
		@param dc A DbControl to use for database access
		@param annotationSetId The ID of the annotation set
		@return A snapshot, or null if the annotation set with the given
			id doesn't exist
		@since 2.14
	*/
	public static AnnotationSetSnapshot createSnapshot(DbControl dc, int annotationSetId)
	{
		AnnotationSetData set = HibernateUtil.loadData(dc.getHibernateSession(), AnnotationSetData.class, annotationSetId);
		if (set == null) return null;
		
		AnnotationSetSnapshot snapshot = new AnnotationSetSnapshot();
		snapshot.init(set);
		return snapshot;
	}
	
	
	/**
		Delete all empty annotation sets. This method is intended to be executed at regular 
		intervals by a cleanup application and will remove annotation sets that have no primary or
		inherited annotations.
		@return The number of deleted annotation sets
	 	@throws BaseException If an empty annotation set could not be deleted
	 	@since 3.2
	*/
	public static synchronized int deleteEmptyAnnotationSets(ProgressReporter progress)
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numDeleted = 0;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			numDeleted = deleteEmptyAnnotationSets(session, progress);
			
			HibernateUtil.commit(tx);
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return numDeleted;
	}
	
	
	/**
		Delete all empty annotation sets. This method is intended to be executed at regular 
		intervals by a cleanup application and will remove annotation sets that have no primary or
		inherited annotations.
		@return The number of deleted annotation sets
	 	@throws BaseException If an empty annotation set could not be deleted
	 	@since 3.2
	*/
	static synchronized int deleteEmptyAnnotationSets(org.hibernate.Session session, ProgressReporter progress)
	{
		int numDeleted = 0;

		if (progress != null)
		{
			progress.display(5, "Loading annotation sets with no annotations...");
		}
		
		// This SQL select annotationset id and type of item for annotation sets
		// that has no annotations
		String sql = "select [aa].[id], [aa].[item_type] from [AnnotationSets] [aa]"+
				" left join [Annotations] [a] on [a].[annotationset_id]=[aa].[id]"+
				" where [a].[annotationset_id] is null"+
				" order by [aa].[item_type]";
		
		org.hibernate.query.Query<Object[]> query = HibernateUtil.createSqlQuery(session, sql, Object[].class);
		List<Object[]> allEmpty = query.list();
		if (progress != null)
		{
			progress.display(25, "Found " + allEmpty.size() + " annotations sets without primary or inherited annotations");
		}
		if (allEmpty.size() == 0) return 0;
		
		// Divide the annotation set ids into one list per table to delete from
		Map<String, List<Integer>> emptyPerTable = new HashMap<String, List<Integer>>();
		int currentType = 0;
		List<Integer> currentIdList = null;
		
		for (Object[] row : allEmpty)
		{
			int id = (Integer)row[0];
			int type = (Integer)row[1];
			if (currentType != type)
			{
				Item itemType = Item.fromValue(type);
				PersistentClass pClass = HibernateUtil.getClassMapping(itemType.getDataClass().getName());
				String table = pClass.getTable().getName();
				currentIdList = new ArrayList<Integer>(10000);
				emptyPerTable.put(table, currentIdList);
			}
			currentIdList.add(id);
		}

		int totalCount = allEmpty.size();
		// Now it is finally time to delete from the "AnnotationSets" table
		// and nullify the reference from the item tables
		String deleteSql = "delete from [AnnotationSets] where [id] in (:ids)";
		org.hibernate.query.Query<?> deleteQuery = HibernateUtil.createSqlQuery(session, deleteSql);

		// Divide the query into chunks since the database driver may not support
		// an arbitrary large number of parameters
		int maxParameters = HibernateUtil.getDbEngine().getMaxParametersInQuery();
		
		for (Map.Entry<String, List<Integer>> entry : emptyPerTable.entrySet())
		{
			String table = entry.getKey();
			List<Integer> ids = entry.getValue();

			String nullifySql = "update [" + table + "] set [annotationset_id] = null where [annotationset_id] in (:ids)";
			org.hibernate.query.Query<?> nullifyQuery = HibernateUtil.createSqlQuery(session, nullifySql);

			int startIndex = 0;
			int endIndex = Math.min(maxParameters, ids.size());
			
			while (startIndex < ids.size())
			{
				List<Integer> sublist = ids.subList(startIndex, endIndex);
				
				if (progress != null)
				{
					int percent = 30+(60*numDeleted)/totalCount;
					progress.display(percent, "Unlinking " + ids.size() + " empty annotation sets from [" + table + "] table...");
				}
				
				deleteQuery.setParameterList("ids", sublist, TypeWrapper.H_INTEGER);
				nullifyQuery.setParameterList("ids", sublist, TypeWrapper.H_INTEGER);
				nullifyQuery.executeUpdate();
				deleteQuery.executeUpdate();
				
				numDeleted += sublist.size();
				startIndex = endIndex;
				endIndex = Math.min(startIndex + maxParameters, ids.size());
			}
		}
		
		if (progress != null)
		{
			progress.display(100, "Deleted " + numDeleted + " unused annotation sets");
		}
		
		return numDeleted;
	}

	
	/**
		The item this annotation set belongs to.
	*/
	private Annotatable item;
	
	/**
		Flag to indicate if this annotation set has been modified 
		and the snapshot must be removed.
	*/
	private boolean snapshotIsInvalid;
	
	/**
		Regular constructor if the item is not known beforehand.
	*/
	AnnotationSet(AnnotationSetData annotationSetData)
	{
		super(annotationSetData);		
	}

	/**
		Special constructor to avoid {@link #getItem()} to hit the database
		when the item is known beforehand.
		@see AnnotatedItem#getAnnotationSet()
	*/
	AnnotationSet(AnnotationSetData annotationSetData, Annotatable item)
	{
		super(annotationSetData);
		setItem(item);
	}
	
	@Override
	AnnotationSetData getData()
	{
		return (AnnotationSetData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Always return TRUE. An annotation set always has a one-to-one relation
		to some item. The annotation set is automatically deleted if that
		item is deleted.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return true;
	}
	/**
		Get the item this annotation set belongs to.
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
		throws BaseException
	{
		Set<ItemProxy> using = super.getUsingItems();
		using.add(new ItemProxy(getItemId(), getItemType()));
		return using;
	}
	
	/**
		READ permission is granted to all users. CREATE, WRITE and DELETE permissions are granted
		if the logged in user has RESTRICTED_WRITE permission on the associated item.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{	
		Annotatable item = null;
		try
		{
			item = getItem();			
		}
		catch (PermissionDeniedException ex)
		{}
		if (item != null && item.hasPermission(Permission.RESTRICTED_WRITE))
		{
			granted |= Permission.grant(Permission.CREATE, Permission.READ, Permission.WRITE, Permission.DELETE);				
		}
		else if (item != null && item.hasPermission(Permission.USE))
		{
			granted |= Permission.grant(Permission.USE);
		}
		else
		{
			granted |= Permission.grant(Permission.READ);				
		}	
		super.initPermissions(granted, denied);
	}
	@Override
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	/**
		Remove the cached snapshot if needed.
	*/
	@Override
	void onBeforeCommit(Action action) 
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Action.CREATE && item != null && item.getId() != 0)
		{
			getData().setItemId(item.getId());
		}
		if ((snapshotIsInvalid && action == Action.UPDATE) || action == Action.DELETE) 
		{
			if (!SnapshotManager.removeSnapshot(getId()))
			{
				throw new BaseException("Could not remove cached snapshot for annotation set[id=" + getId() + "]");
			}
		}
	}
	// -------------------------------------------

	/**
		Get the item this annotation set belongs to.
		@return The <code>Annotatable</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public Annotatable getItem()
		throws PermissionDeniedException, BaseException
	{
		return getItem(null);
	}
	
	/**
		Get the item this annotation set belongs to using a specific 
		DbControl for database access. If the item has already been loaded
		it is re-loaded by the given DbControl.
		
		@param dc An open DbControl, or null to use the current DbControl associated
			with this annotation set
		@return The <code>Annotatable</code> item
		@since 2.9
	*/
	public Annotatable getItem(DbControl dc)
		throws PermissionDeniedException, BaseException
	{
		if (item == null)
		{	
			Integer itemId = getData().getItemId();
			if (dc == null) dc = getDbControl();
			if (itemId == null) 
			{
				itemId = loadItemId(dc.getHibernateSession(), getItemType(), getId());
				getData().setItemId(itemId);
			}
			item = getItemType().getById(dc, itemId);			
		}
		else if (dc != null)
		{
			item = getItemType().getById(dc, item.getId());
		}
		return item;
	}
	
	/**
		Get the ID of the item this annotation set belongs to. This may
		need to hit the database unless the id is already known.
		@since 2.14
	*/
	public int getItemId()
	{
		Integer itemId = getData().getItemId();
		if (itemId == null)
		{
			itemId = loadItemId(getDbControl().getHibernateSession(), getItemType(), getId());
			getData().setItemId(itemId);
		}
		return itemId;
	}
	
	/**
		Set the item this annotation set belongs to.
	*/
	void setItem(Annotatable item)
	{
		this.item = item;
		if (item.getId() != 0) getData().setItemId(item.getId());
	}
	
	/**
		Get the {@link Item} type of the item this annotation set belongs to.
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	/**
		Get the {@link Annotation} object of the specified annotation type.
		If the annotation set doesn't contain an annotation of the specified type
		a new object is created and automatically saved to the database when
		{@link DbControl#commit()} is called. To check if an annotation of a certain
		annotation type exists without creating a new annotation use the
		{@link #hasAnnotation(AnnotationType, Annotation.Source)} method.
		
		<p>
		If the annotation type has enabled project-specific annotations 
		(see {@link AnnotationType#getProjectAnnotations()}), this method 
		will return the annotation for the currently enabled project or
		the default annotation if no specific annotation exists.
		If a new annotation is created it is assigned to the currently
		active project.
		
		@param annotationType The annotation type
		@return An <code>Annotation</code> object
		@throws PermissionDeniedException If a new annotation must be created
			and the logged in user doesn't have write permission
		@throws InvalidDataException If a new annotation must be created
			and the annotation type isn't valid for this type of item
		@throws BaseException If there is another error
	*/
	public Annotation getAnnotation(AnnotationType annotationType)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		DbControl dc = getDbControl();
		
		int primary = Annotation.Source.PRIMARY.ordinal();
		int activeProjectId = annotationType.getProjectAnnotations() ? 
			dc.getSessionControl().getActiveProjectId() : 0;

		// Search for an existing primary annotation
		AnnotationTypeData atd = annotationType.getData();
		AnnotationData found = null;
		for (AnnotationData ad : getData().getAnnotations())
		{
			int projectId = ad.getProjectId();
			if (ad.getSource() == primary && ad.getAnnotationType().equals(atd))
			{
				// This is a primary annotation of the correct type
				// Check for project-specific match
				if (projectId == activeProjectId) // also matches default annotation if no project is active
				{
					found = ad;
					break;
				}
				else if ((projectId == 0))
				{
					// The default value -- keep for now
					found = ad;
				}
			}
		}
		
		Annotation a = null;
		if (found != null) 
		{
			a = dc.getItem(Annotation.class, found);
		}
		else
		{
			// Create a new annotation
			checkPermission(Permission.WRITE);
			annotationType.checkPermission(Permission.USE);
			// AnnotatableProxy items allow annotations of all types
			boolean isAnnotatableProxy = AnnotatableProxy.class.isAssignableFrom(getItemType().getItemClass());
			if (!isAnnotatableProxy && !annotationType.isEnabledForItem(getItemType()) &&  
				!annotationType.isProtocolParameter(getItem().getProtocol()))
			{
				throw new InvalidAnnotationException(annotationType, getItemType());
			}
			
			a = Annotation.getNewPrimary(dc, this, atd);
			dc.saveItemIf(this, a, false);
			setSnapshotInvalid();
		}
		
		return a;
	}
	
	/**
		Get a value for a project-specific annotation using the specified project instead
		of the currently active project. Note that this method behaves differently than
		the {@link #getAnnotation(AnnotationType)} method.
		
		 * If the annotation type is not using project annotations, the project is ignored
		 * If project=null the default annotation  value is returned (new or existing)
		 * If project!=null a project-specific annotation value is returned (new or existing,
		   the default value is never returned)
		   
		The returned annotation is always locked to the project it has been defined for.
		Calling {@link Annotation#setValuesIfDifferent(List, Unit)} will always modify the
		existing value (it will not create a new project-specific value when trying
		to change the default value).

		@param annotationType The annotation type
		@param project A project or null to get the default value
		@since 3.11
	*/
	public Annotation getProjectAnnotation(AnnotationType annotationType, Project project)
	{
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		DbControl dc = getDbControl();
		
		int primary = Annotation.Source.PRIMARY.ordinal();
		int projectId = annotationType.getProjectAnnotations() && project != null ? project.getId() : 0;

		// Search for an existing primary annotation
		AnnotationTypeData atd = annotationType.getData();
		AnnotationData found = null;
		for (AnnotationData ad : getData().getAnnotations())
		{
			if (ad.getSource() == primary && ad.getAnnotationType().equals(atd))
			{
				// This is a primary annotation of the correct type
				// Check for project-specific match
				if (ad.getProjectId() == projectId)
				{
					found = ad;
					break;
				}
			}
		}
		
		Annotation a = null;
		if (found != null) 
		{
			a = dc.getItem(Annotation.class, found);
		}
		else
		{
			// Create a new annotation
			checkPermission(Permission.WRITE);
			annotationType.checkPermission(Permission.USE);
			// AnnotatableProxy items allow annotations of all types
			boolean isAnnotatableProxy = AnnotatableProxy.class.isAssignableFrom(getItemType().getItemClass());
			if (!isAnnotatableProxy && !annotationType.isEnabledForItem(getItemType()) &&  
				!annotationType.isProtocolParameter(getItem().getProtocol()))
			{
				throw new InvalidAnnotationException(annotationType, getItemType());
			}
			
			a = Annotation.getNewPrimary(dc, this, atd);
			dc.saveItemIf(this, a, false);
			setSnapshotInvalid();
		}
		
		a.setProjectId(projectId);
		
		return a;
	}
	
	/**
		Check if the annotation set contains an annotation of the specified type.
		If the annotation type has enabled project-specific annotations 
		(see {@link AnnotationType#getProjectAnnotations()}), this method 
		will only check if a default annotation or a project-specific 
		annotation for the current project exists. Annotations that are specific
		for other projects are ignored.
		
		@param annotationType An <code>AnnotationType</code> object
		@param source The source type of the annotation or null if it doesn't matter
		
		@return TRUE if the annotation set contains an annotation of 
			the specified type, FALSE otherwise
		@since 3.6
	*/
	public boolean hasAnnotation(AnnotationType annotationType, Annotation.Source source)
	{
		if (annotationType == null) return false;
		
		DbControl dc = getDbControl();
		int projectId = annotationType.getProjectAnnotations() ? dc.getSessionControl().getActiveProjectId() : 0;
		
		AnnotationTypeData atd = annotationType.getData();
		for (AnnotationData ad : getData().getAnnotations())
		{
			boolean matchSource = source == null || ad.getSource() == source.ordinal();
			// Matches current project or the default value
			boolean matchProject = ad.getProjectId() == projectId || ad.getProjectId() == 0;
			if (matchSource && matchProject && ad.getAnnotationType().equals(atd))
			{
				return true;
			}
		}
		return false;
	}
	
	/**
		Check if the annotation set contains an annotation of the specified type.
		If the annotation type has enabled project-specific annotations 
		(see {@link AnnotationType#getProjectAnnotations()}), this method 
		will only check if an annotation for the given project exists (or the
		default value if projectId=0). Annotations that are specific
		for other projects are ignored.
		
		@param annotationType An <code>AnnotationType</code> object
		@param source The source type of the annotation or null if it doesn't matter
		@param project A project or null to check for the default value
		
		@return TRUE if the annotation set contains an annotation of 
			the specified type, FALSE otherwise
		@since 3.11
	*/
	public boolean hasProjectAnnotation(AnnotationType annotationType, Annotation.Source source, Project project)
	{
		if (annotationType == null) return false;
		
		DbControl dc = getDbControl();
		int projectId = annotationType.getProjectAnnotations() && project != null ? project.getId() : 0;
		
		AnnotationTypeData atd = annotationType.getData();
		for (AnnotationData ad : getData().getAnnotations())
		{
			boolean matchSource = source == null || ad.getSource() == source.ordinal();
			// Matches current project or the default value
			boolean matchProject = ad.getProjectId() == projectId;
			if (matchSource && matchProject && ad.getAnnotationType().equals(atd))
			{
				return true;
			}
		}
		return false;
	}

	
	/**
		Delete the primary annotation of the specified annotation type.

		If the annotation type has enabled project-specific annotations 
		(see {@link AnnotationType#getProjectAnnotations()}), this method 
		will only remove an annotation if it is specific to the currently
		active project (or the default annotation if no project is active)

		
		@param annotationType The type of the annotation to delete
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission
		@throws InvalidDataException If the annotation type is null
		@throws BaseException If there is another error
	*/
	public void removeAnnotation(AnnotationType annotationType)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		int projectId = getSessionControl().getActiveProjectId();
		removeProjectAnnotation(annotationType, projectId == 0 ? null : Project.getById(getDbControl(), projectId));
	}
	
	/**
		Delete the primary annotation of the specified annotation type.
	
		Note that this method is intended to be used
		with project-specific annotation and will only remove the value
		for the given project (or the default value if projectId=0).
		
		@param annotationType The type of the annotation to delete
		@param project A project or null to remove the default value
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission
		@throws InvalidDataException If the annotation type is null
		@throws BaseException If there is another error
		@since 3.11
	*/
	public void removeProjectAnnotation(AnnotationType annotationType, Project project)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		annotationType.checkPermission(Permission.USE);
		int projectId = annotationType.getProjectAnnotations() && project != null ? project.getId() : 0;
		
		DbControl dc = getDbControl();
		int primary = Annotation.Source.PRIMARY.ordinal();
		AnnotationTypeData atd = annotationType.getData();
		
		Iterator<AnnotationData> it = getData().getAnnotations().iterator();
		while (it.hasNext())
		{
			AnnotationData a = it.next();
			boolean matchSource = a.getSource() == primary;
			boolean matchProject = a.getProjectId() == projectId;
			if (matchSource && matchProject && a.getAnnotationType().equals(atd))
			{
				it.remove();
				if (a.getId() != 0)
				{
					a.getValues().replaceValuesIfDifferent(null, false); // Ensure values are initialized in case logging is needed
					Annotation ann = dc.getItem(Annotation.class, a);
					dc.deleteItem(ann);
					setSnapshotInvalid();
				}
				break;
			}
		}
	}

	/**
		Get a query returning annotations with the specified source.
		
		If the annotation type is using project-specific annotations
		the query will load project-specific values where they exists
		and otherwise default values.

		@param source A source or null to not filter on source
	*/
	public ItemQuery<Annotation> getAnnotations(Annotation.Source source)
	{
		ItemQuery<Annotation> query = Annotation.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("annotationSet"),
				Hql.entity(this)
			)
		);
		query.restrictPermanent(new ProjectSpecificAnnotationsRestriction(getId()));
		if (source != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("source"), 
					Expressions.integer(source.ordinal())
			));
		}
		return query;
	}

	/**
		Get a query returning annotations with the specified source
		and belonging to the specified project (or default
		values if projectId=0).
	
		@param source A source or null to not filter on source
		@param project The project or null to query default values
		@since 3.11
	*/
	public ItemQuery<Annotation> getProjectAnnotations(Annotation.Source source, Project project)
	{
		ItemQuery<Annotation> query = Annotation.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("annotationSet"),
				Hql.entity(this)
		));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("projectId"), 
				Expressions.integer(project != null ? project.getId() : 0)
		));
		if (source != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("source"), 
					Expressions.integer(source.ordinal())
			));
		}
		return query;
	}

	/**
		Inherit or clone an annotation. NOTE! If the annotation is already
		inherited or cloned, and the clone parameter is not matching, the
		existing annotations is deleted and replaced with a new annotation. 
		
		@param annotation The annotation to inherit which must be a primary
			annotation
		@param clone TRUE to clone the values, FALSE to only keep the link to
			the original values
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set or
			use permission for the annotation or if the annotation type is not
			enabled for inheritance
		@throws InvalidDataException If the annotation is null
		@return The new annotation instance (or the old if it already was inherited)
		@since 3.6
	 */
	public Annotation inheritAnnotation(Annotation annotation, boolean clone)
	{
		checkPermission(Permission.WRITE);
		if (annotation == null) throw new InvalidUseOfNullException("annotation");
		if (annotation.getSource() != Annotation.Source.PRIMARY)
		{
			throw new PermissionDeniedException("Annotation is not a primary annotation: " + annotation.getAnnotationType());
		}
		annotation.checkPermission(Permission.USE);
		if (annotation.getAnnotationType().getDisableInheritance())
		{
			throw new PermissionDeniedException("Inheritance has been disabled on " + annotation.getAnnotationType());
		}
		
		// Check if the annotation is already inherited
		DbControl dc = getDbControl();
		AnnotationData toInherit = annotation.getData();
		Annotation a = null;
		Iterator<AnnotationData> it = getData().getAnnotations().iterator();
		while (it.hasNext())
		{
			AnnotationData existing = it.next();
			AnnotationData existingInherited = existing.getInheritedFrom();
			if (existingInherited == null) continue;
			
			if (toInherit.equals(existingInherited)) 
			{
				// Found it...
				a = dc.getItem(Annotation.class, existing);
				
				// Check if the existing annotation is cloned or not
				boolean isCloned = existing.getSource() == Annotation.Source.CLONED.ordinal();
				if (isCloned != clone)
				{
					// We should switch between CLONED/INHERITED so we need to remove the
					// existing annotation
					dc.deleteItem(a);
					it.remove();
					a = null;
				}
				break;
			}
		}

		if (a == null)
		{
			a = Annotation.getNewInherited(dc, this, toInherit, clone);
			dc.saveItemIf(this, a, false);
			setSnapshotInvalid();
		}
		return a;
	}
	
	/**
		Remove an inherited an annotation. The specified annotation can be
		either a CLONED or INHERITED annotation in this set or the PRIMARY
		annotation of another annotation set that is inherited/cloned by
		this set.
		
		@param annotation The annotation to remove
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set
		@throws InvalidDataException If the annotation is null
	*/
	public void removeInheritedAnnotation(Annotation annotation)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (annotation == null) throw new InvalidUseOfNullException("annotation");
		
		DbControl dc = getDbControl();
		AnnotationData toRemove = annotation.getData();
		Iterator<AnnotationData> it = getData().getAnnotations().iterator();
		boolean checkInherited = annotation.getSource() == Annotation.Source.PRIMARY;
		
		while (it.hasNext())
		{
			AnnotationData ad = it.next();
			boolean remove = checkInherited ? toRemove.equals(ad.getInheritedFrom()) : toRemove.equals(ad);
			if (remove)
			{
				Annotation a = dc.getItem(Annotation.class, ad);
				dc.deleteItem(a);
				it.remove();
				setSnapshotInvalid();
				break;
			}
		}
	}
	
	/**
		Check if an annotation is inherited or not by this annotation set.
		@param annotation The annotation to check
	  	@return TRUE if inherited, FALSE otherwise. 
		@throws InvalidDataException If the annotation is null
	*/
	public boolean isInherited(Annotation annotation)
		throws InvalidDataException
	{
		if (annotation == null) throw new InvalidUseOfNullException("annotation");

		AnnotationData inherited = annotation.getData();
		for (AnnotationData ad : getData().getAnnotations())
		{
			if (inherited.equals(ad.getInheritedFrom())) return true;
		}
		return false;
	}
	
	/**
		Utility method for inheriting or cloning all annotations 
		from the given annotation set.
		
		@param annotationSet The annotation set to inherit
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set or
			use permission for the annotation set to inherit
		@throws InvalidDataException If the annotation set is null
		@return The number of added annotations
		@since 3.6
	*/
	public int inheritAll(AnnotationSet annotationSet, boolean clone)
	{
		checkPermission(Permission.WRITE);
		if (annotationSet == null) throw new InvalidUseOfNullException("annotationSet");
		annotationSet.checkPermission(Permission.USE);
		
		// Load already inherited annotations
		Set<AnnotationData> alreadyInherited = new HashSet<AnnotationData>();
		for (AnnotationData a : getData().getAnnotations())
		{
			alreadyInherited.add(a.getInheritedFrom());
		}

		int added = 0;
		int primary = Annotation.Source.PRIMARY.ordinal();
		DbControl dc = getDbControl();

		for (AnnotationData from : annotationSet.getData().getAnnotations())
		{
			AnnotationTypeData at = from.getAnnotationType();
			if (from.getSource() == primary && !at.getDisableInheritance() && !alreadyInherited.contains(from))
			{
				Annotation a = Annotation.getNewInherited(dc, this, from, clone);
				a.disableATPermissionCheck();
				dc.saveItemIf(this, a, false);
				added++;
			}
		}
		
		if (added > 0) setSnapshotInvalid();
		return added;
	}
	
	/**
		Utility method to stop inheriting all annotations in the given
		annotation set.
		@return The number of removed annotations
		@since 3.6
	*/
	public int removeAllInherited(AnnotationSet annotationSet)
	{
		checkPermission(Permission.WRITE);
		if (annotationSet == null) throw new InvalidUseOfNullException("annotationSet");
		
		// Annotations that we should no longer inherit
		Set<AnnotationData> stopInherit = annotationSet.getData().getAnnotations();

		Iterator<AnnotationData> it = getData().getAnnotations().iterator();
		int removed = 0;
		DbControl dc = getDbControl();
		
		while (it.hasNext())
		{
			AnnotationData ad = it.next();
			if (stopInherit.contains(ad.getInheritedFrom()))
			{
				removed++;
				it.remove();
				Annotation a = dc.getItem(Annotation.class, ad);
				dc.deleteItem(a);
				it.remove();
			}
		}

		if (removed > 0) setSnapshotInvalid();
		return removed;
	}

	/**
		Automatically inherit all annotations of one or more specified annotation
		types from parent items that are currntly missing from this annotation
		set
		<p>
		This method will only consider parent items that the logged in user
		has access to with USE permission.
		<p>
		This method may result in inheriting multiple annotations (from different
		parent items) of the specified annotation type.
		
		@param dc The DbControl to use for database access
		@param annotationTypes The annotation type to look for
		@param manager A snapshot manager that is used to look up annotations
		@param preferAnnotationSet If TRUE, the entire annotation set containing the
			found annotation is inherited, otherwise only the found annotation is inherited
			NOTE! Since BASE 3.6 this parameter is ignored
		@since 2.15
	*/
	public int autoInherit(DbControl dc, Collection<AnnotationType> annotationTypes, 
		SnapshotManager manager, boolean preferAnnotationSet)
	{
		if (annotationTypes == null || annotationTypes.isEmpty()) return 0;
		// Find the annotation types that are currently missing from this annotation set
		Set<AnnotationType> missing = new HashSet<AnnotationType>();
		AnnotationSetSnapshot snapshot = manager.getSnapshot(dc, getId());
		AnnotationTypeFilter atFilter = new AnnotationTypeFilter();
		if (snapshot == null)
		{
			// If snapshot==null this annotation set is new and unsaved 
			// and doesn't have any annotations
			missing.addAll(annotationTypes);
		}
		else
		{
			for (AnnotationType at : annotationTypes)
			{
				atFilter.setAnnotationType(at);
				if (manager.findAnnotations(dc, snapshot, atFilter, true).size() == 0)
				{
					missing.add(at);
				}
			}
		}
		if (missing.isEmpty()) return 0;
		ThreadSignalHandler.checkInterrupted();

		// Find all annotated parent items for 
		// which the logged in user has USE permission
		int numInherited = 0;
		Annotatable item = getItem(dc);
		Filter<Annotatable> parentFilter = new Filter<Annotatable>()
		{
			@Override
			public boolean evaluate(Annotatable a)
			{
				return a.isAnnotated() && a.hasPermission(Permission.USE);
			}
		};
		Set<Annotatable> allParents = AnnotationUtil.getAllAnnotatableParentItems(dc, item, parentFilter);

		// For each parent -- inherit any annotation that match the given annotation types
		for (Annotatable parent : allParents)
		{
			ThreadSignalHandler.checkInterrupted();
			AnnotationSet pas = parent.getAnnotationSet();
			snapshot = manager.getSnapshot(dc, pas.getId());
			
			for (AnnotationType at : missing)
			{
				atFilter.setAnnotationType(at);
				List<AnnotationSnapshot> tmp = manager.findAnnotations(dc, snapshot, atFilter, true, false);
				if (tmp.size() > 0)
				{
					numInherited++;
					inheritAnnotation(tmp.get(0).getThisAnnotation(dc), false);
				}
			}
		}
		return numInherited;
	}

	
	/**
		Copy annotations from another annotatable item to this annotation set.
		This method will copy all primary annotations with an annotation type that is 
		valid for this item type. The logged in user doesn't need to have access
		to all annotation types. The copy will still work. The logged in user must
		have write permission on this annotation set.
		<p>
		Project-specific annotations are only copied if they belong to the
		currently active project.
		
		@param item The item to copy annotation from
		@param overwrite TRUE to overwrite existing values
		@since 2.5
	*/
	public void copyFrom(Annotatable item, boolean overwrite)
	{
		checkPermission(Permission.WRITE);
		if (!item.isAnnotated()) return;

		DbControl dc = getDbControl();
		int activeProjectId = dc.getSessionControl().getActiveProjectId();

		// Build map of current primary annotations
		int primary = Annotation.Source.PRIMARY.ordinal();
		Map<AnnotationTypeData, AnnotationData> currentAnnotations = new HashMap<AnnotationTypeData, AnnotationData>();
		for (AnnotationData a : getData().getAnnotations())
		{
			if (a.getSource() != primary) continue; // Skip non-PRIMARY annotations

			int projectId = a.getProjectId();
			AnnotationTypeData at = a.getAnnotationType();
			
			// Skip project-specific annotations for other projects
			if (at.getProjectAnnotations() && projectId != activeProjectId) continue;
			
			currentAnnotations.put(at, a);
		}
		
		int targetItem = getItemType().getValue();
		for (AnnotationData from : item.getAnnotationSet().getData().getAnnotations())
		{
			// Skip non-PRIMARY annotations
			if (from.getSource() != primary) continue; 

			AnnotationTypeData at = from.getAnnotationType();
			// Skip project-specific annotations from other projects
			if (at.getProjectAnnotations() && from.getProjectId() != activeProjectId) continue;
			
			// Check if the target item can hold annotation of this type
			if (!at.getItemTypes().contains(targetItem)) continue;
			
			// Check if there is an existing annotation and skip if it should not overwrite
			AnnotationData current = currentAnnotations.get(at);
			if (current != null && !overwrite) continue;
			
			if (current == null)
			{
				// Create new annotation
				Type valueType = Type.fromValue(at.getValueType());
				current = new AnnotationData();
				current.setAnnotationSet(this.getData());
				current.setAnnotationType(at);
				current.setValues(valueType.newParameterValueData());				
				current.setSource(primary);
				current.setProjectId(at.getProjectAnnotations() ? activeProjectId  : 0);
				
				currentAnnotations.put(at, current);
				Annotation newAnnotation = new Annotation(current, this);
				newAnnotation.disableATPermissionCheck();
				dc.saveItemIf(this, newAnnotation, false);
			}
			
			// Copy the values
			if (current.getValues().replaceValuesIfDifferent(from.getValues().getValues(), false))
			{
				current.setUnit(from.getUnit());
				current.setLastUpdate(new Date());
				setSnapshotInvalid();
			}
		}
		
	}
	
	/**
		Mark the snapshot (if it exists) of this annotation set as
		invalid. The snapshot will be removed when the transaction
		is committed. Once the snapshot has been marked as invalid
		it can't be unmarked.		
		@since 2.14
	*/
	public void setSnapshotInvalid()
	{
		this.snapshotIsInvalid = true;
	}
	
	/**
		Get the status of the 'invalid snapshot' flag.
		@return TRUE if the snapshot has been marked as invalid, 
			FALSE if not
		@since 2.14
	*/
	public boolean isSnapshotInvalid()
	{
		return snapshotIsInvalid;
	}
	
	/**
		Restriction for matching project-specific annotations against
		the currently active project. Default values that are not
		overridden are also included.
		@since 3.10
	*/
	static class ProjectSpecificAnnotationsRestriction
		implements Restriction
	{
		
		private final int annotationSetId;
		
		public ProjectSpecificAnnotationsRestriction(int annotationSetId) 
		{
			this.annotationSetId = annotationSetId;
		}
		
		@Override
		public String toQl(Query query, DbControl dc) 
			throws BaseException 
		{
			int activeProjectId = dc.getSessionControl().getActiveProjectId();
			String root = query.getRootAlias();
			
			// If no project is enabled we only return default values
			String ql = "("+ root + ".projectId="+activeProjectId;
			if (activeProjectId != 0)
			{
				// If a project is enabled we return the project-specific values
				// AND default values that are NOT overridden by default values
				ql += " OR (" + root + ".projectId=0 AND "+root+".id NOT IN "+
					"(select b.overrideId from AnnotationData b "+
					"where b.projectId="+activeProjectId+" AND b.annotationSet="+ annotationSetId + "))";
			}
			ql += ")";
			return ql;
		}
		
		@Override
		public Collection<? extends QueryElement> getChildren()
		{
			return null;
		}

	}
	
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			// Do not add any filters.
		}
	}
}
