/*
	$Id: BasicBatcher.java 6468 2014-06-10 11:27:02Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.IdentifiableData;
import net.sf.basedb.core.data.BatchableData;

import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.List;
import java.util.Date;
import java.util.Set;

import org.hibernate.mapping.Table;
import org.hibernate.metadata.ClassMetadata;

/**
	This is an abstract class for batching functionallity.
	Add all objects and then call <code>flush</code> to
	send the batch to the database. <code>BasicBatcher</code> can
	also have a limit on the batch size, when the batch is
	full <code>BasicBatcher</code> will automatically call
	<code>flush</code> to send the current batch to 
	the database.

	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2014-06-10 13:27:02 +0200 (tis, 10 juni 2014) $
*/
public abstract class BasicBatcher<D extends BatchableData>
	extends AbstractBatcher
	implements AccessControlled
{

	
	/**
		Log all parameter bindings to prepared statement.
	*/
	protected static final org.slf4j.Logger logParam = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.batcher.bind.BasicBatcher");

	/**
		So we don't always have to call logParam.debug()
	*/
	protected static final boolean debugEnabled = logParam.isDebugEnabled();

	/**
		Counts the number of inserts done by the batcher.
	*/
	private int insertCountTotal;

	/**
		The class of data objects accepted by this batcher.
	*/
	private final Class<D> dataClass;
	
	/**
		The entity name as known to Hibernate.
	*/
	private final String entityName;

	/**
		The logged in user's permission to this item. Calculated by
		the {@link #initPermissions(int, int)} method.
	*/
	private int permissions;

	/**
		Hibernate metadata for the batchable object.
	*/
	private final ClassMetadata metaData;
	
	/**
		The table where items of this type are stored.
	*/
	private final Table table;
	
	/**
		The name of the ID property. Usually this is 'id'.
	*/
	private final String idPropertyName;
	
	// SQL prepared statements
	private final PreparedStatement insertStatement;
	private final PreparedStatement updateStatement;
	private final PreparedStatement deleteStatement;

	private final String insertSql;
	private final String updateSql;
	private final String deleteSql;
	
	// Order of parameters in the SQL statements
	private final List<BatchedPropertyInfo> insertPropertyOrder;
	private final List<BatchedPropertyInfo> updatePropertyOrder;
	
	// Keep track of the amount of SQL commands in the batch.
	private int insertCount;
	private int updateCount;
	private int deleteCount;

	private int totalCount;
	
	/**
		Constructor that initiate everything.

		@param dataClass The class of the data objects that are accepted by this batcher
		@param entityName The entity name of the data as known by Hibernate, use null
			to use the dataClass as the entity name
		@throws BaseException If there is an error
	*/
	BasicBatcher(org.hibernate.Session session, Class<D> dataClass, String entityName)
		throws BaseException
	{
		super();
		assert dataClass != null : "dataClass == null";
		
		this.dataClass = dataClass;
		this.entityName = entityName == null ? dataClass.getName() : entityName;
		this.metaData = HibernateUtil.getClassMetadata(this.entityName);
		this.idPropertyName = metaData.getIdentifierPropertyName();
		this.table = HibernateUtil.getClassMapping(this.entityName).getTable();
		this.permissions = 0;

		this.insertPropertyOrder = BatchUtil.getInsertPropertyOrder(this.entityName);
		this.updatePropertyOrder = BatchUtil.getUpdatePropertyOrder(this.entityName);
		
		insertSql = BatchUtil.getInsertSql(this.entityName);
		updateSql = BatchUtil.getUpdateSql(this.entityName);
		deleteSql = BatchUtil.getDeleteSql(this.entityName);
		
		try
		{
			java.sql.Connection c = HibernateUtil.getConnection(session);
			insertStatement = c.prepareStatement(insertSql);
			updateStatement = c.prepareStatement(updateSql);
			deleteStatement = c.prepareStatement(deleteSql);
		}
		catch (SQLException ex)
		{
			logSql.error(ex.getMessage(), ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}

	/*
		From the AccessControlled interface
		-------------------------------------------
	*/
	/**
		Checks if the logged in user has the specified permission on
		this item. The default implementation only checks the role
		keys for permissions. A subclass may not override this method
		since it would make it possible to bypass security checks.
		<p>
		Subclasses that needs to check other keys, should override the {@link
		#initPermissions(int,int)} method instead.

		@param permission A value from the {@link Permission} class
		@return <code>TRUE</code> if the user has the permission,
			<code>FALSE</code> otherwise
	*/
	@Override
	public final boolean hasPermission(Permission permission)
	{
		return Permission.hasPermission(permissions, permission);
	}
	/**
		Checks if the logged in user has the specified permission on
		this item. If not, a {@link PermissionDeniedException} is thrown.
		@param permission A value from the {@link Permission} class
		@throws PermissionDeniedException If the logged in user doesn't
			have the requested permission
	*/
	@Override
	public final void checkPermission(Permission permission)
		throws PermissionDeniedException
	{
		if (!hasPermission(permission))
		{
			throw new PermissionDeniedException(permission, this.getType().toString());
		}
	}
	@Override
	public final Set<Permission> getPermissions()
	{
		return Permission.fromInt(permissions);
	}
	// -------------------------------------------
	/*
		From the Batcher interface
		-------------------------------------------
	*/
	@Override
	public void flush()
		throws BaseException
	{
		flushInsert();
		flushUpdate();
		flushDelete();
	}
	// -------------------------------------------
	/*
		From the AbstractBatcher class
		-------------------------------------------
	*/
	/**
		Close all open SQL statements.
	*/
	@Override
	void onBeforeClose()
		throws BaseException
	{
		super.onBeforeClose();
		try
		{
			if (insertStatement != null) insertStatement.close();
			if (updateStatement != null) updateStatement.close();
			if (deleteStatement != null) deleteStatement.close();
		}
		catch (SQLException ex)
		{
			logSql.error(ex.getMessage(), ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
		}
		if (totalCount > 0 && !getDbControl().isRolledBack())
		{
			analyzeTable(table.getCatalog(), table.getSchema(), table.getName());
		}
	}
	// -------------------------------------------

	/**
		Initialise the logged in user's permissions for this batcher.
		The default implementation checks the role keys. 
		<p>
		Subclasses that needs to check other keys or properties, such 
		as the {@link RawDataBatcher} should override 
		this method. The subclass should calculate additional permissions 
		to be granted or denied, and combine those with whatever was passed
		as parameters. Use the binary OR operator ( | ) to combine
		the permissions. Finally the subclass must call 
		<code>super.initPermissions(granted, denied)</code>.
		
		@param granted Permissions that have been granted by the subclass
		@param denied Permissions that have been denied by the subclass
		@throws BaseException If the permissions couldn't be initialised
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		int rolePermissions = getSessionControl().getRolePermissions(getType());
		if (Permission.hasPermission(rolePermissions, Permission.DENIED))
		{
			granted = 0;
		}
		else
		{
			PluginPermission pp = getPluginPermissions();
			granted |= rolePermissions | pp.granted;
			denied |= pp.denied & ~pp.granted;
		}
		this.permissions = granted & ~denied; // granted AND NOT denied
	}
	
	PluginPermission getPluginPermissions()
	{
		return getSessionControl().getPluginPermissions(getType());
	}

	/**
		Insert a data object. The data object will not be synchronized with the database. The object values
		may be modified.
		@param data Data object to be inserted
		@throws InvalidDataException If the data is null or it doesn't validate
			against the {@link #validate(BatchableData)} method
		@throws PermissionDeniedException If the logged in user doesn't
			have create permission
		@throws BaseException If there is another error
	*/
	public void insert(D data)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.CREATE);
		if (isClosed()) throw new ConnectionClosedException();
		if (data == null) throw new InvalidUseOfNullException("data");
		if (data.getClass() != dataClass) 
		{
			throw new InvalidDataException("Invalid data class. This batcher only accepts '"+
				dataClass.getName()+"' objects not '"+data.getClass().getName()+"'"); 
		}
		validate(data);
		onBeforeCommit(data, Transactional.Action.CREATE);
		addToBatch(data, insertStatement, insertPropertyOrder);

		// If buffer size is set, then flush when full
		if (++insertCount == getBatchSize())
		{
			flushInsert();
		}
	}
	
	/**
		Update a data object. The data object will not be synchronized with the database. The object values
		may be modified.
		@param data Data object to be updated
		@throws InvalidDataException If the data is null or it doesn't validate
			against the {@link #validate(BatchableData)} method
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws BaseException If there is another error
	*/
	public void update(D data)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (isClosed()) throw new ConnectionClosedException();
		if (data == null) throw new InvalidUseOfNullException("data");
		if (data.getClass() != dataClass) 
		{
			throw new InvalidDataException("Invalid data class. This batcher only accepts '"+
				dataClass.getName()+"' objects not '"+data.getClass().getName()+"'"); 
		}
		validate(data);
		setPropertyValue(data, "version", data.getVersion() + 1);
		onBeforeCommit(data, Transactional.Action.UPDATE);
		addToBatch(data, updateStatement, updatePropertyOrder);

		// If buffer size is set, then flush when full
		if (++updateCount == getBatchSize())
		{
			flushUpdate();
		}
	}
	
	/**
		Delete a data object.
		@param data Data object to be deleted
		@throws InvalidDataException If the data is null
		@throws PermissionDeniedException If the logged in user doesn't
			have delete permission
		@throws BaseException If there is another error
	*/
	public void delete(D data)
		throws InvalidDataException, PermissionDeniedException, BaseException
	{
		if (data == null) throw new InvalidUseOfNullException("data");
		if (data.getClass() != dataClass) 
		{
			throw new InvalidDataException("Invalid data class. This batcher only accepts '"+
				dataClass.getName()+"' objects not '"+data.getClass().getName()+"'"); 
		}
		delete(data.getId());
	}
	
	/**
		Delete a data object by id.
		@param id Id of the object to be deleted
		@throws PermissionDeniedException If the logged in user doesn't
			have delete permission
		@throws BaseException If there is another error
	*/
	public void delete(int id)
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.DELETE);
		if (isClosed()) throw new ConnectionClosedException();

		try
		{
			deleteStatement.setInt(1, id);
			deleteStatement.addBatch();
		}
		catch (SQLException ex)
		{
			logSql.error(deleteSql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
		
		// If buffer size is set, then flush when full
		if (++deleteCount == getBatchSize())
		{
			flushDelete();
		}
	}
	
	/**
		Flush the SQL insert command buffer. Sends the remaining
		SQL commands in the batch to the database for 
		execution.
		@throws BaseException Thrown if there is a database error.
	*/
	public void flushInsert()
		throws BaseException
	{
		if (isClosed() || insertStatement == null) return;
		try
		{
			insertStatement.executeBatch();
			insertCountTotal += insertCount;
			totalCount += insertCount;
			insertCount = 0;
		}
		catch (SQLException ex)
		{
			logSql.error(insertSql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}
	
	/**
		Flush the SQL insert command buffer. Sends the remaining
		SQL commands in the batch to the database for 
		execution.
		@throws BaseException Thrown if there is a database error.
	*/
	public void flushUpdate()
		throws BaseException
	{
		if (isClosed() || updateStatement == null) return;
		try
		{
			updateStatement.executeBatch();
			totalCount += updateCount;
			updateCount = 0;
		}
		catch (SQLException ex)
		{
			logSql.error(updateSql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}
	
	/**
		Flush the SQL delete command buffer. Sends the remaining
		SQL commands in the batch to the database for 
		execution.
		@throws BaseException Thrown if there is a database error.
	*/
	public void flushDelete()
		throws BaseException
	{
		if (isClosed() || deleteStatement == null) return;
		try
		{
			deleteStatement.executeBatch();
			totalCount += deleteCount;
			deleteCount = 0;
		}
		catch (SQLException ex)
		{
			logSql.error(deleteSql, ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}
	
	/**
		Add the data object to the batch represented by the prepared statement
		@param data Data object to get property values from
		@param ps The prepared statement to add the data object to
		@param propertyOrder The property names of the parameter order in the prepared statement
		@throws DatabaseException If there is an error
	*/
	private void addToBatch(D data, PreparedStatement ps, List<BatchedPropertyInfo> propertyOrder)
		throws DatabaseException
	{
		updateLastAccess();
		try
		{
			if (debugEnabled) logParam.debug("Adding entity to prepared statement: " + data);
			int index = 1;
			for (BatchedPropertyInfo pi : propertyOrder)
			{
				String propertyName = pi.getName();
				int sqlTypeCode = pi.getSqlTypeCode();
				if ("oldversion".equals(propertyName))
				{
					if (debugEnabled) logParam.debug("Binding oldversion = "+(data.getVersion()-1)+" to parameter index " + index);
					ps.setInt(index, data.getVersion()-1);
				}
				else if (idPropertyName.equals(propertyName))
				{
					if (debugEnabled) logParam.debug("Binding id = "+data.getId()+" to parameter " + index);
					ps.setInt(index, data.getId());
				}
				else
				{
					Object propertyValue = getPropertyValue(data, propertyName);
					if (propertyValue instanceof IdentifiableData) // If association insert id
					{
						if (debugEnabled) logParam.debug("Binding "+propertyName+" = "+((IdentifiableData)propertyValue).getId()+" to parameter " + index);
						ps.setInt(index, ((IdentifiableData)propertyValue).getId());
					}
					else if (propertyValue instanceof Date)
					{
						// Date value need special handling since not all JDBC drivers (Postgres) now 
						// how to translate a Date into java.sql.Date/Time/TimeStamp objects
						if (debugEnabled) logParam.debug("Binding "+propertyName+" = '"+propertyValue+"' to parameter " + index);
						Date date = (Date)propertyValue;
						if (sqlTypeCode == Types.DATE)
						{
							ps.setDate(index, new java.sql.Date(date.getTime()));
						}
						else if (sqlTypeCode == Types.TIME)
						{
							ps.setTime(index, new Time(date.getTime()));
						}
						else if (sqlTypeCode == Types.TIMESTAMP)
						{
							ps.setTimestamp(index, new Timestamp (date.getTime()));
						}
						else
						{
							// This sholdn't happen but just in case...
							ps.setObject(index, propertyValue, sqlTypeCode);
						}
					}
					else if (propertyValue instanceof String)
					{
						// String values are mapped to CLOB on some databases...
						// setObject(index, String, Types.CLOB) doesn't work since a String can't be cast to a Clob
						if (debugEnabled) logParam.debug("Binding "+propertyName+" = '"+propertyValue+"' to parameter " + index);
						ps.setString(index, (String)propertyValue);
					}
					else
					{
						if (debugEnabled) logParam.debug("Binding "+propertyName+" = '"+propertyValue+"' to parameter " + index);
						ps.setObject(index, propertyValue, sqlTypeCode);
					}
				}
				index++;
			}
			ps.addBatch();
		}
		catch (SQLException ex)
		{
			logSql.error(ex.getMessage(), ex);
			LogUtil.logSQLExceptionChain(logSql, ex);
			throw new BaseException(ex);
		}
	}

	/**
		Set a protected property value on a data object using Hibernate
		built-in reflection methods.
		@param data The data object to set the property on
		@param propertyName The name of the property, ie. same
			as <code>getXxx</code> method name minus the <code>get</code>
			part
		@param value The value to give to the property
	*/
	void setPropertyValue(D data, String propertyName, Object value)
	{
		metaData.setPropertyValue(data, propertyName, value);
	}
	
	/**
		Get the value of a protected property value on a data object using 
		Hibernate built-in reflection methods.
		@param data The data object to set the property on
		@param propertyName The name of the property, ie. same
			as <code>getXxx</code> method name minus the <code>get</code>
			part
		@return The value of the property
	*/
	Object getPropertyValue(D data, String propertyName)
	{
		return metaData.getPropertyValue(data, propertyName);
	}
	
	/**
		Called after the object has been validated just before the SQL is
		added to the batch queue.

		@param data Object to update
		@param action The action
	*/
	void onBeforeCommit(D data, Transactional.Action action)
		throws BaseException
	{}

	/**
		Validate the data, since no validation is possible in set methods
		of the data object. Ie. this is very similar to case 2 validation for
		items.
		@param data Data object to validate
		@throws InvalidDataException If the data object is invalid
	*/
	public abstract void validate(D data)
		throws InvalidDataException;

	/**
		Get the type of item supported by the batcher. The returned
		value is one of the values defined in the {@link Item} enumeration.

		@return A value indicating the type of item
		@see Item
	*/
	public abstract Item getType();
	
	/**
	  	Gets the total number of inserts done by the batcher 
	*/
	public int getTotalInsertCount()
	{
		return insertCountTotal;
	}
}

