/**
	$Id: BioPlateEvent.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;

import net.sf.basedb.core.data.BioMaterialEventData;
import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioPlateEventData;
import net.sf.basedb.core.data.BioPlateEventParticipantData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	A bioplate event represents something that happened simultaneously to the biomaterial
	on one or more bioplates. Individual {@link BioMaterialEvent}:s may be linked with
	a bioplate event. The biomaterial events are then locked and most properties
	are set on the bioplate event. The description field is mapped to the
	comment field of the biomaterial events but it is possible to overwrite with a specific comment.

	@author Nicklas
	@since 2.17
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (mån, 22 maj 2017) $
*/
public class BioPlateEvent
	extends CommonItem
	implements Registered, Transactional
{

	
	/**
		The type of item represented by this class.
		@see Item#BIOPLATEEVENT
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOPLATEEVENT;

	/**
		Create a new <code>BioPlateEvent</code> item.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param eventType The event type of the event (must not be null)
		@return The new <code>BioPlateEvent</code> item
		@throws BaseException If there is an error
	*/
	public static BioPlateEvent getNew(DbControl dc, BioPlateEventType eventType)
		throws BaseException
	{
		BioPlateEvent event = dc.newItem(BioPlateEvent.class);
		event.setName("New bioplate event");
		event.setEventType(eventType);
		event.getData().setEntryDate(new Date());
		SessionControl sc = dc.getSessionControl();
		event.setOwner(User.getById(dc, sc.getLoggedInUserId()));
		return event;
	}

	/**
		Get a <code>BioPlateEvent</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioPlateEvent</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static BioPlateEvent getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioPlateEvent event = dc.loadItem(BioPlateEvent.class, id);
		if (event == null) throw new ItemNotFoundException("BioPlateEvent[id="+id+"]");
		return event;
	}
	
	/**
		Get a query that returns bioplate event items.
		@return An {@link ItemQuery} object.
	*/
	public static ItemQuery<BioPlateEvent> getQuery()
	{
		return new ItemQuery<BioPlateEvent>(BioPlateEvent.class);
	}

	private boolean needSynchronize = false;
	BioPlateEvent(BioPlateEventData data)
	{
		super(data);
	}

	@Override
	BioPlateEventData getData()
	{
		return (BioPlateEventData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	/**
		Set the description. This is copied to the comment property 
		on all related biomaterial events that has a null comment 
		(when the transaction is committed)
	*/
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		super.setDescription(description);
		needSynchronize = true;
	}
	// -----------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	@Override
	void onBeforeCommit(Action action)
		throws NotLoggedInException, BaseException
	{
		super.onBeforeCommit(action);
		if (needSynchronize && action != Action.DELETE)
		{
			synchronizeAllBioMaterialEvents();
		}
		if (action == Action.DELETE)
		{
			// Unlink all biomaterial events
			org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(
				getDbControl().getHibernateSession(), "CLEAR_PARTICIPANT_ON_BIOMATERIALEVENTS_FOR_BIOPLATEEVENT");
			/*
				UPDATE BioMaterialEventData bme
				SET bme.bioPlateEventParticipant = null
				WHERE bme.bioPlateEventParticipant =ANY (
					SELECT bep FROM BioPlateEventParticipantData bep
					WHERE bep.event = :event
			)
			*/
			query.setParameter("event", getId(), TypeWrapper.H_INTEGER);
			query.executeUpdate();
		}
	}
	// -----------------------------------------
	
	/**
		Get the {@link Protocol} describing the procedure used in this event.
		@return A <code>Protocol</code> item, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to the protocol
		@throws BaseException If there is another error
	*/
	public Protocol getProtocol()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Protocol.class, getData().getProtocol());
	}

	/**
		Set the {@link Protocol} that describes the procedure used in this event.
		The protocol is also updated on all related biomaterial events (when the
		transaction is committed).
		@param protocol A <code>Protocol</code> item, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission on this event or use permission for the protocol
	*/
	public void setProtocol(Protocol protocol)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (protocol != null) protocol.checkPermission(Permission.USE);
		getData().setProtocol(protocol == null ? null : protocol.getData());
		needSynchronize = true;
	}

	/**
		Get the {@link Hardware} that was used in this event.
	 	@return A <code>Hardware</code> item, or null if not known
	 	@throws PermissionDeniedException If the logged in user doesn't have 
	 		read permission to the hardware
	 	@throws BaseException If there is another error
	*/
	public Hardware getHardware()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Hardware.class, getData().getHardware());
	}
	
	/**
	 	Set the {@link Hardware} that was used in this event. The hardware
	 	 is also updated on all related biomaterial events (when the
		transaction is committed).
	 	@param hardware A <code>Hardware</code> item, or null if not known
	 	@throws PermissionDeniedException If the logged in user doesn't have
	 		write permission on this event or use permission for the hardware
	*/
	public void setHardware(Hardware hardware)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (hardware != null) hardware.checkPermission(Permission.USE);
		getData().setHardware(hardware == null ? null : hardware.getData());
		needSynchronize = true;
	}

	/**
		Get the {@link Kit} that was used in this event.
		@return A <code>Kit</code> item, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission to the kit
		@throws BaseException If there is another error
		@since 3.7
	*/
	public Kit getKit()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Kit.class, getData().getKit());
	}
	
	/**
		Set the {@link Kit} that was used in this event. The kit
	 	is also updated on all related biomaterial events (when the
		transaction is committed).
		
		@param kit A <code>Kit</code> item, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission on this event or use permission for the kit
		@since 3.7
	*/
	public void setKit(Kit kit)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (kit != null) kit.checkPermission(Permission.USE);
		getData().setKit(kit == null ? null : kit.getData());
		needSynchronize = true;
	}
	

	/**
		Get the date this event was done in the lab.
		@return A <code>Date</code> object, or null if not known
	*/
	public Date getEventDate()
	{
		return DateUtil.copy(getData().getEventDate());
	}
	
	/**
		Set the date this event was done in the lab. The date
		 is also updated on all related biomaterial events (when the
		transaction is committed).
		@param eventDate The date, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission on this event
	*/
	public void setEventDate(Date eventDate)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setEventDate(DateUtil.setNullableDate(eventDate, "eventDate"));
		needSynchronize = true;
	}

	/**
		Get the event type of this event.
	*/
	public BioPlateEventType getEventType()
	{
		return getDbControl().getItem(BioPlateEventType.class, getData().getEventType());
	}
	
	void setEventType(BioPlateEventType eventType)
	{
		if (eventType == null) throw new InvalidUseOfNullException("eventType");
		getData().setEventType(eventType.getData());
	}
	
	/**
		Check if the given bioplate is a participant of this event.
		@param plate A bioplate
		@return TRUE if the plate is a participant, FALSE otherwise
	*/
	public boolean isParticipant(BioPlate plate)
	{
		return plate != null && findParticipant(plate.getData()) != null;
	}
	
	/**
		Get the participant object representing the given bioplate in this
		event.
		
		@param bioPlate The bioplate
		@return A BioPlateEventParticipant object or null if the bioplate is not a partipant
	*/
	public BioPlateEventParticipant getParticipant(BioPlate bioPlate)
	{
		if (bioPlate == null) return null;
		DbControl dc = getDbControl();
		BioPlateEventParticipant participant = dc.getItem(BioPlateEventParticipant.class, 
				findParticipant(bioPlate.getData()));
		return participant;
	}
	
	/**
		Add a bioplate as a participant in this event. If the bioplate has already been
		registered as a participant the existing information is updated. The participant
		object is automatically saved to the database if the bioplate event is already saved
		or has been put in the commit queue.
		
		@param bioPlate The bioplate to add as a participant (null is not allowed)
		@param role The rolw of the plate
		@param index The index (if the event requires ordering of the plates)
		@return A new or existing bioplate event participant object
	*/
	public BioPlateEventParticipant addParticipant(BioPlate bioPlate, String role, int index)
	{
		checkPermission(Permission.WRITE);
		if (bioPlate == null) throw new InvalidUseOfNullException("bioPlate");
		
		DbControl dc = getDbControl();
		BioPlateEventParticipant participant = dc.getItem(BioPlateEventParticipant.class, 
				findParticipant(bioPlate.getData()));
		if (participant == null)
		{
			participant = BioPlateEventParticipant.getNew(dc, this, bioPlate);
			dc.saveItemIf(this, participant, false);
		}
		participant.setIndex(index);
		participant.setRole(role);
		return participant;
	}
	
	/**
		Remove the given bioplate from this event. Removing a participant
		will automatically unlink related biomaterial events from this
		event.
		@param bioPlate The bioplate
		@throws PermissionDeniedException If the logged in user 
			doesn't have write permission
	*/
	public void removeParticipant(BioPlate bioPlate)
	{
		checkPermission(Permission.WRITE);
		if (bioPlate == null) return;
		BioPlateEventParticipantData participant = findParticipant(bioPlate.getData());
		if (participant != null) 
		{
			getData().getParticipants().remove(participant);
			if (participant.getId() != 0) 
			{
				getDbControl().deleteItem(new BioPlateEventParticipant(participant));
			}
		}
	}
	
	/**
		Get a query returning all participants in the event.
	*/
	public ItemQuery<BioPlateEventParticipant> getParticipants()
	{
		return BioPlateEventParticipant.getQuery(this);
	}
	
	/**
		Get a query returning all biomaterial events linked to this
		event.
	*/
	public ItemQuery<BioMaterialEvent> getBioMaterialEvents()
	{
		ItemQuery<BioMaterialEvent> query = 
			new ItemQuery<BioMaterialEvent>(BioMaterialEvent.class, null);
		query.joinPermanent(Hql.innerJoin(null, "bioPlateEventParticipant", Item.BIOPLATEEVENTPARTICIPANT.getAlias(), true));
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property(Item.BIOPLATEEVENTPARTICIPANT.getAlias(), "event"), 
					Hql.entity(this)
				)
		);
		return query;
	}
	
	/**
		Find the participant object for the given bioplate.
		@param plate The bioplate
		@return The participant object, or null if the plate isn't a
			participant
	*/
	private BioPlateEventParticipantData findParticipant(BioPlateData plate)
	{
		for (BioPlateEventParticipantData participant : getData().getParticipants())
		{
			if (participant.getBioPlate().equals(plate))
			{
				return participant;
			}
		}
		return null;
	}
	
	/**
		Synchronize all properties on the child biomaterial
		events that are related to this event.
	*/
	void synchronizeAllBioMaterialEvents()
	{
		for (BioPlateEventParticipantData participant : getData().getParticipants())
		{
			for (BioMaterialEventData event : participant.getBioMaterialEvents())
			{
				synchronizeProperties(event, null);
			}
		}
	}
	
	/**
		Synchronize all related properties between this event and the
		biomaterial event. If a participant is given, link the
		biomaterial event to it.
	*/
	void synchronizeProperties(BioMaterialEventData event, BioPlateEventParticipantData participant)
	{
		if (participant != null)
		{
			event.setBioPlateEventParticipant(participant);
			participant.getBioMaterialEvents().add(event);
		}

		BioPlateEventData data = getData();
		if (event.getComment() == null) 
		{
			event.setComment(data.getDescription());
		}
		event.setEntryDate(data.getEntryDate());
		event.setEventDate(data.getEventDate());
		event.setUser(data.getOwner());
		event.setProtocol(data.getProtocol());
		event.setHardware(data.getHardware());
		event.setKit(data.getKit());
	}
	
}
