/*
	$Id: BioSource.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.BioSourceData;
import net.sf.basedb.core.hibernate.TypeWrapper;

import java.util.Date;
import java.util.Set;

/**
	This class represent biosource items. A biosource has
	information about the source of material used in an experiment,
	such as a cell line or patient. A biosource can be used to
	create {@link Sample}:s.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (mån, 22 maj 2017) $
*/
public class BioSource
	extends BioMaterial
	implements Registered
{

	/**
		The type of item represented by this class.
		@see Item#BIOSOURCE
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOSOURCE;
	
	/**
		Create a new <code>BioSource</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>BioSource</code> item
		@throws BaseException If there is an error
	*/
	public static BioSource getNew(DbControl dc)
		throws BaseException
	{
		BioSource bs = dc.newItem(BioSource.class);
		bs.setName("New biosource");
		bs.getData().setEntryDate(new Date());
		return bs;
	}

	/**
		Get an <code>BioSource</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioSource</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static BioSource getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioSource bs = dc.loadItem(BioSource.class, id);
		if (bs == null) throw new ItemNotFoundException("BioSource[id="+id+"]");
		return bs;
	}

	/**
		Get a query that returns biosource items.
		@return An {@link ItemQuery} object.
	*/
	public static ItemQuery<BioSource> getQuery()
	{
		return new ItemQuery<BioSource>(BioSource.class);
	}
	
	
	BioSource(BioSourceData bioSourceData)
	{
		super(bioSourceData);
	}
	
	@Override
	BioSourceData getData()
	{
		return (BioSourceData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no {@link Sample} has been created from this biosource
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return countSamples() > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Sample}:s created from this biosource
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SAMPLES_FOR_BIOSOURCE", Integer.class, "s.id");
			/*
				SELECT {1}
				FROM SampleData s
				WHERE s.parent = :biosource
			*/
		query.setParameter("biosource", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.SAMPLE, query);
		return using;
	}
	// -------------------------------------------

	/**
		Create a new {@link Sample} from this biosource.
		@return The new <code>Sample</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this biosource
		@throws BaseException If there is another error
	*/
	public Sample newSample()
		throws PermissionDeniedException, BaseException
	{
		Sample s= Sample.getNew(getDbControl());
		s.getCreationEvent().setSource(this);
		return s;
	}
	
	/**
		Get a query that returns all samples created from this biosource.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Sample> getSamples()
	{
		ItemQuery<Sample> query = Sample.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("parent"), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Counts all samples that have been created from this biosource. The count
		includes samples that the logged in user doesn't have access to.
		@return The number of samples
	 	@throws BaseException If counting the samples fails. 
	 */
	public long countSamples()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SAMPLES_FOR_BIOSOURCE", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM SampleData s
			WHERE s.parent = :biosource
		*/
		query.setParameter("biosource", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query);
	}
	
	/**
		Count the number of child samples. Since biosources can't be
		pooled this method always return 0 if pooled=TRUE
		@since 2.16
	*/
	@Override
	public long countChildren(boolean pooled)
	{
		return pooled ? 0 : countSamples();
	}
	
}

