/*
	$Id: DataResultIterator.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.ResultIterator;
import net.sf.basedb.core.data.BasicData;

/**
	Return the results of a {@link DataQuery} as an iterator. This type
	of iterator is only used for batchable items like reporter and raw data.
	It is expected that the query uses the stateless Hibernate session available
	from {@link DbControl#getStatelessSession()}. If not,
	the returned objects are automatically disconnected from the Hibernate session
	to avoid memory problems and bypassing permission checks.
	
	@author Samuel, Nicklas
	@version 2.0
	@see DataQuery#iterate(DbControl)
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fre, 14 sep. 2012) $
*/
public class DataResultIterator<I extends BasicData>
	implements ResultIterator<I>
{
	/**
		The internal iterator of BasicData objects.
	*/
	private final ScrollIterator<I> data;

	/**
		The Hibernate Session that was used in the query, null if a stateless
		session is used.
	*/
	private final org.hibernate.Session session;

	/**
		The SessionControl
	*/
	private final SessionControl sc;
	
	/**
		The total number of items.
	*/
	private final long totalCount;

	/**
		The class of the data objects returned by the query (and this object). 
	*/
	private final Class<I> dataClass;

	/**
		The type of items returned by the query.
	*/
	private final Item itemType;

	DataResultIterator(ScrollIterator<I> data, SessionControl sc, org.hibernate.Session session, 
		Class<I> dataClass, long totalCount)
	{
		assert data != null : "data == null";
		this.data = data;
		this.sc = sc;
		this.session = session;
		this.dataClass = dataClass;
		this.itemType = Item.fromDataClass(dataClass);
		this.totalCount = totalCount;
	}

	/**
		From the QueryResult interface
		---------------------------------
	*/
	@Override
	public long getTotalCount()
	{
		return totalCount;
	}
	// ---------------------------------
	/**
		From the ResultIterator interface
		---------------------------------
	*/
	@Override
	public void close()
	{
		data.close();
	}
	@Override
	public boolean isClosed()
	{
		return data.isClosed();
	}
	// ---------------------------------
	/*
		From the Iterator interface
		---------------------------
	*/
	@Override
	public boolean hasNext()
	{
		return data.hasNext();
	}
	@Override
	public I next()
	{
		sc.updateLastAccess();
		I dataObject = data.next();
		if (session != null)
		{
			HibernateUtil.evictData(session, dataObject);
		}
		return dataObject;
	}
	/**
		Not supported.
		@throws UnsupportedOperationException Always
	*/
	@Override
	public void remove()
	{
		throw new UnsupportedOperationException();
	}
	// ---------------------------------
	
	/**
		Get the type of items contained in this list.
		@return An {@link Item} object or null if not known
	*/
	public Item getItemType()
	{
		return itemType;
	}

}
