/*
  $Id: DiskUsage.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.DiskUsageData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

/**
	This class is used to represent the disk usage of
	{@link DiskConsumable} items. The creation, updating
	and deletion of items of this type is handled entirely by Base
	using the data package.
	For each <code>DiskConsuming</code> item there is a corresponding
	item of this type.
	
	@author Enell, Nicklas
	@version 2.0
	@see DiskUsageData
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (mån, 22 maj 2017) $
*/
public class DiskUsage
	extends BasicItem
{
	/**
		The type of item represented by this class.
		@see Item#DISKUSAGE
		@see #getType()
	*/
	public static final Item TYPE = Item.DISKUSAGE;

	
	/**
		Get a query configured to retrieve disk usage for the specified user.
		
		@param user The user to retreive disk usage for, null is allowed if
			the logged in user has generic READ permission for disk usage in which case
			all items will be returned
		@param group The group to retrieve the disk usage for
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<DiskUsage> getQuery(User user, Group group)
	{
		ItemQuery<DiskUsage> query = null;
		if (user != null)
		{
			query = new ItemQuery<DiskUsage>(DiskUsage.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("user"), 
					Hql.entity(user)
				)
			);
		}
		else
		{
			query = new ItemQuery<DiskUsage>(DiskUsage.class);
		}
		if (group != null)
		{
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("group"), 
					Hql.entity(group)
				)
			);
		}
		return query;
	}
	
	public static DiskUsageStatistics getStatistics(DbControl dc)
	{
		return new DiskUsageStatistics(dc);
	}
	
		
	/**
		The item this diks usage information belongs to.
	*/
	private DiskConsumable item;
	
	/**
		Create a new <code>DiskUsage</code> item.
		
		@param data The data class for this item. 
	*/
	DiskUsage(DiskUsageData data)
	{
		super(data);
	}
	
	@Override
	DiskUsageData getData()
	{
		return (DiskUsageData)getBasicData();
	}

	/*
	 From the Identifiable interface
	 ---------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
	 From the BasicItem class
	 ---------------------------------------------
	*/
	/**
		@return Always return true. If the DiskUsage is created it is used. 
	*/
	@Override
	public boolean isUsed()
	{
		return true;
	}
	/**
		Get the item this disk usage information belongs to.
		@since 2.2, 2.13
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		int id = 0;
		if (item != null)
		{
			id = item.getId();
		}
		else
		{
			Item type = getItemType();
			DbControl dc = getDbControl();
			String hql = "SELECT item.id"+
				" FROM "+type.getDataClass().getName()+" item"+
				" WHERE item.diskUsage = :diskUsage";
			org.hibernate.query.Query<Integer> query = HibernateUtil.createQuery(dc.getHibernateSession(), hql, Integer.class);
			query.setParameter("diskUsage", getId(), TypeWrapper.H_INTEGER);
			id = HibernateUtil.loadData(query);
		}
		using.add(new ItemProxy(id, getItemType()));
		return using;
	}
	/**
		READ permission is granted if the logged in user is the owner of this
		disk usage. All other permissions are denied.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (getSessionControl().getLoggedInUserId() == getData().getUser().getId())
		{
			granted |= Permission.grant(Permission.READ);
		}
		denied |= Permission.deny(Permission.WRITE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	
	/**
		Get the item this disk usage holds information for
		@return The <code>DiskConsumable</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
		@since 2.2
	*/
	public DiskConsumable getItem()
		throws PermissionDeniedException, BaseException
	{
		if (item == null)
		{	
			Item type = getItemType();
			DbControl dc = getDbControl();
			String hql = "SELECT item"+
				" FROM "+type.getDataClass().getName()+" item"+
				" WHERE item.diskUsage = :diskUsage";
			org.hibernate.query.Query<BasicData> query = HibernateUtil.createQuery(dc.getHibernateSession(), hql, BasicData.class);
			query.setParameter("diskUsage", getId(), TypeWrapper.H_INTEGER);
			item = (DiskConsumable)dc.getItem(type.getItemClass(), HibernateUtil.loadData(query));
		}
		return item;
	}

	/**
		Get the {@link Item} type of the item this annotation set belongs to.
		@since 2.2
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}

	
	/**
		Get the {@link QuotaType} the {@link DiskConsumable} item is using.
		@return The @{QuotaType} associated with this <code>DiskUsage</code> item.
		@throws BaseException If there is any error
	*/
	public QuotaType getQuotaType()
		throws BaseException
	{
		return getDbControl().getItem(QuotaType.class, getData().getQuotaType());
	}
	
	/**
		Get the location of the {@link DiskConsumable} item.
		@return The {@link Location} of this <code>DiskUsage</code> item.
	*/
	public Location getLocation()
	{
		return Location.fromValue(getData().getLocation());
	}

	/**
		Get the number of bytes the {@link DiskConsumable}
		item is using.
		@return The number of bytes of this <code>DiskUsage</code> item.
	*/
	public long getBytes()
	{
		return getData().getBytes();
	}
	
	/**
		Get the user that owns the {@link DiskConsumable} item.
		@return The <code>user</code>
		@throws BaseException If there is any error 
	*/
	public User getUser()
		throws BaseException
	{
		return getDbControl().getItem(User.class, getData().getUser());
	}
	
	/**
		Get the group that owns the {@link DiskConsumable} item or <code>null</code>
		if ther is none.
		@return The <code>group</code>
		@throws BaseException If there is any error 
	*/
	public Group getGroup()
		throws BaseException
	{
		return getDbControl().getItem(Group.class, getData().getGroup());
	}
}
