/*
	$Id: DoubleParameterType.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.DoubleParameterValueData;
import java.util.List;

/**
	This class represent a parameter type that is a double.
	This type defines a {@link #getLowerLimit() lower limit}
	and an {@link #getUpperLimit upper limit} for the range of
	allowed values.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fre, 14 sep. 2012) $
*/
public class DoubleParameterType
	extends ParameterType<Double>
{
	/**
		The lower limit of the parameter.
	*/
	private final Double lowerLimit;

	/**
		The upper limit of the parameter.
	*/
	private final Double upperLimit;

	/**
		Create a new double parameter type, without any limits and allowing null
		values.
	*/
	public DoubleParameterType()
	{
		this(null, null, null, false);
	}
	
	/**
		Create a new double parameter type.
		@param lowerLimit The lowest allowed value, or null to have no limit
		@param upperLimit The highest allowed value, or null to have no limit
	 	@param defaultValue Default value for the parameter, or null. 
		@param notNull FALSE if nulls values are allowed, TRUE otherwise
	*/
	public DoubleParameterType(Double lowerLimit, Double upperLimit, Double defaultValue, boolean notNull)
	{
		super(Type.DOUBLE, Double.class, defaultValue, notNull);
		this.lowerLimit = lowerLimit;
		this.upperLimit = upperLimit;
	}
	
	/**
	 	Create a new double parameter type with a list of allowed values.
	  	@param lowerLimit The lowest allowed value, or null to have no limit.
	  	@param upperLimit The highest allowed value, or null to have no limit.
	  	@param defaultValue The default value of the parameter.
	  	@param notNull FALSE if nulls values are allowed, TRUE otherwise
	  	@param multiplicity Number of values that can be tied to this parameter, or null to have no limit.
	  	@param width Width of the input field for the parameter in the GUI.
	  	@param height Heigth of the input field for the parameter in the GUI.
	  	@param items List of allowed values..
	*/
	public DoubleParameterType(Double lowerLimit, Double upperLimit, Double defaultValue, boolean notNull, 
		int multiplicity, int width, int height, List<Double> items)
	{
		super(Type.DOUBLE, Double.class, defaultValue, notNull, multiplicity, width, height, items);
		this.lowerLimit = lowerLimit;
		this.upperLimit = upperLimit;
	}

	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return "DoubleParameterType: Lower " + lowerLimit + ", Upper " + upperLimit;
	}
	// -------------------------------------------

	/*
		From the ParameterType class
		-------------------------------------------
	*/
	/**
		Checks if the value is within the range given by the upper and lower 
		limits if those are given.
		@param value The value to test
		@throws InvalidDataException If the value is outside 
			the range of allowed values
	*/
	@Override
	void validateValue(String name, Double value)
		throws InvalidDataException
	{
		if (((lowerLimit != null) && (value.compareTo(lowerLimit) < 0))
			||
			((upperLimit != null) && (value.compareTo(upperLimit) > 0)))
		{
			if (lowerLimit != null && upperLimit != null)
			{
				throw new NumberOutOfRangeException(name, value, lowerLimit, upperLimit);
			}
			else if (lowerLimit != null)
			{
				throw new NumberOutOfRangeException(name, value, lowerLimit, false);
			}
			else if (upperLimit != null)
			{
				throw new NumberOutOfRangeException(name, value, upperLimit, true);
			}
		}
	}
	/**
		Create a new {@link DoubleParameterValueData} object.
	*/
	@Override
	DoubleParameterValueData newParameterValueData()
	{
		return new DoubleParameterValueData();
	}
	// -------------------------------------------


	/**
		Get the lowst valid value of the parameter.
		@return Lowst valid value or null if none is set.
	*/
	public Double getLowerLimit()
	{
		return lowerLimit;
	}
	
	/**
		Get the highest valid value of the parameter.
		@return Highest valid value or null if none is set.
	*/
	public Double getUpperLimit()
	{
		return upperLimit;
	}
}
