/*
	$Id: DynamicExtraValueQuery.java 6127 2012-09-14 07:28:35Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.util.jep.JepFunction;
import net.sf.basedb.util.jep.PositionFunction;

/**
	Represents a query for extra values in the dynamic database. A single
	query object may query against data in one bioassayset at a time.
	<p>
	This type of query is executed by the {@link #iterate(DbControl)} method
	and returns the result as a {@link DynamicResultIterator}.
	
	@author Nicklas
	@version 2.0
	@see ExtraValue#getValues()
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fre, 14 sep. 2012) $
*/
public class DynamicExtraValueQuery
	extends DynamicQuery
{

	private final Select[] defaultSelects;

	/**
		Create a new dynamic query.
		@param cube The cube to query against
		@param rootTable The root table of the query
		@param defaultSelects The columns in the root table that are
			selected by default if no other selections are specfied before
			the query is executed
	*/
	DynamicExtraValueQuery(DataCube cube, VirtualTable rootTable, VirtualColumn... defaultSelects)
	{
		super(cube, rootTable, null);
		this.defaultSelects = new Select[defaultSelects.length];
		for (int i = 0; i < defaultSelects.length; ++i)
		{
			this.defaultSelects[i] = Dynamic.select(defaultSelects[i]);
		}
	}
	
	/*
		From the AbstractSqlQuery class
		-------------------------------------------
	*/
	/**
		Default selection: column, position, extraValue
	*/
	@Override
	Select[] getDefaultSelects()
	{
		return defaultSelects;
	}
	// -------------------------------------------
	/*
		From the DynamicQuery class
		-------------------------------------------
	*/
	/**
		The following JEP functions can be used:
		<ul>
		<li>pos(): {@link PositionFunction}
		</ul>
		@since 2.15
	*/
	@Override
	public JepFunction[] getJepFunctions(DbControl dc, boolean restrictions)
	{
		List<JepFunction> functions = new ArrayList<JepFunction>();
		functions.add(new PositionFunction());
		return functions.toArray(new JepFunction[functions.size()]);
	}
	// -------------------------------------------

}
