/**
	$Id: LogControl.java 7522 2018-11-06 14:50:29Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.lang.ref.WeakReference;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;

import org.hibernate.Cache;

import net.sf.basedb.core.data.ChangeHistoryData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.log.TransactionDetails;

/**
	Gives information about the current transaction and gives log
	implementations limited access to the database.

	@author Nicklas
	@version 2.13
	@base.modified $Date: 2018-11-06 15:50:29 +0100 (tis, 06 nov. 2018) $
*/
public class LogControl
{

	private final WeakReference<DbControl> dc;
	private TransactionDetails transactionDetails;
	private org.hibernate.StatelessSession session;
	// Used for batch inserting into the "ChangeHistoryDetails" table
	private PreparedStatement insertIntoDetails;
	private final int batchSize;
	private int batchCount;
	
	LogControl(DbControl dc)
	{
		this.dc = new WeakReference<DbControl>(dc);
		// Use a higher batch size here since we are dealing with small objects
		this.batchSize = Config.getInt("db.batch-size", 50) * 10;
	}
	
	/**
		Get information about the current transaction.
		@return A TransactionDetails object
	*/
	public TransactionDetails getTransactionDetails()
	{
		if (transactionDetails == null) 
		{
			DbControl tmp = dc.get();
			if (tmp == null || tmp.isClosed()) throw new ConnectionClosedException();
			transactionDetails = new TransactionDetails(this, tmp.getSessionControl());
		}
		return transactionDetails;
	}
	
	/**
		Add a main entry the change history table. Each transaction
		usually results in one such entry. Changes to individual entities
		are logged with {@link #log(ChangeHistoryDetailData)}.
		
		@param history A fully initialized ChangeHistoryData object
	*/
	public void log(ChangeHistoryData history)
	{
		getSession().insert(history);
	}
	
	/**
		Adds information about a modified entity to the change history
		table. The details may not be sent to the database immediately but
		can be batched for more efficient use of resources.
		
		@param detail A fully initialized ChangeHistoryDetailData object
			including a link to {@link ChangeHistoryData} that has 
			been logged with {@link #log(ChangeHistoryData)}
	*/
	public void log(ChangeHistoryDetailData detail)
	{
		try
		{
			if (insertIntoDetails == null)
			{
				// Create the batcher with SQL from the 
				// configuration files
				org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();
				String sql = PredefinedQuery.getQueryString("DBLOG_INSERT_INTO_CHANGEHISTORYDETAILS");
				sql = sql.replace('[', dialect.openQuote()).replace(']', dialect.closeQuote());
				Connection c = HibernateUtil.getConnection(dc.get().getHibernateSession());
				insertIntoDetails = c.prepareStatement(sql);
			}
			
			insertIntoDetails.setInt(1, detail.getChangeHistory().getId());
			insertIntoDetails.setInt(2, detail.getChangeType());
			insertIntoDetails.setInt(3, detail.getItemId());
			insertIntoDetails.setInt(4, detail.getItemType());
			insertIntoDetails.setString(5, detail.getChangeInfo());
			insertIntoDetails.setString(6, detail.getOldValue());
			insertIntoDetails.setString(7, detail.getNewValue());
			insertIntoDetails.addBatch();
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		if (++batchCount == batchSize)
		{
			flush();
		}
	}
	
	/**
		Creates a Hibernate HQL query with a known return type. Use this to get information
		from the database that is relevent for the logging implementation.
		The query uses the stateless session, which means that there
		is no first- or second-level cache and on support for initialization
		of proxies. We recommend that queries are used to return
		scalar values only, not entities.
		@return A Query
		@since 3.12
	*/
	public <R> org.hibernate.query.Query<R> createHqlQuery(String hql, Class<R> returnType)
	{
		return HibernateUtil.createQuery(getSession(), hql, returnType);
	}

	/**
		Flushes any batched changes to the database.
	*/
	public void flush()
	{
		if (batchCount > 0)
		{
			Cache c = HibernateUtil.getSessionFactory().getCache();
			if (c != null) c.evictDefaultQueryRegion();
			try
			{
				insertIntoDetails.executeBatch();
				batchCount = 0;
			}
			catch (SQLException ex)
			{
				throw new BaseException(ex);
			}
		}
	}
	
	/**
		Creates a Hibernate SQL query. Use this to get information
		from the database that is relevent for the logging 
		implementation. The query uses the stateless session.
		@return A query
		@deprecated In 3.12. use {@link #createSqlQuery(String, Class)} instead
	*/
	@Deprecated
	public org.hibernate.query.Query<?> createSqlQuery(String sql)
	{
		return HibernateUtil.createSqlQuery(getSession(), sql);
	}
	
	/**
		Creates a Hibernate SQL query with a known return type. Use this to get information
		from the database that is relevent for the logging 
		implementation. The query uses the stateless session.
		@return A query
		@since 3.12
	*/
	public <R> org.hibernate.query.Query<R> createSqlQuery(String sql, Class<R> returnType)
	{
		return HibernateUtil.createSqlQuery(getSession(), sql, returnType);
	}
	
	private org.hibernate.StatelessSession getSession()
	{
		if (session == null)
		{
			DbControl realDc = dc.get();
			if (realDc == null) throw new ConnectionClosedException();
			session = realDc.getStatelessSession();
		}
		return session;
	}
	
	
}
