/*
	$Id: ProjectKey.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.hibernate.TypeWrapper;

import java.util.Map;
import java.util.List;
import java.util.Set;
import java.util.HashSet;

/**
	This class is used to get information about which projects
	a {@link Shareable} item has been shared to.
	<p>
	A project key is immutable and the project/permission combinations 
	cannot be changed once the key has been created.
	This allows the core to resuse project keys for all items with the same
	combination of project/permission values without fear that someone
	else may change the permissions.
	<p>
	Project keys can be used with {@link Shareable} items
	to share them with projects. If the logged in user has
	set an active project (see {@link SessionControl#setActiveProject(Project)})
	new items will automatically be shared to that project.
	<p>
	To create a new project key use the {@link ProjectPermissions} object and pass 
	the information to  the {@link #getNewOrExisting(DbControl, ProjectPermissions)}
	method.
	<p>
	Another option is to use the {@link MultiPermissions} object which allows
	you to modifiy the sharing information of a whole bunch of shareable items
	in one go. The <code>MultiPermissions</code> class takes care of all
	the cumbersome work of creating new keys and assigning those to the items
	for you.

	@author Nicklas
	@version 2.0
	@see Shareable
	@see SessionControl#setActiveProject(Project)
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (mån, 22 maj 2017) $
*/
public class ProjectKey
	extends Key
{
	/**
		The type of item represented by this class.
		@see Item#PROJECTKEY
		@see #getType()
	*/
	public static final Item TYPE = Item.PROJECTKEY;
	
	/**
		Create a new or load an existing <code>ProjectKey</code> when you
		have a combination of project/permission values. If the database
		contains a project key with the same combination that project key is
		loaded and returned, otherwise a new project key is created.
		If no existing key is found, the new key will immediately be saved
		and committed to the database using another transaction. This ensures
		that no two keys contains identical permissions. 

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param projectPermissions A <code>ProjectPermissions</code> object
			holding the permissions
		@return The <code>ProjectKey</code> item
		@throws BaseException If there is an error
	*/
	public static ProjectKey getNewOrExisting(DbControl dc, ProjectPermissions projectPermissions)
		throws BaseException
	{
		if (projectPermissions == null || projectPermissions.size() == 0)
		{
			throw new InvalidDataException("projectPermissions is null or empty");
		}
		int id = getNewOrExistingId(null, projectPermissions.getProjectPermissions());
		return getById(dc, id);
	}

	/**
		Get a <code>ProjectKey</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>ProjectKey</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static ProjectKey getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		ProjectKey pk = dc.loadItem(ProjectKey.class, id);
		if (pk == null) throw new ItemNotFoundException("ProjectKey[id="+id+"]");
		return pk;
	}

	/**
		Delete all keys that are currently not used by any item. This method 
		is intended to be executed at regular intervals by a cleanup application.
	 	@throws BaseException If there is some kind of error. 
	*/
	public static synchronized int deleteUnusedProjectKeys()
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numDeleted = 0;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
					"FIND_USED_PROJECTKEY_IDS", Integer.class);
			/*
				SELECT DISTINCT sd.projectKey.id
				FROM net.sf.basedb.core.data.ShareableData sd
				WHERE NOT sd.projectKey IS NULL
			*/
			Set<Integer> used = new HashSet<Integer>(HibernateUtil.loadList(query, null));
			if (used.size() == 0) used.add(0);
			org.hibernate.query.Query<ProjectKeyData> keyQuery = HibernateUtil.getPredefinedQuery(session, 
					"SELECT_UNUSED_PROJECTKEYS", ProjectKeyData.class);
			/*
				SELECT pk FROM ProjectKeyData pk 
				WHERE pk.id NOT IN (:used)
			*/
			keyQuery.setParameterList("used", used, TypeWrapper.H_INTEGER);
			List<ProjectKeyData> unused = HibernateUtil.loadList(keyQuery, null);
			for (ProjectKeyData pk : unused)
			{
				HibernateUtil.deleteData(session, pk);
				numDeleted++;
			}
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
		}
		return numDeleted;
	}
	
	ProjectKey(ProjectKeyData projectKeyData)
	{
		super(projectKeyData);
	}

	@Override
	ProjectKeyData getData()
	{
		return (ProjectKeyData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Checks if:
		<ul>
		<li>A {@link Shareable} item is using this key.
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<BasicData> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SHAREABLE_ITEMS_FOR_PROJECTKEY", BasicData.class);
		/*
			SELECT s
			FROM net.sf.basedb.core.data.ShareableData s
			WHERE s.projectKey = :projectKey
		*/
		query.setParameter("projectKey", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query) != null || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Shareable} items shared to this key
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
	
		// Shareable items
		org.hibernate.query.Query<BasicData> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SHAREABLE_ITEMS_FOR_PROJECTKEY", BasicData.class);
			/*
				SELECT s
				FROM net.sf.basedb.core.data.ShareableData s
				WHERE s.projectKey = :projectKey
			*/
		query.setParameter("projectKey", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, query);
		return using;
	}
	/**
		WRITE permission is always denied. USE and CREATE
		permission is always granted.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		denied |= Permission.deny(Permission.WRITE);
		granted |= Permission.grant(Permission.USE, Permission.CREATE);
		super.initPermissions(granted, denied);
	}
	@Override
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	// -------------------------------------------

	/**
		Get the permissions for a project.
		@param project The <code>Project</code> for which we want to get the permission
		@return A <code>Set</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the project is null
		@see Permission
	*/
	public Set<Permission> getPermissions(Project project)
		throws InvalidDataException
	{
		if (project == null) throw new InvalidUseOfNullException("project");
		return Permission.fromInt(IntegerUtil.getInt(getData().getProjects().get(project.getData())));
	}
	
	/**
		Get the ID:s of all projects that are shared with this project key
		having the specified permission.
		@param permission The permission the projects must have
		@return A set with the ID:s of all matching projects, or an empty
			set if there are no matches
	*/
	public Set<Integer> getProjectIds(Permission permission)
	{
		Set<Integer> projects = new HashSet<Integer>();
		for (Map.Entry<ProjectData, Integer> entry : getData().getProjects().entrySet())
		{
			int permissions = entry.getValue();
			if (Permission.hasPermission(permissions, permission))
			{
				projects.add(entry.getKey().getId());
			}
		}
		return projects;
	}
	
	/**
		Find the ID of an existing key with exactly the same
		project/permission combination. If no existing key is found a new one is created.
		@param projectPermissions A <code>Map</code> holding the project/permission combinations
		@return The ID of the new or existing key
	*/
	static synchronized int getNewOrExistingId(org.hibernate.Session session, Map<ProjectData,Integer> projectPermissions)
		throws BaseException
	{
		org.hibernate.Transaction tx = null; 
		if (session == null)
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
		}
		try
		{
			// 1. Find all keys with the same number of entries as we
			// have project/permission combinations
			int numPermissions = projectPermissions.size();
			org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, "GET_PROJECTKEY_IDS_FOR_COUNT", Integer.class);
			/*
				SELECT pk.keyId FROM ProjectKeys pk
				GROUP BY pk.keyId
				HAVING count(pk.projectId) = :numPermissions
			*/
			query.setParameter("numPermissions", numPermissions, TypeWrapper.H_INTEGER);
			List<Integer> candidates = HibernateUtil.loadList(query, null);
			if (candidates.size() > 0)
			{
		
				// 2. Check that the candidates have the same project/permission combinations
				StringBuilder sb = new StringBuilder();
				int i = 0;
				for (Map.Entry<ProjectData,Integer> entry : projectPermissions.entrySet())
				{
					if (i > 0) sb.append(" OR ");
					i++;
					sb.append("(pk.projectId=").append(entry.getKey().getId()).
						append(" AND pk.permission=").append(entry.getValue()).append(") ");
				}
		
				query = HibernateUtil.createQuery(session,
					"SELECT pk.keyId FROM ProjectKeys pk "+
					"WHERE pk.keyId IN (:candidates) AND ("+sb.toString()+") "+
					"GROUP BY pk.keyId "+
					"HAVING COUNT(pk.projectId) = :numPermissions", Integer.class
				);
				query.setParameterList("candidates", candidates, TypeWrapper.H_INTEGER);
				query.setParameter("numPermissions", numPermissions, TypeWrapper.H_INTEGER);
				candidates = HibernateUtil.loadList(query, null);
			}
			if (candidates.size() == 0)
			{
				// No more candidates, create a new project key
				ProjectKeyData data = new ProjectKeyData();
				data.getProjects().putAll(projectPermissions);
				HibernateUtil.saveData(session, data);
				if (tx != null) HibernateUtil.commit(tx);
				return data.getId();
			}
			else
			{
				if (tx != null) HibernateUtil.commit(tx);
				return candidates.get(0);
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		finally
		{
			if (tx != null && session != null) HibernateUtil.close(session);
		}
	}
}




