/*
	$Id: Reporter.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.List;

import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.data.ReporterTypeData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.PermissionRestriction;

import org.hibernate.query.NativeQuery;
import org.hibernate.metadata.ClassMetadata;

/**
	This is an utilitiy class for reporters.

	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (mån, 22 maj 2017) $
*/
public class Reporter
{

	private static ClassMetadata metaData = HibernateUtil.getClassMetadata(ReporterData.class.getName());

	/**
		Get a new instance of a reporter, and setting the external ID.
		The external ID has to be unique for the reporters in the system.
		
		@param externalId The external id of the new reporter
		@return A new instance of a {@link ReporterData} class
		@throws InvalidDataException If the external id is null
	 	@throws BaseException If there is some other error. 
	*/
	public static ReporterData getNew(String externalId)
		throws InvalidDataException, BaseException
	{
		if (externalId == null) throw new InvalidUseOfNullException("externalId");
		ReporterData rd = new ReporterData();
		rd.setExternalId(externalId);
		rd.setName("New reporter");
		return rd;
	}

	/**
		Get a {@link ReporterData} object when you know the ID.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load.
		@return The <code>ReporterData</code> object.
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static ReporterData getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (!dc.getSessionControl().hasPermission(Permission.READ, Item.REPORTER))
		{
			throw new PermissionDeniedException(Permission.READ, "Reporter");
		}
		ReporterData rd = HibernateUtil.loadData(dc.getStatelessSession(), ReporterData.class, id);
		if (rd == null) throw new ItemNotFoundException("Reporter[id="+id+"]");
		return rd;
	}
	
	/**
		Get a {@link ReporterData} object when you know the ID, possibly
		using the cloned reporter annotations of the given experiment.
		If the experiment doesn't have cloned reporter information this
		method is the same as {@link #getById(DbControl, int)}. If
		the experiment has cloned information, all non-cloned properties
		are set to null.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param id The ID of the reporter
		@param experiment The experiment which may have cloned reporter information
		@return The <code>ReporterData</code> object
		@since 3.1
	*/
	public static ReporterData getById(DbControl dc, int id, Experiment experiment)
	{
		VirtualDb vdb = experiment == null ? null : experiment.getVirtualDb();
		if (vdb != null && vdb.hasClonedReporters())
		{
			String sql = getClonedSelectSql(vdb);
			sql += " where " + HibernateUtil.quote("id") + "=:id";
			NativeQuery<ReporterData> q = HibernateUtil.createSqlQuery(dc.getStatelessSession(), sql.toString(), ReporterData.class);
			q.setParameter("id", id, TypeWrapper.H_INTEGER);
			ReporterData rd = HibernateUtil.loadData(q);
			if (rd == null) throw new ItemNotFoundException("Reporter[id="+id+"]");
			return rd;
		}
		else
		{ 
			return getById(dc, id);
		}
	}
	
	/**
		Get a {@link ReporterData} object when you know the external id. Use
		this method only when you need to get a few reporters. If you need
		to get many reporter use the {@link ReporterBatcher#getByExternalId(String)}
		method instead. It has better performance since it resuses a single
		query instead of creating a new one each time as this method does.
		
		@param dc The <code>DbControl</code> which will be used for
			database access.
		@param externalId The external id of the reporter to load.
		@return The <code>ReporterData</code> object.
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static ReporterData getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		if (!dc.getSessionControl().hasPermission(Permission.READ, Item.REPORTER))
		{
			throw new PermissionDeniedException(Permission.READ, "Reporter");
		}
		org.hibernate.StatelessSession session = dc.getStatelessSession();
		org.hibernate.query.Query<ReporterData> query = HibernateUtil.getPredefinedQuery(session, "GET_REPORTER_FOR_EXTERNAL_ID", ReporterData.class);
		query.setParameter("externalId", externalId, TypeWrapper.H_STRING);
		
		ReporterData rd = HibernateUtil.loadData(query);
		if (rd == null) throw new ItemNotFoundException("Reporter[externalId="+externalId+"]");
		return rd;
	}
	
	/**
		Get a {@link ReporterData} object when you know the external id, possibly
		using the cloned reporter annotations of the given experiment. 
		If the experiment doesn't have cloned reporter information this
		method is the same as {@link #getByExternalId(DbControl, String)}.
		If	the experiment has cloned information, all non-cloned properties
		are set to null.
		<p>
		
		Use this method only when you need to get a few reporters. If you need
		to get many reporters use a query from the {@link #getDynamicQuery(Experiment)}
		method instead. It has better performance since it resuses a single
		query instead of creating a new one each time as this method does.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param externalId The external id of the reporter
		@param experiment The experiment which may have cloned reporter information
		@return The <code>ReporterData</code> object
		@since 3.1
	*/
	public static ReporterData getByExternalId(DbControl dc, String externalId, Experiment experiment)
	{
		if (externalId != null) externalId = externalId.trim();
		VirtualDb vdb = experiment == null ? null : experiment.getVirtualDb();
		if (vdb != null && vdb.hasClonedReporters())
		{
			String sql = getClonedSelectSql(vdb);
			sql += " where " + HibernateUtil.quote("external_id") + "=:id";
			NativeQuery<ReporterData> q = HibernateUtil.createSqlQuery(dc.getStatelessSession(), sql.toString(), ReporterData.class);
			q.setParameter("id", externalId, TypeWrapper.H_STRING);
			ReporterData rd = HibernateUtil.loadData(q);
			if (rd == null) throw new ItemNotFoundException("Reporter[externalId="+externalId+"]");
			return rd;
		}
		else
		{ 
			return getByExternalId(dc, externalId);
		}
	}

	/**
		Get a proxy reporter object. This method is usefult when linking
		reporters to other items, and you are sure that a reporter with the given ID
		exists. Since this method doesn't hit the database, there are some things to 
		be aware of:
		<ul>
		<li>There is no check to see if a reporter with the given ID exists or
			not
		<li>The returned object doesn't contain any reporter information
			except the ID.
		</ul>
		
		@param reporterId The ID of the reporter
		@return A reporter object
		@since 2.8
	*/
	public static ReporterData getProxy(int reporterId)
	{
		return new ReporterData(reporterId);
	}

	/**
		Get a query object configured to retrieve reporters.
		@return A {@link DataQuery} object
	*/
	public static DataQuery<ReporterData> getQuery()
	{
		DataQuery<ReporterData> query = new DataQuery<ReporterData>(ReporterData.class, null);
		// We only allow reporters if there is an excplicit READ permission
		query.restrictPermanent(new PermissionRestriction(Permission.READ, Item.REPORTER));
		return query;
	}

	/**
		Get a dynamic query object configured to retrieve reporter information.
		You must add columns/expressions to select with
		{@link net.sf.basedb.core.query.Query#select(net.sf.basedb.core.query.Select)}
		@return A {@link DynamicReporterQuery} object
		@since 2.4
	*/
	public static DynamicReporterQuery getDynamicQuery()
	{
		DynamicReporterQuery query = new DynamicReporterQuery();
		// We only allow reporters if there is an excplicit READ permission
		query.restrictPermanent(new PermissionRestriction(Permission.READ, Item.REPORTER));
		return query;
	}
	
	/**
		Get a dynamic query object configured to retrieve reporter
		information from either the master reporter table or a
		cloned reporter table in the given experiment. 
		You must add columns/expressions to select with
		{@link net.sf.basedb.core.query.Query#select(net.sf.basedb.core.query.Select)}
		@return A {@link DynamicReporterQuery} object
		@since 3.1
	*/
	public static DynamicReporterQuery getDynamicQuery(Experiment experiment)
	{
		if (experiment != null && experiment.getVirtualDb().hasClonedReporters())
		{
			return new DynamicReporterQuery(experiment.getVirtualDb());
		}
		else
		{
			return new DynamicReporterQuery();
		}
	}
	
	/**
		Get the reporter type item from a reporter data object.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
	 	@param reporter The reporter to get the reporter type data for. 
		@return A {@link ReporterType} object.
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static ReporterType getReporterType(DbControl dc, ReporterData reporter)
		throws PermissionDeniedException, BaseException
	{
		ReporterTypeData rtd = (ReporterTypeData)metaData.getPropertyValue(reporter, "reporterType");
		return dc.getItem(ReporterType.class, rtd);
	}
	
	/**
		Set the reporter type on a reporter data object. The reporter must be passed
		to {@link BasicBatcher#update(net.sf.basedb.core.data.BatchableData)} for the changes to be saved in the database.
		@param reporter The reporter data object
		@param reporterType The new reporter type or null
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission on the reporter type
		@throws BaseException If there is another error
	*/
	public static void setReporterType(ReporterData reporter, ReporterType reporterType)
		throws PermissionDeniedException, BaseException
	{
		if (reporterType != null) reporterType.checkPermission(Permission.USE);
		metaData.setPropertyValue(reporter, "reporterType", reporterType == null ? null : reporterType.getData());
	}
	
	/**
		Generate an SQL statement for selecting ALL reporter properties from a cloned
		table. Properties that have not been cloned should be included as 'null'
		selects.
	*/
	private static String getClonedSelectSql(VirtualDb vdb)
	{
		ReporterCloneTemplate template = vdb.getReporterCloneTemplate();
		List<ClonableProperty> all = ClonableProperty.getAll();
		StringBuilder sql = new StringBuilder();
		sql.append("select ");
		for (ClonableProperty cp : all)
		{
			if (template.hasClonableProperty(cp.getName()))
			{
				sql.append(HibernateUtil.quote(cp.getColumn()));
			}
			else
			{
				sql.append("null as " + HibernateUtil.quote(cp.getColumn()));
			}
			sql.append(",");
		}
		sql.append("null as ").append(HibernateUtil.quote("reportertype_id"));
		sql.append(" from ").append(VirtualTable.CLONED_REPORTERS.getQualifiedTableName(vdb));
		return sql.toString();
	}
}
