/*
	$Id: ReporterListJoin.java 6127 2012-09-14 07:28:35Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;

import net.sf.basedb.core.query.Join;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.util.EqualsHelper;

/**
	Represents a join to a reporter list in a dynamic query:
	<code>leftTable JOIN ReporterListScores ON ...</code>
	<p>
	This class is used by, for example, {@link 
	DynamicSpotQuery#joinReporterList(ReporterList, JoinType)} to create the 
	{@link Join} query element that is needed. After the reporter list
	has been join it can be used in expressions in the query.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.ExtraValueExpression
	@see net.sf.basedb.core.query.Dynamic#extraValue(ExtraValue)
	@base.modified $Date: 2012-09-14 09:28:35 +0200 (fre, 14 sep. 2012) $
*/
class ReporterListJoin
	implements Join
{
	
	private final ReporterList reporterList;
	private final JoinType joinType;
	private final String leftAlias;
	private final String leftColumn;
	private final RealTable right;
	
	/**
		Create a new join to a reporter list using the given VirtualTable
		as the left table. The given VirtualColumn should hold the reporter
		id values.
		
		@param reporterList The reporter list to join
		@param joinType The type of join
		@param left The left table 
		@param leftColumn The column in the left table holding reporter id:s
		@since 3.0
	*/
	ReporterListJoin(ReporterList reporterList, JoinType joinType, VirtualTable left, VirtualColumn leftColumn)
	{
		this.reporterList = reporterList;
		this.joinType = joinType;
		this.leftAlias = left.getAlias();
		this.leftColumn = leftColumn.getColumn();
		this.right = RealTable.REPORTER_LIST_SCORES;
	}
	
	/**
		Create a new join to a reporter list using the given RealTable
		as the left table. The given leftColumn should hold the reporter
		id values.
		
		@param reporterList The reporter list to join
		@param joinType The type of join
		@param left The left table 
		@param leftColumn The column in the left table holding reporter id:s
		@since 3.0
	*/
	ReporterListJoin(ReporterList reporterList, JoinType joinType, RealTable left, String leftColumn)
	{
		this.reporterList = reporterList;
		this.joinType = joinType;
		this.leftAlias = left.getAlias();
		this.leftColumn = leftColumn;
		this.right = RealTable.REPORTER_LIST_SCORES;
	}
	
	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	@Override
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String rightAlias = HibernateUtil.quote(right.getAlias()+reporterList.getId());
			
			StringBuilder sb = new StringBuilder();
			sb.append(joinType.getJoinExpression()).append(" ").append(right.getQualifiedTableName());
			sb.append(" AS ").append(rightAlias);
			sb.append(" ON ");
			sb.append(rightAlias).append(".").append(HibernateUtil.quote("reporter_id"));
			sb.append(" = ").append(leftAlias).append(".").append(leftColumn);
			sb.append(" AND ").append(rightAlias).append(".").append("reporterlist_id");
			sb.append(" = ").append(reporterList.getId());
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ReporterListJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	/**
		@return Always null
	*/
	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -------------------------------------------
	/*
	 	From the Join interface
	 	-------------------------------------------
	*/
	@Override
	public String getThetaJoin(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			StringBuilder sb = new StringBuilder();
			sb.append(", ").append(right.getQualifiedTableName());
			sb.append(" ").append(HibernateUtil.quote(right.getAlias()+reporterList.getId()));
			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}

	@Override
	public String getThetaJoinCondition(Query query, DbControl dc) 
		throws BaseException 
	{
		if (query.getQueryType() == QueryType.SQL)
		{
			String rightAlias = HibernateUtil.quote(right.getAlias()+reporterList.getId());
			StringBuilder sb = new StringBuilder();
			sb.append(rightAlias).append(".").append(HibernateUtil.quote("reporter_id"));
			sb.append(" = ").append(leftAlias).append(".").append(leftColumn);
			if (joinType == JoinType.LEFT) sb.append("(+)");
			sb.append(" AND ").append(rightAlias).append(".").append("reporterlist_id");
			sb.append(" = ").append(reporterList.getId());
			if (joinType == JoinType.LEFT) sb.append("(+)");

			return sb.toString();
		}
		else
		{
			throw new UnsupportedOperationException("ThetaJoin operation not supported by "+query.getQueryType()+" query.");
		}
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		String rightAlias = right.getAlias()+reporterList.getId();

		StringBuilder sb = new StringBuilder();
		sb.append(joinType.getJoinExpression()).append(" ").append(right.getTableName());
		sb.append(" AS ").append(rightAlias);
		sb.append(" ON ");
		
		sb.append(rightAlias).append(".").append("reporter_id");
		sb.append(" = ").append(leftAlias).append(".").append(leftColumn);
		sb.append(" AND ").append(rightAlias).append(".").append("reporterlist_id");
		sb.append(" = ").append(reporterList.getId());
		return sb.toString();
	}
	/**
		This join is equal to another ReporterListJoin if they 
		join to the same reporter list with the same join type.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		ReporterListJoin o = (ReporterListJoin)other;
		return EqualsHelper.equals(reporterList, o.reporterList) && joinType == o.joinType;
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(reporterList, joinType);
	}
	// -------------------------------------------
	
}
