/*
	$Id: Sample.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.data.SampleData;
import net.sf.basedb.core.hibernate.TypeWrapper;

import java.util.Set;

/**
	This class is used to represent sample items. A sample is the
	actual sample taken from a {@link BioSource}. A sample
	can also be created by combining other samples into a new sample.
	A process is known as pooling. It is also possible to create
	standalone samples, with no reference to a biosource or to other
	samples.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (mån, 22 maj 2017) $
*/
public class Sample
	extends MeasuredBioMaterial
{

	/**
		The type of item represented by this class.
		@see Item#SAMPLE
		@see #getType()
	*/
	public static final Item TYPE = Item.SAMPLE;
	
	/**
		Create a new <code>Sample</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>Sample</code> item
		@throws BaseException If there is an error
	*/
	public static Sample getNew(DbControl dc)
		throws BaseException
	{
		return getNew(dc, null);
	}
	
	/**
		Create a new <code>Sample</code> item as part of a
		master bioplate event.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param master The master bioplate event participant or null
		@return The new <code>Sample</code> item
		@throws BaseException If there is an error
		@since 2.17
	*/
	public static Sample getNew(DbControl dc, BioPlateEventParticipant master)
		throws BaseException
	{
		Sample s = dc.newItem(Sample.class);
		s.setName("New sample");
		s.setBioPlateEventParticipant(master);
		return s;
}

	/**
		Get an <code>Sample</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Sample</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Sample getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Sample s = dc.loadItem(Sample.class, id);
		if (s == null) throw new ItemNotFoundException("Sample[id="+id+"]");
		return s;
	}

	/**
		Get a query that returns samples.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Sample> getQuery()
	{
		return new ItemQuery<Sample>(Sample.class);
	}
	
	
	Sample(SampleData sampleData)
	{
		super(sampleData);
	}
	
	@Override
	SampleData getData()
	{
		return (SampleData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Subtypable interface
		-----------------------------
	*/
	@Override
	@SupportsPushAnnotations
	@SubtypableRelatedItems({Item.PROTOCOL, Item.KIT, Item.HARDWARE, Item.BIOSOURCE, Item.SAMPLE})
	public ItemSubtype getItemSubtype()
	{
		return super.getItemSubtype();
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that:
		<ul>
		<li>no {@link Extract}:s has been created from this item
		<li>no {@link Sample}:s has been created from this item
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return countExtracts() > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Extract}:s created from this sample
		<li>Pooled samples
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Integer> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_EXTRACTS_FOR_SAMPLE", Integer.class, "e.id");
			/*
				SELECT {1}
				FROM ExtractData e
				WHERE e.parent = :sample
			*/
		query.setParameter("sample", getId(), TypeWrapper.H_INTEGER);
		addUsingItems(using, Item.EXTRACT, query);
		return using;
	}
	// -------------------------------------------

	/**
		Create a new {@link Extract} from this sample.
		@param usedQuantity The quantity used from this sample or null
			if not known
		@return The new <code>Extract</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this sample
		@throws BaseException If there is another error
	*/
	public Extract newExtract(Float usedQuantity)
		throws PermissionDeniedException, BaseException
	{
		return newExtract(usedQuantity, null);
	}

	/**
		Create a new {@link Extract} from this sample as part of a
		master bioplate event.
		@param usedQuantity The quantity used from this sample or null
			if not known
		@param master The master bioplate event participant or null
		@return The new <code>Extract</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			use permission for this sample
		@throws BaseException If there is another error
		@since 2.17
	*/
	public Extract newExtract(Float usedQuantity, BioPlateEventParticipant master)
		throws PermissionDeniedException, BaseException
	{
		Extract e = Extract.getNew(getDbControl(), master);
		e.getCreationEvent().setSource(this).setUsedQuantity(usedQuantity);
		return e;
	}

	
	/**
		Get a query that returns all extracts created from this sample.
		@return An {@link ItemQuery} object
	 	@throws BaseException If the query couldn't be created. 
	*/
	public ItemQuery<Extract> getExtracts()
		throws BaseException
	{
		ItemQuery<Extract> query = Extract.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("parent"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Creates a query that returns all child (pooled) samples that
		has this sample as one of it's sources.
		@return A query
		@since 2.10
	*/
	public ItemQuery<Sample> getChildSamples()
	{
		ItemQuery<Sample> query = Sample.getQuery();
		query.joinPermanent(Hql.innerJoin("creationEvent", Item.BIOMATERIALEVENT.getAlias()));
		query.joinPermanent(Hql.innerJoin(Item.BIOMATERIALEVENT.getAlias(), "sources", "src"));
		query.restrictPermanent(Restrictions.eq(Hql.index("src", null), Hql.entity(this)));
		return query;
	}
	
	
	/**
		Counts all extracts that have been created from this sample. The count
		includes extracts that the logged in user doesn't have access to.
		@return The number of extracts
	 	@throws BaseException If there is some kind of error. 
	 */
	public long countExtracts()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_EXTRACTS_FOR_SAMPLE", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM ExtractData e
			WHERE e.parent = :sample
		*/
		query.setParameter("sample", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query);
	}

	/**
		Counts all samples that have been created from this sample. The count
		includes samples that the logged in user doesn't have access to.
		@return The number of samples
	 	@throws BaseException If there is some kind of error.
	 	@since 2.16
	*/
	public long countSamples()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.query.Query<Long> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SOURCEEVENTS_FOR_BIOMATERIAL", Long.class, "count(*)");
		/*
			SELECT {1}
			FROM BioMaterialEventData bme
			JOIN bme.sources src
			WHERE index(src) = :bioMaterial
		*/
		query.setParameter("bioMaterial", getId(), TypeWrapper.H_INTEGER);
		return HibernateUtil.loadData(query);
	}

	/**
		Count the number of child samples or extracts.
		@param samples TRUE = count child samples, FALSE = count child extracts
		@since 2.16
	*/
	@Override
	public long countChildren(boolean samples)
	{
		return samples ? countSamples() : countExtracts();
	}

	
}
