/*
	$Id: Update.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.hibernate.mapping.Table;
import org.hibernate.resource.transaction.spi.TransactionStatus;

import net.sf.basedb.core.authentication.LoginRequest;
import net.sf.basedb.core.data.AnnotatableData;
import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.core.data.BioMaterialEventData;
import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioPlateTypeData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.ClientData;
import net.sf.basedb.core.data.ContextData;
import net.sf.basedb.core.data.DerivedBioAssayData;
import net.sf.basedb.core.data.ExperimentData;
import net.sf.basedb.core.data.FileData;
import net.sf.basedb.core.data.FileServerData;
import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.data.ItemListData;
import net.sf.basedb.core.data.ItemSubtypeData;
import net.sf.basedb.core.data.JobData;
import net.sf.basedb.core.data.KitData;
import net.sf.basedb.core.data.MeasuredBioMaterialData;
import net.sf.basedb.core.data.NewsData;
import net.sf.basedb.core.data.ParameterValueData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.QuantityData;
import net.sf.basedb.core.data.RawBioAssayData;
import net.sf.basedb.core.data.RemovableData;
import net.sf.basedb.core.data.ReporterCloneTemplateData;
import net.sf.basedb.core.data.RoleKeyData;
import net.sf.basedb.core.data.RootRawBioAssayData;
import net.sf.basedb.core.data.SchemaVersionData;
import net.sf.basedb.core.data.SessionData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.data.UnitData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.UserDeviceData;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.core.dbengine.TableInfo.ForeignKeyInfo;
import net.sf.basedb.core.hibernate.TableInfoWork;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.Values;

/**
	This class contains static methods used to update items already in the database.
	
	<table>
	<caption>Schema version updates</caption>
	<tr>
		<th>Schema version</th>
		<th>Description</th>
	</tr>
	
	<tr>
		<td>90</td>
		<td>
			The schema as released in BASE 2.17. For information about schema changes before
			this check the BASE 2.17 code.
		</td>
	</tr>
	
	<tr>
		<td>100</td>
		<td>
			The schema as released in BASE 3.0. We don't list all changes here.
		</td>
	</tr>
	<tr>
		<td>101 and 102</td>
		<td>
			Fixes an issue with duplicate {@link FileSetMemberData} entries
			due to a bug in the upgrade program. Make sure that the unique
			constraint that should prevent duplicates exists.
		</td>
	</tr>
	<tr>
		<td>103</td>
		<td>
			Added {@link ReporterCloneTemplateData} and related classes. No
			special update is required.
		</td>
	</tr>
	<tr>
		<td>104</td>
		<td>
			Make {@link ExperimentData}  {@link AnnotatableData}. No
			special update is required.
		</td>
	</tr>
	<tr>
		<td>105</td>
		<td>
			Added {@link ItemSubtypeData#getPushAnnotations()}. All existing subtypes
			are set to 'false'.
		</td>
	</tr>
	<tr>
		<td>106</td>
		<td>
			Added {@link BioPlateData#getEntryDate()} and {@link BioPlateData#getEventDate()}.
			All existing bioplates get a null value.
		</td>
	</tr>
	<tr>
		<td>107</td>
		<td>
			Added {@link BioPlateTypeData#getStorageType()}
			All existing bioplate types get a null value.
		</td>
	</tr>
	<tr>
		<td>108</td>
		<td>
			Added {@link BioPlateData#getSection()}, {@link BioPlateData#getTray()}
			and {@link BioPlateData#getPosition()}
			All existing bioplates get null values.
		</td>
	</tr>
	<tr>
		<td>109 and 110</td>
		<td>
			Added {@link DerivedBioAssayData#getPhysicalBioAssays()} and {@link DerivedBioAssayData#getParents()}
			and removed <code>DerivedBioAssayData.getPhysicalBioAssay()</code> and 
			<code>DerivedBioAssayData.getParent()</code>. Existing derived bioassays are updated in a two-step
			procedure. The first step is to copy existing information to the new tables and the second step is
			to drop columns that are no longer needed.
		</td>
	</tr>
	<tr>
		<td>111</td>
		<td>
			Not an actual database schema change, but we need to remove annotation sets that have been
			created without any annotations by the annotation importer and BASE 1-&gt;2 migration.
		</td>
	</tr>
	<tr>
		<td>112</td>
		<td>
			Not an actual database schema change, but we need to fix the parent_id and parent_type
			columns in the BioMaterials table to really reflect what is in the BioMaterialEventSources2 
			table. For more info see BASE ticket: http://base.thep.lu.se/ticket/1745
		</td>
	</tr>
	<tr>
		<td>113</td>
		<td>
			Made {@link Hardware}, {@link Software}, {@link Tag} and <code>BioMaterialList</code>
			{@link Annotatable} items. No special database update is needed. Only increase 
			the schema version.
		</td>
	</tr>
	<tr>
		<td>114</td>
		<td>
			Added {@link ChangeHistoryDetailData#getOldValue()} and {@link ChangeHistoryDetailData#getNewValue()}.
			An update that changes the log format for annotation changes is needed to make log entries
			using the same format. 
		</td>
	</tr>
	<tr>
		<td>115</td>
		<td>
			Added {@link AnnotationTypeData#getDisableLogOfValues()}. The update should set
			the value for all existing annotation types to false.
		</td>
	</tr>
	
	<tr>
		<td>116</td>
		<td>
			Added {@link JobData#getExternalId()} and made {@link JobData} a {@link ShareableData} item. 
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	
	<tr>
		<td>117</td>
		<td>
			Added {@link FileServerData#getRootPath()} and {@link FileServerData#getSshFingerprint()}.
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>118</td>
		<td>
			Added {@link SchemaVersionData#getAppId()}.
			The update will set the app id for the current entry to 'net.sf.basedb.core'
			and the update the schema version.
		</td>
	</tr>
	<tr>
		<td>119</td>
		<td>
			Added {@link AnnotationTypeData#getDisableInheritance()}.
			The update will set the value to FALSE for all existing annotation types.
		</td>
	</tr>
	<tr>
		<td>120</td>
		<td>
			Added {@link ItemListData} and removed <code>BioMaterialListData</code>.
			The update will convert all existing biomaterial lists to item lists
			and make other changes as are requried to keep things working.
		</td>
	</tr>
	<tr>
		<td>121</td>
		<td>
			No actual schema change, but we need to fix incorrect item lists that
			has got a '0' member due to ticket http://base.thep.lu.se/ticket/1945.
		</td>
	</tr>
	<tr>
		<td>122</td>
		<td>
			Added {@link RootRawBioAssayData}. Existing raw bioassay -- Experiment links
			are moved over to this item.
		</td>
	</tr>
	<tr>
		<td>123</td>
		<td>
			Remove {@link AnnotationSetData}.getInheritedSets(). Existing inherited
			annotations sets will be converted to inherited annotations.
		</td>
	</tr>
	<tr>
		<td>124</td>
		<td>
			Inherited annotations are moved from the "InheritedAnnotations" table
			to the "Annotations" table.
		</td>
	</tr>
	<tr>
		<td>125</td>
		<td>
			Experimental factor annotations for an experiment which are found as either
			primary or inherited annotations on the raw bioassays in the experiment are
			inherited by the {@link RootRawBioAssayData} items. This means that existing
			experiments will continue to have access to the experimental factors as before.
		</td>
	</tr>
	<tr>
		<td>126</td>
		<td>
			Fixes the conversion formula for Fahrenheit. Existing annotation values
			that are re-calculated to the corrent result.
		</td>
	</tr>
	<tr>
		<td>127</td>
		<td>
			Added {@link ParameterValueData#isMasked()}. All existing parameters
			are set to false.
		</td>
	</tr>
	<tr>
		<td>128</td>
		<td>
			Added {@link KitData} and other things related to it.
		</td>
	</tr>
	<tr>
		<td>129 and 130</td>
		<td>
			Added {@link RemovableData#getRemovedBy()} and removed <code>RemovableData.isRemoved()</code>.
			Existing items that has been flagged for removal are updated by setting the <code>removedBy</code>
			property to either the owner of the item or, if the item doesn't have an owner, to the root
			user. In a second step the <code>removed</code> columns are dropped from all tables.
			Note that this must be done in {@link #adjustExistingItems(ProgressReporter, int, String, String)}
			since otherwise the installation will try to create new items in table that still has the
			'removed' column but it will not accept 'null' values and the installation will fail.
		</td>
	</tr>
	<tr>
		<td>131</td>
		<td>
			Not an actual database schema change, but we need to fix issues with old date values
			in the change history table. See http://base.thep.lu.se/ticket/1989 for more
			information.
		</td>
	</tr>
	<tr>
		<td>132</td>
		<td>
			Not an actual database schema change, but we need to fix issues with 
			annotation entries that have no values. See http://base.thep.lu.se/ticket/1992 
			for more information.
		</td>
	</tr>
	<tr>
		<td>133</td>
		<td>
			Added {@link SessionData#getAuthenticationMethod()}. 
			No special database update is needed. Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>134</td>
		<td>
			Made {@link Project} annotatable. No special database update is needed. 
			Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>135</td>
		<td>
			Added support for project specific annotations. Added
			{@link AnnotationTypeData#getProjectAnnotations()}. Existing
			annotation types are set to FALSE. 
			Added {@link AnnotationData#getProjectId()} and {@link AnnotationData#getOverrideId()}.
			Existing annotations set both values to 0.
		</td>
	</tr>
	<tr>
		<td>136</td>
		<td>
			Added {@link JobData#getNode()} and {@link JobData#getItemSubtype()}.
			Job entries which have a server value containing '[...]' are updated
			so that the part within '[...]' is moved to the node column.
		</td>
	</tr>
		<tr>
		<td>137</td>
		<td>
			Added {@link FileServerData#getFingerprintType()}.
			File server entries which have a SSH fingerprint value are updated
			to "MD5".
		</td>
	</tr>
	<tr>
		<td>138</td>
		<td>
			Added {@link NewsData#isSticky()}. The update should set
			the value for all existing news items to false.
		</td>
	</tr>
	<tr>
		<td>139</td>
		<td>
			Added {@link UserDeviceData} and related changes in {@link UserData},
			{@link ClientData} and {@link SessionData} classes.
		</td>
	</tr>
	<tr>
		<td>140</td>
		<td>
			Added {@link SessionData#getLocation()}, {@link SessionData#getLocationLatitude()}
			and {@link SessionData#getLocationLongitude()}. No special database update is needed. 
			Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>141</td>
		<td>
			Added {@link FileServerData#getSshPrivateKey()}, {@link FileServerData#getSshPrivateKeyPassword()}
			and {@link FileServerData#getSshPrivateKeyFormat()}. No special database update is needed. 
			Only increase the schema version.
		</td>
	</tr>
	<tr>
		<td>142</td>
		<td>
			Added the default password login form. It needs to be disabled by default which is
			triggered by the schema version update. No actual database change.
		</td>
	</tr>
	<tr>
		<td>143 and 144</td>
		<td>
			Remvoving support for secondary storage. {@link FileData}.action has been deleted.
			The update works in two step and the first step is to set FileData.location=0 (offline) 
			on all files with FileData.location=2 (secondary) and remove filters on the action column
			The second step is to drop the action column.
		</td>
	</tr>
	<tr>
		<td>145</td>
		<td>
			Added {@link AnnotationTypeData#isIdentifier()}.
			The update will set the value to FALSE for all existing annotation types.
		</td>
	</tr>
	<tr>
		<td>146 and 147</td>
		<td>
			Removed support for spot images:
			<ul>
			<li>SpotImageCreator plug-in was removed. The entry in the database is marked as disabled.
			<li>The "SpotImages" table is dropped from the database.
			</ul>
			
		</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
*/
public final class Update
{
	/**
		Log core events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(Update.class);

	/**
		Update the database by modifying existing items to follow new requirements.
		
		@param progress An object implementing the {@link ProgressReporter}
			interface
		@param rootLogin The root user login
		@param rootPassword The root user password
	 */
	public static synchronized void updateDatabase(ProgressReporter progress, String rootLogin, String rootPassword)
		throws BaseException
	{
		if (progress != null) progress.display(0, "Updating database...");
		org.hibernate.Session session = null;
		try
		{
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			LoginRequest loginRequest = new LoginRequest(rootLogin, rootPassword);
			sc.login(loginRequest);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			int schemaVersion = getSchemaVersionInTransaction(session, SchemaVersionData.BASE_APP_ID);
			final int total_progress_steps = 1 + Install.NEW_SCHEMA_VERSION - schemaVersion;
			final float progress_step = 100F / total_progress_steps;
			float progress_current = progress_step;
			
			// The 101 and 102 updates are only needed when updating from a BASE 3.0.0 installation
			if (schemaVersion < 101)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 101...");
				schemaVersion = updateToSchemaVersion101(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 102)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 102...");
				schemaVersion = updateToSchemaVersion102(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 106)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 106...");
				schemaVersion = setSchemaVersionInTransaction(session, 106);
				progress_current += 4 * progress_step;
			}
			
			if (schemaVersion < 108)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 108...");
				schemaVersion = setSchemaVersionInTransaction(session, 108);
				progress_current += 2 * progress_step;
			}
			
			if (schemaVersion < 109)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 110...");
				schemaVersion = updateToSchemaVersion110(session, schemaVersion);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 111)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 111...");
				schemaVersion = updateToSchemaVersion111(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 112)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 112...");
				schemaVersion = updateToSchemaVersion112(session, progress);
				progress_current += progress_step;
			}

			if (schemaVersion < 113)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 113...");
				// Schemaversion 113 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 113);
				progress_current += progress_step;
			}

			if (schemaVersion < 114)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 114...");
				schemaVersion = updateToSchemaVersion114(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 117)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 117...");
				// Schemaversion 115 - 117 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 117);
				progress_current += 3 * progress_step;
			}
			
			if (schemaVersion < 119)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 119...");
				// Schemaversion 118-119 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 119);
				progress_current += 2 * progress_step;
			}

			if (schemaVersion < 120)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 120...");
				schemaVersion = updateToSchemaVersion120(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 121)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 121...");
				schemaVersion = updateToSchemaVersion121(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 122)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 122...");
				schemaVersion = updateToSchemaVersion122(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 123)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 123...");
				schemaVersion = updateToSchemaVersion123(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 124)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 124...");
				schemaVersion = updateToSchemaVersion124(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 125)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 125...");
				schemaVersion = updateToSchemaVersion125(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 126)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 126...");
				schemaVersion = updateToSchemaVersion126(session);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 130)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 130...");
				// Schemaversion 127-130 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 130);
				progress_current += 4 * progress_step;
			}
			
			if (schemaVersion < 131)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 131...");
				schemaVersion = updateToSchemaVersion131(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 132)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 132...");
				schemaVersion = updateToSchemaVersion132(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 133)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 133...");
				schemaVersion = setSchemaVersionInTransaction(session, 133);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 135)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 135...");
				// Schemaversion 134+135 only updates the version number
				schemaVersion = setSchemaVersionInTransaction(session, 135);
				progress_current += 2 * progress_step;
			}
			
			if (schemaVersion < 136)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 136...");
				schemaVersion = updateToSchemaVersion136(session, progress);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 137)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 137...");
				schemaVersion = setSchemaVersionInTransaction(session, 137);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 138)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 138...");
				schemaVersion = setSchemaVersionInTransaction(session, 138);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 140)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 140...");
				schemaVersion = setSchemaVersionInTransaction(session, 140);
				progress_current += 2 * progress_step;
			}
			
			if (schemaVersion < 141)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 141...");
				schemaVersion = setSchemaVersionInTransaction(session, 141);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 142)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 142...");
				schemaVersion = setSchemaVersionInTransaction(session, 142);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 144)
			{
				// Note! 143 and 144 is a two-step update due to DROP COLUMN
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 144...");
				schemaVersion = updateToSchemaVersion144(session, schemaVersion);
				progress_current += 2 * progress_step;
			}
			
			if (schemaVersion < 145)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 145...");
				schemaVersion = setSchemaVersionInTransaction(session, 145);
				progress_current += progress_step;
			}
			
			if (schemaVersion < 147)
			{
				if (progress != null) progress.display((int)(progress_current), "--Updating schema version: " + schemaVersion + " -> 147...");
				schemaVersion = updateToSchemaVersion147(session, schemaVersion);
				progress_current += progress_step;
			}
		
			sc.logout();
			if (progress != null) progress.display(100, "Database updated successfully.");
		}
		catch (BaseException ex)
		{
			if (progress != null) progress.display(100, "Database update failed: " + ex.getMessage()+"\n");
			throw ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	
	/**
		Verify the root user login and password. This method is called first 
		in the update procedure, which means that tables, etc. are usually
		still from the old version.
	*/
	static void verifyRootLoginAndPassword(String login, String password)
	{
	
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			int schemaVersion = 0;
			try
			{
				org.hibernate.query.Query<Integer> query = HibernateUtil.createQuery(session, 
						"SELECT sv.schemaVersion FROM SchemaVersionData sv", Integer.class);
				schemaVersion = HibernateUtil.loadData(query);
			}
			catch (Throwable t)
			{}
			
			if (schemaVersion > 0)
			{
				org.hibernate.query.Query<Object[]> userQuery = HibernateUtil.createQuery(session, 
					"SELECT usr.id, usr.login FROM UserData usr WHERE usr.systemId='" + User.ROOT + "'", Object[].class);
				Object[] user = userQuery.uniqueResult();
				if (user == null)
				{
					throw new LoginException("User ROOT not found");
				}
				if (!login.equals(user[1]))
				{
					throw new LoginException("User '" + login + "' is not the ROOT user");
				}
				
				int id = (Integer)user[0];
				
				if (schemaVersion > 90)
				{
					// BASE 3
					org.hibernate.query.Query<String> pwdQuery = HibernateUtil.createQuery(session, 
						"SELECT cryptedPassword FROM PasswordData pwd WHERE pwd.id="+id, String.class);
					String cryptedPassword = pwdQuery.uniqueResult();
					if (!User.checkPassword(password, cryptedPassword))
					{
						throw new LoginException("Incorrect password for user '" + login + "'");
					}
				}
				else if (schemaVersion == 90)
				{
					// BASE 2.17
					org.hibernate.query.Query<String> pwdQuery = HibernateUtil.createSqlQuery(session, 
						"SELECT [md5password] FROM [Passwords] WHERE [id]="+id, String.class);
					String md5Password = pwdQuery.uniqueResult();
					if (!MD5.getHashString(password).equals(md5Password))
					{
						throw new LoginException("Incorrect password for user '" + login + "'");
					}
				}
				// We let all other conditions pass, since they will 
				// usually generate an error later
			}
		}
		finally
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (session != null) HibernateUtil.close(session);
		}
	}
	
	/**
		Get the current schema version number as it is stored in the database.
		@return The schema version number, or 0 if no schem version is found
	*/
	static int getSchemaVersionInTransaction(org.hibernate.Session session, String appId)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		int schemaVersion = 0;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			schemaVersion = getSchemaVersion(session, appId);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("getSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	static int getSchemaVersion(org.hibernate.Session session, String appId)
		throws BaseException
	{
		org.hibernate.query.Query<SchemaVersionData> query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SCHEMA_VERSION_FOR_APP", SchemaVersionData.class);
		query.setParameter("appId", appId, TypeWrapper.H_STRING);
		SchemaVersionData data = HibernateUtil.loadData(query);
		int schemaVersion = data == null ? 0 : data.getSchemaVersion();
		return schemaVersion;
	}

	/**
		Update the database with a new schema version number. This method should
		be called from an open transaction.
		
		@param schemaVersion The new schema version number
	 */
	static void setSchemaVersion(org.hibernate.Session session, String appId, int schemaVersion, int build)
		throws BaseException
	{
		org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(session, "SET_SCHEMA_VERSION_FOR_APP");
		query.setParameter("schemaVersion", schemaVersion, TypeWrapper.H_INTEGER);
		query.setParameter("build", build, TypeWrapper.H_INTEGER);
		query.setParameter("appId", appId, TypeWrapper.H_STRING);
		int modified = HibernateUtil.executeUpdate(query);
		if (modified == 0)
		{
			// Insert
			SchemaVersionData sv = new SchemaVersionData();
			sv.setBuild(build);
			sv.setSchemaVersion(schemaVersion);
			sv.setAppId(appId);
			HibernateUtil.saveData(session, sv);
		}
	}
	
	private static void setSchemaVersion(org.hibernate.Session session, int schemaVersion)
		throws BaseException
	{
		setSchemaVersion(session, SchemaVersionData.BASE_APP_ID, schemaVersion, Application.getBuild());
	}
	
	/**
		Update the schema version version number using a new transaction.
		@param schemaVersion The new schema version number
	 */
	private static int setSchemaVersionInTransaction(org.hibernate.Session session, int schemaVersion)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			setSchemaVersion(session, schemaVersion);
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("setSchemaVersion: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Adjust the existing items in the database to be compatible with the latest mappings.
	 	No update to the schema version should be done here. The code must also  
	 	consider the case that the update failes at a later stage and that the same
	 	update is executed again.
	 	
	  	@param progress An object implementing the {@link ProgressReporter}
	 		interface
	  	@param rootLogin The root user login
	  	@param rootPassword The root user password
	*/
	public static synchronized void adjustExistingItems(ProgressReporter progress, int schemaVersion, String rootLogin, String rootPassword)
		throws BaseException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.Session session = null;
		try
		{
			Config.setProperty("extensions.disabled", "true");
			Application.start(false, false, false);

			if (progress != null) progress.display(50, "Checking existing items...");
			// Test root user account
			SessionControl sc = Application.newSessionControl(null, null, null);
			LoginRequest loginRequest = new LoginRequest(rootLogin, rootPassword);
			sc.login(loginRequest);
			if (sc.getLoggedInUserId() != SystemItems.getId(User.ROOT))
			{
				throw new PermissionDeniedException("User '" + rootLogin + "' is not the root account.");
			}
			
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			// List of schema changes that need to be executed
			// after the main transaction has ended.
			List<SchemaChange> schemaChanges = new ArrayList<SchemaChange>();
			
			if (schemaVersion < 115)
			{
				// Set destroyed=false on all bioplates
				org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedQuery(session, "SET_DISABLE_LOG_FOR_ANNOTATION_TYPES");
				/*
					UPDATE AnnotationTypeData at
					SET at.disableLogOfValues = false
					WHERE at.disableLogOfValues IS NULL
				*/
				query.executeUpdate();
			}
			
			if (schemaVersion < 118)
			{
				// Set appId=net.sf.basedb.core on the only existing schema version entry
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE SchemaVersionData sv " + 
					"SET sv.appId = :appId WHERE sv.appId IS NULL");
				query.setParameter("appId", SchemaVersionData.BASE_APP_ID, TypeWrapper.H_STRING);
				query.executeUpdate();
			}
			
			if (schemaVersion < 119)
			{
				// Set disableInheritance=false on all annotation types
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE AnnotationTypeData at " +
					"SET at.disableInheritance = false " +
					"WHERE at.disableInheritance IS NULL");
				query.executeUpdate();
			}
			
			if (schemaVersion < 124)
			{
				// Set source=0 on all existing "Annotations"
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE AnnotationData a " +
					"SET a.source = 0 " +
					"WHERE a.source IS NULL");
				query.executeUpdate();
			}
			
			if (schemaVersion < 130)
			{
				// Load tables names and set a flag if it has an 'owner' column or not
				Map<String, Boolean> tables = new LinkedHashMap<String, Boolean>();
				for (Item item : Metadata.getRemovableItems())
				{
					String tableName = HibernateUtil.getClassMapping(item.getDataClass().getName()).getRootTable().getName();
					// Multiple items may be mapped to the same table (eg. BioMaterials) -- Only add each table once
					if (!tables.containsKey(tableName))
					{
						// Check if 'removed' column exists
						// It is not present in new tables (eg. Kits)
						TableInfo tblInfo = getTableInfo(session, tableName);
						if (tblInfo.findColumn("removed") != null)
						{
							tables.put(tableName, tblInfo.findColumn("owner") != null);
						}
					}
				}
				
				// UPDATE all tables and set 'removed_by' to 'owner' or 'root'
				int rootId = SystemItems.getId(User.ROOT);
				for (Map.Entry<String, Boolean> entry : tables.entrySet())
				{
					String tableName = entry.getKey();
					org.hibernate.query.Query<?> query;
					if (entry.getValue())
					{
						String sql = "UPDATE [" + tableName + "] SET [removed_by]=[owner], [removed]=false WHERE [removed]=true";
						query = HibernateUtil.createSqlQuery(session, sql);
					}
					else
					{
						String sql = "UPDATE [" + tableName + "] SET [removed_by]=:root, [removed]=false WHERE [removed]=true";
						query = HibernateUtil.createSqlQuery(session, sql);
						query.setParameter("root", rootId, TypeWrapper.H_INTEGER);
					}
					int updated = query.executeUpdate();
					// Schedule removal of the 'removed' column
					schemaChanges.add(new DropColumnChange(tableName, "removed"));
				}
			}
			
			if (schemaVersion < 135)
			{
				// Set projectAnnotations=false on all annotation types
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE AnnotationTypeData at " +
					"SET at.projectAnnotations = false " +
					"WHERE at.projectAnnotations IS NULL");
				query.executeUpdate();
				
				// Set project_id=0 and override_id=0 on all annotations
				query = HibernateUtil.createQuery(session, 
					"UPDATE AnnotationData a " +
					"SET a.projectId = 0, a.overrideId = 0 " +
					"WHERE a.projectId IS NULL OR a.overrideId IS NULL");
				query.executeUpdate();
			}
			
			if (schemaVersion < 137)
			{
				// Set fingerprintType=MD5 on all file servers with a fingerprint
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE FileServerData fs " +
					"SET fs.fingerprintType = 'MD5' " +
					"WHERE NOT fs.sshFingerprint IS NULL");
				query.executeUpdate();
			}
			
			if (schemaVersion < 138)
			{
				// Set sticky=false on all news
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE NewsData n " + 
					"SET n.sticky = false " +
					"WHERE n.sticky IS NULL");
				query.executeUpdate();
			}

			if (schemaVersion < 139)
			{
				// Set supportsDeviceVerification = false on all clients (except on the web client)
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE ClientData c " + 
					"SET c.supportsDeviceVerification = false " +
					"WHERE c.supportsDeviceVerification IS NULL");
				query.executeUpdate();

				query = HibernateUtil.createQuery(session, 
					"UPDATE ClientData c " + 
					"SET c.supportsDeviceVerification = true " +
					"WHERE c.externalId = 'net.sf.basedb.clients.web'");
				query.executeUpdate();
			}
			
			if (schemaVersion < 145)
			{
				// Set identifier=false on all annotation types
				org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
					"UPDATE AnnotationTypeData at " +
					"SET at.identifier = false " +
					"WHERE at.identifier IS NULL");
				query.executeUpdate();
			}

			
			//	Commit the changes
			HibernateUtil.commit(tx);
			
			for (SchemaChange change : schemaChanges)
			{
				change.performChange(session);
			}
			log.info("adjustExistingItems: OK");	
			
		}
		catch (BaseException | SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			if (progress != null) progress.display(100, "The adjustment of the existing items failed: " + ex.getMessage()+"\n");
			log.info("adjustExistingItems: FAILED");
			throw ex instanceof SQLException ? new BaseException((SQLException)ex) : (BaseException)ex;
		}
		finally
		{
			if (session != null) HibernateUtil.close(session);
			Application.stop();
		}
	}
	

	/**
	 	Remove duplicate file set members.
	 	
		@return The new schema version (=101)
	*/
	private static int updateToSchemaVersion101(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 101;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);

			org.hibernate.query.Query<FileSetMemberData> query = HibernateUtil.createQuery(session, 
				"SELECT fsm FROM FileSetMemberData fsm", FileSetMemberData.class);
			List<FileSetMemberData> members = HibernateUtil.loadList(query, null);
			Set<String> allMembers = new HashSet<String>();
			for (FileSetMemberData member : members)
			{
				int fileSetId = member.getFileSet().getId();
				int fileTypeId = member.getDataFileType().getId();
				int fileId = member.getFile().getId();
				String unique = fileSetId + "." + fileTypeId + "." + fileId;
				if (!allMembers.add(unique))
				{
					// This combination has been seen before
					session.delete(member);
				}
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion101: OK");
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion101: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Add unique index on FileSetMembers(fileset_id,datafiletype_id,file_id)
	 	
		@return The new schema version (=102)
	*/
	private static int updateToSchemaVersion102(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 102;
		try
		{
	
			// Try to drop just to make sure that we don't create duplicated indexes
			dropIndex(session, true, "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			// Create the index
			createIndex(session, true, "FileSetMembers_uniquefile", "FileSetMembers", "fileset_id", "datafiletype_id", "file_id");
			
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);
	
			log.info("updateToSchemaVersion102: OK");
		}
		catch (SQLException ex)
		{
			log.error("updateToSchemaVersion102: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (BaseException ex)
		{
			log.error("updateToSchemaVersion102: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Move derived bioassay parent items to different tables. Get rid of
	 	old columns in the DerivedBioAssays table.
	 	
		@return The new schema version (=110)
	*/
	private static int updateToSchemaVersion110(org.hibernate.Session session, int currentSchemaVersion)
		throws BaseException
	{
		final int schemaVersion = 110;
		try
		{
			if (currentSchemaVersion < 109)
			{
				// First step is to copy existing information to the new tables
				org.hibernate.Transaction tx = null;
				try
				{
					tx = HibernateUtil.newTransaction(session);
					
					// Copy parent physical and derived bioassay information to the
					// new tables: ParentPhysicalBioAssays and ParentDerivedBioAssays
					org.hibernate.query.Query<?> query = HibernateUtil.getPredefinedSQLQuery(session, "COPY_PARENT_PHYSICAL_BIOASSAYS");
					/*
						INSERT INTO ParentPhysicalBioAssays (physicalbioassay_id, derivedbioassay_id)
						SELECT bioassay_id, id FROM DerivedBioAssays
					*/
					query.executeUpdate();
					query = HibernateUtil.getPredefinedSQLQuery(session, "COPY_PARENT_DERIVED_BIOASSAYS");
					/*
						INSERT INTO ParentDerivedBioAssays (parentbioassay_id, derivedbioassay_id)
						SELECT parent_id, id FROM DerivedBioAssays WHERE NOT parent_id IS NULL
					*/
					query.executeUpdate();
					
					// Set is_root flag on derived bioassays
					query = HibernateUtil.getPredefinedSQLQuery(session, "SET_ISROOT_ON_DERIVED_BIOASSAYS");
					/*
						UPDATE DerivedBioAssays SET is_root = (parent_id IS NULL)
					*/
					query.executeUpdate();
					
					// Remove table filters, etc. that use the old properties
					cleanContextFromProperty(session, Item.DERIVEDBIOASSAY, "physicalBioAssay", null, true);
					cleanContextFromProperty(session, Item.DERIVEDBIOASSAY, "parent", null, true);
					
					setSchemaVersion(session, 109);
	
					// Commit the changes
					HibernateUtil.commit(tx);
				}
				finally
				{
					if (tx != null && tx.getStatus() != TransactionStatus.COMMITTED) HibernateUtil.rollback(tx);
				}
			}
			
			// Second step is to drop the old columns that are no longer used in BASE 3.2
			dropColumn(session, "DerivedBioAssays", "parent_id", null);
			dropColumn(session, "DerivedBioAssays", "bioassay_id", null);
				
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);
	
			log.info("updateToSchemaVersion110: OK");
		}
		catch (SQLException ex)
		{
			log.error("updateToSchemaVersion110: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (RuntimeException ex)
		{
			log.error("updateToSchemaVersion110: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Remove annotation sets that have no annotation values and no inherited annotations.
		@return The new schema version (=111)
	*/
	private static int updateToSchemaVersion111(org.hibernate.Session session, final ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 111;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			AnnotationSet.deleteEmptyAnnotationSets(session,
					new ProgressReporter() 
					{
						@Override
						public void display(int percent, String message) 
						{
							progress.append("\n\t  "+message);
						}
						
						@Override
						public void append(String message) 
						{
							progress.append(message);
						}
					}
				);
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion111: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion111: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Ensure that parent_id and parent_type in BioMaterials table is correct
	 	when compared to BioMaterialEventSources2 table.
		@return The new schema version (=112)
	*/
	private static int updateToSchemaVersion112(org.hibernate.Session session, final ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 112;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Get all BioMaterial that has null in parent_id and/or parent_type
			String hql = "select bm.creationEvent from MeasuredBioMaterialData bm" +
					" where bm.parent is null or bm.parentType is null";
			
			org.hibernate.query.Query<BioMaterialEventData> query = HibernateUtil.createQuery(session, hql, BioMaterialEventData.class);
			List<BioMaterialEventData> events = HibernateUtil.loadList(query, null);
			int i = 0;
			for (BioMaterialEventData event : events)
			{
				BioMaterialEvent.ensureProductParentCorrectness(event);
				i++;
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion112: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion112: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	
	/**
	 	Ensure that old log entries for annotation changes use the same text format
	 	that is used after this update.
		@return The new schema version (=114)
	*/
	private static int updateToSchemaVersion114(org.hibernate.Session session, final ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 114;
		org.hibernate.Transaction tx = null;
		org.hibernate.StatelessSession statelessSession = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			statelessSession = HibernateUtil.newStatelessSession(session);
			// Get all ChangeHistoryDetail entries that are related to annotation changes
			// (eg those that start with "Annotation create|updated|deleted:"
			String hql = "select ch from ChangeHistoryDetailData ch" +
					" where ch.changeInfo is not null";
			String count = "select count(*) from ChangeHistoryDetailData ch" +
					" where ch.changeInfo is not null";
			
			String fixSql = "update [ChangeHistoryDetails] set [change_info] = :ci, [change_type] = :ct where [id] = :id";
			org.hibernate.query.Query<?> fixQuery = HibernateUtil.createSqlQuery(statelessSession, fixSql);

			org.hibernate.query.Query<Long> countQuery = HibernateUtil.createQuery(statelessSession, count, Long.class);
			long numItems = HibernateUtil.loadData(countQuery);
			
			org.hibernate.query.Query<ChangeHistoryDetailData> query = HibernateUtil.createQuery(statelessSession, hql, ChangeHistoryDetailData.class);
			Iterator<ChangeHistoryDetailData> changes = HibernateUtil.loadIterator(query, null);
			
			int i = 0;
			int delta = (int)numItems / 50;
			while (changes.hasNext())
			{
				ChangeHistoryDetailData change = changes.next();
				String info = change.getChangeInfo();
				HibernateUtil.evictData(session, change);
				boolean fixIt = true;
				if (info.startsWith("Annotation created: "))
				{
					// Annotation created: <name> --> ADD + Annotation[<name>]
					fixQuery.setParameter("ci", "Annotation[" + info.substring("Annotation created: ".length()) + "]", TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.ADD.getValue(), TypeWrapper.H_INTEGER);
				}
				else if (info.startsWith("Annotation updated: "))
				{
					// Annotation updated: <name> --> UPDATE + Annotation[<name>]
					fixQuery.setParameter("ci", "Annotation[" + info.substring("Annotation updated: ".length()) + "]", TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.UPDATE.getValue(), TypeWrapper.H_INTEGER);
				}
				else if (info.startsWith("Annotation deleted: "))
				{
					// Annotation deleted: <name> --> REMOVE + Annotation[<name>]
					fixQuery.setParameter("ci", "Annotation[" + info.substring("Annotation deleted: ".length()) + "]", TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.REMOVE.getValue(), TypeWrapper.H_INTEGER);
				}
				else if (info.startsWith("File created: "))
				{
					// File created: <name> --> ADD + File[<name>]
					fixQuery.setParameter("ci", "File[" + info.substring("File created: ".length()) + "]", TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.ADD.getValue(), TypeWrapper.H_INTEGER);
				}
				else if (info.startsWith("File updated: "))
				{
					// File updated: <name> --> UPDATE + File[<name>]
					fixQuery.setParameter("ci", "Annotation[" + info.substring("File updated: ".length()) + "]", TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.UPDATE.getValue(), TypeWrapper.H_INTEGER);
				}
				else if (info.startsWith("File deleted: "))
				{
					// File deleted: <name> --> REMOVE + File[<name>]
					fixQuery.setParameter("ci", "File[" + info.substring("File deleted: ".length()) + "]", TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.REMOVE.getValue(), TypeWrapper.H_INTEGER);
				}
				else if (info.startsWith("Updated: "))
				{
					// Updated: <list-of-properties> --> UPDATE + property (one entry for each in the list)
					String[] properties = info.substring("Updated: ".length()).split(", ");
					if (properties.length > 1)
					{
						for (int p = 1; p < properties.length; p++)
						{
							// Create new entries for the extra properties
							ChangeHistoryDetailData clone = new ChangeHistoryDetailData();
							clone.setChangeHistory(change.getChangeHistory());
							clone.setChangeInfo(properties[p]);
							clone.setChangeType(ChangeType.UPDATE.getValue());
							clone.setItemId(change.getItemId());
							clone.setItemType(change.getItemType());
							HibernateUtil.saveData(session, clone);
						}
						session.flush();
						session.clear();
					}
					fixQuery.setParameter("ci", properties[0], TypeWrapper.H_STRING);
					fixQuery.setParameter("ct", ChangeType.UPDATE.getValue(), TypeWrapper.H_INTEGER);
				}
				else
				{
					fixIt = false;
				}
				
				if (fixIt)
				{
					fixQuery.setParameter("id", change.getId(), TypeWrapper.H_INTEGER);
					fixQuery.executeUpdate();
				}
				
				i++;
				if (progress != null && i % delta == 0) 
				{
					progress.append(".");
				}
				
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion114: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion114: FAILED", ex);
			throw ex;
		}
		finally
		{
			HibernateUtil.close(statelessSession);
		}
		return schemaVersion;
	}
	
	/**
	 	Transfer existing biomaterial lists to item lists.
		@return The new schema version (=120)
	*/
	private static int updateToSchemaVersion120(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 120;
		org.hibernate.Transaction tx = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			String[] itemListFields = {
				"[id]", "[version]", "[external_id]", "[member_type]", "[size]", 
				"[annotationset_id]", "[name]", "[description]", "[removed]",
				"[itemkey_id]", "[projectkey_id]", "[owner]"
			};
			
			String combinedListFields = Values.getString(Arrays.asList(itemListFields), ", ", true);
			
			String transferSql = "insert into [ItemLists] (" + combinedListFields + 
					", [disable_syncfilters], [disable_manual]) " +
					" select " + combinedListFields + ", false, false from [BioMaterialLists]";
			
			org.hibernate.query.Query<?> query = HibernateUtil.createSqlQuery(session, transferSql);
			int numLists = query.executeUpdate();
		
			transferSql = "insert into [ItemListMembers] ([list_id], [item_id]) " +
					"select [list_id], [biomaterial_id] from [BioMaterialListMembers]";
			query = HibernateUtil.createSqlQuery(session, transferSql);
			int numMembers = query.executeUpdate();
			
			dropTable(session, "BioMaterialListMembers");
			dropTable(session, "BioMaterialLists");
			
			// Fix property filters so they use the new property name (§itemLists)
			String propertyFilterSql = "update [PropertyFilters] set [property] = :newValue where [property] = :oldValue";
			query = HibernateUtil.createSqlQuery(session, propertyFilterSql);
			// @bioMaterialLists --> §itemLists
			query.setParameter("oldValue", "@bioMaterialLists", TypeWrapper.H_STRING);
			query.setParameter("newValue", "§itemLists", TypeWrapper.H_STRING);
			int numFilters = query.executeUpdate();

			// @bioMaterial.bioMaterialLists --> $mbm.§itemLists
			query.setParameter("oldValue", "@bioMaterial.bioMaterialLists", TypeWrapper.H_STRING);
			query.setParameter("newValue", "$mbm.§itemLists", TypeWrapper.H_STRING);
			numFilters += query.executeUpdate();

			// Update the schema version number
			setSchemaVersion(session, schemaVersion);

			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion120: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion120: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Remove all '0' members from item lists and decrease the size of
	 	the list with 1 where it is found.
		@return The new schema version (=121)
	*/
	private static int updateToSchemaVersion121(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 121;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Get all ItemList that has a '0' member
			String hql = "select lst from ItemListData lst" +
				" where 0 in elements(lst.members)";
			
			org.hibernate.query.Query<ItemListData> query = HibernateUtil.createQuery(session, hql, ItemListData.class);
			List<ItemListData> lists = HibernateUtil.loadList(query, null);
			for (ItemListData list : lists)
			{
				list.getMembers().remove(0);
				list.setSize(list.getMembers().size());
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
	
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion121: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion121: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Move all entries from the "ExperimentRawBioAssays" table to "RootRawBioAssays".
		@return The new schema version (=122)
	*/
	private static int updateToSchemaVersion122(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 122;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Get all entries in the "ExperimentRawBioAssays" table
			String sql = "select [er].[experiment_id], [er].[rawbioassay_id], [rba].[name]"+
				" from [ExperimentRawBioAssays] [er]"+
				" inner join [RawBioAssays] [rba] on [rba].[id] = [er].[rawbioassay_id]"; 
			
			org.hibernate.query.Query<Object[]> query = HibernateUtil.createSqlQuery(session, sql, Object[].class);
			List<Object[]> list = HibernateUtil.loadList(query, null);
			for (Object[] entry : list)
			{
				int experiment_id = (Integer)entry[0];
				int rawbioassay_id = (Integer)entry[1];
				String name = (String)entry[2];
				
				// Move over to the new table "RootRawBioAssays"
				RootRawBioAssayData rootRba = new RootRawBioAssayData();
				rootRba.setName(name);
				rootRba.setExperiment(HibernateUtil.loadData(session, ExperimentData.class, experiment_id));
				rootRba.setRawBioAssay(HibernateUtil.loadData(session, RawBioAssayData.class, rawbioassay_id));
				session.save(rootRba);
			}
			
			// Drop the old table
			dropTable(session, "ExperimentRawBioAssays");

			// Update the schema version number
			setSchemaVersion(session, schemaVersion);	
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion122: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion122: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Convert all entries from the "InheritedAnnotationSets" table 
	 	to entries in the "InheritedAnnotations" table.
		@return The new schema version (=123)
	*/
	private static int updateToSchemaVersion123(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 123;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Copy entries in the "IneritedAnnotationSets" table to "InheritedAnnotations"
			// The query is complicated by the fact that we need to avoid duplicates
			// and skip annotations types with inheritance disabled
			String sql = "insert into [InheritedAnnotations] ([annotationset_id], [annotation_id])"+ 
				" select [ia].[annotationset_id], [a].[id] from [InheritedAnnotationSets] [ia]"+
				" inner join [Annotations] [a] on [ia].[inherited_id]=[a].[annotationset_id]"+
				" inner join [AnnotationTypes] [at] on [a].[annotationtype_id] = [at].[id]"+
				" where [at].[disable_inherit] = false"+
				" and not exists ("+
				"   select 1 from [InheritedAnnotations] [i]"+
				"   where [i].[annotationset_id]=[ia].[annotationset_id] and [i].[annotation_id]=[a].[id])";
			
			org.hibernate.query.Query<?> query = HibernateUtil.createSqlQuery(session, sql);
			HibernateUtil.executeUpdate(query);
		
			// Drop the old table
			dropTable(session, "InheritedAnnotationSets");
	
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);	
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion123: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion123: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Inherited annotations in the "InheritedAnnotations" table must be
	 	converted to entries in the "Annotations" table with source=1 (INHERITED)
		@return The new schema version (=124)
	*/
	private static int updateToSchemaVersion124(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 124;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Drop NOT NULL on Annotations.value_id
			dropNotNullConstraint(session, "Annotations", "value_id", "int");

			// Drop UNIQUE constraint on Annotations.annotationset_id/annotationtype_id
			dropIndex(session, true, "Annotations", "annotationset_id", "annotationtype_id");
			
			// Load inherited annotations and their annnotation type
			String sql = "select [i].[annotation_id], [i].[annotationset_id], [a].[annotationtype_id]" +
				" from [InheritedAnnotations] [i]" +
				" inner join [Annotations] [a] on [a].[id] = [i].[annotation_id]";
			
			org.hibernate.query.Query<Object[]> query = HibernateUtil.createSqlQuery(session, sql, Object[].class);
			List<Object[]> list = HibernateUtil.loadList(query, null);
			
			for (Object[] entry : list)
			{
				int annotation_id = (Integer)entry[0];
				int annotationset_id = (Integer)entry[1];
				int annotationtype_id = (Integer)entry[2];
				AnnotationData a = HibernateUtil.loadData(session, AnnotationData.class, annotation_id);
				AnnotationSetData as = HibernateUtil.loadData(session, AnnotationSetData.class, annotationset_id);
				AnnotationTypeData at = HibernateUtil.loadData(session, AnnotationTypeData.class, annotationtype_id);
				
				AnnotationData inherited = new AnnotationData();
				inherited.setAnnotationSet(as);
				inherited.setInheritedFrom(a);
				inherited.setAnnotationType(at);
				inherited.setSource(Annotation.Source.INHERITED.ordinal());
				HibernateUtil.saveData(session, inherited);
			}
			
			// Drop the old table
			dropTable(session, "InheritedAnnotations");
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);	
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion124: OK");
		}
		catch (SQLException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion124: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion124: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	
	/**
	 	Experimental factors in existing experiments are inherited to
	 	the {@link RootRawBioAssayData} items from the {@link RawBioAssayData}
	 	parents.
		@return The new schema version (=125)
	*/
	private static int updateToSchemaVersion125(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 125;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Experimental factors that are found as annotations 
			// on raw bioassays in the experiment. The query will load the
			// id of the root raw bioassay and the annotation
			String sql = "select [root].[id] as rid, [a].[id] as aid" +
				" from [ExperimentalFactors] [xf]" +
				" inner join [RootRawBioAssays] [root] on [root].[experiment_id]=[xf].[experiment_id]" +
				" inner join [RawBioAssays] [rba] on [rba].[id]=[root].[rawbioassay_id]"+
				" inner join [Annotations] [a] on [a].[annotationset_id]=[rba].[annotationset_id]"+
				" and [xf].[annotationtype_id]=[a].[annotationtype_id]";
			
			org.hibernate.query.Query<Object[]> query = HibernateUtil.createSqlQuery(session, sql, Object[].class);
			List<Object[]> list = new ArrayList<Object[]>(HibernateUtil.loadList(query, null));
			
			for (Object[] entry : list)
			{
				int rootrawbioassay_id = (Integer)entry[0];
				int annotation_id = (Integer)entry[1];
				RootRawBioAssayData root = HibernateUtil.loadData(session, RootRawBioAssayData.class, rootrawbioassay_id);
				AnnotationData a = HibernateUtil.loadData(session, AnnotationData.class, annotation_id);
				
				AnnotationSetData as = root.getAnnotationSet();
				if (as == null)
				{
					as = new AnnotationSetData();
					root.setAnnotationSet(as);
					as.setItemId(root.getId());
					as.setItemType(Item.ROOTRAWBIOASSAY.getValue());
					session.save(as);
				}
				AnnotationData inherited = new AnnotationData();
				inherited.setAnnotationSet(as);
				inherited.setAnnotationType(a.getAnnotationType());
				if (a.getInheritedFrom() != null)
				{
					inherited.setInheritedFrom(a.getInheritedFrom());
				}
				else
				{
					inherited.setInheritedFrom(a);
				}
				inherited.setSource(Annotation.Source.INHERITED.ordinal());
				session.save(inherited);
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);	
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion125: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion125: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Fix the conversion formula for Fahrenheit.
		@return The new schema version (=126)
	*/
	private static int updateToSchemaVersion126(org.hibernate.Session session)
		throws BaseException
	{
		final int schemaVersion = 126;
		org.hibernate.Transaction tx = null;
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			QuantityData temperature = HibernateUtil.loadData(session, QuantityData.class, SystemItems.getId(Quantity.TEMPERATURE));
			
			org.hibernate.query.Query<UnitData> query = HibernateUtil.getPredefinedQuery(session, 
					"GET_UNIT_WITH_NAME_FOR_QUANTITY", UnitData.class);
			query.setParameter("name", "Fahrenheit", TypeWrapper.H_STRING);
			query.setParameter("quantity", temperature.getId(), TypeWrapper.H_INTEGER);
			UnitData fahrenheit = HibernateUtil.loadData(query);
			
			Unit.changeReferenceFactorAndOffset(null, session, fahrenheit, fahrenheit.getReferenceFactor(), 255.3722);
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);	
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion126: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion126: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	/**
	 	Fix the old_values for date annotations in the change history.
	 	The fix will scan for all log entries related to DATE annotation
	 	types. By ordering the changes by timestamp it should be possible
	 	to set the old_value of UPDATE entries to the new_value of the previous
	 	entry for the same item and annotation.
		@return The new schema version (=131)
	*/
	private static int updateToSchemaVersion131(org.hibernate.Session session, ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 131;
		org.hibernate.Transaction tx = null;
		org.hibernate.StatelessSession statelessSession = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			statelessSession = HibernateUtil.newStatelessSession(session);
			
			// Load all DATE annotation types
			org.hibernate.query.Query<String> nameQuery = HibernateUtil.createQuery(statelessSession, 
				"SELECT at.name FROM AnnotationTypeData at WHERE at.valueType = " + Type.DATE.getValue(), String.class);
			List<String> names = HibernateUtil.loadList(nameQuery, null);
			
			// Load the entries from ChangeHistoryDetails for each annotation type ordered by entry time
			org.hibernate.query.Query<ChangeHistoryDetailData> query = HibernateUtil.createQuery(statelessSession, 
				"SELECT chd FROM ChangeHistoryDetailData chd "+
				"INNER JOIN chd.changeHistory ch "+
				"WHERE chd.changeInfo = :info "+
				"ORDER BY ch.time", ChangeHistoryDetailData.class);
			
			// This query is used to update the 'old_value' for a single change history entry
			org.hibernate.query.Query<?> fixQuery = HibernateUtil.createSqlQuery(statelessSession,
				"update [ChangeHistoryDetails] "+
				"set [old_value] = :value "+
				"where [id] = :id");
			
			Map<String, ChangeHistoryDetailData> oldEntries = new HashMap<String, ChangeHistoryDetailData>();
			for (String name : names)
			{
				if (progress != null) progress.append(".");
				// Entries related to annotations have the format: Annotation[name]
				query.setParameter("info", "Annotation["+name+"]", TypeWrapper.H_STRING);
				
				List<ChangeHistoryDetailData> changes = HibernateUtil.loadList(query, null);
				for (ChangeHistoryDetailData change : changes)
				{
					String key = change.getItemType()+":"+change.getItemId();
					ChangeType changeType = ChangeType.fromValue(change.getChangeType());
					
					// This is an UPDATE entry with no 'old_value'...
					// ...we may want to fix this if we have a previous value
					if (changeType == ChangeType.UPDATE && change.getOldValue() == null)
					{
						ChangeHistoryDetailData old = oldEntries.get(key);
						if (old != null)
						{
							// Yes.. update it!
							fixQuery.setParameter("value", old.getNewValue(), TypeWrapper.H_STRING);
							fixQuery.setParameter("id", change.getId(), TypeWrapper.H_INTEGER);
							fixQuery.executeUpdate();
						}
					}
					
					// Store the current entry if it may be of interest to a later entry
					if (change.getNewValue() != null)
					{
						oldEntries.put(key, change);
					}
					else
					{
						oldEntries.remove(key);
					}
				}
				oldEntries.clear();
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion131: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion131: FAILED", ex);
			throw ex;
		}
		finally
		{
			HibernateUtil.close(statelessSession);
		}
		return schemaVersion;
	}

	/**
	 	Delete annotation entries that have no values.
		@return The new schema version (=132)
	*/
	private static int updateToSchemaVersion132(org.hibernate.Session session, ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 132;
		org.hibernate.Transaction tx = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			
			// Queries for deleting from Annotations+ParameterValues
			// that doesn't have any corresponding entry in one of the data values tables
			org.hibernate.query.Query<?> deleteAnnotations = HibernateUtil.createSqlQuery(session, 
				"delete from [Annotations] [a] where [a].[value_id] in (:ids)");
			org.hibernate.query.Query<?> deleteInherited = HibernateUtil.createSqlQuery(session, 
				"delete from [Annotations] [a] where [a].[inherited_id] in ("+
				"select [b].[id] from [Annotations] [b] where [b].[value_id] in (:ids))");
			org.hibernate.query.Query<?> deleteValues = HibernateUtil.createSqlQuery(session, 
				"delete from [ParameterValues] [p] where [p].[id] in (:ids)");
			
			for (Type t : Type.values())
			{
				// Load "value_id" for all annotation entries that has no values
				// Note that Annotations.value_id is null for inherited annotations
				// so we must LEFT JOIN the data value table
				org.hibernate.query.Query<Integer> query = HibernateUtil.createSqlQuery(session, 
					"select [a].[value_id] from [Annotations] [a] " +
					"inner join [AnnotationTypes] [at] on [at].[id]=[a].[annotationtype_id] "+
					"left join [" + t.getTableName() + "] [v] on [v].[id]=[a].[value_id] "+
					"where [at].[value_type] = "+t.getValue()+
					" and [v].[id] is null and not [a].[value_id] is null", Integer.class);
				
				List<Integer> valueIds = HibernateUtil.loadList(query, null);
				//System.out.println("Deleting["+t.name()+"]: " + valueIds);
				if (valueIds.size() > 0)
				{
					deleteInherited.setParameterList("ids", valueIds, Type.INT.getTypeWrapper().getHibernateType());
					deleteAnnotations.setParameterList("ids", valueIds, Type.INT.getTypeWrapper().getHibernateType());
					deleteValues.setParameterList("ids", valueIds, Type.INT.getTypeWrapper().getHibernateType() );
					int inherited = deleteInherited.executeUpdate();
					int annotations = deleteAnnotations.executeUpdate();
					int values = deleteValues.executeUpdate();
					//System.out.println(t+": "+ inherited + " inherited + " + annotations+" annotations + "+values + " values");
				}
			}

			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion132: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion132: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	Update the Jobs table by splitting the server value into a server+node value
	 	for all entries that contain '[..]'. That part withing the brackets are removed
	 	from the server column and moved to the node column.
		@return The new schema version (=136)
	*/
	private static int updateToSchemaVersion136(org.hibernate.Session session, ProgressReporter progress)
		throws BaseException
	{
		final int schemaVersion = 136;
		org.hibernate.Transaction tx = null;
		org.hibernate.StatelessSession statelessSession = null;
		
		try
		{
			tx = HibernateUtil.newTransaction(session);
			statelessSession = HibernateUtil.newStatelessSession(session);
			
			// Load the entries from ChangeHistoryDetails for each annotation type ordered by entry time
			org.hibernate.query.Query<JobData> query = HibernateUtil.createQuery(statelessSession, 
				"SELECT j FROM JobData j ", JobData.class);
			
			// This query is used to update the 'server' and 'node' for a single job entry
			org.hibernate.query.Query<?> fixQuery = HibernateUtil.createSqlQuery(statelessSession,
				"update [Jobs] "+
				"set [server] = :server, [node] = :node "+
				"where [id] = :id");
			
			Iterator<JobData> it = HibernateUtil.loadIterator(query, null);
			while (it.hasNext())
			{
				JobData job = it.next();
				String server = job.getServer();
				int startIndex = server.indexOf('[');
				int endIndex = server.lastIndexOf(']');
				if (startIndex > 0 && endIndex > startIndex)
				{
					String node = Values.getStringOrNull(server.substring(startIndex+1, endIndex).trim());
					server = Values.getStringOrNull(server.substring(0, startIndex).trim());
					if (node != null && server != null)
					{
						fixQuery.setParameter("server", server, TypeWrapper.H_STRING);
						fixQuery.setParameter("node", node, TypeWrapper.H_STRING);
						fixQuery.setParameter("id", job.getId(), TypeWrapper.H_INTEGER);
						fixQuery.executeUpdate();
					}
				}
			}
			
			// Update the schema version number
			setSchemaVersion(session, schemaVersion);
			
			// Commit the changes
			HibernateUtil.commit(tx);
			log.info("updateToSchemaVersion136: OK");
		}
		catch (RuntimeException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			log.error("updateToSchemaVersion136: FAILED", ex);
			throw ex;
		}
		finally
		{
			HibernateUtil.close(statelessSession);
		}
		return schemaVersion;
	}

	/**
	 	Update the Files table by setting location=0 (offline) on 
	 	all files with location=2 (secondary), remove the 'action'
	 	column and all stored contexts using the 'action' column.
	 	Update the QuotaValues table by removing all entries with
	 	location=2. Remove the 'secondary_location' column from the
	 	QuotaTypes table.
		@return The new schema version (=144)
	*/
	private static int updateToSchemaVersion144(org.hibernate.Session session, int currentSchemaVersion)
		throws BaseException
	{
		final int schemaVersion = 144;
		try
		{
			if (currentSchemaVersion < 143)
			{
				// First step is to update/delete existing information to the new tables
				org.hibernate.Transaction tx = null;
				try
				{
					tx = HibernateUtil.newTransaction(session);
					
					// Update 'location'
					org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
						"UPDATE FileData f " + 
						"SET f.location=0 " +
						"WHERE f.location=2");
					query.executeUpdate();
			
					// Remove rows in QuotaValues with location=2
					query = HibernateUtil.createSqlQuery(session, 
						"DELETE FROM [QuotaValues] WHERE [location]=2");
					query.executeUpdate();
					
					// Remove table filters, etc. that use the 'action' property
					cleanContextFromProperty(session, Item.FILE, "action", null, true);
					
					setSchemaVersion(session, 143);
	
					// Commit the changes
					HibernateUtil.commit(tx);
				}
				finally
				{
					if (tx != null && tx.getStatus() != TransactionStatus.COMMITTED) HibernateUtil.rollback(tx);
				}
			}
			
			// Drop the 'Files.action' and 'QuotaTypes.secondary_location' columns
			dropColumn(session, "Files", "action", null);
			dropColumn(session, "QuotaTypes", "secondary_location", null);
			
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);
	
			log.info("updateToSchemaVersion144: OK");
		}
		catch (SQLException ex)
		{
			log.error("updateToSchemaVersion144: FAILED", ex);
			throw new BaseException(ex);
		}
		catch (RuntimeException ex)
		{
			log.error("updateToSchemaVersion144: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}
	
	/**
	 	
		@return The new schema version (=147)
	*/
	private static int updateToSchemaVersion147(org.hibernate.Session session, int currentSchemaVersion)
		throws BaseException
	{
		final int schemaVersion = 147;
		try
		{
			
			if (currentSchemaVersion < 146)
			{
				// First step is to update existing information
				org.hibernate.Transaction tx = null;
				try
				{
					tx = HibernateUtil.newTransaction(session);
					
					org.hibernate.query.Query<?> query = HibernateUtil.createQuery(session, 
							"UPDATE PluginDefinitionData " + 
							"SET disabled = true, name = CONCAT(name, ' (deprecated)') " +
							"WHERE className='net.sf.basedb.plugins.SpotImageCreator'");
						query.executeUpdate();
					
					setSchemaVersion(session, 146);
					
					// Commit the changes
					HibernateUtil.commit(tx);
				}
				finally
				{
					if (tx != null && tx.getStatus() != TransactionStatus.COMMITTED) HibernateUtil.rollback(tx);
				}
			}

			// Drop the 'SpotImages' table
			dropTable(session, "SpotImages");
			
			// Update the schema version number
			setSchemaVersionInTransaction(session, schemaVersion);

			log.info("updateToSchemaVersion147: OK");
		}
		catch (RuntimeException ex)
		{
			log.error("updateToSchemaVersion147: FAILED", ex);
			throw ex;
		}
		return schemaVersion;
	}

	
	/**
	 	Vefify and update the remaining quantity of all biomaterials.
		@return The number of biomaterial fixed
	*/
	public static int recalculateRemainingQuantity(ProgressReporter progress)
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numUpdated = 0;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			// For temporary storing the new remaining quantities
			Map<Integer, Double> remaining = new HashMap<Integer, Double>(); 
			
			// Load used quantity for all events
			org.hibernate.query.Query<Object[]> eventQuery = HibernateUtil.createQuery(session, 
				"SELECT evt.bioMaterial.id, evt.usedQuantity FROM BioMaterialEventData evt WHERE NOT evt.usedQuantity IS NULL", Object[].class);
			
			// Load used quantity for sources to all events
			org.hibernate.query.Query<Object[]> sourcesQuery = HibernateUtil.createQuery(session, 
				"SELECT src.bioMaterial.id, src.usedQuantity FROM BioMaterialEventSourceData src WHERE NOT src.usedQuantity IS NULL", Object[].class);
	
			progress.display(10, "Loading biomaterial items...\n");
			
			List<Object[]> events = HibernateUtil.loadList(eventQuery, null);
			events.addAll(HibernateUtil.loadList(sourcesQuery, null));
	
			// Calculate the remaining quantity for each biomaterial
			// o[0] = biomaterial id, o[1] = used quantity			
			for (Object[] o : events)
			{
				Integer id = (Integer)o[0];
				Float usedQuantity = (Float)o[1];
				Double currentRemaining = remaining.get(id);
				if (currentRemaining == null) currentRemaining = 0.0;
				currentRemaining = currentRemaining - usedQuantity.doubleValue();
				remaining.put(id, currentRemaining);
			}
			progress.display(30, "Found "+ remaining.size() + " items. Checking remaining quantity...\n");
			
			// Update the remaining quantity
			org.hibernate.query.Query<?> updateQuery = HibernateUtil.createQuery(session, 
				"UPDATE MeasuredBioMaterialData mbm " +
				"SET mbm.remainingQuantity = :remain "+
				"WHERE mbm = :bioMaterial");
			int numProcessed = 0;
			for (Map.Entry<Integer, Double> entry : remaining.entrySet())
			{
				MeasuredBioMaterialData bioMaterial = HibernateUtil.loadData(session, MeasuredBioMaterialData.class, entry.getKey());
				Double remainingQuantity = entry.getValue();
				Float currentRemaining = bioMaterial.getRemainingQuantity();
				if (currentRemaining == null || Math.abs(currentRemaining-remainingQuantity) > Math.abs(currentRemaining / 10000))
				{
					updateQuery.setParameter("bioMaterial", bioMaterial.getId(), TypeWrapper.H_INTEGER);
					updateQuery.setParameter("remain", remainingQuantity.floatValue(), TypeWrapper.H_FLOAT);
					updateQuery.executeUpdate();
					numUpdated++;
				}
				numProcessed++;
				if (numProcessed % 100 == 0)
				{
					progress.display(30 + numProcessed * 60 / remaining.size(), "Processed " + numProcessed + " biomaterial items...                   ");
				}
			}
			progress.append("\n");
			if (numUpdated == 0)
			{
				progress.display(100, "Found " + numUpdated + " items to fix\n");
			}
			else
			{
				progress.display(95, "Found " + numUpdated + " items to fix. Comitting transaction...\n");
				// Commit the changes
				HibernateUtil.commit(tx);
				progress.display(100, "Completeted update of " + numUpdated + " items\n");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			throw ex;
		}
		return numUpdated;
	}

	/**
	 	Vefify and update the free wells of all bioplates.
		@return The number of bioplates fixed
	*/
	public static int recalculateFreeWells(ProgressReporter progress)
		throws BaseException
	{
		org.hibernate.Session session = null;
		org.hibernate.Transaction tx = null;
		int numUpdated = 0;
		try
		{
			session = HibernateUtil.newSession();
			tx = HibernateUtil.newTransaction(session);
			
			progress.display(10, "Updating bioplates...\n");
			
			// Subquery that counts the number of free wells on
			// a bioplate by checking the "BioMaterials.biowell_id" column
			String countSql = 
				"select count(*) from [BioWells] bw " + 
				"left join [BioMaterials] bm on bm.[biowell_id]=bw.[id] " +
				"where bw.[bioplate_id] = bp.[id] and bm.[id] is null";
			
			// Query to update the free wells on plates that have an incorrect value
			org.hibernate.query.NativeQuery<?> updateQuery = HibernateUtil.createSqlQuery(session, 
				"update [BioPlates] bp "+
				"set [free_wells] = (" + countSql + ") "+
				"where bp.[free_wells] <> ("+countSql+")");
			
			numUpdated = updateQuery.executeUpdate();
			if (numUpdated == 0)
			{
				progress.display(100, "Found " + numUpdated + " bioplates to fix\n");
			}
			else
			{
				progress.display(95, "Found " + numUpdated + " bioplates to fix. Comitting transaction...\n");
				// Commit the changes
				HibernateUtil.commit(tx);
				progress.display(100, "Completed update of " + numUpdated + " bioplates\n");
			}
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx); 
			throw ex;
		}
		return numUpdated;
	}


	private static TableInfo getTableInfo(org.hibernate.Session session, final String tableName)
		throws SQLException
	{
		return HibernateUtil.doReturningWork(session, new TableInfoWork(new Table(tableName)));
	}
	
	private static void dropColumn(org.hibernate.Session session, String tableName, String columnName, ProgressReporter progress)
		throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.query.Query<?> query = null;
		try
		{
			DbEngine engine = HibernateUtil.getDbEngine();
			
			log.debug("Dropping column: " + tableName + "." + columnName);
			tx = HibernateUtil.newTransaction(session);
			
			// Drop foreign keys where this column is used
			TableInfo info = getTableInfo(session, tableName);
			for (ForeignKeyInfo fk : info.getForeignKeys())
			{
				if (fk.getFkColumns().contains(columnName))
				{
					log.debug("Dropping foreign key: " + tableName + "." + fk.getName() + "=" + fk.getFkColumns());
					String sql = engine.getDropForeignKeySql(null, null, tableName, fk.getName());
					query = HibernateUtil.createSqlQuery(session, sql);
					query.executeUpdate();
				}
			}
			
			if (progress != null)
			{
				progress.display(90, "  --alter table " + tableName + " drop column " + columnName);
			}
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_COLUMN", tableName, columnName);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void dropIndex(org.hibernate.Session session, boolean unique, String tableName, String... columnNames)
			throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.query.Query<?> query = null;
		try
		{
			log.debug("Dropping index: " + tableName + ".[" + Arrays.toString(columnNames) + "]");
			tx = session.getTransaction().getStatus() == TransactionStatus.ACTIVE ? null : HibernateUtil.newTransaction(session);
			
			TableInfo info = getTableInfo(session, tableName);
			String indexName = info.findIndexName(null, new HashSet<String>(Arrays.asList(columnNames)));

			if (indexName != null)
			{
				DbEngine engine = HibernateUtil.getDbEngine();

				// Drop foreign keys using any of the column names in the index
				List<ForeignKeyInfo> dropped = new ArrayList<ForeignKeyInfo>();
				if (engine.dropForeignKeysUsedInIndex())
				{
					for (ForeignKeyInfo fk : info.getForeignKeys())
					{
						for (String colName : columnNames)
						{
							if (fk.getFkColumns().contains(colName))
							{
								dropped.add(fk);
								String fkSql = engine.getDropForeignKeySql(null, null, tableName, fk.getName());
								query = HibernateUtil.createSqlQuery(session, fkSql);
								query.executeUpdate();
								break;
							}
						}
					}
				}
				
				String sql = engine.getDropIndexSql(null, null, tableName, indexName, unique);
				query = HibernateUtil.createSqlQuery(session, sql);
				query.executeUpdate();
				
				// Re-created dropped foreign keys
				for (ForeignKeyInfo fk : dropped)
				{
					String fkSql = engine.getCreateForeignKeySql(null, null, tableName, fk.getName(), fk.getFkColumns(), fk.getRefName(), fk.getRefColumns());
					query = HibernateUtil.createSqlQuery(session, fkSql);
					query.executeUpdate();
				}

			}
			// Only commit if we started a new transaction
			if (tx != null) HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void createIndex(org.hibernate.Session session, boolean unique, String indexName, String tableName, String... columnNames)
			throws SQLException
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.query.Query<?> query = null;
		try
		{
			log.debug("Creating index: " + tableName + ".[" + Arrays.toString(columnNames) + "]");
			tx = HibernateUtil.newTransaction(session);
			DbEngine engine = HibernateUtil.getDbEngine();
			String sql = engine.getCreateIndexSql(null, null, tableName, indexName, new LinkedHashSet<String>(Arrays.asList(columnNames)), unique);
			query = HibernateUtil.createSqlQuery(session, sql);
			query.executeUpdate();
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void dropNotNullConstraint(org.hibernate.Session session, String tableName, String columnName, String mySqlDataType)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.query.Query<?> query = null;
		try
		{
			log.debug("Dropping NOT NULL on column: " + tableName + "." + columnName);
			tx = session.getTransaction().getStatus() == TransactionStatus.ACTIVE ? null : HibernateUtil.newTransaction(session);
			
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_NOT_NULL_CONSTRAINT", tableName, columnName, mySqlDataType);
			query.executeUpdate();
			
			// Only commit if we started a new transaction
			if (tx != null) HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}

	private static void dropTable(org.hibernate.Session session, String tableName)
	{
		org.hibernate.Transaction tx = null;
		org.hibernate.query.Query<?> query = null;
		try
		{
			log.debug("Dropping table: " + tableName);
			tx = session.getTransaction().getStatus() == TransactionStatus.ACTIVE ? null : HibernateUtil.newTransaction(session);
			query = HibernateUtil.getPredefinedSQLQuery(session, "DROP_TABLE", tableName);
			query.executeUpdate();
			// Only commit if we started a new transaction
			if (tx != null) HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}
	
	
	/**
		Remove context-related information for the given item type.
	*/
	private static void removeContext(org.hibernate.Session session, int itemType)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			log.debug("Removing context for item: " + itemType);
			tx = HibernateUtil.newTransaction(session);
			
			org.hibernate.query.Query<ContextData> ctxQuery = HibernateUtil.createQuery(session, 
				"SELECT ctx FROM ContextData ctx WHERE ctx.itemType = :itemType", ContextData.class);
			ctxQuery.setParameter("itemType", itemType, TypeWrapper.H_INTEGER);
			for (ContextData ctx : HibernateUtil.loadList(ctxQuery, null))
			{
				session.delete(ctx);
			}

			// DELETE FROM PluginDefinitionGuiContexts WHERE item_type=?
			org.hibernate.query.Query<?> deleteQuery = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PLUGINGUICONTEXTS_FOR_ITEMTYPE");
			deleteQuery.setParameter("itemType", itemType, TypeWrapper.H_INTEGER);
			deleteQuery.executeUpdate();
			
			log.debug("Removing role key for item: " + itemType);
			// DELETE FROM PluginKeys WHERE item_type=?
			deleteQuery = HibernateUtil.getPredefinedSQLQuery(session, 
				"DELETE_PLUGINKEYS_FOR_ITEMTYPE");
			deleteQuery.setParameter("itemType", itemType, TypeWrapper.H_INTEGER);
			deleteQuery.executeUpdate();

			org.hibernate.query.Query<RoleKeyData> keyQuery = HibernateUtil.createQuery(session, 
				"SELECT key FROM RoleKeyData key WHERE key.itemType = :itemType", RoleKeyData.class);
			keyQuery.setParameter("itemType", itemType, TypeWrapper.H_INTEGER);
			for (RoleKeyData key : HibernateUtil.loadList(keyQuery, null))
			{
				session.delete(key);
			}
			
			HibernateUtil.commit(tx);
		}
		catch (BaseException ex)
		{
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
	}
	
	private static void cleanContextFromProperty(org.hibernate.Session session, Item itemType, 
		String propertyName, String replacementPropertyName, boolean forceRemoveFilter)
	{
		org.hibernate.Transaction tx = null;
		try
		{
			log.debug("Cleaning context for item: " + itemType + "; property=" + propertyName + "; replacement=" + replacementPropertyName);
			tx = session.getTransaction().getStatus() == TransactionStatus.ACTIVE ? null : HibernateUtil.newTransaction(session);
			
			org.hibernate.query.Query<ContextData> query = HibernateUtil.createQuery(session, 
				"select ctx from ContextData ctx where ctx.itemType = " + itemType.getValue(), ContextData.class);
			List<ContextData> contexts = HibernateUtil.loadList(query, null);
			
			for (ContextData ctx : contexts)
			{
				// Sort property
				if (ctx.getSortProperty() != null)
				{
					ctx.setSortProperty(replaceInCommaSeparatedList(ctx.getSortProperty(), propertyName, replacementPropertyName));
				}
				
				// "columns" settings
				Map<String, String> settings = ctx.getSettings();
				String columns = settings.get("columns");
				if (columns != null)
				{
					columns = replaceInCommaSeparatedList(columns, propertyName, replacementPropertyName);
					if (columns == null)
					{
						settings.remove("columns");
					}
					else
					{
						settings.put("columns", columns);
					}
				}
					
				// Filter
				Map<String, PropertyFilterData> filters = ctx.getPropertyFilters();
				List<String> filterProperties = new ArrayList<String>(filters.keySet());
				for (String property : filterProperties)
				{
					if (property.startsWith(propertyName))
					{
						PropertyFilterData old = filters.remove(property);
						if (replacementPropertyName != null)
						{
							filters.put(replacementPropertyName + property.substring(propertyName.length()), old);
						}
					}
				}
			}
			
			// Only commit if we started a new transaction
			if (tx != null) HibernateUtil.commit(tx);
		}
		catch (RuntimeException ex)
		{
			// Only rollback if we started the transaction
			if (tx != null) HibernateUtil.rollback(tx);
			throw ex;
		}
		
	}
	
	
	private static String replaceInCommaSeparatedList(String list, String find, String replacement)
	{
		List<String> elements = new ArrayList<String>(Arrays.asList(list.split(",")));
		String result = list;
		
		Iterator<String> it = elements.iterator();
		boolean hasMatched = false;
		int index = 0;
		while (it.hasNext())
		{
			String value = it.next();
			if (value != null && value.startsWith(find))
			{
				// A match, replace or remove
				hasMatched = true;
				if (replacement != null)
				{
					elements.set(index, replacement + value.substring(find.length()));
				}
				else
				{
					it.remove();
				}
			}
		}
		if (hasMatched)
		{
			if (elements.size() == 0)
			{
				result = null;
			}
			else
			{
				result = Values.getString(elements, ",", true);
			}
		}
		return result;
	}

	/**
		Action that need to make some change to the database schem after the
		main transaction in {@link Update#adjustExistingItems(ProgressReporter, int, String, String)}
		has finished.
		@since 3.7
	*/
	static abstract class SchemaChange
	{
		abstract void performChange(org.hibernate.Session session)
			throws SQLException;
	}

	
	/**
		Action that drops a column in a table.
		@since 3.7
	*/
	static class DropColumnChange
		extends SchemaChange
	{

		private final String tableName;
		private final String columnName;
		
		public DropColumnChange(String tableName, String columnName) 
		{
			this.tableName = tableName;
			this.columnName = columnName;
		}
		
		@Override
		void performChange(org.hibernate.Session session)
			throws SQLException
		{
			dropColumn(session, tableName, "removed", null);			
		}
		
	}
	
}
