/*
	$Id: BioMaterialEventData.java 6989 2015-11-02 07:46:40Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.Map;
import java.util.HashMap;

/**
	This class is used to register events for measured biomaterials.

	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.BioMaterialEvent
	@see <a href="../../../../../../html/developer/api/data_api.biomaterials.html">Developer documentation: Biomaterials LIMS</a>
	@base.modified $Date: 2015-11-02 08:46:40 +0100 (mån, 02 nov. 2015) $
	@hibernate.class table="`BioMaterialEvents`" lazy="true"
*/
public class BioMaterialEventData
	extends BasicData
	implements RegisteredData, LoggableData
{

	public BioMaterialEventData()
	{}
	
	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	@Override
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------
	
	private int eventType;
	/**
		Get the type of event, 1=creation event, 2=hybridization event, 3=other event
		@hibernate.property column="`event_type`" type="int" not-null="true" update="false"
	*/
	public int getEventType()
	{
		return eventType;
	}
	public void setEventType(int eventType)
	{
		this.eventType = eventType;
	}

	private BioPlateEventParticipantData bioPlateEventParticipant;
	/**
		Get the bioplate event that is the master event for this event. Null if
		no bioplate event exists for this event.
		@hibernate.many-to-one column="`bioplateeventparticipant_id`" not-null="false" outer-join="false" update="false"
		@since 2.17
	*/
	public BioPlateEventParticipantData getBioPlateEventParticipant()
	{
		return bioPlateEventParticipant;
	}
	public void setBioPlateEventParticipant(BioPlateEventParticipantData bioPlateEventParticipant)
	{
		this.bioPlateEventParticipant = bioPlateEventParticipant;
	}
	
	private MeasuredBioMaterialData bioMaterial;
	/**
		Get the biomaterial produced or affected by this event.
		@hibernate.many-to-one column="`biomaterial_id`" not-null="false" outer-join="false" update="false"
	*/
	public MeasuredBioMaterialData getBioMaterial()
	{
		return bioMaterial;
	}
	public void setBioMaterial(MeasuredBioMaterialData bioMaterial)
	{
		this.bioMaterial = bioMaterial;
	}
	
	private Float usedQuantity;
	/**
		Get the used quantity of the biomaterial affected by this event.
		This can be both positive and negative.
		@hibernate.property column="`used_quantity`" type="float" not-null="false" 
	*/
	public Float getUsedQuantity()
	{
		return usedQuantity;
	}
	public void setUsedQuantity(Float usedQuantity)
	{
		this.usedQuantity = usedQuantity;
	}
	
	
	private PhysicalBioAssayData bioassay;
	/**
		Get the physical bioassay that was created by this event.
		@hibernate.many-to-one column="`physicalbioassay_id`" not-null="false" outer-join="false" update="false"
	*/
	public PhysicalBioAssayData getPhysicalBioAssay()
	{
		return bioassay;
	}
	public void setPhysicalBioAssay(PhysicalBioAssayData bioassay)
	{
		this.bioassay = bioassay;
	}

	private UserData user;
	/**
		Get the user that is responsible for this event, typically this is the user that
		entered the information in the database.
		@hibernate.many-to-one column="`user_id`" not-null="true" outer-join="false" update="false"
	*/
	public UserData getUser()
	{
		return user;
	}
	public void setUser(UserData user)
	{
		this.user = user;
	}

	private ProtocolData protocol;
	/**
		Get the protocol used in this event.
		@hibernate.many-to-one column="`protocol_id`" not-null="false" outer-join="false"
	*/
	public ProtocolData getProtocol()
	{
		return protocol;
	}
	public void setProtocol(ProtocolData protocol)
	{
		this.protocol = protocol;
	}

	private HardwareData hardware;
	/**
	 	Get the hardware used in this event
	 	@hibernate.many-to-one column="`hardware_id`" not-null="false" outer-join="false"
	*/
	public HardwareData getHardware()
	{
		return this.hardware;
	}
	public void setHardware(HardwareData hardware)
	{
		this.hardware = hardware;
	}
	
	private KitData kit;
	/**
		Get the kit used in this event.
		@since 3.7
		@hibernate.many-to-one column="`kit_id`" not-null="false" outer-join="false"
	*/
	public KitData getKit()
	{
		return kit;
	}
	public void setKit(KitData kit)
	{
		this.kit = kit;
	}

	private Date eventDate;
	/**
		Get the date this event happened in the lab.
		@hibernate.property column="`event_date`" type="date" not-null="false"
	*/
	public Date getEventDate()
	{
		return eventDate;
	}
	public void setEventDate(Date eventDate)
	{
		this.eventDate = eventDate;
	}

	/**
		The maximum length of the comment about this event.
	*/
	public static final int MAX_COMMENT_LENGTH = 65535;
	private String comment;
	/**
		Get the comment about this event.
		@hibernate.property column="`comment`" type="text" not-null="false"
	*/
	public String getComment()
	{
		return comment;
	}
	public void setComment(String comment)
	{
		this.comment = comment;
	}
	
	private Map<BioMaterialData, BioMaterialEventSourceData> sources;
	/**
		Get a map containing information about the source biomaterials used
		in the event. This is only used for events that create child biomaterial or
		physical bioassay. This is the inverse end. See
		{@link BioMaterialEventSourceData#getEvent()}
		@hibernate.map inverse="true" lazy="true" cascade="all"
		@hibernate.collection-key column="`event_id`"
		@hibernate.index-many-to-many column="`biomaterial_id`" class="net.sf.basedb.core.data.BioMaterialData"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.BioMaterialEventSourceData"
		@since 3.0
	*/
	public Map<BioMaterialData, BioMaterialEventSourceData> getSources()
	{
		if (sources == null) 
		{
			sources = new HashMap<BioMaterialData, BioMaterialEventSourceData>();
		}
		return sources;
	}
	void setSources(Map<BioMaterialData, BioMaterialEventSourceData> sources)
	{
		this.sources = sources;
	}
	
}
