/*
 $Id $

	Copyright (C) 2017 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.basedb.core.data;

import java.util.Date;
import java.util.Set;

/**
	This class holds information about a verified user device.
	The {@link #getToken()} is randomly generated token that
	each device is supposed to keep track of and submit together
	with the login info.
 
	@author Nicklas
	@since 3.12
	@hibernate.class table="`UserDevices`" lazy="false"
*/
public class UserDeviceData
	extends BasicData
	implements RegisteredData, NameableData
{
	public UserDeviceData()
	{}
	
	private UserData user;
	/**
		The user that this device is associated with.
		@hibernate.many-to-one column="`user_id`" not-null="true" update="false" outer-join="false"
	*/
	public UserData getUser()
	{
		return user;
	}
	public void setUser(UserData user)
	{
		this.user = user;
	}
		
	private ClientData client;
	/**
		Get the client application this device is associated with.
		@hibernate.many-to-one column="`client_id`" not-null="true" update="false" outer-join="false"
	*/
	public ClientData getClient()
	{
		return client;
	}
	public void setClient(ClientData client)
	{
		this.client = client;
	}
	
	/*
		From the RegisteredData interface
		-------------------------------------------
	*/
	private Date entryDate;
	@Override
	public Date getEntryDate()
	{
		return entryDate;
	}
	public void setEntryDate(Date entryDate)
	{
		this.entryDate = entryDate;
	}
	// -------------------------------------------
	/*
		From the NameableData interface
		-------------------------------------------
	*/	
	private String name;
	@Override
	public String getName()
	{
		return name;
	}
	
	@Override
	public void setName(String name)
	{
		this.name = name;
	}
	
	private String description;
	@Override
	public String getDescription()
	{
		return description;
	}
	
	@Override
	public void setDescription(String description)
	{
		this.description = description;
	}
	// -------------------------------------------

	
	/**
		The maximum length of the device token that can be stored in the database.
		@see #setToken(String)
	*/
	public static final int MAX_TOKEN_LENGTH = 255;
	private String token;
	/**
		Get the device token.
		@hibernate.property column="`token`" type="string" length="255" not-null="true" update="false"
	*/
	public String getToken()
	{
		return token;
	}
	public void setToken(String token)
	{
		this.token = token;
	}

	/**
		The maximum length of the user agent that can be stored in the database.
		@see #setUserAgent(String)
	*/
	public static final int MAX_USER_AGENT_LENGTH = 65535;
	private String userAgent;
	/**
		Get a string with more information about the device (eg. User-Agent from a web browser).
		@hibernate.property column="`user_agent`" type="text" not-null="false"
	*/
	public String getUserAgent()
	{
		return userAgent;
	}
	public void setUserAgent(String userAgent)
	{
		this.userAgent = userAgent;
	}
	
	private Date lastUsed;
	/**
		Get the date and time the user last logged in from this device.
		@hibernate.property column="`last_used`" type="timestamp" not-null="true"
	*/
	public Date getLastUsed()
	{
		return lastUsed;
	}
	
	public void setLastUsed(Date lastUsed)
	{
		this.lastUsed = lastUsed;
	}

	/**
		The maximum length of the remote ID that can be stored in the database.
		@see #setLastRemoteId(String)
	*/
	public static final int MAX_REMOTE_ID_LENGTH = 255;
	private String remoteId;
	/**
		Get the remote id (=ip address) of the last login from this device.
		@hibernate.property column="`lastremote_id`" type="string" length="255" not-null="false"
	*/
	public String getLastRemoteId()
	{
		return remoteId;
	}
	public void setLastRemoteId(String remoteId)
	{
		this.remoteId = remoteId;
	}

	/**
		The maximum length of the location that can be stored in the database.
		@see #setLocation(String)
	*/
	public static final int MAX_LOCATION_LENGTH = 255;
	private String location;
	/**
		Get the location of the user (if known).
		@hibernate.property column="`location`" type="string" length="255" not-null="false"
	*/
	public String getLocation()
	{
		return location;
	}
	public void setLocation(String location)
	{
		this.location = location;
	}

	private Float locationLat;
	/**
		Get the location latitude of the user (if known).
		@hibernate.property column="`location_lat`" type="float" not-null="false"
	*/
	public Float getLocationLatitude()
	{
		return locationLat;
	}
	public void setLocationLatitude(Float locationLat)
	{
		this.locationLat = locationLat;
	}
	
	private Float locationLong;
	/**
		Get the location longitude of the user (if known).
		@hibernate.property column="`location_long`" type="float" not-null="false"
	*/
	public Float getLocationLongitude()
	{
		return locationLong;
	}
	public void setLocationLongitude(Float locationLong)
	{
		this.locationLong = locationLong;
	}
	
	private Set<SessionData> sessions;
	/**
		This is the inverse end but we do not map with inverse="true" cascade="delete"
		since we want Hibernate to set SessionData.device_id = null instead of deleting
		sessions.
		@see SessionData#getDevice()
		@hibernate.set lazy="true" 
		@hibernate.collection-key column="`device_id`"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.SessionData"
	*/
	Set<SessionData> getSessions()
	{
		return sessions;
	}
	void setSessions(Set<SessionData> sessions)
	{
		this.sessions = sessions;
	}


}
