/**
	$Id: SavePointWork.java 6468 2014-06-10 11:27:02Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Savepoint;

import org.hibernate.jdbc.ReturningWork;


/**
	A wrapper work that protects a parent work with a savepoint. 
	Before executing the parent work, a {@link Savepoint} is
	created on the {@link Connection}. If the parent work
	results throws an {@link SQLException} the connection is
	rolled back to the savepoint. Other exceptions and a normal
	completion just releases the savepoint.

	@author nicklas
	@since 3.1
	@base.modified $Date: 2014-06-10 13:27:02 +0200 (tis, 10 juni 2014) $
*/
public class SavePointWork<R>
	implements ReturningWork<R>
{

	private final ReturningWork<R> work;
	
	/**
		Create a new work item.
		@param work The parent work
	*/
	public SavePointWork(ReturningWork<R> work)
	{
		this.work = work;
	}
	
	/**
		From the Work interface
		-----------------------
	*/
	@Override
	public R execute(Connection connection) 
		throws SQLException
	{
		Savepoint savepoint = null;
		R result = null;
		try
		{
			savepoint = connection.setSavepoint();
			result = work.execute(connection);
		}
		catch (SQLException ex)
		{
			if (savepoint != null) 
			{
				connection.rollback(savepoint);
				savepoint = null;
			}
			throw ex;
		}
		finally
		{
			if (savepoint != null) 
			{
				connection.releaseSavepoint(savepoint);
			}
		}
		return result;
	}
	// --------------------------

}
