/*
	$Id: AbstractPlugin.java 7714 2019-05-21 11:37:09Z nicklas $

	Copyright (C) 2005 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.PluginParameter;

import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Collection;
import java.util.List;
import java.util.ArrayList;

/**
	 This is an abstract base class useful for developing plugins.
	 It's main purpose is to help with validating and storing request
	 parameters into the plugin or job configuration.
 
	 @base.modified $Date: 2019-05-21 13:37:09 +0200 (tis, 21 maj 2019) $
	 @author Enell, Nicklas
	 @version 2.0
*/
public abstract class AbstractPlugin 
	implements Plugin
{
	protected SessionControl sc = null;
	
	protected ParameterValues configuration = null;
	
	protected ParameterValues job = null;
	
	private DbControl logDc;
	private File logfile;
	private PrintWriter logger;
	
	/**
		Section definition for grouping annotation options.
		@since 2.5
	*/
	protected static final PluginParameter<String> annotationSection = new PluginParameter<String>(
		"annotationSection",
		"Annotations",
		"Options related to annotations.",
		null
	);
	
	
	/**
		Create a new <code>AbstractPlugin</code>.
	*/
	protected AbstractPlugin()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	/**
		Store copies of the session control, plugin and job configuration. These
		are available to subclasses in the {@link #sc}, {@link #configuration}
		and {@link #job} variables. If a subclass overrides this method it is 
		recommended that it also calls <code>super.init(sc, configuration, job)</code>.
	*/
	@Override
	public void init(SessionControl sc, ParameterValues configuration, ParameterValues job)
		throws BaseException
	{
		this.sc = sc;
		this.configuration = configuration;
		this.job = job;
	}
	/**
		Clears the variables set by the <code>init</code> method. If a subclass 
		overrides this method it is  recommended that it also calls <code>super.done()</code>.
	*/
	@Override
	public void done()
	{
		closeLogFile();
		configuration = null;
		job = null;
		sc = null;
	}
	/**
		Returns TRUE, since that is how the plugins used to work before this method
		was introduced.
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	/**
		Returns FALSE, since that is how the plugins used to work before this method
		was introduced.
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
		Return null, which gives the plugin the same permissions as the logged in user.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		return null;
	}
	// -------------------------------------------

	/**
		Validate the parameter values in a request against a list of plugin parameter definitions.
		The validation is done by calling the {@link net.sf.basedb.core.ParameterType#validate(String, List)}
		for each parameter in the list of parameter definitions. If all parameters are valid
		this method returns null, otherwise a list of exceptions that ocurred during the validation.
		This list can be directly propagated to the {@link Response#setError(String, List)}
		method. For example:
		<pre class="code">
// <b>In Plugin.invoke() method:</b>
List&lt;PluginParameter&lt;?&gt;&gt; parameters = getParametersForCommand(request.getCommand());
List&lt;Throwable&gt; errors = 
   validateRequestParameters(parameters, request);
if (errors != null)
{
   response.setError(errors.size()+
      " invalid parameter(s) were found in the request", errors);
   return;
}
</pre>

		@param parameters A list containing the parameter definitions
		@param request The request
		@return A list of exceptions, or null if all parameters are valid
	*/
	protected List<Throwable> validateRequestParameters(List<PluginParameter<?>> parameters, Request request)
	{
		List<Throwable> errors = new ArrayList<Throwable>();
		for (PluginParameter<?> pp : parameters)
		{
			try
			{
				if (pp.getParameterType() != null)
				{
					pp.getParameterType().validate(pp.getName(), request.getParameterValues(pp.getName()));
				}
			}
			catch (Throwable t)
			{
				errors.add(t);
			}
		}
		return errors.size() == 0 ? null : errors;
	}
	
	/**
		Copy a parameter value from a {@link Request} to a {@link ParameterValues}
		object. This method makes it easy for a subclass to directly store a request
		parameter into either the plugin or job configuration. Use this method when
		the parameter valus is a single value.
	
		@param values The <code>ParameterValues</code> object where the value should
			be stored, use either {@link #configuration} or {@link #job}
		@param request The <code>Request</code> object
		@param parameter The <code>PluginParameter</code> object for the parameter
		@throws BaseException If the parameter value is invalid
	*/
	@SuppressWarnings({"unchecked"})
	protected <T> void storeValue(ParameterValues values, Request request, PluginParameter<T> parameter)
		throws BaseException
	{
		if (parameter == null) return;
		values.setValue(parameter.getName(), parameter.getParameterType(), (T)request.getParameterValue(parameter.getName()));
	}

	/**
		Store a value in a {@link ParameterValues} object. This method is useful when 
		a plugin needs to store a value that is not part of the request.
		
		@param <T> The type of the parameter
		@param values The <code>ParameterValues</code> object where the value should
			be stored, use either {@link #configuration} or {@link #job}
		@param parameter The <code>PluginParameter</code> object for the parameter
		@param value The value to save
		@throws BaseException If the parameter value is invalid
	*/
	protected <T> void storeValue(ParameterValues values, PluginParameter<T> parameter, T value)
	{
		if (parameter == null) return;
		values.setValue(parameter.getName(), parameter.getParameterType(), value);
	}
	
	/**
		Copy a list of parameter values from a {@link Request} to a {@link ParameterValues}
		object. This method makes it easy for a subclass to directly store a request
		parameter into either the plugin or job configuration. Use this method when the 
		parameter value is a list of values.

		@param values The <code>ParameterValues</code> object where the value should
			be stored, use either {@link #configuration} or {@link #job}
		@param request The <code>Request</code> object
		@param parameter The <code>ParameterType</code> object for the parameter
		@throws BaseException If the parameter value is invalid
	*/
	@SuppressWarnings({"unchecked"})
	protected <T> void storeValues(ParameterValues values, Request request, PluginParameter<T> parameter)
		throws BaseException
	{
		if (parameter == null) return;
		values.setValues(parameter.getName(), parameter.getParameterType(), (java.util.List<T>)request.getParameterValues(parameter.getName()));
	}
	
	/**
		Get the job that is executing this plugin or null.
		@param dc The DbControl to use for database access
		@return The job or null
	*/
	protected Job getCurrentJob(DbControl dc)
	{
		return job == null || job.getId() == 0 ? null : Job.getById(dc, job.getId());
	}
	
	/**
		Get the plug-in configuration that this is currently in use or null.
		@param dc The DbControl to use for database access
		@return The configuration or null
		@since 2.5
	*/
	protected PluginConfiguration getCurrentConfiguration(DbControl dc)
	{
		return configuration == null || configuration.getId() == 0 ? 
			null : PluginConfiguration.getById(dc, configuration.getId());
	}
	
	/**
		The name of the parameter that asks if annotations should be copied or not.
		@see #getCopyAnnotationsParmeter(String, String, Boolean)
		@since 2.5
	*/
	protected static final String COPY_ANNOTATIONS = "copyAnnotations";
	
	/**
		Parameter definition that asks if annotations should be copied from the
		plug-in definition/configuration to other items.
		@param label The label to use for the parameter or null to use the default label
			(Copy annotations)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default value
	*/
	protected PluginParameter<Boolean> getCopyAnnotationsParmeter(String label, String description, Boolean defaultValue)
	{
		if (label == null) label = "Copy annotations";
		if (description == null) 
		{
			description = "Select this option to copy annotations from the plug-in to the item.";
		}
		return new PluginParameter<Boolean>(
				COPY_ANNOTATIONS, label, description, defaultValue, 
				new BooleanParameterType(defaultValue, false)
		);
	}
	
	/**
		The name of the parameter that asks if annotations should be overwritten or not.
		@see #getOverwriteAnnotationsParameters(String, String, Boolean)
		@since 2.5
	*/
	protected static final String OVERWRITE_ANNOTATIONS = "overwriteAnnotations";
	
	/**
		Parameter definition that asks if copied annotations should overwrite existing
		annotations or not.
		@param label The label to use for the parameter or null to use the default label
			(Overwrite existing)
		@param description The description to use for the parameter or null to use the
			default description
		@param defaultValue The default value 
		@since 2.5
	*/
	protected PluginParameter<Boolean> getOverwriteAnnotationsParameters(String label, String description, Boolean defaultValue)
	{
		if (label == null) label = "Overwrite existing";
		if (description == null) 
		{
			description = "If set, existing annotations will be overwritten with new values.";
		}
		return new PluginParameter<Boolean>(
				OVERWRITE_ANNOTATIONS, label, description, defaultValue, 
				new BooleanParameterType(defaultValue, false)
		);
	}
	
	/**
		Get the value for a job or configuration parameter. This method first
		checks the job parameters. If there is no value it will check the
		configuration parameters.
		@param name The name of the parameter
		@return The parameter value, or null
		@since 2.15
	*/
	protected Object getJobOrConfigurationValue(String name)
	{
		Object value = null;
		if (job != null) value = job.getValue(name);
		if (value == null && configuration != null) value = configuration.getValue(name);
		return value;
	}
	
	/**
		Clone a plug-in parameter and set the default value to the value that is
		already stored in the job or configuration parameters.
		@param pp The parameter to clone
		@return The cloned parameter
		@since 2.15
	*/
	@SuppressWarnings("unchecked")
	protected <T> PluginParameter<T> cloneParameterWithDefaultValue(PluginParameter<T> pp)
	{
		T defaultValue = (T)getJobOrConfigurationValue(pp.getName());
		PluginParameter<T> clone = new PluginParameter<T>(
				pp.getName(), 
				pp.getLabel(),
				pp.getDescription(), 
				defaultValue, 
				pp.getParameterType());
		return clone;
	}
	
	/**
		Create a log file in the BASE file system for logging 
		detailed information from a plug-in. The log file will
		be created if it doesn't exists. The log file is handled 
		in it's own transaction. Call {@link #closeLogFile()} to 
		close the log file and commit changes to the database.
		If this method is called when a log file is open, it
		does nothing.
		
		@param path The path to log to, if null this method does nothing
		@since 2.8
	*/
	protected void createLogFile(String path)
	{
		if (logfile != null || path == null) return;
		
		logDc = sc.newDbControl();
		try
		{
			Path p = new Path(path, Path.Type.FILE);
			File f = File.getByPath(logDc, p, true);
			f.setMimeType("text/plain");
			Charset utf8 = StandardCharsets.UTF_8;
			f.setCharacterSet(utf8.name());
			if (!f.isInDatabase()) logDc.saveItem(f);
			logger = new PrintWriter(
					new OutputStreamWriter(f.getUploadStream(false), utf8), 
					true);
			logfile = f;
		}
		finally
		{
			// If we failed creating the logfile we must clean up the DbControl
			if (logfile == null && logDc != null) 
			{
				logDc.close();
				logDc = null;
				logger = null;
			}
		}
	}
	
	
	/**
		Close the log file created by {@link #createLogFile(String)}.
		@since 2.8
	*/
	protected void closeLogFile()
	{
		try
		{
			if (logger != null) logger.close();
			if (logDc != null) logDc.commit();
		}
		catch (Throwable t)
		{}
		finally
		{
			logger = null;
			logfile = null;
			if (logDc != null) logDc.close();
			logDc = null;
		}
	}
	
	/**
		Log a message to the log file created by {@link #createLogFile(String)}.
		If no log file has been created, this method does nothing.
		
		@param message The message to log
		@since 2.8
	*/
	protected void log(String message)
	{
		if (logger == null) return;
		logger.println(message);
		logger.flush();
	}
	
	/**
		Log a message and stack trace to the log file created by 
		{@link #createLogFile(String)}. If not log file exists, 
		this method does nothing.
		
		@param message The message to log
		@param t A stack trace
		@since 2.8
	*/
	protected void log(String message, Throwable t)
	{
		if (logger == null) return;
		if (t == null)
		{
			logger.println(message);
		}
		else
		{
			logger.println(message + ": " + t.getMessage());
			t.printStackTrace(logger);
		}
		logger.flush();
	}
	
	
}
