/**
	$Id: DelegatingSignalHandler.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.signal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
	A signal handler implementation that can delegate to other signal handlers,
	depending on which signal it receives. Multiple signal handlers may be registered
	for each signal. The order in which they are called is not specified, but it is
	probably different from the order they are registered.
	
	@author nicklas
	@version 2.6
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (ons, 09 apr. 2014) $
*/
public class DelegatingSignalHandler
	extends AbstractSignalHandler
{
	/**
		Log signals processing.
	*/
	private static final org.slf4j.Logger logger = 
		org.slf4j.LoggerFactory.getLogger(DelegatingSignalHandler.class);
	
	/**
		Holds the registered signal handlers.
	*/
	private Map<Signal, Set<SignalHandler>> handlers;

	/**
		List holding received signals.
		@since 2.16
	*/
	private final List<Signal> received;
	
	/**
		Create a new delegating signal handler. Signal handlers to
		delegate to should be registered with 
		{@link #registerSignalHandler(SignalHandler)}.
	*/
	public DelegatingSignalHandler()
	{
		super();
		handlers = new HashMap<Signal, Set<SignalHandler>>();
		received = Collections.synchronizedList(new ArrayList<Signal>());
	}
	
	/**
		Create a new delegating signal handler and register the
		given handlers as delegates.
		@param handlers An array with signal handlers that signal should
			be delegated to
		@since 2.16
	*/
	public DelegatingSignalHandler(SignalHandler... handlers)
	{
		this();
		for (SignalHandler sh : handlers)
		{
			registerSignalHandler(sh);
		}
	}
	
	/*
		From the SignalHandler interface
		-------------------------------------------
	*/
	/**
		@return A collection containing all signals that have at least one 
			registered handler
	*/
	@Override
	public Collection<Signal> getSupportedSignals()
	{
		return Collections.unmodifiableSet(handlers.keySet());
	}
	
	/**
		Check if at least one handler has been registered for the given signal.
	*/
	@Override
	public boolean supports(Signal signal)
	{
		return handlers.containsKey(signal);
	}
	
	/**
		Handler the given signal by invoking {@link SignalHandler#handleSignal(Signal)}
		on each registered signal handler. The order of the invokation is not
		predictable.
		@throws UnsupportedSignalException If no handler has been registered for the
			given signal
	*/
	@Override
	public void handleSignal(Signal signal)
	{
		logger.debug("Got signal: " + signal);
		Set<SignalHandler> all = handlers.get(signal);
		if (all == null) 
		{
			logger.debug("Signal not supported " + signal);
			throw new UnsupportedSignalException(signal);
		}
		received.add(signal);
		for (SignalHandler handler : all)
		{
			logger.debug("Sending signal " + signal + " to: " + handler);
			handler.handleSignal(signal);
		}
	}
	// -------------------------------------------

	/**
		Register a signal handler to delegate to. The handler is registered
		for all signals it supports.
		@param handler The handler to register
	*/
	public void registerSignalHandler(SignalHandler handler)
	{
		logger.debug("Registering signal handler: " + handler);
		for (Signal signal : handler.getSupportedSignals())
		{
			Set<SignalHandler> temp = handlers.get(signal);
			if (temp == null)
			{
				temp = new HashSet<SignalHandler>();
				handlers.put(signal, temp);
			}
			temp.add(handler);
		}
	}
	
	/**
		Unregister a signal handler.
		@param handler The handler to unregister
	*/
	public void unregisterSignalHandler(SignalHandler handler)
	{
		logger.debug("Unregistering signal handler: " + handler);
		Iterator<Map.Entry<Signal, Set<SignalHandler>>> it = handlers.entrySet().iterator();
		while (it.hasNext())
		{
			Map.Entry<Signal, Set<SignalHandler>> entry = it.next();
			Set<SignalHandler> registered = entry.getValue();
			// Remove the registered handler...
			if (registered.remove(handler))
			{
				// ... if no more handlers for this signal, remove the entry
				if (registered.size() == 0) it.remove();
			}
		}
	}

	/**
		Check if any signals has been received by this handler.
		@return TRUE if at least one signal has been received, FALSE otherwise
		@since 2.16
	*/
	public boolean hasReceivedSignals()
	{
		return received.size() > 0;
	}
	
	/**
		Get the list of received signals. Calling this method clears the internal
		list and if no signals are received in the meantime, the next call to
		this method will return null.
		@return A list with signals in the order they were
			received, or null if no signals has been received
		@since 2.16
	*/
	public List<Signal> getReceivedSignals()
	{
		List<Signal> toReturn = null;
		if (received.size() > 0)
		{
			toReturn = new ArrayList<Signal>(received);
			received.clear();
		}
		return toReturn;
	}
	
	/**
		Check if the given signal has been received by this signal handler.
		@param signal The signal to check for
		@return TRUE if the signal has been received, FALSE otherwise
		@since 2.16
	*/
	public boolean hasReceived(Signal signal)
	{
		return received.contains(signal);
	}

	
}
