/**
	$Id: ColorGenerator.java 4515 2008-09-11 20:08:14Z jari $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.awt.Color;

import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.NumberOutOfRangeException;

/**
	This class is used to generate colors for numerical values. Given a
	numeric range mith a start, mid and max value and one color for each
	of these values you can create new colors for values within the specified
	range.
	<p>
	Example:<br>
	We have the range and colors: <code>-1 (BLUE), 0 (WHITE) , +1 (YELLOW)</code>
	The value -0.5 is converted to a lighter blue and the value +0.5 is converted
	to a ligther yellow. Values falling outside the min or max values are
	assigned the endpoint colors. In this case, BLUE and YELLOW.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2008-09-11 22:08:14 +0200 (tors, 11 sep. 2008) $
*/
public class ColorGenerator
{

	/**
		Constant holding the value of <code>1 / ln(2)</code>. It is used
		to calculate the two-based logarithm of values using the formula:
		<code>log2(x) = ln(x) / ln(2)</code>
	*/
	public static final float INVLN2 = (float)(1 / Math.log(2));
	
	/**
		Convert a hexadecimal color string such as one used in HTML
		to a Color object. The hexadecimal string must have 6 characters
		in the range 0-9 or A-F (case ignored).
		@param hexRGB The hexadecimal color string
		@return A Color object
		@throws InvalidDataException If the string cannot be converted
	*/
	public static Color fromHex(String hexRGB)
	{
		if (hexRGB == null || !hexRGB.matches("[0-9A-Fa-f]{6}"))
		{
			throw new InvalidDataException("Not a hex RGB value: " + hexRGB);
		}
		int r = Integer.parseInt(hexRGB.substring(0, 2), 16);
		int g = Integer.parseInt(hexRGB.substring(2, 4), 16);
		int b = Integer.parseInt(hexRGB.substring(4, 6), 16);
		return new Color(r, g, b);
	}

	/**
		Generate a hex color string from a Color object. Useful for
		outputting into HTML pages.
		
		@param color The color object
		@return The hexadecimal representation of the color
	*/
	public static String toHex(Color color)
	{
		StringBuilder sb = new StringBuilder(6);
		int r = color.getRed();
		if (r < 16) sb.append("0");
		sb.append(Integer.toHexString(r));

		int g = color.getGreen();
		if (g < 16) sb.append("0");
		sb.append(Integer.toHexString(g));
	
		int b = color.getBlue();
		if (b < 16) sb.append("0");
		sb.append(Integer.toHexString(b));
		return sb.toString();
	}
	
	private Color minColor;
	private Color midColor;
	private Color maxColor;
	private float minValue;
	private float midValue;
	private float maxValue;
	private boolean log;
	
	/**
		The RGB components of the min color. [0] = red; [1] = green; [2] = blue
	*/
	private final int[] offsetMin;

	/**
		The RGB components of the midpoint color. [0] = red; [1] = green; [2] = blue
	*/
	private final int[] offsetMid;
	
	/**
		The difference of the RGB components of the min and mid
		colors. [0] = red; [1] = green; [2] = blue
	*/
	private final int[] deltaMinMid;

	/**
		The difference of the RGB components of the mid and max
		colors. [0] = red; [1] = green; [2] = blue
	*/
	private final int[] deltaMidMax;
	
	/**
		The inverted difference between the mid and min values:
		1 / (midValue - minValue)
	*/
	private final float rangeMinMid;

	/**
		The inverted difference between the max and mid values:
		1 / (maxValue - midValue)
	*/
	private final float rangeMidMax;
	
	/**
		Create a new color generator.
		@param minColor The color for the minimum endpoint
		@param midColor The color for the midpoint
		@param maxColor The color for the maximum endpoint
		@param minValue The value for the minimum endpoint
		@param midValue The value for the midpoint
		@param maxValue The value for the maximum endpoint
		@param log If values should be logarithmised before beeing assigned a 
			color or not. Note the values in this constructor are not affected by this
			setting.
		@throws InvalidDataException If not the minValue &lt;= midValue &lt;= maxValue 
	*/
	public ColorGenerator(Color minColor, Color midColor, Color maxColor, 
		float minValue, float midValue, float maxValue, boolean log)
	{
		// Check that minValue <= midValue <= maxValue
		if (minValue > midValue)
		{
			throw new NumberOutOfRangeException("minValue", minValue, midValue, true);
		}
		if (midValue > maxValue)
		{
			throw new NumberOutOfRangeException("midValue", midValue, maxValue, true);			
		}
		
		// Store paramters
		this.minColor = minColor;
		this.midColor = midColor;
		this.maxColor = maxColor;
		this.minValue = minValue;
		this.midValue = midValue;
		this.maxValue = maxValue;
		this.log = log;
		
		// Precalculate some settings
		this.offsetMin = new int[3];
		offsetMin[0] = minColor.getRed();
		offsetMin[1] = minColor.getGreen();
		offsetMin[2] = minColor.getBlue();

		this.offsetMid = new int[3];
		offsetMid[0] = midColor.getRed();
		offsetMid[1] = midColor.getGreen();
		offsetMid[2] = midColor.getBlue();
		
		this.deltaMinMid = new int[3];
		for (int i = 0; i < 3; ++i)
		{
			deltaMinMid[i] = offsetMid[i] - offsetMin[i];
		}
		
		this.deltaMidMax = new int[3];
		deltaMidMax[0] = maxColor.getRed() - offsetMid[0];
		deltaMidMax[1] = maxColor.getGreen() - offsetMid[1];
		deltaMidMax[2] = maxColor.getBlue() - offsetMid[2];
		
		this.rangeMinMid = midValue == minValue ? 0 : 1 / (midValue - minValue);
		this.rangeMidMax = maxValue == midValue ? 0 : 1 / (maxValue - midValue);
	}
	
	/**
		Get the color for the value. If the log flag was true when creating the
		generator the value is first logarithmised. Note that negative values
		can't be logarithmised so null is returned in those cases.
		<p>
		If the value is lower than the min endpoint the min endpoint color
		is returned. If the value is higher than the max endpoint the
		max endpoint color is returned. Otherwise a new color is calculated
		based on the difference between the value and the given points.
		
		@param value The value
		@return A color object or null if no color could be generated
	*/
	public Color getColor(float value)
	{
		if (Float.isNaN(value)) return null;
		
		if (log)
		{
			if (value <= 0) return null;
			value = ((float)Math.log(value) * INVLN2);
		}
		
		Color newColor = null;
		if (value >= maxValue)
		{
			newColor = maxColor;
		}
		else if (value <= minValue)
		{
			newColor = minColor;
		}
		else
		{
			float factor;
			int[] delta;
			int[] offset;
			if (value >= midValue)
			{
				delta = deltaMidMax;
				offset = offsetMid;
				factor = (value - midValue) * rangeMidMax;
			}
			else
			{
				delta = deltaMinMid;
				offset = offsetMin;
				factor = (value - minValue) * rangeMinMid;
			}
			int r = (int)(offset[0] + factor * delta[0]);
			int g = (int)(offset[1] + factor * delta[1]);
			int b = (int)(offset[2] + factor * delta[2]);
			newColor = new Color(r, g, b);
		}
		return newColor;
	}
	
}
