/*
	$Id: Enumeration.java 7333 2017-04-11 06:43:47Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.LinkedList;

import net.sf.basedb.core.Nameable;

/**

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-04-11 08:43:47 +0200 (tis, 11 apr. 2017) $
*/
public class Enumeration<K, V>
	implements Serializable
{

	private static final long serialVersionUID = 1222952912319368574L;

	/**
		Create a new enumeration from a list of basic items.
		The key of the enumeration is the id of each item
		and the value is the name.
		@param items A collection with items
		@param noneOption If not null, this options will be inserted first in the
			enumeration with an empty string as the key
	*/
	public static Enumeration<String, String> fromItems(Collection<? extends Nameable> items, String noneOption)
	{
		return fromItems(items, noneOption, "");
	}

	/**
		Create a new enumeration from a list of basic items.
		The key of the enumeration is the id of each item
		and the value is the name.
		@param items A collection with items
		@param noneOption If not null, this options will be inserted first in the
			enumeration with an empty string as the key
		@param noneValue Value for the 'none' option (empty string by default)
		@since 3.11
	*/
	public static Enumeration<String, String> fromItems(Collection<? extends Nameable> items, String noneOption, String noneValue)
	{
		Enumeration<String, String> e = new Enumeration<String, String>();
		if (noneOption != null)
		{
			e.add(noneValue == null ? "" : noneValue, noneOption);
		}
		for (Nameable item : items)
		{
			e.add(Integer.toString(item.getId()), item.getName());
		}
		e.lock();
		return e;
	}

	/**
		Create a new enumeration from a list strings. The key
		and value are the same for each entry. A null value in
		the array will be translated to key="" and value="none"
		@since 3.11
	*/
	public static Enumeration<String, String> from(String... values)
	{
		Enumeration<String, String> e = new Enumeration<String, String>();
		for (String v : values)
		{
			if (v == null)
			{
				e.add("", "none");
			}
			else
			{
				e.add(v, v);
			}
		}
		e.lock();
		return e;
	}

	
	private final List<Entry<K, V>> entries;
	private boolean locked;

	public Enumeration()
	{
		entries = new LinkedList<Entry<K, V>>();
		locked = false;
	}
	
	public boolean isLocked()
	{
		return locked;
	}
	
	public void lock()
	{
		this.locked = true;
	}
	
	public int size()
	{
		return entries.size();
	}
	
	public int add(K key, V value)
	{
		if (locked) throw new UnsupportedOperationException("add; The enumeration has been locked for modifications.");
		int index = size();
		entries.add(new Entry<K, V>(key, value));
		return index;
	}
	
	public K getKey(int index)
	{
		return entries.get(index).getKey();
	}
	
	public V getValue(int index)
	{
		return entries.get(index).getValue();
	}
	
	public Entry<K, V> getEntry(int index)
	{
		return entries.get(index);
	}
	
	/**
		Get a list with all entries that equals the given key.
		@return A list which may be empty if no entries are found
		@since 3.2.4
	*/
	public List<Entry<K, V>> getEntriesByKey(K key)
	{
		List<Entry<K, V>> result = new LinkedList<Enumeration.Entry<K,V>>();
		for (Entry<K, V> e : entries)
		{
			if (EqualsHelper.equals(e.key, key)) result.add(e);
		}
		return result;
	}
	
	/**
		Get a list with all entries that equals the given value.
		@return A list which may be empty if no entries are found
		@since 3.2.4
	*/
	public List<Entry<K, V>> getEntriesByValue(V value)
	{
		List<Entry<K, V>> result = new LinkedList<Enumeration.Entry<K,V>>();
		for (Entry<K, V> e : entries)
		{
			if (EqualsHelper.equals(e.value, value)) result.add(e);
		}
		return result;
	}
	
	public void remove(int index)
	{
		if (locked) throw new UnsupportedOperationException("remove; The enumeration has been locked for modifications.");
		entries.remove(index);
	}
	
	/**
		This method will sort the enumration on the keys using 
		{@link Collections#sort(List, Comparator)}. 
		The comparator will try to use the keys 'compareTo' method 
		if it implements the Comparable interface
		otherwilse it will treat the key as a String.
	*/
	public void sortKeys()
	{
		Collections.sort(entries, new Comparator<Entry<K, V>>()
		{
			@Override
			@SuppressWarnings("unchecked")
			public int compare(Entry<K, V> o1, Entry<K, V> o2)
			{
				K key1 = o1.getKey();
				K key2 = o2.getKey();
				if (key1 instanceof Comparable)
				{
					return ((Comparable<K>) key1).compareTo(key2);
				}
				else
				{
					return key1.toString().compareTo(key2.toString());
				}
			}
		});
	}
	
	/**
		This method will sort the enumration on the values using 
		{@link Collections#sort(List, Comparator)}. 
		The comparator will try to use the values 'compareTo' method 
		if it implements the Comparable interface
		otherwilse it will treat the value as a String.
	*/	
	public void sortValues()
	{
		Collections.sort(entries, new Comparator<Entry<K, V>>()
		{
			@Override
			@SuppressWarnings("unchecked")
			public int compare(Entry<K, V> o1, Entry<K, V> o2)
			{
				V value1 = o1.getValue();
				V value2 = o2.getValue();
				if (value1 instanceof Comparable)
				{
					return ((Comparable<V>) value1).compareTo(value2);
				}
				else
				{
					return value1.toString().compareTo(value2.toString());
				}
			}
		});
	}
	public static class Entry<K, V>
	{
		private final K key;
		private final V value;
		
		private Entry(K key, V value)
		{
			this.key = key;
			this.value = value;
		}
		
		public K getKey()
		{
			return key;
		}
		
		public V getValue()
		{
			return value;
		}
		
	}

	
	
}
