/**
	$Id: InputStreamSplitter.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
	This class can be used to split an input stream into one or more 
	additional output streams. When data is read from this input stream
	the bytes are also written to the specified output streams with one of the
	{@link java.io.OutputStream#write(int)} or 
	{@link java.io.OutputStream#write(byte[], int, int)} methods.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (fre, 13 aug. 2010) $
*/
public class InputStreamSplitter
	extends InputStream
{

	private int numRead;
	
	/**
		The source input stream to read from.
	*/
	private final InputStream in;
	
	/**
		Streams to copy the data to.
	*/
	private final OutputStream[] copyTo;
	
	/**
		If the {@link #skip(long)} method should copy
		to the output streams or no.
	*/
	private final boolean copySkipped;
	
	/**
		If the remainder of the input stream should be copied
		if it is closed before the end.
	*/
	private final boolean readToEnd;
	
	/**
		Create a new input stream splitter. Skipped bytes are not copied
		and the remainder of the stream is not copied if closed.
		@param in The input stream to read from
		@param copyTo The output streams to copy data to
	*/
	public InputStreamSplitter(InputStream in, OutputStream... copyTo)
	{
		this(in, false, false, copyTo);
	}

	/**
		Create a new input stream splitter. Skipped bytes are not copied
		and the remainder of the stream is not copied if closed.
		@param in The input stream to read from. Null is not allowed
	 	@param copySkipped  TRUE if skipped bytes should be copied, FALSE otherwise
	 	@param readToEnd If the remainder of the input stream should be copied
			if it is closed before the end
		@param copyTo The output streams to copy data to. Null is not allowed
		@since 2.4
	*/
	public InputStreamSplitter(InputStream in, boolean copySkipped, boolean readToEnd, OutputStream... copyTo)
	{
		if (in == null) throw new NullPointerException("in");
		if (copyTo == null) throw new NullPointerException("copyTo");
		this.in = in;
		this.copyTo = copyTo;
		this.copySkipped = copySkipped;
		this.readToEnd = readToEnd;
	}
	
	/*
		From the InputStream class
		-------------------------------------------
	*/
	@Override
	public int available() 
		throws IOException
	{
		return in.available();
	}
	@Override
	public void close() 
		throws IOException
	{
		if (readToEnd)
		{
			FileUtil.read(this);
		}
		in.close();
		for (OutputStream out : copyTo)
		{
			out.flush();
		}
	}
	@Override
	public synchronized void mark(int readlimit)
	{}
	@Override
	public boolean markSupported()
	{
		return false;
	}
	@Override
	public int read() 
		throws IOException
	{
		int result = in.read();
		if (result >= 0) 
		{
			++numRead;
			copy(result);
		}
		return result;
	}
	@Override
	public int read(byte[] b, int off, int len) 
		throws IOException
	{
		int result = in.read(b, off, len);
		if (result > 0) 
		{
			numRead += result;
			copy(b, off, result);
		}
		return result;
	}
	@Override
	public int read(byte[] b) 
		throws IOException
	{
		return read(b, 0, b.length);
	}
	@Override
	public synchronized void reset() 
		throws IOException
	{
		throw new IOException("reset/mark is not supported");
	}
	@Override
	public long skip(long n)
		throws IOException
	{
		long result = 0;
		if (copySkipped)
		{
			// This will call the read(byte[], int, int) method
			result = super.skip(n);
		}
		else
		{
			result = in.skip(n);
		}
		return result;
	}
	// -------------------------------------------

	/**
		Get the number of bytes that has been read or skipped 
		from the input stream.
		@since 2.4
	*/
	public long getNumRead()
	{
		return numRead;
	}
	
	/**
		Write data to all output streams.
	*/
	private void copy(int b)
		throws IOException
	{
		for (OutputStream out : copyTo)
		{
			out.write(b);
		}
	}
	
	/**
		Write data to all output streams.
	*/
	private void copy(byte[] b, int off, int len)
		throws IOException
	{
		for (OutputStream out : copyTo)
		{
			out.write(b, off, len);
		}
	}

}
