/*
	$Id: JarClassLoader.java 7718 2019-05-22 07:56:58Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Vector;
import java.util.jar.JarFile;
import java.util.jar.JarEntry;
import java.util.jar.Manifest;

import java.util.jar.Attributes;
import java.util.jar.Attributes.Name;
import java.io.File;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.net.MalformedURLException;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.InvalidDataException;

/**
	A class loader implementation that loads classes from JAR files.
	Each JAR file requires a separate instance of this class. Use
	the {@link #getInstance(String)} method to get an existing or create
	a new instance for a specific JAR file. If the classes in the JAR file
	requires other classes in another JAR file to work, the paths to those
	JAR files must be listed in the <code>Class-Path</code> attribute in the
	<code>META-INF/MANIFEST.MF</code> file. For example:
	<pre class="code">
Manifest-Version: 1.0
Class-Path: OtherJarPlugin.jar
</pre> 

	If more than one JAR is needed separate them with one or more spaces.
	Note! It is only the <code>Class-Path</code> entry for the JAR file passed
	to the {@link #getInstance(String)} method that is checked. The manifest
	file is not checked for the other JARs.
	<p>
	This class loader will by default first look in the specified JAR files 
	for a class or resource and only if not found delegate to the parent class
	loader. This behaviour can be changed by calling {@link #setDelegateFirst(boolean)}
	or by setting <code>X-Delegate-First : true</code> in the <code>MANIFEST.MF</code>
	file in the main JAR file.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2019-05-22 09:56:58 +0200 (ons, 22 maj 2019) $
*/
public final class JarClassLoader
	extends ClassLoader
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(JarClassLoader.class);
	
	/**
		A map of all loaded class loaders.
	*/
	private static final HashMap<String, JarClassLoader> classLoaders = 
		new HashMap<String, JarClassLoader>();

	/**
		Get a class loader for the specified JAR file. If a class loader for
		the specified file already exists, that class loader is returned, otherwise
		a new class loader is created.
		
		@param jarPath The path to a JAR file
		@return A class loader
		@throws IOException If the jar file can't be loaded
	*/
	public static final ClassLoader getInstance(String jarPath)
		throws IOException
	{
		return getInstance(jarPath, false);
	}

	/**
		Get a class loader for the specified JAR file, optionally unloading an the old
		one if the JAR file has been modified. A new class loader is created if no
		class loader exists or if <code>autoUnload</code> is <code>true</code> and the JAR file 
		has changed since the existing class loader was created. 
		
		@param jarPath The path to a JAR file
		@param autoUnload If TRUE the old class loaded will automatically be unloaded if the
			JAR file or any one it depends on (listed in the Class-Path attribute in the manifest file)
			has been modified (if the timestamp and/or size) is different
		@return A class loader
		@throws IOException If the jar file can't be loaded
		@since 2.4
	*/
	public static final ClassLoader getInstance(String jarPath, boolean autoUnload)
		throws IOException
	{
		JarClassLoader cl = classLoaders.get(jarPath);
		if (cl != null && autoUnload)
		{
			// Check if JAR file exists and has same timestamp and size
			if (cl.hasChanged(true)) cl = null;
		}
		if (cl == null)
		{
			cl = new JarClassLoader(jarPath);
			synchronized (classLoaders)
			{
				classLoaders.put(jarPath, cl);
			}
		}
		return cl;
	}
	
	/**
	 	Get a new class loader for the specified jar file.  
	  	@param jarPath The path to the jar file
	  	@return A class loader object
	  	@throws IOException If the jar file can't be loaded
	*/
	public static final ClassLoader newInstance(String jarPath)
		throws IOException
	{
		return new JarClassLoader(jarPath);
	}
	
	/**
		Check if a class loader for the given JAR file exists.
		@param jarPath The path to the JAR file
		@return TRUE if a class loader exists, FALSE otherwise
	*/
	public static final boolean exists(String jarPath)
	{
		return classLoaders.containsKey(jarPath);
	}
	
	/**
		The main JAR file to load classes from.
	*/
	private final File mainJarFile;

	/**
		Contains mappings from class names to the JAR file in which 
		the class implementation can be found.
	*/
	private final Map<String, List<File>> classPath;

	/**
		If we should delegate to the parent class loader first.
	*/
	private boolean delegateFirst;
	
	/**
		For keeping track of jar files listed by Class-Path entry
		in the manifest file.
	*/
	private final Map<File, JarInfo> jarFiles;
		
	/**
		Class loaders for extensions.
	*/
	private final List<JarClassLoaderProxy> proxyLoaders;
	
	/**
		The timestamp of the JAR file.
	*/
	private final long jarTimeStamp;
	
	/**
		The size of the JAR file.
	*/
	private final long jarSize;
	
	
	/**
		Create a new JAR file class loader.
		@param jarPath The path to the JAR file
		@throws InvalidDataException If the JAR file can't be loaded
		@throws IOException If there is another IO-related error
	*/
	private JarClassLoader(String jarPath)
		throws IOException
	{
		super(Thread.currentThread().getContextClassLoader());
		log.debug("Creating JarClassLoader: file=" + jarPath);
		mainJarFile = new File(jarPath);
		if (!mainJarFile.exists() || !mainJarFile.isFile()) 
		{
			throw new InvalidDataException("JAR file not found: " + jarPath);
		}
		this.jarTimeStamp = mainJarFile.lastModified();
		this.jarSize = mainJarFile.length();
		this.delegateFirst = false;
		classPath = new HashMap<String, List<File>>();
		jarFiles = new HashMap<File, JarInfo>();
		proxyLoaders = new ArrayList<JarClassLoaderProxy>();
		loadJarFile(mainJarFile, true);
	}

	@Override
	public String toString() 
	{
		return super.toString() + "[" + mainJarFile + "]";
	}

	/*
		From the ClassLoader class
		-------------------------------------------
	*/
	@Override
	protected Class<?> findClass(String name)
		throws ClassNotFoundException
	{
		log.debug("findClass: " + name);
		List<File> files = classPath.get(classNameToPath(name));
		if (files == null || files.isEmpty())
		{
			throw new ClassNotFoundException(name);
		}
		File file = files.get(0);
		byte[] b = loadClassData(file, name);
		int i = name.lastIndexOf('.');
		if (i > 0)
		{
			String packageName = name.substring(0, i);
			if (!isPackageDefined(packageName))
			{
				JarInfo info = jarFiles.get(file);
				Manifest mf = info != null ? info.manifest : null;
				definePackage(packageName, mf);
			}
		}
		Class<?> clazz = defineClass(name, b, 0, b.length);
		return clazz;
	}
	
	// Safety limit to avoid endless loops in case parent class loaders are incorrectly linked
	private static final int MAX_PARENTS = 20;
	private boolean isPackageDefined(String name)
	{
		ClassLoader c = this;
		for (int i = 0; c != null && i < MAX_PARENTS; i++)
		{
			if (c.getDefinedPackage(name) != null) return true;
			c = c.getParent();
		}
		return false;
	}
	
	@Override
	protected URL findResource(String name)
	{
		log.debug("findResource: " + name);
		URL url = null;
		try
		{
			List<File> files = classPath.get(name);
			if (files != null && files.size() > 0)
			{
				if (log.isDebugEnabled()) log.debug("Found: " + files);
				String newUrl = "jar:"+files.get(0).toURI().toURL()+"!/"+name;
				url = new URL(newUrl);
			}
			else
			{
				log.debug("Not found: " + name);
			}
		}
		catch (MalformedURLException ex)
		{
			ex.printStackTrace();
		}
		return url;
	}
	
	@Override
	protected Enumeration<URL> findResources(String name)
	{
		log.debug("findResources: " + name);
		Enumeration<URL> url = null;
		try
		{
			List<File> files = classPath.get(name);
			if (files != null && files.size() > 0)
			{
				if (log.isDebugEnabled()) log.debug("Found: " + files);
				Vector<URL> v = new Vector<URL>(files.size());
				for (File f : files)
				{
					String newUrl = "jar:"+f.toURI().toURL()+"!/"+name;
					v.add(new URL(newUrl));
				}
				url = v.elements();
			}
			else
			{
				log.debug("Not found: " + name);
			}
		}
		catch (MalformedURLException ex)
		{
			ex.printStackTrace();
		}
		return url;
	}
	
	@Override
	protected Class<?> loadClass(String name, boolean resolve)
		throws ClassNotFoundException
	{
		if (log.isDebugEnabled()) log.debug("loadClass: " + name);
		
		// 1. Check if the class has already been loaded
		Class<?> c = findLoadedClass(name);
		if (log.isDebugEnabled()) log.debug("loaded class (" + name + "): " + c);
		
		ClassLoader system = ClassLoader.getSystemClassLoader();
		ClassLoader parent = getParent();
		
		// 2. Check the system class loader
		if (c == null && parent != system) 
		{
			c = loadClassInternal(system, name);
			if (log.isDebugEnabled()) log.debug("system class (" + name + "): " + c);
		}
		
		// 3. Delegate to parent class loader
		if (c == null && delegateFirst) 
		{
			c = loadClassInternal(parent, name);
			if (log.isDebugEnabled()) log.debug("parent class (" + name + "): " + c);
		}

		// 4. Look for the class in this class path
		if (c == null)
		{
			try
			{
				c = findClass(name);
			}
			catch (ClassNotFoundException ex)
			{}
			if (log.isDebugEnabled()) log.debug("my class (" + name + "): " + c);
		}
		
		// 5. Side-load from proxied JAR files (eg. other extensions)
		Iterator<JarClassLoaderProxy> it = proxyLoaders.iterator();
		while (c == null && it.hasNext())
		{
			JarClassLoaderProxy proxyLoader = it.next();
			c = proxyLoader.findClass(name);
			if (log.isDebugEnabled()) log.debug("proxy class (" + proxyLoader.jarPath + "; " + name + "): " + c);
		}
		
		// 6. Delegate to parent class loader if it hasn't been done
		if (c == null && !delegateFirst) 
		{
			c = loadClassInternal(parent, name);
			if (log.isDebugEnabled()) log.debug("parent class (" + name + "): " + c);
		}
		
		if (c == null)
		{
			// Class not found
			throw new ClassNotFoundException(name);
		}
		
		if (resolve) resolveClass(c);
		return c;
	}
	
	@Override
	public URL getResource(String name)
	{
		if (log.isDebugEnabled()) log.debug("getResource: " + name);
		URL url = null;
		ClassLoader parent = getParent();
		
		// 1. Check with parent loader if delegating
		if (delegateFirst)
		{
			if (parent != null) url = parent.getResource(name);
		}
		
		// 2. Check with this class loader
		if (url == null) url = findResource(name);

		// 3. Side-load from proxied JAR files (eg. other extensions)
		if (url == null)
		{
			Iterator<JarClassLoaderProxy> it = proxyLoaders.iterator();
			while (url == null && it.hasNext())
			{
				JarClassLoaderProxy proxyLoader = it.next();
				url = proxyLoader.findResource(name);
			}
		}

		// 3. Check with parent loader if not done so already
		if (url == null && !delegateFirst)
		{
			if (parent != null) url = parent.getResource(name);
		}
		return url;
	}
	
	@Override
	public Enumeration<URL> getResources(String name)
		throws IOException
	{
		if (log.isDebugEnabled()) log.debug("getResources: " + name);
		ClassLoader parent = getParent();
		Vector<URL> resources = new Vector<URL>();
		
		Enumeration<URL> parentResources = null;
		if (parent != null) 
		{
			parentResources = parent.getResources(name);
		}
		Enumeration<URL> myResources = findResources(name);

		if (delegateFirst) addResources(resources, parentResources);
		addResources(resources, myResources);
		Iterator<JarClassLoaderProxy> it = proxyLoaders.iterator();
		while (it.hasNext())
		{
			JarClassLoaderProxy proxyLoader = it.next();
			addResources(resources, proxyLoader.findResources(name));
		}

		if (!delegateFirst) addResources(resources, parentResources);
		
		return resources.elements();
	}
	// -------------------------------------------

	/**
		If the class loader should delegate to the parent class loader
		before trying to find the class by it's own.
		@param delegateFirst TRUE to delegate to parent first, FALSE to
			only delegate if the class is not found
		@since 2.13
	*/
	public void setDelegateFirst(boolean delegateFirst)
	{
		this.delegateFirst = delegateFirst;
	}
	
	/**
		If this class loader delegates to the parent class loader
		before or after trying to find the class by itself.
		@since 2.13
	*/
	public boolean getDelegateFirst()
	{
		return delegateFirst;
	}
	
	/**
		Check if the JAR file this class loader loads is classes
		from has changed since this class loader was created. 
		@param checkSecondary TRUE to also check secondary JAR files
			listed in the Class-Path entry of the manifest file
		@since 2.8
	*/
	public boolean hasChanged(boolean checkSecondary)
	{
		if (log.isDebugEnabled())
		{
			log.debug("Checking if JAR has changed: " + mainJarFile);
		}
		boolean isSame = mainJarFile.exists() && 
			mainJarFile.lastModified() == jarTimeStamp && mainJarFile.length() == jarSize;
		if (!isSame) 
		{
			log.debug("Main JAR file has changed: " + mainJarFile);
			return true;
		}
		if (checkSecondary)
		{
			for (JarInfo info : jarFiles.values())
			{
				File jarFile = info.jarFile;
				if (jarFile.equals(mainJarFile)) continue; // We have already checked this one
				
				if (log.isDebugEnabled())
				{
					log.debug("Checking if secondary JAR has changed: " + jarFile);
				}
				if (jarFile.exists())
				{
					isSame = info.existed && 
						jarFile.lastModified() == info.lastModified && 
						jarFile.length() == info.size;
				}
				else
				{
					isSame = !info.existed;
				}
				if (!isSame) 
				{
					log.debug("Secondary JAR file has changed: " + jarFile);
					return true;
				}
			}
			for (JarClassLoaderProxy proxy : proxyLoaders)
			{
				if (log.isDebugEnabled())
				{
					log.debug("Checking if proxy JAR has changed: " + proxy.jarPath);
				}
				isSame = !proxy.hasChanged();
				if (!isSame) 
				{
					log.debug("Proxy JAR has changed: " + proxy.jarPath);
					return true;
				}
			}
		}
		
		if (log.isDebugEnabled()) log.debug("JAR has not changed: " + mainJarFile);
		return false;
	}
	
	/**
		Convert a class name to a file path. Ie. replace dots with slahses and add .class
		to the end: net.sf.basedb.util.JarClassLoader -&gt; net/sf/basedb/util/JarClassLoader.class
	*/
	private String classNameToPath(String className)
	{
		return className.replaceAll("\\.", "/")+".class";
	}
	
	/**
		Open the specified JAR file, list all entries and put them in the 
		{@link #classPath} mapping.
		@param file The JAR file to open
		@param followClassPath If the MANIFEST file should be checked for
			a <code>Class-Path</code> entry that lists other JAR files
			that also should be checked.
		@throws IOException If there is an error reading the JAR files
	*/
	private void loadJarFile(File file, boolean followClassPath)
		throws IOException
	{
		log.debug("loadJarFile: file=" + file);
		JarInfo info = new JarInfo(file);
		jarFiles.put(file, info);
		if (!file.exists() || !file.isFile())
		{
			log.warn("File not found: " + file);
			return;
		}
		JarFile jarFile = new JarFile(file);
		try
		{
			Manifest manifest = jarFile.getManifest();
			info.manifest = manifest;
			Enumeration<JarEntry> entries = jarFile.entries();
			while (entries.hasMoreElements())
			{
				JarEntry jarEntry = entries.nextElement();
				if (jarEntry.isDirectory())
				{
					/* #### CONTINUE-STATEMENT #### */
					continue;
				}
				String name = jarEntry.getName();
				log.trace("Adding: " + name);
				List<File> list = classPath.get(name);
				if (list == null)
				{
					list = new ArrayList<File>();
					classPath.put(name, list);
				}
				list.add(file);
			}
			if (followClassPath && manifest != null)
			{
				File directory = file.getParentFile();
				Attributes attr = manifest.getMainAttributes();
				if (attr != null)
				{
					String cp = attr.getValue(Attributes.Name.CLASS_PATH);
					log.debug("Class path of JAR file '" + file + "': " + cp);
					if (cp != null)
					{
						String[] cps = cp.split("\\s+");
						for (int i = 0; i < cps.length; ++i)
						{
							if (cps[i] != null && !cps[i].trim().equals(""))
							{
								File classPathFile = new File(directory, cps[i]);
								if (!classPathFile.exists())
								{
									log.debug("Referenced JAR file doesn't exist: '" + cps[i]);
									// See if the requested file exists under META-INF
									JarEntry entry = jarFile.getJarEntry("META-INF/" + cps[i]);
									if (entry != null)
									{
										log.debug("Extracting JAR from META-INF/" + cps[i] + " to jar.cache");
										// Extract it to userfiles directory (if we have to)
										classPathFile = new File(Application.getUserFilesDirectory(), 
												"jar.cache/" + file.getName() + "/" + classPathFile.getName());
										boolean exists = classPathFile.exists();
										boolean sameTime = exists && classPathFile.lastModified() == entry.getTime();
										boolean sameSize = exists && classPathFile.length() == entry.getSize();
										if (!exists || !sameTime || !sameSize)
										{
											classPathFile.getParentFile().mkdirs();
											InputStream in = null;
											OutputStream out = null;
											try
											{
												in = jarFile.getInputStream(entry);
												out = FileUtil.getOutputStream(classPathFile);
												FileUtil.copy(in, out);
											}
											finally
											{
												FileUtil.close(in);
												FileUtil.close(out);
											}
										}
									}
									loadJarFile(classPathFile, false);
								}
								else
								{
									proxyLoaders.add(new JarClassLoaderProxy(classPathFile.getAbsolutePath()));
								}
							}
						}
					}
					if (file == mainJarFile)
					{
						delegateFirst = Values.getBoolean(attr.getValue("X-Delegate-First"), delegateFirst);
					}
				}
			}
		}
		finally
		{
			FileUtil.close(jarFile);
		}
	}
	
	/**
		Load the byte[] of the given class. 
		
		@param file The JAR file in which the class implmentation is located
		@param name The name of the class using regular naming convention,
			ie. net.sf.basedb.util.JarClassLoader
		@throws ClassNotFoundException If the class can't be loaded
	*/
	private byte[] loadClassData(File file, String name)
		throws ClassNotFoundException
	{
		log.debug("loadClassData: " + name + " from file: " + file);
		InputStream in = null;
		JarFile jarFile = null;
		try
		{
			jarFile = new JarFile(file);
			// Get the jar entry and open an input stream to read the file
			JarEntry jarEntry = jarFile.getJarEntry(classNameToPath(name));
			if (jarEntry == null)
			{
				throw new ClassNotFoundException(name+" in file "+file);
			}
			in = jarFile.getInputStream(jarEntry);
			int classSize = (int)jarEntry.getSize();
			byte[] b = new byte[classSize];
			int totalRead = 0;
			while (totalRead < classSize)
			{
				int numRead = in.read(b, totalRead, b.length - totalRead);
				if (numRead == -1) break;
				totalRead += numRead;
				log.debug("loadClassData: size=" + classSize + "; loaded=" + totalRead);
			}
			if (totalRead != classSize) 
			{
				throw new ClassFormatError("Could not read the complete class for " + 
					name + ": " + totalRead + " bytes; expected " + classSize + " bytes");
			}
			return b;
		}
		catch (Throwable ex)
		{
			throw new ClassNotFoundException(name, ex);
		}
		finally
		{
			FileUtil.close(in);
			FileUtil.close(jarFile);
		}
	}
	/**
		Define the package with the given name using information from a 
		manifest for vendor, title and version. The package should not
		already exist.
		@param name The name of the package
		@param mf An optional manifest
		@return A package
		@since 2.16
	*/
	private Package definePackage(String name, Manifest mf)
	{
		String specTitle = null;
		String specVersion = null;
		String specVendor = null;
		String implTitle = null;
		String implVersion = null;
		String implVendor = null;
		if (mf != null)
		{
			// Check if attributes are defined for the given package path
			String path = name.replace('.', '/') + "/";
			Attributes attr = mf.getAttributes(path);
			if (attr != null) 
			{
				specTitle = attr.getValue(Name.SPECIFICATION_TITLE);
				specVersion = attr.getValue(Name.SPECIFICATION_VERSION);
				specVendor = attr.getValue(Name.SPECIFICATION_VENDOR);
				implTitle = attr.getValue(Name.IMPLEMENTATION_TITLE);
				implVersion = attr.getValue(Name.IMPLEMENTATION_VERSION);
				implVendor = attr.getValue(Name.IMPLEMENTATION_VENDOR);
			}
			// Check global attributes
			attr = mf.getMainAttributes();
			if (attr != null) 
			{
			    if (specTitle == null) specTitle = attr.getValue(Name.SPECIFICATION_TITLE);
			    if (specVersion == null) specVersion = attr.getValue(Name.SPECIFICATION_VERSION);
			    if (specVendor == null) specVendor = attr.getValue(Name.SPECIFICATION_VENDOR);
			    if (implTitle == null) implTitle = attr.getValue(Name.IMPLEMENTATION_TITLE);
			    if (implVersion == null) implVersion = attr.getValue(Name.IMPLEMENTATION_VERSION);
			    if (implVendor == null) implVendor = attr.getValue(Name.IMPLEMENTATION_VENDOR);
			}
		}
		return definePackage(name, specTitle, specVersion, specVendor, implTitle, implVersion, implVendor, null);
	}
	
	private Class<?> loadClassInternal(ClassLoader loader, String name)
	{
		Class<?> c = null;
		if (loader != null)
		{
			try
			{
				c = loader.loadClass(name);
			}
			catch (ClassNotFoundException ex)
			{}
		}
		return c;
	}
	
	private void addResources(Vector<URL> resources, Enumeration<URL> e)
	{
		if (e == null) return;
		while (e.hasMoreElements())
		{
			resources.add(e.nextElement());
		}
	}
	
	static class JarInfo
	{
		final File jarFile;
		Manifest manifest;
		final boolean existed;
		final long lastModified;
		final long size;
		
		JarInfo(File jarFile)
		{
			this.jarFile = jarFile;
			this.existed = jarFile.exists();
			this.size = jarFile.length();
			this.lastModified = jarFile.lastModified();
		}
	}
	
	/**
		A proxy class loader is typically needed when one extension
		depends on another extension. The proxy is needed since
		we want to use lazy initialization of the other class loaders.
		They may not be needed immediately. The proxy will also make it
		possible to avoid infinite recursion in case two extensions 
		depend on each other.
		<p>
		
		Once a real class loader has been aquired it will be kept 
		until the main class loader is discared. Note that the 
		{@link JarClassLoader#hasChanged(boolean)} is not used for 
		automatic reloading since that may cause class-cast exceptions.
		<p>
		
		The proxy is also used to limit the search for classes to 
		classes that are directly handled directly by the other class 
		loader. The system and parent class loaders are not searched again
		since they are already searched in the main class loader.

		@since 3.10
	*/
	static class JarClassLoaderProxy
	{
		
		final String jarPath;
		private JarClassLoader loader;
		private boolean isInitialized;
		
		/**
			Creates a proxy class loader for the given JAR file.
		*/
		JarClassLoaderProxy(String jarPath)
		{
			this.jarPath = jarPath;
			this.isInitialized = false;
		}
		
		/**
			Initialize the proxy.
		*/
		private synchronized void init()
		{
			if (isInitialized) return;
			if (log.isDebugEnabled()) log.debug("JarClassLoaderProxy.init[" + jarPath + "]");
			
			try
			{
				loader = (JarClassLoader)JarClassLoader.getInstance(jarPath, true);
			}
			catch (IOException ex)
			{}
			isInitialized = true;
		}
		
		
		Class<?> findClass(String name)
		{
			if (log.isDebugEnabled()) log.debug("JarClassLoaderProxy.findClass[" + jarPath + "]: " + name);
			if (!isInitialized) init();
			
			Class<?> c = null;
			if (loader != null)
			{
				try
				{
					// 1. Check if the class has already been loaded
					c = loader.findLoadedClass(name);
					if (log.isDebugEnabled()) log.debug("JarClassLoaderProxy.findClass[" + jarPath + "]: loaded class (" + name + "): " + c);
					
					// 2. Otherwise try to load the class without searching parent or system class loaders
					if (c == null)
					{
						c = loader.findClass(name);
						if (log.isDebugEnabled()) log.debug("JarClassLoaderProxy.findClass[" + jarPath + "]: found class (" + name + "): " + c);
					}
				}
				catch (ClassNotFoundException ex)
				{}
			}
			return c;
		}
		
		URL findResource(String name)
		{
			if (log.isDebugEnabled()) log.debug("JarClassLoaderProxy.findResource[" + jarPath + "]: " + name);
			if (!isInitialized) init();
			
			URL url = null;
			if (loader != null)
			{
				url = loader.findResource(name);
			}
			return url;
		}
		
		Enumeration<URL> findResources(String name)
		{
			if (log.isDebugEnabled()) log.debug("JarClassLoaderProxy.findResources[" + jarPath + "]: " + name);
			if (!isInitialized) init();
			
			Enumeration<URL> url = null;
			if (loader != null)
			{
				url = loader.findResources(name);
			}
			return url;
		}
		
		boolean hasChanged()
		{
			if (!isInitialized) return false;
			if (loader != null) return loader.hasChanged(true);
			return new File(jarPath).exists();
		}
	}
}
