/**
	$Id: SocketUtil.java 5804 2011-10-12 12:45:07Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.Writer;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.nio.channels.ServerSocketChannel;


/**
	This class contains some useful methods when working with sockets. Mainly
	for easy reading and writing to/from the sockets, and also for closing
	sockets without throwing exceptions.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2011-10-12 14:45:07 +0200 (ons, 12 okt. 2011) $
*/
public class SocketUtil
{

	private static InetAddress LOCAL_HOST = null;
	static
	{
		try
		{
			LOCAL_HOST = InetAddress.getLocalHost();
		}
		catch (Throwable t)
		{
			t.printStackTrace();
		}
	}
	
	/**
		Send output to a socket.
		@param socket The socket to send to
		@param output The output to send
		@param shutdownOutput TRUE to close the output stream, FALSE to leave it open
		@throws IOException If there is an error
	 */
	public static void send(Socket socket, String output, boolean shutdownOutput)
		throws IOException
	{
		PrintWriter out = new PrintWriter(socket.getOutputStream());
		out.println(output);
		out.flush();
		if (shutdownOutput) socket.shutdownOutput();
	}

	/**
		Read input from a socket.
		@param socket The socket to read from
		@param shutdownInput If the input stream should be closed or not
		@return The data that was read
		@throws IOException If there is an error
	*/
	public static String read(Socket socket, boolean shutdownInput)
		throws IOException
	{
		StringBuilder input = new StringBuilder();
		BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		boolean more = true;
		while (more)
		{
			String line = in.readLine();
			if (line == null)
			{
				more = false;
			}
			else
			{
				if (input.length() > 0) input.append("\n");
				input.append(line);
			}
		}
		if (shutdownInput) socket.shutdownInput();
		return input.toString();
	}
	
	/**
		Close a socket without throwing any exceptions.
		@param socket The socket to close
	*/
	public static void close(Socket socket)
	{
		if (socket == null) return;
		try
		{
			socket.close();
		}
		catch (IOException ex)
		{}
	}
	
	/**
		Close a server socket channel without throwing any exceptions.
		@param socket The socket to close
	*/
	public static void close(ServerSocketChannel socket)
	{
		if (socket == null) return;
		try
		{
			socket.close();
		}
		catch (IOException ex)
		{}
	}	
	
	/**
		Close a server socket without throwing any exceptions.
		@param socket The socket to close
	*/
	public static void close(ServerSocket socket)
	{
		if (socket == null) return;
		try
		{
			socket.close();
		}
		catch (IOException ex)
		{}
	}	

	
	/**
		Close a <code>Writer</code> without throwing any exceptions.
		@param out The writer to close
	*/
	public static void close(Writer out)
	{
		if (out == null) return;
		try
		{
			out.close();
		}
		catch (IOException ex)
		{}
	}
	
	/**
		Close a <code>Reader</code> without throwing any exceptions.
		@param in The reader to close
	*/
	public static void close(Reader in)
	{
		if (in == null) return;
		try
		{
			in.close();
		}
		catch (IOException ex)
		{}
	}
	
	/**
		Close an <code>InputStream</code> without throwing any exceptions.
		@param in The stream to close
	*/
	public static void close(InputStream in)
	{
		if (in == null) return;
		try
		{
			in.close();
		}
		catch (IOException ex)
		{}
	}
	
	/**
		Check if the specfied address is the local host or not.
		@param address The address to check
		@return TRUE if the address represents the local host, FALSE otherwise
	 */
	public static boolean isLocalHost(InetAddress address)
	{
		return address.equals(LOCAL_HOST) || address.isLoopbackAddress();
	}
		
	/**
		Get the local address of the local host.
	*/
	public static InetAddress getLocalHost()
	{
		return LOCAL_HOST;
	}
	
	/**
		Get the external address of the local host.
	*/
	public static InetAddress getPublicLocalHost()
	{
		String hostName = LOCAL_HOST.getCanonicalHostName();
		InetAddress adress = LOCAL_HOST;
		try
		{
			adress = InetAddress.getByName(hostName);
		}
		catch (UnknownHostException ex)
		{
			ex.printStackTrace();
		}
		return adress;
	}
	
	
}
