/**
	$Id: BaseFileWriter.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.basefile;

import java.io.PrintWriter;
import java.io.Writer;
import java.util.regex.Pattern;

import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.util.export.TableWriter;

/**
	Wraps a writer stream and provides methods for easier
	creation of BASEfiles. All methods that start with
	<code>base</code> are used for BASEfile specified formatting.
	It is also possible to use the regular {@link PrintWriter}
	methods for special cases.
	<p>
	A BASEfile is divided in sections. Each section has 
	a header part and a data part. To start a new section
	use {@link #baseBeginSection(String)}. When a section has been 
	started use {@link #basePrintHeader(String, Object)} or one
	of the overloaded methods to print headers for that section.
	To end the header part and start the data part, use
	{@link #baseBeginDataPart()} and then {@link #basePrintData(Object...)}
	to print data lines. When all data has been printed
	either start a new section or end the current section 
	with {@link #baseEndSection()}.
	
	@author Nicklas
	@version 2.12
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (fre, 13 aug. 2010) $
*/
public class BaseFileWriter
	extends TableWriter
{
	private String metadataPrefix = "";
	
	private String currentSection;
	private boolean hasBom = false;
	private boolean headerPart = false;
	private boolean dataPart = false;
	
	/**
		Create a new BASEfile writer that is writing it's output to the
		given writer. The file prefix 'BASEfile' is automatically written.
		@param out The writer to write the data to
	*/
	public BaseFileWriter(Writer out)
	{
		super(out);
	}
	
	/**
		@return The name of the current section, or null if no
			section has been started.
	*/
	public String getCurrentSection()
	{
		return currentSection;
	}
	
	/**
		@return TRUE if this writer is currently in the header part of a section,
			FALSE if no section is started or if the writer is in the data part
	*/
	public boolean isWritingHeaders()
	{
		return headerPart;
	}
	
	/**
		@return TRUE if this writer is currently in the data part of a section,
			FALSE if no section is started or if the writer is in the header part
	*/
	public boolean isWritingData()
	{
		return dataPart;
	}
	
	protected void setMode(boolean headerPart, boolean dataPart)
	{
		this.headerPart = headerPart;
		this.dataPart = dataPart;
	}
	
	/**
		The metadata prefix is written to the beginning of 
		each line that is not a data line. The default value
		is an empty string. 
		<p>
		Tip! Settting this value to '#' makes it possible to send
		a BASEfile to gnuplot.
		@return The current value of the metadata prefix
	*/
	public String getMetadataPrefix()
	{
		return metadataPrefix;
	}
	
	/**
		Change the metadata prefix string.
		@param metaDataPrefix The new metadata prefix
	*/
	public void setMetadataPrefix(String metaDataPrefix)
	{
		this.metadataPrefix = metaDataPrefix;
	}
	
	/**
		Write the beginning-of-file marker (BASEfile) to the output.
		This method must be called before it is possible to writer other
		data to the file.
	*/
	public void baseWriteBom()
	{
		println(getMetadataPrefix() + "BASEfile");
		hasBom = true;
	}
	
	/**
		Start a new section in the BASEfile. If the current
		section hasn't been ended, {@link #baseEndSection()}
		is automatically called to end it. After this call the
		BASEfile is ready for writing headers to the new section.
		<p>
		If the beginning-of-file-marker hasn't been written the
		{@link #baseWriteBom()} is automatically called.
		
		@param name The name of the section
	*/
	public void baseBeginSection(String name)
	{
		if (!hasBom) baseWriteBom();
		baseEndSection();
		println(getMetadataPrefix() + "section\t" + baseEscape(name));
		setMode(true, false);
		currentSection = name;
	}
	
	/**
		Ends the current section by writing an empty line.
	*/
	public void baseEndSection()
	{
		if (isWritingHeaders()) 
		{
			println(getMetadataPrefix() + "%");
			println();
		}
		else if (isWritingData())
		{
			println();
		}
		setMode(false, false);
		currentSection = null;
	}

	/**
		Prints a header in the BASEfile. This call is ignored if
		no section has been started or if the current section is
		currently in data writing mode. Line breaks (\n) and
		carriage returns (\r) in the value are converted to
		the literal string values '\n' and '\r'.
		@param key The header key, 'section' is not allowed
		@param value The value, if null the configured null value
			is used, otherwise the toString() method is called
		@throws InvalidDataException If the key is 'section'
	*/
	public <T> void basePrintHeader(String key, Object value)
	{
		if (!isWritingHeaders()) return;
		if ("section".equals(key))
		{
			throw new InvalidDataException("'section' is not allowed as a parameter");
		}
		String valueToWrite = value == null ? getNullValue() : value.toString();
		println(getMetadataPrefix() + key + "\t" + baseEscape(valueToWrite));
	}

	/**
		End the header part of the current section (by writing a single % on a line 
		by itself) and start the data part. The call to this method is ignored if 
		no section has been started or if the data part of the current section
		has alredy been started.
	*/
	public void baseBeginDataPart()
	{
		if (isWritingData() || !isWritingHeaders()) return;
		println(getMetadataPrefix() + "%");
		setMode(false, true);
	}
	
	/**
		Print a data line to the BASEfile. Each value is separated by the
		data separator string ('tab' by default). Null values are replaced 
		with the null value string (empty string by default). Non-null objects
		are converted to strings values by calling their toString() method.
		<p>
		This call is ignored if the data part has not been started.
		@param data The values to write
		@see #isWritingData()
		@see #baseBeginDataPart()
	*/
	public void basePrintData(Object... data)
	{
		if (!isWritingData()) return;
		tablePrintData(data);
	}
	
	private Pattern r = Pattern.compile("\r");
	private Pattern n = Pattern.compile("\n");

	/**
		Escape new line and carrige return characters in the input
		to the literal string '\n' and '\r' in the output.
		@param in The input string
	*/
	public String baseEscape(String in)
	{
		in = r.matcher(in).replaceAll("\\\\r");
		in = n.matcher(in).replaceAll("\\\\n");
		return in;
	}
	
}
