/**
	$Id: DataWriter.java 7714 2019-05-21 11:37:09Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.StandardCharsets;

import net.sf.basedb.core.File;
import net.sf.basedb.util.encode.TabCrLfEncoderDecoder;
import net.sf.basedb.util.export.TableWriter;

/**
	Writer implementation for writing BFS data files. A data file is
	a matrix of data. Columns are separated by tabs. Values are escaped
	using the regular rules ({@link TabCrLfEncoderDecoder}). All
	rows must have an equal number of columns. The number of columns is
	defined by the first call to {@link #bfsPrintData(Object...)}
	or by calling {@link #setColumnCount(int)}.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2019-05-21 13:37:09 +0200 (tis, 21 maj 2019) $
*/
public class DataWriter
	extends TableWriter
{

	/**
		Utility method for creating a data writer when you have an
		output stream.
		@param out The output stream the writer should print to
		@param filename Optional, the name of the file the output stream
			is printing to
	*/
	public static DataWriter create(OutputStream out, String filename)
	{
		if (out == null) throw new NullPointerException("out");
		DataWriter writer = new DataWriter(
				new OutputStreamWriter(out, StandardCharsets.UTF_8));
		writer.setFilename(filename);
		return writer;
	}
	
	/**
		Utility method for creating a data writer to a file in the BASE
		file system. The character set and MIME type on the file are automatically 
		updated.
		@param file The file in the BASE file system
	*/
	public static DataWriter create(File file)
	{
		if (file == null) throw new NullPointerException("file");
		file.setCharacterSet("UTF-8");
		file.setMimeType("text/plain");
		return create(file.getUploadStream(false), file.getName());
	}
	
	/**
		Utility method for creating a data writer to a file in the native
		file system. If the file doesn't exists, it is created.
		@param file The file in the native file system
	*/
	public static DataWriter create(java.io.File file)
		throws IOException
	{
		if (file == null) throw new NullPointerException("file");
		if (!file.createNewFile() && !file.exists()) 
		{
			throw new FileNotFoundException(file.toString());
		}
		return create(new FileOutputStream(file), file.getName());
	}

	private String filename;
	private boolean lockedColumns;
	private int columnCount = -1;
	
	/**
		Create a new BFS data writer.
		@param out The parent writer which this writer will print to
	*/
	public DataWriter(Writer out)
	{
		super(out);
		super.setDataSeparator("\t");
		super.setNullValue("");
		super.setEncoder(new TabCrLfEncoderDecoder(false));
	}

	/**
		Get the file name that this writer is printing to.
		@return The file name or null if not known
	*/
	public String getFilename()
	{
		return filename;
	}

	/**
		Set the file name that this writer is printing to.
	*/
	public void setFilename(String filename)
	{
		this.filename = filename;
	}
	
	/**
		Get the number of data columns. This information is only
		available after the first line of data has been printed
		or if a specific number of columns has been set.
		
		@return The number of columns, or -1 if this information
			is not yet known
	*/
	public int getColumnCount()
	{
		return columnCount;
	}
	
	/**
		Set the number of data columns. This can only be changed
		as long as no data has been written to the stream.
		@param columns The number of data columns, if 0 or negative the
			first call to {@link #bfsPrintData(Object...)} determines 
			the number of columns
	*/
	public void setColumnCount(int columns)
	{
		if (lockedColumns) 
		{
			throw new IllegalStateException("Can't change the number of columns since data " +
				"has already been written in file: " + getFilename());
		}
		this.columnCount = columns;
	}
	
	/**
		Checks if the number of columns has been locked or not. The number
		of required columns are locked the first time a call to
		{@link #bfsPrintData(Object...)}	is made.
	*/
	public boolean isLockedColumns()
	{
		return lockedColumns;
	}
	
	/**
		Print a data line. The number of columns must be the same for all calls
		to this method. Normally, the first call determines the required number
		of columns, but this can also be set with {@link #setColumnCount(int)}.
		
		@param data The data to print
		@throws IllegalStateException If headers has not been printed
		@throws IllegalArgumentException If the data array doesn't match the
			required number of columns
		@throws NullPointerException If the data array is null
	*/
	public void bfsPrintData(Object... data)
	{

		if (data == null) throw new NullPointerException("'data' in file: " + getFilename());
		if (columnCount <= 0) columnCount = data.length;
		if (data.length != columnCount)
		{
			throw new IllegalArgumentException("Data array has " + data.length + " columns; " +
					"; expected " + columnCount + " in file: " + getFilename());
		}
		lockedColumns = true;
		if (data.length > 0)
		{
			tablePrintData(data);
		}
		else
		{
			print("\n");
		}
	}
	
}
