package net.sf.basedb.util.excel;

import java.time.ZoneId;
import java.time.temporal.ChronoField;
import java.util.Date;

/**
	Excel formatter implementation that automatically selects a format
	depending on the type/class of the value.
	
	* Null values are created as empty cells: {@link ExcelValue#emptyCell()}
	* Integer and Long values are created as numeric cells with no decimals: {@link ExcelValue#asInt(Number)}
	* Other numeric values are created as numeric cell without format specification: {@link ExcelValue#asNumber(Number, int)}
	* Date values with time at midnight are created as numeric cells with date formatting applied: {@link ExcelValue#asDate(Date, String)} 
	* Date values with other times are created as numeric cells with timestamp formatting applied: {@link ExcelValue#asTimestamp(Date, String)}
	* Boolean values are created as numeric cells (0 or 1) with BOOLEAN format: {@link ExcelValue#asBoolean(Boolean)}
	* All other values are created as string values via their toString() method: {@link ExcelValue#asString(String)}
 
	@author nicklas
	@since 3.15
*/
public class AutoFormatter
	implements ExcelFormatter<Object, Object>
{

	private String dateFormat;
	private String timestampFormat;
	
	public AutoFormatter() 
	{}
	
	/**
		Set the data format template to use for date values (=dates with all time components at 0).
		If not set, the default from {@link ExcelValue#asDate(Date, String)} is used.
	*/
	public void setDateFormat(String dateFormat)
	{
		this.dateFormat = dateFormat;
	}
	
	/**
		@see #setDateFormat(String)
	*/
	public String getDateFormat()
	{
		return dateFormat;
	}
	
	/**
		Set the data format template to use for timestamp values. If not set, the default
		from {@link ExcelValue#asTimestamp(Date, String)} is used.
	*/
	public void setTimstampFormat(String timestampFormat)
	{
		this.timestampFormat = timestampFormat;
	}
	
	/**
		@see #setTimstampFormat(String)
	*/
	public String getTimstampFormat()
	{
		return timestampFormat;
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	@Override
	public ExcelValue<Object> toExcelValue(Object value) 
	{
		ExcelValue ev = null;
		if (value == null)
		{
			ev = ExcelValue.emptyCell();
		}
		else if (value instanceof Number)
		{
			Number n = (Number)value;
			if (value instanceof Integer || value instanceof Long)
			{
				ev = ExcelValue.asInt(n);
			}
			else
			{
				ev = ExcelValue.asNumber(n, -1);
			}
		}
		else if (value instanceof Date)
		{
			Date d = (Date)value;
			boolean atMidnight = d.toInstant().atZone(ZoneId.systemDefault()).get(ChronoField.MINUTE_OF_DAY) == 0;
			if (atMidnight)
			{
				ev = ExcelValue.asDate(d, dateFormat);
			}
			else
			{
				ev = ExcelValue.asTimestamp(d, timestampFormat);
			}
		}
		else if (value instanceof Boolean)
		{
			ev = ExcelValue.asBoolean((Boolean)value);
		}
		else
		{
			ev = ExcelValue.asString(value.toString());
		}
		return ev;
	}
	
}
