/**
	$Id: AdvancedFieldConverter.java 5381 2010-08-05 11:16:18Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.ExtraValue.CoordinateType;
import net.sf.basedb.util.Values;

/**
	Field converter implementation that gives more detailed control how
	fields are converted and used in the export. The implementation 
	builds on the {@link StandardFieldConverter} which is used as
	a fallback. The advanced options can be accessed by using formula-like
	field "names" as shown below. In most cases the "title" parameter
	is optional. If not given, the name is used as the title.
	
	<ul>
	<li>=property("name", ["title"]): Lookup a static property with the given
		name. Can be used on assay, reporter and spot (=raw data properties)
		fields.
	<li>=factor("name" [,"title"]): Lookup an experimental factor with the 
		given name or external id. Can be used in 
		{@link #getAssayField(String, boolean)}.
	<li>=annotation("name" [,"title"]): Lookup an assay annotation with the
		given name or external id. Can be used in 
		{@link #getAssayField(String, boolean)}. 
	<li>=xtra("name" [,"title"]): Lookup an extra value with the given name
		or external id. Can be used in {@link #getReporterField(String, boolean)}
		and {@link #getSpotField(String, boolean)}.
	<li>=formula("name" [,"title"]): Lookup a formula with the given name.
		Can be used in {@link #getReporterField(String, boolean)} and 
		{@link #getSpotField(String, boolean)}.
	<li>=jep("jep expression", "title" [,"datatype"]): Use the given JEP formula
		to calculate the exported value. Can be used in {@link 
		#getReporterField(String, boolean)} and {@link #getSpotField(String, boolean)}.
		The "title" is required. The "datatype" is optional. If not given "string" will
		be used. Valid values are constants from the {@link Type} enumeration 
		(case-insensitive).
	</ul>
	
	Field names that doesn't start with an equal character (=) are passed on to
	the {@link StandardFieldConverter}.
	
	@author Nicklas
	@since 2.16
	@base.modified $Date: 2010-08-05 13:16:18 +0200 (tors, 05 aug. 2010) $
*/
public class AdvancedFieldConverter
	extends StandardFieldConverter
{

	private Map<String, Pattern> patterns;
	
	/**
		Create a new advanced field converter.
		@param dc A DbControl to use for database access
		@param source The source bioassay set
	*/
	public AdvancedFieldConverter(DbControl dc, BioAssaySet source)
	{
		super(dc, source);
		this.patterns = new HashMap<String, Pattern>();
	}
	/*
		From the ExportableFieldConverter interface
		-------------------------------------------
	*/
	@Override
	public AssayField getAssayField(String name, boolean required)
	{
		AssayField af = null;
		if (name.startsWith("="))
		{
			af = getAdvancedAssayField(name, required);
		}
		else
		{
			af = super.getAssayField(name, required);
		}
		return af;
	}

	@Override
	public DynamicField getReporterField(String name, boolean required)
	{
		DynamicField df = null;
		if (name.startsWith("="))
		{
			df = getAdvancedReporterField(name, required);
		}
		else
		{
			df = super.getReporterField(name, required);
		}
		return df;
	}

	@Override
	public DynamicField getSpotField(String name, boolean required)
	{
		DynamicField df = null;
		if (name.startsWith("="))
		{
			df = getAdvancedSpotField(name, required);
		}
		else
		{
			df = super.getSpotField(name, required);
		}
		return df;
	}
	// ------------------------------------
	
	/**
		Convert an expression to an assay field. Supported expressions are
		listed below. The "title" parameter is optional in all cases.
		<ul>
		<li>=property("name" [,"title"]): Get the value of the property with the
			given name.
		<li>=factor("name" [,"title"]): Get the value of the experimental factor
			with the given name or external id.
		<li>=annotation("name" [,"title"]): Get the value of the assay annotation
			with the given name or external id.
		</ul>
	*/
	protected AssayField getAdvancedAssayField(String expression, boolean required)
	{
		AssayField af = null;
		Pattern p = null;
		Matcher m = null;
		
		// Check: =property("name", "title")
		p = getPattern("property", 1, 1);	
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			af = getAssayPropertyField(name, title);
			if (af == null && required) 
			{
				throw new ItemNotFoundException("Unknown property in assay field expression: " + expression);
			}
			return af;
		}
		
		// Check: =factor("name", "title")
		p = getPattern("factor", 1, 1);	
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			af = getExperimentalFactorField(name, title);
			if (af == null && required) 
			{
				throw new ItemNotFoundException("Unknown experimental factor in assay field expression: " + expression);
			}
			return af;
		}

		// Check: =annotation("name", "title")
		p = getPattern("annotation", 1, 1);	
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			af = getAssayAnnotationField(name, title);
			if (af == null && required) 
			{
				throw new ItemNotFoundException("Unknown annotation in assay field expression: " + expression);
			}
			return af;
		}
		
		if (af == null && required)
		{
			throw new InvalidDataException("Invalid assay field expression: " + expression);
		}
		return af;
	}
	
	/**
		Convert an expression to a reporter field. Supported expressions are
		listed below. The "title" parameter is optional in the "jep" expression.
		<ul>
		<li>=property("name" [,"title"]): Get the value of the reporter property 
			with the given name.
		<li>=formula("name" [,"title"]): Get the value of the formula 
			with the given name. The formula should only use
			properties that are related to the reporter and it's position.
		<li>=xtra("name" [,"title"]): Get the value of the extra value with
			the given name or external id. The extra value should be a 
			position-based extra value.
		<li>=jep("expression", "title" [,"datatype"]): Calculate the value using the
			given JEP expression. A title is required. The datatype is optional
			(default is STRING). Valid datatypes are those defined by the
			{@link Type} enumeration.
		</ul>
	*/
	protected DynamicField getAdvancedReporterField(String expression, boolean required)
	{
		DynamicField df = null;
		Pattern p = null;
		Matcher m = null;
		
		// Check: =property("name", "title")
		p = getPattern("property", 1, 1);	
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			df = getReporterProperty(name, title);
			if (df == null && required) 
			{
				throw new ItemNotFoundException("Unknown reporter property in expression: " + expression);
			}
			return df;
		}
		
		// Check: =formula("formula", "title")
		p = getPattern("formula", 1, 1);
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			df = getFormulaField(name, title);
			if (df == null && required) 
			{
				throw new ItemNotFoundException("Unknown formula in expression: " + expression);
			}
			return df;
		}

		// Check: =xtra("name", "title")
		p = getPattern("xtra", 1, 1);
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			df = getExtraValueField(name, CoordinateType.POSITION, title);
			if (df == null && required) 
			{
				throw new ItemNotFoundException("Unknown extra value in expression: " + expression);
			}
			return df;
		}

		// Check: =jep("name", "title", "datatype")
		p = getPattern("jep", 2, 1);
		m = p.matcher(expression);
		if (m.matches())
		{
			String jep = m.group(1);
			String title = m.group(2);
			String dataType = Values.getString(m.group(3), "STRING");
			df = ExportableFieldFactory.jep(jep, title, Type.valueOf(dataType.toUpperCase()), null, null);
			return df;
		}

		if (df == null && required)
		{
			throw new InvalidDataException("Invalid reporter field expression: " + expression);
		}
		
		return df;
	}
	
	/**
		Convert an expression to a spot field. Supported expressions are
		listed below. The "title" parameter is optional in the "jep" expression.
		<ul>
		<li>=property("name" [,"title"]): Get the value of the raw data property 
			with the given name.
		<li>=formula("name" [,"title"]): Get the value of the formula 
			with the given name. The formula should only use
			properties that are related to the spot and raw data.
		<li>=xtra("name" [,"title"]): Get the value of the extra value with
			the given name or external id. The extra value should be a 
			spot-based extra value.
		<li>=jep("expression", "title" [,"datatype"]): Calculate the value using the
			given JEP expression. A title is required. The datatype is optional
			(default is STRING). Valid datatypes are those defined by the
			{@link Type} enumeration.
		</ul>
	*/
	protected DynamicField getAdvancedSpotField(String expression, boolean required)
	{
		DynamicField df = null;
		Pattern p = null;
		Matcher m = null;
		
		// Check: =property("name", "title")
		p = getPattern("property", 1, 1);	
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			df = getRawDataProperty(name, title);
			if (df == null && required) 
			{
				throw new ItemNotFoundException("Unknown raw data property in expression: " + expression);
			}
			return df;
		}
		
		// Check: =formula("formula", "title")
		p = getPattern("formula", 1, 1);
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			df = getFormulaField(name, title);
			if (df == null && required) 
			{
				throw new ItemNotFoundException("Unknown formula in expression: " + expression);
			}
			return df;
		}
		
		// Check: =xtra("name", "title")
		p = getPattern("xtra", 1, 1);
		m = p.matcher(expression);
		if (m.matches())
		{
			String name = m.group(1);
			String title = Values.getString(m.group(2), name);
			df = getExtraValueField(name, CoordinateType.SPOT, title);
			if (df == null && required) 
			{
				throw new ItemNotFoundException("Unknown extra value in expression: " + expression);
			}
			return df;
		}
		
		// Check: =jep("name", "title", "datatype")
		p = getPattern("jep", 2, 1);
		m = p.matcher(expression);
		if (m.matches())
		{
			String jep = m.group(1);
			String title = m.group(2);
			String dataType = Values.getString(m.group(3), "STRING");
			df = ExportableFieldFactory.jep(jep, title, Type.valueOf(dataType.toUpperCase()), null, null);
			return df;
		}
		
		if (df == null && required)
		{
			throw new InvalidDataException("Invalid reporter field expression: " + expression);
		}
		return df;
	}

	
	private Pattern getPattern(String function, int requiredParameters, int optionalParameters)
	{
		Pattern p = patterns.get(function);
		if (p == null)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("\\=").append(function).append("\\("); // matches: =function(
			for (int i = 0; i < requiredParameters; ++i)
			{
				if (i > 0) sb.append(","); // matches: ,
				sb.append("\\s*\\\"(.*?)\\\"\\s*");  // matches: "...."
			}
			for (int i = 0; i < optionalParameters; ++i)
			{
				sb.append("(?:,\\s*\\\"(.*?)\\\"\\s*)?");
			}
			sb.append("\\)\\s*"); // matches: )
			p = Pattern.compile(sb.toString());
			patterns.put(function, p);
		}
		return p;
	}
}
