/**
	$Id: ExportableFieldFactory.java 5590 2011-03-16 11:48:47Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.util.formatter.Formatter;

/**
	Utility class with factory methods that helps with the creation
	of exportable fields.
	
	@author Nicklas
	@version 2.12
	@base.modified $Date: 2011-03-16 12:48:47 +0100 (ons, 16 mars 2011) $
*/
public class ExportableFieldFactory
{
	/**
		Create an annotation assay field.
		@param at The annotation type
		@param formatter A formatter or null
		@see AnnotationAssayField
	*/
	public static AnnotationAssayField annotation(AnnotationType at, Formatter<?> formatter)
	{
		return annotation(at, null, formatter);
	}

	/**
		Create an annotation assay field, overriding the title
		of the annotation type.
		@param at The annotation type
		@param title A custom title for the assay field, or null to use the 
			name of the annotation type
		@param formatter A formatter or null
		@see AnnotationAssayField
		@since 2.15
	*/
	public static AnnotationAssayField annotation(AnnotationType at, String title, Formatter<?> formatter)
	{
		AnnotationAssayField field = new AnnotationAssayField();
		field.setAnnotationType(at);
		field.setFormatter(formatter);
		field.setTitle(title);
		return field;
	}

	
	/**
		Create a property assay field. This field is used to export, for
		example, name and description, from the bioassay.
		@param property The property to export, see {@link Metadata#getPropertyPath(String, boolean)}
			for more information about the syntax of this parameter.
		@param title The column title
		@param formatter A formatter or null
		@return An {@link AssayPropertyField} object
		@since 2.15
	*/
	public static AssayPropertyField assayProperty(String property, String title,
		Type type, Formatter<?> formatter)
	{
		AssayPropertyField field = new AssayPropertyField();
		field.setProperty(property);
		field.setTitle(title);
		field.setFormatter(formatter);
		field.setType(type);
		return field;
	}
	
	/**
		Create a simple dynamic field.
		
		@param e The expression to export
		@param title The column title
		@param averageMethod Average method to use or null to use the
			average method specified by the bioassay set
		@param formatter A formatter or null
		@see SimpleDynamicField
		@since 2.15
	*/
	public static SimpleDynamicField simple(Expression e, String title, 
		Type type, Formula.AverageMethod averageMethod, Formatter<?> formatter)
	{
		SimpleDynamicField field = new SimpleDynamicField();
		field.setExpression(e);
		field.setTitle(title);
		field.setType(type);
		field.setAverageMethod(averageMethod);
		field.setFormatter(formatter);
		return field;
	}
	
	/**
		Create a JEP dynamic field.
		
		@param jep The JEP expression
		@param title The column title
		@param type The data type of the expression
		@param averageMethod Average method to use or null to use the
			average method specified by the bioassay set
		@param formatter A formatter or null
	*/
	public static JepDynamicField jep(String jep, String title, Type type,
			Formula.AverageMethod averageMethod, Formatter<?> formatter)
	{
		JepDynamicField field = new JepDynamicField();
		field.setTitle(title);
		field.setJep(jep);
		field.setAverageMethod(averageMethod);
		field.setFormatter(formatter);
		field.setType(type);
		return field;
	}
	
	/**
		Create a simple dynamic field that exports the channel intensity
		of the given channel. This method uses {@link Dynamic#column(VirtualColumn)}
		and {@link VirtualColumn#channelRaw(int)} to generate the expression. It
		uses the average method specified by the bioassay set.
		
		@param channel The channel number
		@param title The column title. If null, the title is generated as 'intensity' + channel
		@param formatter A formatter for numeric data or null
	*/
	public static SimpleDynamicField channel(int channel, String title, 
		Formatter<?> formatter)
	{
		if (title == null) title = "Ch " + channel;
		Expression e = Dynamic.column(VirtualColumn.channelRaw(channel)); 
		return simple(e, title, Type.FLOAT, null, formatter);
	}
	
	/**
		Create a simple dynamic field that exports a raw data value.
		This method uses {@link Dynamic#rawData(String)} to generate the
		expression and uses the average method specified by the raw data
		property.
		@param property The raw data property
		@param title The column title, or null to use the title of 
			the property
		@param formatter A formatter or null
	*/
	public static SimpleDynamicField rawData(RawDataProperty property, String title,
		Formatter<?> formatter)
	{
		String name = property.getName();
		if (title == null) title = property.getTitle();
		return simple(Dynamic.rawData(name), title, property.getType(), 
				property.getAverageMethod(), formatter);
	}
	
	/**
		Create a simple dynamic field that exports a raw data value.
		This method uses {@link Dynamic#rawData(String)} to generate the
		expression.
		@param property The raw data property
		@param title The column title, or null to use the property
			as title
		@param averageMethod Average method to use or null to use 
			{@link net.sf.basedb.core.Formula.AverageMethod#NONE}
		@param formatter A formatter or null
		@since 2.15
	*/
	public static SimpleDynamicField rawData(String property, String title, Type type,
		Formula.AverageMethod averageMethod, Formatter<?> formatter)
	{
		if (title == null) title = property;
		if (averageMethod == null) averageMethod = Formula.AverageMethod.NONE;
		return simple(Dynamic.rawData(property), title, type, averageMethod, formatter);
	}

	
	/**
		Create a simple dynamic field that exports reporter annotation values.
		This method uses {@link Dynamic#reporter(String)} to generate the
		expression and uses the average method specified by the reporter
		property.
		@param property The reporter property
		@param title The column title, or null to use the title of 
			the property
		@param formatter A formatter or null
	*/
	public static SimpleDynamicField reporter(ExtendedProperty property, String title,
			Formatter<?> formatter)
	{
		String name = property.getName();
		if (title == null) title = property.getTitle();
		return simple(Dynamic.reporter(name), title, property.getType(), 
				property.getAverageMethod(), formatter);
	}

	/**
		Create a simple dynamic field that exports reporter annotation values.
		This method uses {@link Dynamic#reporter(String)} to generate the
		expression.
		@param property The reporter property
		@param title The column title, or null to use the property
			as title
		@param averageMethod Average method to use or null to use 
			{@link net.sf.basedb.core.Formula.AverageMethod#NONE}
		@param formatter A formatter or null
		@since 2.15
	*/
	public static DynamicField reporter(String property, String title, Type type,
		Formula.AverageMethod averageMethod, Formatter<?> formatter)
	{
		if (title == null) title = property;
		if (averageMethod == null) averageMethod = Formula.AverageMethod.NONE;
		return simple(Dynamic.reporter(property), title, type, averageMethod, formatter);
	}

	
	/**
		Create a JEP dynamic field that exports the result of 
		the JEP expression in a formula. This method doesn't
		check the type of the formula or the number of expressions
		stored in the formula. It just takes the first expression
		it finds.
		
		@param f The formula
		@param title The column title, or null to use the
			name of the formula as the title
		@param formatter A formatter or null
	*/
	public static JepDynamicField formula(Formula f, String title, Formatter<?> formatter)
	{
		String expression = f.getFormula(0);
		if (title == null) title = f.getName();
		return jep(expression, title, f.getValueType(), f.getAverageMethod(), formatter);
	}
	
	/**
		Create a JEP dynamic field that exports an extra value.
		@param extra The extra value type
		@param title The column title, or null to use the
			name of the formula as the title
		@param formatter A formatter or null
	*/
	public static JepDynamicField extraValue(ExtraValueType extra, String title, Formatter<?> formatter)
	{
		String expression = "xtra('" + extra.getExternalId() + "')";
		if (title == null) title = extra.getName();
		return jep(expression, title, extra.getValueType(), extra.getAverageMethod(), formatter);
	}
	
}
