/**
	$Id: JepDynamicField.java 5385 2010-08-13 10:48:17Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.util.BioAssaySetUtil;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.formatter.Formatter;

/**
	A dynamic field implementation that uses a JEP expression to
	generate the expression that is needed to select the data. One 
	instance of this class is considered to be equal to another if 
	they are configured with the same jep expression and title.

	@author Nicklas
	@version 2.12
	@base.modified $Date: 2010-08-13 12:48:17 +0200 (fre, 13 aug. 2010) $
*/
public class JepDynamicField
	implements DynamicField
{
	private String jep;
	private String title;
	private Formula.AverageMethod averageMethod;
	private Formatter<?> formatter;
	private Type type;
	
	/**
		Create a new object. Before it can be used it must be
		configured.
	 */
	public JepDynamicField()
	{}
	
	/*
		From the DynamicField interface
		-------------------------------
	*/
	/**
		Converts the configured JEP expression using the
		{@link BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)}
		method. If 'forAverage' is TRUE the confgured {@link net.sf.basedb.core.Formula.AverageMethod} is 
		allowed to transform	the expression. If no average method has been configured the 
		average method specified by the {@link IntensityTransform} from
		the 'source' bioassay set is used.
	*/
	@Override
	public Expression getExpression(DbControl dc, DynamicQuery query, BioAssaySet source, boolean forAverage)
	{
		Expression e = BioAssaySetUtil.createJepExpression(dc, jep, query);
		if (forAverage)
		{
			Formula.AverageMethod avg = averageMethod == null ?
					source.getIntensityTransform().getAverageMethod() : averageMethod;
			e = avg.getAverageExpression(e);
		}
		return e;
	}
	@Override
	public String getTitle()
	{
		return title;
	}
	@Override
	public Type getType()
	{
		return type;
	}
	@Override
	public Formatter<?> getFormatter()
	{
		return formatter;
	}
	// -----------------------------------
	
	/**
		A jep dynamic field is equal to another field if it uses the same
		jep expression and has the same title.
		@since 2.15
	*/
	@Override
	public boolean equals(Object obj)
	{
		if (this == obj) return true;
		if (!(obj instanceof JepDynamicField)) return false;
		JepDynamicField other = (JepDynamicField)obj;
		return EqualsHelper.equals(this.getJep(), other.getJep()) 
			&& EqualsHelper.equals(this.getTitle(), other.getTitle());
	}
	
	/**
		@since 2.15
	*/
	@Override
	public int hashCode()
	{
		return 13 * EqualsHelper.hashCode(getJep()) + 
			EqualsHelper.hashCode(getTitle());
	}
	/**
	 	@since 2.15
	*/
	@Override
	public String toString()
	{
		return "JepDynamicField(title=" + getTitle() + "; expression=" + getJep() + ")";
	}
	
	/**
		Get the configured JEP expression.
	*/
	public String getJep()
	{
		return jep;
	}

	/**
		Set the expression used to select data in the query. The
		expression is parsed and converted to an {@link Expressions}
		with the {@link BioAssaySetUtil#createJepExpression(DbControl, String, DynamicQuery)}
		method.
	*/
	public void setJep(String jep)
	{
		this.jep = jep;
	}

	/**
		Set the column header title.
	*/
	public void setTitle(String title)
	{
		this.title = title;
	}
	
	/**
		Set the data type of the exported field.
		@since 2.15
	*/
	public void setType(Type type)
	{
		this.type = type;
	}
	
	/**
		Get the average method to use if the exporter is exporting
		averaged data. If no average method has been configured, the
		everage method specified by the {@link IntensityTransform}
		of the source bioassay set is used.
	*/
	public Formula.AverageMethod getAverageMethod()
	{
		return averageMethod;
	}
	
	/**
		Set the average method to use when exporting averaged data.
	*/
	public void setAverageMethod(Formula.AverageMethod averageMethod)
	{
		this.averageMethod = averageMethod;
	}
	
	/**
		Set a formatter to use for formatting the exported values.
	*/
	public void setFormatter(Formatter<?> formatter)
	{
		this.formatter = formatter;
	}


}
