/**
	$Id: ExtensionContext.java 6030 2012-03-28 10:38:04Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions;

import net.sf.basedb.util.extensions.Registry.RegisteredExtension;

/**
	Invokation context for an extension. Most
	calls are forwarded to the invokation context for the
	extension point this extension belongs to.
	
	@author nicklas
	@version 2.7
	@base.modified $Date: 2012-03-28 12:38:04 +0200 (ons, 28 mars 2012) $
 */
class ExtensionContext<A extends Action>
	extends InvokationContext<A>
{
	private final ExtensionPointContext<A> mainContext;
	private final RegisteredExtension<A> extension;
	private Renderer<? super A> renderer;
	private boolean hasCreatedRenderer;
	
	ExtensionContext(ExtensionPointContext<A> mainContext, RegisteredExtension<A> extension)
	{
		this.mainContext = mainContext;
		this.extension = extension;
	}
	/*
		From the InvokationContext object
		---------------------------------
	*/
	@Override
	public ClientContext getClientContext()
	{
		return mainContext.getClientContext();
	}

	@Override
	public ExtensionPoint<? super A> getExtensionPoint()
	{
		return mainContext.getExtensionPoint();
	}
	
	@Override
	public Extension<A> getExtension()
	{
		return extension;
	}
	
	/**
		The class loader is the class loader that loaded the action
		factory of the extension.
	*/
	@Override
	public ClassLoader getClassLoader()
	{
		return extension.getClassLoader();
	}
	
	@Override
	public Object getAttribute(String name)
	{
		return mainContext.getRegistry().getAttribute(extension.getId(), name);
	}

	@Override
	protected Registry getRegistry()
	{
		return mainContext.getRegistry();
	}
	
	@Override
	protected void prepareRendererFactory()
	{
		RendererFactory<? super A> factory = extension.getRendererFactory();
		if (factory != null) 
		{
			try
			{
				factory.prepareContext(this);
			}
			catch (Throwable t)
			{
				handleError("Error preparing renderer factory for extension '" +
						extension.getId() + "': " + factory, t);
			}
		}
	}
	
	/**
		If the extension point allows it and this extension
		specifies a renderer factory, create a new renderer from
		it. Otherwise, get the renderer from the extension point.
	*/
	@Override
	protected Renderer<? super A> getRenderer()
	{
		if (!hasCreatedRenderer)
		{
			ExtensionPoint<? super A> ep = mainContext.getExtensionPoint();
			if (ep.allowRendererOverride())
			{
				RendererFactory<? super A> factory = extension.getRendererFactory();
				if (factory != null)
				{
					try
					{
						renderer = factory.getRenderer(this);
					}
					catch (Throwable t)
					{
						handleError("Error creating renderer for for extension '" +
								extension.getId() + "': " + factory, t);
					}
				}
			}
			if (renderer == null)
			{
				renderer = mainContext.getRenderer();
			}
			hasCreatedRenderer = true;
		}
		return renderer;
	}

	/**
		Call {@link Extension#getActionFactory()} and then
		{@link ActionFactory#getActions(InvokationContext)} on the
		returned factory.
	*/
	@Override
	protected A[] getActions()
	{
		A[] actions = null;
		ActionFactory<? extends A> factory = extension.getActionFactory();
		try
		{
			actions = factory.getActions(this);
		}
		catch (Throwable t)
		{
			handleError("Error creating actions for extension '" +
					extension.getId() + "' using factory: " + factory, t);
		}
		return actions;
	}
	
	@Override
	protected void setError(A action, Throwable t)
	{
		extension.setError(action, t);
	}
	
	@Override
	protected void clearError()
	{
		extension.clearError();
	}

	@Override
	protected ErrorHandler<? super A> getErrorHandler()
	{
		return mainContext.getErrorHandler();
	}
	// ----------------------------------------
	
	boolean prepareActionFactory()
	{
		boolean prepared = false;
		ActionFactory<? extends A> factory = extension.getActionFactory();
		try
		{
			prepared = factory.prepareContext(this);
		}
		catch (Throwable t)
		{
			handleError("Error preparing action factory for extension '" +
					extension.getId() + "': " + factory, t);
		}
		return prepared;
	}
	
	private void handleError(String message, Throwable t)
	{
		getErrorHandler().handleError(this, null, message, t);
	}

	void handleError(A action, String message, Throwable t)
	{
		getErrorHandler().handleError(this, action, message, t);
	}


}
