/**
	$Id:RendererFactory.java 4187 2008-03-20 11:15:25Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions;

import net.sf.basedb.util.extensions.xml.XmlLoader;

/**
	A renderer factory is a factory that can create {@link Renderer}
	objects. An extension point may provide a default factory 
	({@link ExtensionPoint#getRendererFactory()}). If allowed by the 
	extension point an extension may provide it's own factory
	implementation ({@link Extension#getRendererFactory()}.

	<p>
	Renderer factory implementations must be thread-safe, since a 
	single instance of the factory may have to serve multiple requests
	at the same time.
	
	<p>
	Note 1! A single factory instance may be used by more than one 
	extension/extension point. This is at the control of the client application.
	
	<p>
	Note 2! The BASE web-client uses the {@link XmlLoader} to 
	define extensions. The <code>XmlLoader</code> always create a
	separate factory instance for each extension/extension point. 
	The <code>XmlLoader</code> also requires that the factories has 
	a public no-argument constructor.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public interface RendererFactory<A extends Action>
{

	/**
		This method is called once for each request/use of an
		extension and should be used by a factory to initialise the context 
		with resources that the actions may need. With the BASE web-client 
		this means that it is possible to add scripts or stylesheets that 
		is needed by the extension. See 
		{@link net.sf.basedb.clients.web.extensions.JspContext}.
		
		<p>
		Note! This method has no return value as opposed to 
		{@link ActionFactory#prepareContext(InvokationContext)}. The 
		simple reason is that once we get to the point of rendering it
		is already known that the extension is enabled.
		
		@param context The current invokation context
	*/
	public void prepareContext(InvokationContext<? extends A> context);
	
	/**
		Get a renderer for a given context. This method may
		create a new instance or use an existing instance. If
		an existing instance is used, it must be thread-safe, since
		multiple threads may use the same renderer.
		
		@param context The current invokation context
		@return A renderer instance
	*/
	public Renderer<? super A> getRenderer(InvokationContext<? extends A> context);
	
}
