/**
	$Id: BfsImporter.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.PositionBatcher;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SpotBatcher;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.ChainedProgressReporter;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bfs.AnnotationModel;
import net.sf.basedb.util.bfs.AnnotationParser;
import net.sf.basedb.util.bfs.BfsParser;
import net.sf.basedb.util.bfs.DataParser;
import net.sf.basedb.util.bfs.EventHandler;
import net.sf.basedb.util.bfs.EventType;
import net.sf.basedb.util.bfs.ExtraFileImporter;
import net.sf.basedb.util.bfs.InputStreamLocator;
import net.sf.basedb.util.bfs.MetadataModel;
import net.sf.basedb.util.bfs.MetadataParser;
import net.sf.basedb.util.bfs.MultiEventHandler;
import net.sf.basedb.util.bfs.MetadataModel.Section;

/**
	Imports spot data from a serial or matrix BFS. Before it
	can be used various configuration properties must be set.
	<ul>
	<li>An open DbControl: {@link #setDbControl(DbControl)}
	<li>The parent transaction: {@link #setTransformation(Transformation)}
	<li>A parser for the metadata file: {@link #setMetadataParser(MetadataParser)}
	<li>A locator implementation for locating files that are referenced in the
		metadata file: {@link #setInputStreamLocator(InputStreamLocator)}
	</ul>
	<p>

	The import is started with {@link #doImport()}. This will start by parsing the
	metadata file. Additional files that are referenced in the [files] section
	are opened via the {@link InputStreamLocator}. File entries that start with 'x-'
	are considered to be extra files that should be attached to the child bioassay
	set. But this is only done if an {@link ExtraFileImporter} has been specified.
	<p>
	
	The section [settings] is used to control some aspects of the parsing. If
	the data is going into the same datacube the section is not needed. If a 
	new datacube is needed the importer needs to know how to map the spot data
	to positions/reporters and to assays. We support three cases:
	<ol>
	<li>The data uses the same datacube as the parent
	<li>The data needs a new datacube but there is a one-to-one relation
		with the parent assays
	<li>The data needs a new datacube and there may be more than one
		parent assay for each child assay
	</ol>
	
	<p>
	A new datacube is created by including a <code>new-data-cube</code>
	entry in the settings section. Multiple parents per assay is enabled
	by including a <code>multi-assay-parents</code> entry.
	
	<p>
	The child bioassay set is by default assumed to use the same {@link IntensityTransform}
	as the parent bioassay set. If the imported data uses a different transform
	this should be specified by the <code>transform</code> setting. Allowed
	values are those defined by the IntensityTransform enumeration 
	(<code>NONE</code>, <code>LOG2</code> or <code>LOG10</code>).
	
	<p>
	The importer needs 'rdata' and 'pdata' files in all cases. The ID column in the
	rdata file should be the position number. If the data is using the same datacube
	(case 1) no more columns are needed. The order of the positions doesn't matter 
	as long as it matches the spot data files. All positions must be unique. If the data
	needs a new datacube (case 2 and 3) the rdata file also needs one of 'External ID' 
	or 'Internal ID' columns which are used to map positions to reporters.
	<p>
	
	The meaning of the ID column in the pdata file depends on the case. In case 1
	and 2 the ID is the internal ID of the parent assay. No more columns are needed.
	In case 3, the ID column is just a unique positive integer. In this case, the
	pdata file also needs a second column 'Parent ID', which should be a comma-separated
	list of the internal assay ID values of the parent assays.
	
	<p>
	The [sdata] section lists all spot data columns. The importer is looking for 
	entries of the form 'Ch N' where N is 1, 2, ... etc. Exactly one entry for each
	channel is required.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (mån, 20 apr. 2015) $
*/
public class BfsImporter
{
	private boolean autoCloseParsers;
	private DbControl dc;
	private ProgressReporter progress;
	private MetadataParser metadataParser;
	private Transformation transformation;
	private InputStreamLocator locator;
	private ExtraFileImporter extraFileImporter;

	/**
		Create a new importer object.
	*/
	public BfsImporter()
	{
		this.autoCloseParsers = true;
	}
	
	/*
		public configuration properties
		-------------------------------
	*/
	/**
		If this option is set then all parsers are automatically closed
		when all data has been read to them. This setting is enabled by default.
	*/
	public void setAutoCloseParsers(boolean autoClose)
	{
		this.autoCloseParsers = autoClose;
	}

	/**
		Set's the DbControl that should be used to get data from the database.
	*/
	public void setDbControl(DbControl dc)
	{
		this.dc = dc;
	}
	
	/**
		Get the current DbControl.
	*/
	public DbControl getDbControl()
	{
		return dc;
	}
	
	/**
		Set the progress reporter that is used to report progress. Call
		{@link #setProgress(int, String)} to update the current status.
	*/
	public void setProgressReporter(ProgressReporter progress)
	{
		this.progress = progress;
	}
	
	/**
		Get the progress reporter.
	*/
	public ProgressReporter getProgressReporter()
	{
		return progress;
	}

	/**
		Set the destination transformation. A child bioassay set is created.
		@param transformation The transformation
	*/
	public void setTransformation(Transformation transformation)
	{
		this.transformation = transformation;
	}
	
	/**
		Get the destination transformation.
	*/
	public Transformation getTransformation()
	{
		return transformation;
	}

	/**
		Set the metadata file parser. The metadata file should contains
		information about other files and their contents.
	*/
	public void setMetadataParser(MetadataParser parser)
	{
		this.metadataParser = parser;
	}
	
	public MetadataParser getMetadataParser()
	{
		return metadataParser;
	}
	
	/**
		Set the input stream locator that is responsible for opening files
		that are referenced from the metadata file.
	*/
	public void setInputStreamLocator(InputStreamLocator locator)
	{
		this.locator = locator;
	}
	
	public void setExtraFileImporter(ExtraFileImporter extraFileImporter)
	{
		this.extraFileImporter = extraFileImporter;
	}

	/*
		Helper methods
		--------------
	*/
	/**
		Update the progress of the export.
		@see ProgressReporter
	*/
	protected void setProgress(int percent, String message)
	{
		if (progress != null)
		{
			progress.display(percent, message);
		}
	}
	// ----------------------------

	/**
		Start the import.
		@return The created bioassay set or null if the BASEfile didn't contain
			spot data
	*/
	public BioAssaySet doImport()
		throws IOException
	{
		// Check required configuration parameters
		MetadataParser metadataParser = getMetadataParser();
		DbControl dc = getDbControl();
		Transformation t = getTransformation();
		if (metadataParser == null) throw new InvalidUseOfNullException("metadataParser");
		if (dc == null) throw new InvalidUseOfNullException("dbControl");
		if (t == null) throw new InvalidUseOfNullException("transformation");
		
		// Get optional and derived parameters
		BioAssaySet parent = t.getSource();
		int numChannels = parent.getRawDataType().getChannels();
		String metadataFile = metadataParser.getFilename();
		IntensityTransform transform = parent.getIntensityTransform();
		ChainedProgressReporter chainedProgress = progress == null ? 
			null : new ChainedProgressReporter(progress);
		
		setProgress(0, "Initializing BFS importer");
		
		// Parse the metadata file
		MetadataModel metadata = new MetadataModel();
		metadataParser.parse(metadata);
		if (autoCloseParsers) metadataParser.close();
		
		// Serial or matrix file?
		boolean isSerial = "serial".equals(metadata.getSubtype());

		Section sdata = metadata.getSection("sdata");
		boolean hasSdata = sdata != null;
		BioAssaySet child = null;
		
		if (hasSdata)
		{
			MetadataModel.Section settings = metadata.getSection("settings");
			boolean newDataCube = false;
			boolean multiAssayParents = false;
			
			if (settings != null)
			{
				newDataCube = Values.getBoolean(settings.getValue("new-data-cube"));
				if (newDataCube)
				{
					multiAssayParents = Values.getBoolean(settings.getValue("multi-assay-parents"));
				}
				String intensityTransform = settings.getValue("transform");
				if (intensityTransform != null)
				{
					transform = IntensityTransform.valueOf(intensityTransform.toUpperCase());
				}
			}
		
			// Parse the assay annotations file
			String pdataFile = metadata.getFile("pdata");
			if (pdataFile == null) 
			{
				throw new InvalidDataException("Missing 'pdata' entry in [files] section in metadata file: " + metadataFile);
			}
			AnnotationParser pdataParser = AnnotationParser.create(
					locator.getInputStream(pdataFile), pdataFile, locator.getSize(pdataFile));
			AnnotationModel pdataModel = new AnnotationModel();
			pdataParser.parse(pdataModel);
			if (autoCloseParsers) pdataParser.close();

			int parentIndex = pdataModel.getColumnIndex("Parent ID");
			if (multiAssayParents && parentIndex == -1)
			{
				throw new InvalidDataException("Missing column 'Parent ID' in rdata file: " + pdataFile);
			}
		
		
			// Create child bioassay set and bioassays
			int numAssays = pdataModel.getRowCount();
			int nameIndex = pdataModel.getColumnIndex("Name");
			ThreadSignalHandler.checkInterrupted();
			setProgress(5, "Creating " + numAssays + " child assays");
	
			child = t.newProduct(newDataCube ? "new" : null, "new", false);
			child.setIntensityTransform(transform);
			dc.saveItem(child);
			BioAssay[] children = new BioAssay[numAssays];
			for (int i = 0; i < numAssays; ++i)
			{
				BioAssay ba;
				if (multiAssayParents)
				{
					String tmp = pdataModel.getData(i, parentIndex);
					Integer[] parentIds = Values.getInt(tmp.split(","));
					List<BioAssay> parents = new ArrayList<BioAssay>();
					for (int parentAssayId : parentIds)
					{
						parents.add(BioAssay.getById(dc, parentAssayId));
					}
					ba = child.newBioAssay(parents);
				}
				else
				{
					int parentAssayId = pdataModel.getRowId(i);
					if (newDataCube)
					{
						ba = child.newBioAssay(Collections.singleton(BioAssay.getById(dc, parentAssayId)));
					}
					else
					{
						ba = child.newBioAssay(BioAssay.getById(dc, parentAssayId));
					}
				}
				if (nameIndex >= 0) ba.setName(pdataModel.getData(i, nameIndex));
				children[i] = ba;
				dc.saveItem(ba);
			}
		
			// Reporter annotations file has the position values for each row
			String rdataFile = metadata.getFile("rdata");
			if (rdataFile == null) 
			{
				throw new InvalidDataException("Missing 'rdata' entry in [files] section in metadata file: " + metadataFile);
			}
			AnnotationParser rdataParser = AnnotationParser.create(
					locator.getInputStream(rdataFile), rdataFile, locator.getSize(rdataFile));
			AnnotationModel rdataModel = new AnnotationModel();
			rdataModel.setHeaders(rdataParser.parseToBof());
			int posIndex = rdataModel.getColumnIndex("ID");
			int externalIdIndex = rdataModel.getColumnIndex("External ID");
			int internalIdIndex = rdataModel.getColumnIndex("Internal ID");
			if (posIndex == -1) 
			{
				throw new InvalidDataException("Missing column 'ID' in rdata file: " + rdataFile);
			}
			if (newDataCube && externalIdIndex == -1 && internalIdIndex == -1)
			{
				throw new InvalidDataException("Missing 'External ID' and " +
						"'Internal ID' columns in rdata file: " + rdataFile);
			}

			// Load 'sdata' information
			int numSdata = sdata.getEntryCount();
			int[] chIndex = new int[numChannels];
			for (int i = 0; i < numChannels; ++i)
			{
				chIndex[i] = sdata.getIndex("Ch " + (i+1));
				if (chIndex[i] == -1) 
				{
					throw new InvalidDataException("Missing 'Ch "+ (i+1) + 
						"' entry in [sdata] section in file: " + metadataFile);
				}
			}
		
			// Initialize a synchronized parser and load data files
			ThreadSignalHandler.checkInterrupted();
			setProgress(10, "Parsing " + numSdata + " spot data files...");
			SynchronizedSpotDataParser mainParser = new SynchronizedSpotDataParser();
			mainParser.setMasterParser(rdataParser);
			int numSdataFiles = isSerial ? numAssays : numSdata;
			int numDataColumns = isSerial ? numSdata : numAssays;
			for (int i = 1; i <= numSdataFiles; ++i)
			{
				String sdataFile = metadata.getFile("sdata" + i);
				if (sdataFile == null) 
				{
					throw new InvalidDataException("Missing 'sdata" + i + 
							"' entry in [files] section in metadata file: " + metadataFile);
				}
				DataParser dp = DataParser.create(
					locator.getInputStream(sdataFile), sdataFile, locator.getSize(sdataFile));
				dp.setNumColumns(numDataColumns);
				dp.parseToBof();
				mainParser.addDataParser(dp);
			}
		
			// Set up event handlers
			SpotBatcher spotBatcher = child.getSpotBatcher();
			MultiEventHandler masterEventHandler = new MultiEventHandler();
			PositionBatcher posBatcher = null;
			if (newDataCube)
			{
				posBatcher = child.getPositionBatcher();
			}
			// The position handler extracts the position value and is used by other handlers
			PositionEventHandler posHandler = 
				new PositionEventHandler(0, posIndex, posBatcher, internalIdIndex, externalIdIndex);
			masterEventHandler.addEventHandler(posHandler);
			// The spot intensity handler extracts spot intensity values for all channels
			if (isSerial)
			{
				masterEventHandler.addEventHandler(new SerialSpotIntensityEventHandler(spotBatcher, posHandler, children, chIndex));
			}
			else
			{
				masterEventHandler.addEventHandler(new MatrixSpotIntensityEventHandler(spotBatcher, posHandler, children, chIndex));
			}
			if (chainedProgress != null)
			{
				chainedProgress.setRange(15, 90);
				masterEventHandler.addEventHandler(
					new ProgressEventHandler(chainedProgress, rdataParser.getFileSize()));
			}
			
			// Start parsing
			mainParser.parse(masterEventHandler);
			if (autoCloseParsers) mainParser.close();
		}
		
		// Import extra files
		if (extraFileImporter != null)
		{
			MetadataModel.Section files = metadata.getSection("files");
			if (chainedProgress != null)
			{
				chainedProgress.setRange(90, 100);
			}
			int fileCount = files.getEntryCount();
			BasicItem linkWithFile = child == null ? t : child;
			for (int i = 0; i < fileCount; ++i)
			{
				ThreadSignalHandler.checkInterrupted();
				MetadataModel.SectionEntry entry = files.getEntry(i);
				if (entry.getKey().startsWith("x-"))
				{
					String filename = entry.getValue();
					if (chainedProgress != null)
					{
						chainedProgress.display((100 * i) / fileCount, "Importing file: " + filename);
					}
					File file = extraFileImporter.importExtraFile(filename, locator);
					if (file != null)
					{
						AnyToAny ata = AnyToAny.getNew(dc, linkWithFile, file, entry.getKey().replace("x-", ""), false);
						dc.saveItem(ata);
					}
				}
			}
		}
		
		// Final check: did we import anything?
		if (!hasSdata && extraFileImporter == null)
		{
			throw new InvalidDataException("Missing [sdata] section and " +
				"extra file importer in metadata file:" + metadataFile);
		}
		return child;
	}

	static class ProgressEventHandler
		implements EventHandler
	{

		private ProgressReporter progress;
		private long totalBytes;
		private long nextReport;
		private long reportInterval;
		
		ProgressEventHandler(ProgressReporter progress, long totalBytes)
		{
			this.progress = progress;
			this.totalBytes = totalBytes;
			this.nextReport = 0;
			this.reportInterval = Math.max(totalBytes / 20, 100);
		}
		
		@SuppressWarnings("rawtypes")
		@Override
		public void handleEvent(EventType eventType, Object eventData, BfsParser parser)
		{
			if (eventType == SynchronizedSpotDataParser.DATA_EVENT)
			{
				long parsedBytes = parser.getParsedBytes();
				if (parsedBytes > nextReport)
				{
					nextReport = parsedBytes + reportInterval;
					int percent = totalBytes == -1 ? -1 : (int)((100 * parsedBytes) / totalBytes);
					progress.display(percent, "Importing spot data. Parsed " + parser.getCurrentLine() + 
						" lines so far.");
				}
			}
		}
		
	}
	
}
