/**
	$Id: MAParser.java 7665 2019-03-20 13:31:03Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.util.List;

import net.sf.basedb.util.basefile.BaseFileParser;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	Spot intensity parser implementation for 2-channel data 
	that looks for M/A values in the spot data and converts them to ch1/ch2
	intensity values. The required assay fields are:
	l2ratio1_2 (M) and l10intgmean1_2 (A).
	<p>
	NOTE! The column names for the M/A values can be redefined by 
	calling {@link BaseFileParser#setRedefinedColumnName(String, String, String)}.
	
	@author Nicklas
	@version 2.14
	@base.modified $Date: 2019-03-20 14:31:03 +0100 (ons, 20 mars 2019) $
*/
public class MAParser
	implements SpotIntensityParser
{

	private int mIndex;
	private int aIndex;
	
	private static final String mColumn = "l2ratio1_2";
	private static final String aColumn = "l10intgmean1_2";
	
	/**
		Creates a new MA-parser.
	*/
	public MAParser()
	{}
	
	/* 
		From the SpotIntensityParser interface
		--------------------------------------
	*/
	@Override
	public boolean hasRequiredAssayFields(BaseFileParser parser, FlatFileParser ffp,
			List<String> assayFields)
	{
		int mIndex = assayFields.indexOf(parser.getRedefinedColumnName("spots", mColumn));
		int aIndex = assayFields.indexOf(parser.getRedefinedColumnName("spots", aColumn));
		return mIndex >= 0 && aIndex >= 0;
	}
	@Override
	public void beginSection(BaseFileParser parser, FlatFileParser ffp, 
			List<String> assayFields)
	{
		mIndex = assayFields.indexOf(parser.getRedefinedColumnName("spots", mColumn));
		aIndex = assayFields.indexOf(parser.getRedefinedColumnName("spots", aColumn));
	}

	@Override
	public boolean setIntensities(Data data, float[] intensities, int firstIndex)
	{
		Float m = data.getFloat(mIndex + firstIndex, null, true);
		if (m == null) return false;
		Float a = data.getFloat(aIndex + firstIndex, null, true);
		if (a == null) return false;
		// int2 = 10^a / 2^(0.5*m)
		// int1 = int2 * 2^m
		intensities[1] = (float)(Math.pow(10, a) / Math.pow(2, 0.5 * m));
		intensities[0] = (float)(intensities[1] * Math.pow(2, m));
		return true;
	}
	// ---------------------------------------

}
