/**
	$Id: PositionEventHandler.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.PositionBatcher;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.bfs.BfsParser;
import net.sf.basedb.util.bfs.EventHandler;
import net.sf.basedb.util.bfs.EventType;
import net.sf.basedb.util.bfs.MultiEventHandler;
import net.sf.basedb.util.importer.spotdata.SynchronizedSpotDataParser.SynchronizedData;

/**
	Event handler implementation intended to be used with a 
	{@link SynchronizedSpotDataParser} to extract the position
	number from the parsed data. When cooperation is needed
	between multiple event handlers (eg. using the {@link MultiEventHandler}
	class, make sure that this event handler receives the event
	before any other handler that needs the position value.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (mån, 20 apr. 2015) $
*/
public class PositionEventHandler
	implements EventHandler

{
	
	private final int parserIndex;
	private final int posIndex;
	private final PositionBatcher posBatcher;
	private final int reporterIndex;
	private final int externalIndex;
	private ReporterBatcher reporterBatcher;
	private int currentPosition;
	
	/**
		Create a new event handler. The position value is expected to
		be found in the data column given by the two indexes. The
		first index is the parser index. Usually this should be 0=the 
		master parser. The second index is the column holding the position
		value, eg. the ID column which should also be 0.
		
		@param parserIndex The index of the parser, a value between 0 and
			the number of synchronized data parsers
		@param posIndex The index of the data column holding the position number, 
			a value between 0 and the number of columns in the parsed file
		@param posBatcher An optional position batcher if a position/reporter
			needs to be created
		@param reporterIndex The index of the data column holding the internal
			reporter id, ignored if no position batcher is used, -1 if the
			external id index should be used instead of the internal id
		@param externalIndex The index of the data column holding the external
			reporter id, only used if reporterIndex=-1 is specified
	*/
	public PositionEventHandler(int parserIndex, int posIndex, 
		PositionBatcher posBatcher, int reporterIndex, int externalIndex)
	{
		this.parserIndex = parserIndex;
		this.posIndex = posIndex;
		this.posBatcher = posBatcher;
		this.reporterIndex = reporterIndex;
		this.externalIndex = externalIndex;
		if (reporterIndex == -1 && posBatcher != null)
		{
			reporterBatcher = ReporterBatcher.getNew(posBatcher.getDbControl());
		}
	}
	
	/**
		Get the position number from the given parser/column. An exception is thrown
		if a valid integer value is not found.
	*/
	@SuppressWarnings("rawtypes")
	@Override
	public void handleEvent(EventType eventType, Object eventData, BfsParser parser)
	{
		if (eventType == SynchronizedSpotDataParser.DATA_EVENT)
		{
			SynchronizedData[] sdata = (SynchronizedData[])eventData;
			SynchronizedData posData = sdata[parserIndex];
			String[] rdata = posData.getData();
			String value = rdata[posIndex];
			Integer position = Values.getInteger(value, null);
			if (position == null)
			{
				BfsParser posParser = posData.getParser();
				throw new BaseException("Missing or invalid value for 'position'" +
						" at line " + posParser.getCurrentLine() + 
						", column " + posIndex +
						", in file '" + posParser.getFilename() + "': " + value);
			}
			if (posBatcher != null)
			{
				ReporterData reporter = null;
				if (reporterIndex != -1)
				{
					value = rdata[reporterIndex];
					Integer reporterId = Values.getInteger(value, null);
					if (reporterId != null) reporter = Reporter.getProxy(reporterIndex);
				}
				else
				{
					value = Values.getStringOrNull(rdata[externalIndex]);
					if (value != null) reporter = reporterBatcher.getByExternalId(value);
				}
				posBatcher.insert(position, reporter);
			}
			currentPosition = position;
		}
	}

	/**
		Get the current position.
	*/
	public int getCurrentPosition()
	{
		return currentPosition;
	}
}
