package net.sf.basedb.util.listable;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.SyncFilter;

/**
	A simple cache implementation for holding the result of source item to target item transformation.

	@author nicklas
	@since 3.16
*/
public class TransformCache 
{

	/**
		Get a key that is able to identify a given transformation. We need the source and target
		item types as well as the transformation direction and all source items ids. It
		is recommended that the sourceIds set is a {@link SortedSet}. If not, a new sorted set
		is automatically created.
	*/
	public static CacheKey getKey(Item sourceType, Item targetType, SyncFilter.SourceItemTransform transform, Set<Integer> sourceIds)
	{
		SortedSet<Integer> sortedSrc = sourceIds instanceof SortedSet ? (SortedSet<Integer>)sourceIds : new TreeSet<>(sourceIds);
		return new CacheKey(sourceType, targetType, transform, sortedSrc);
	}
	
	private final Map<CacheKey, CacheEntry> cache;
	private final long timeoutInMillis;
	private long nextCleanup;
	
	/**
		Creates a new cache with the specified timeout.
	*/
	public TransformCache(int timeoutInMinutes)
	{
		this.cache = Collections.synchronizedMap(new HashMap<>());
		this.timeoutInMillis = timeoutInMinutes * 60000;
		this.nextCleanup = System.currentTimeMillis() + timeoutInMillis;
	}
	
	/**
		Get a cached entry if. Null is returned if there is no
		entry or if the existing entry is too old.
	*/
	public Set<Integer> get(CacheKey key)
	{
		CacheEntry entry = cache.get(key);
		Set<Integer> target = null;
		if (entry != null)
		{
			if (entry.getTimeout() < System.currentTimeMillis())
			{
				cache.remove(key);
			}
			else
			{
				target = entry.getTargetIds();
			}
		}
		return target;
	}
	
	/**
		Store a new entry into the cache.
	*/
	public void store(CacheKey key, Set<Integer> targetList)
	{
		CacheEntry entry = new CacheEntry(System.currentTimeMillis()+timeoutInMillis, targetList);
		cache.put(key, entry);
		if (System.currentTimeMillis() > nextCleanup) clean();
	}
	
	void clean()
	{
		synchronized (cache)
		{
			long time = System.currentTimeMillis();
			Iterator<Entry<CacheKey, CacheEntry>> it = cache.entrySet().iterator();
			while (it.hasNext())
			{
				if (it.next().getValue().getTimeout() < time) it.remove();
			}
			nextCleanup = time + timeoutInMillis;
		}
	}
	
	public static class CacheKey
	{
		private final Item sourceType;
		private final Item targetType;
		private final SyncFilter.SourceItemTransform transform;
		private final String sourceIdHash;
		
		CacheKey(Item sourceType, Item targetType, SyncFilter.SourceItemTransform transform, SortedSet<Integer> sourceIds)
		{
			this.sourceType = sourceType;
			this.targetType = targetType;
			this.transform = transform;
			this.sourceIdHash = SyncFilter.toMd5(sourceIds);
		}

		@Override
		public int hashCode() 
		{
			return sourceType.hashCode() + 3 * targetType.hashCode() + 7 * transform.hashCode() + 11 * sourceIdHash.hashCode();
		}

		@Override
		public boolean equals(Object obj) 
		{
			if (!(obj instanceof CacheKey)) return false;
			CacheKey other = (CacheKey)obj;
			return sourceType == other.sourceType && targetType == other.targetType && 
				transform == other.transform && sourceIdHash.equals(other.sourceIdHash);
		}

		@Override
		public String toString() 
		{
			return "CacheKey[" + sourceType.name() + "->" + targetType.name() + "; " + transform.name() + "; " + sourceIdHash + "]";
		}		
	}
	
	
	static class CacheEntry
	{
		private final Set<Integer> targetList;
		private final long timeout;
		
		CacheEntry(long timeout, Set<Integer> targetList)
		{
			this.targetList = targetList;
			this.timeout = timeout;
		}
		
		long getTimeout()
		{
			return timeout;
		}
		
		Set<Integer> getTargetIds()
		{
			return targetList;
		}
	}
}
