/**
	$Id$

	Copyright (C) 2012 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.util.extensions.ClientContext;
import net.sf.basedb.util.extensions.ExtensionsInvoker;
import net.sf.basedb.util.extensions.Registry;
import net.sf.basedb.util.extensions.manager.Settings;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.extensions.ChildNodeLoaderAction;

/**
	Wrapper for node loaders that automatically hooks into the
	extension system when loading child nodes for a parent node.
	It works by first delegating all method calls to the parent
	node loader, but the {@link #loadChildNodes(DbControl, OverviewContext, Node)}
	contains additional logic to query the extension system and
	call {@link ChildNodeLoaderAction#createChildNodes(DbControl, OverviewContext, Node)}
	for any action created by the an extension.

	@author Nicklas
	@since 3.2
*/
public class ExtensionChildNodeLoader<I>
	implements NodeLoader<I>
{

	private final NodeLoader<I> parent;
	
	/**
		Wrap the given parent node loader.
	*/
	public ExtensionChildNodeLoader(NodeLoader<I> parent)
	{
		this.parent = parent;
	}
	/*
		From the NodeLoader interface
		-----------------------------
	*/
	/**
		Call the same method on the 'parent' node loader.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		return parent.createForwardNode(dc, context, parentNode);
	}

	/**
		Call the same method on the 'parent' node loader.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createPropertyNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		return parent.createPropertyNode(dc, context, parentNode);
	}

	/**
		Call the same method on the 'parent' node loader. If one or more
		child nodes are added to the parent node, the 
		{@link #loadChildNodes(DbControl, OverviewContext, Node)} is called
		for each of the new nodes.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node childNode)
	{
		return parent.createReverseNode(dc, context, childNode);
	}

	/**
		Call the same method on the 'parent' node loader.
		@return The node that the parent node loader creates (may be null)
	*/
	@Override
	public Node createRootNode(DbControl dc, OverviewContext context, I item)
	{
		return parent.createRootNode(dc, context, item);
	}

	/**
		Call the same method on the 'parent' node loader. Then,
		check the extension system for extensions to the 'net.sf.basedb.util.overview.loader'
		extension point and ask the factories for {@link ChildNodeLoaderAction}:s and let
		each of them create their child nodes.
		
		@return TRUE if new nodes were added to the parent node, FALSE if not
	*/
	@Override
	public boolean loadChildNodes(DbControl dc, OverviewContext context, Node node)
	{
		if (node.isChildrenLoaded()) return false;
		// Delegate to parent first
		parent.loadChildNodes(dc, context, node);
		
		// Create a context with the parent node as the current item
		ClientContext clientContext = new ClientContext(dc, node);
		
		// Find and use any registered extensions
		Registry registry = Application.getExtensionsManager().getRegistry();
		Settings settings = Application.getExtensionsManager().getSettings();
		ExtensionsInvoker<ChildNodeLoaderAction> invoker = 
			registry.useExtensions(clientContext, settings, "net.sf.basedb.util.overview.loader");
		
		for (ChildNodeLoaderAction a : invoker)
		{
			a.createChildNodes(dc, context, node);
		}
		node.setChildrenLoaded();
		return node.numChildren() > 0;
	}
	// --------------------------------

	@Override
	public String toString()
	{
		return super.toString() + "-->" + parent.toString();
	}
	
}
