/*
	$Id: CompoundMapper.java 7703 2019-04-11 07:31:26Z nicklas $
	
	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.parser;

import java.text.NumberFormat;
import java.util.Date;
import java.util.List;

import net.sf.basedb.core.Type;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.parser.FlatFileParser.Data;

/**
	A compound mapper that gets its value by invoking multiple other mappers
	which are set in the constructor. This type of mapper is used to combine
	constant strings with column values.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
*/
public class CompoundMapper
	implements Mapper
{
	private final List<Mapper> mappers;
	private final NumberFormat parser;
	private final Formatter<Date> dateParser;
	private final boolean nullIfException;
	
	/**
		Create a new compound mapper.
		@param mappers A list of other mappers that will be invoked in 
			the order they appear in the list
	*/
	public CompoundMapper(List<Mapper> mappers)
	{
		this(mappers, null, null, false);
	}
	
	/**
		Create a new compound mapper, using a specific number formatter as it's parser.
		@param mappers A list of other mappers that will be invoked in 
			the order they appear in the list
		@param parser The number format to use or null to use Float.valueOf()
		@param nullIfException If TRUE, the mapper returns null for unparsable numeric
			values, otherwise an excption is thrown
		@since 2.4
	*/
	public CompoundMapper(List<Mapper> mappers, NumberFormat parser, boolean nullIfException)
	{
		this(mappers, parser, null, nullIfException);
	}
	
	/**
		Create a new compound mapper, using a specific number formatter as it's parser.
		@param mappers A list of other mappers that will be invoked in 
			the order they appear in the list
		@param parser The date format to use 
		@param nullIfException If TRUE, the mapper returns null for unparsable numeric
			values, otherwise an excption is thrown
		@since 3.15
	*/
	public CompoundMapper(List<Mapper> mappers, NumberFormat parser, Formatter<Date> dateParser, boolean nullIfException)
	{
		this.mappers = mappers;
		this.parser = parser;
		this.dateParser = dateParser;
		this.nullIfException = nullIfException;
	}

	/*
		From the Mapper interface
		-------------------------------------------
	*/
	@Override
	@Deprecated
	public String getValue(Data data)
	{
		return getString(data);
	}
	/**
		Get a value by invoking each mapper in the order they appear in
		the list and concatenating the result.
		@return The concatenated value
	*/
	@Override
	public String getString(Data data)
	{
		StringBuilder sb = new StringBuilder();
		for (Mapper m : mappers)
		{
			sb.append(m.getString(data));
		}
		return sb.toString();
	}
	@Override
	public Integer getInt(Data data)
	{
		if (mappers.size() == 1) return mappers.get(0).getInt(data);
		String stringValue = getString(data);
		Integer intValue = null;
		if (stringValue != null)
		{
			intValue = (Integer)Type.INT.parseString(stringValue, parser, nullIfException);
		}
		return intValue;
	}
	@Override
	public Long getLong(Data data)
	{
		if (mappers.size() == 1) return mappers.get(0).getLong(data);
		String stringValue = getString(data);
		Long longValue = null;
		if (stringValue != null)
		{
			longValue = (Long)Type.LONG.parseString(stringValue, parser, nullIfException);
		}
		return longValue;
	}
	@Override
	public Float getFloat(Data data)
	{
		if (mappers.size() == 1) return mappers.get(0).getFloat(data);
		String stringValue = getString(data);
		Float floatValue = null;
		if (stringValue != null)
		{
			floatValue = (Float)Type.FLOAT.parseString(stringValue, parser, nullIfException);
		}
		return floatValue;
	}
	@Override
	public Double getDouble(Data data)
	{
		if (mappers.size() == 1) return mappers.get(0).getDouble(data);
		String stringValue = getString(data);
		Double doubleValue = null;
		if (stringValue != null)
		{
			doubleValue = (Double)Type.DOUBLE.parseString(stringValue, parser, nullIfException);
		}
		return doubleValue;
	}
	@Override
	public Date getDate(Data data)
	{
		if (mappers.size() == 1) return mappers.get(0).getDate(data);
		String stringValue = getString(data);
		Date dateValue = null;
		if (stringValue != null)
		{
			dateValue = (Date)Type.DATE.parseString(stringValue, dateParser, nullIfException);
		}
		return dateValue;
	}
	// -------------------------------------------
	
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		for (Mapper m : mappers)
		{
			sb.append(m.toString());
		}
		return sb.toString();
	}
	// -------------------------------------------
}

