/*
	$Id: FlatFileParser.java 7738 2019-08-14 07:16:18Z nicklas $
	
	Copyright (C) 2005 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Nicklas Nordborg, Martin Svensson, Gregory Vincic
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.
	
	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.parser;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.nio.charset.Charset;
import java.text.NumberFormat;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.regex.PatternSyntaxException;

import org.apache.poi.ss.usermodel.Cell;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.Collections;
import java.util.Date;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.InputStreamTracker;
import net.sf.basedb.util.NumberFormatUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.charset.CharsetUtil;
import net.sf.basedb.util.excel.XlsxToCsvUtil;
import net.sf.basedb.util.excel.XlsxToCsvUtil.SheetInfo;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.NumberFormatFormatter;
import net.sf.basedb.util.jep.JepFunction;

/**

	This class can be used to parse data from flat text files and from
	Excel workbooks in xlsx format. If the file is a text file it must 
	follow a few simple rules:
	<ul>
	<li>Data must be organised into columns, with one record per line
	<li>Each data column must be separated by some special character
		or character sequence not occuring in the data, for example
		a tab or a comma. Data in fixed-size columns cannot be parsed.
	<li>Data may optionally be preceeded by a data header, ie.
		the names of the columns
	<li>The data header may optionally be preceeded by file headers.
		A file header is something that can be split in a name-value
		pair.
	<li>The file may contain comments, which are ignored by the parser
	<li>The file contain section where each section can contain a 
		header and/or a data part
	</ul>
	<p>
	<b>Example</b><br>
<pre class="code">
# Example of a parsable file, not actual format of a GenePix file
section info
Type=GenePix Results 1.3
DateTime=2002/09/04 13:59:48
Scanner=GenePix 4000B [83306]

section data
Block	Column	Row	Name	ID
1	1	1	"Ly68_Lymphocyte antigen 68"	"M000205_01"
1	2	1	"Bag1_Bcl2-associated athanogene 1"	"M000209_01"
1	3	1	"Rps16_Ribosomal protein S16"	"M000213_01"
1	4	1	"Col4a1_Procollagen, type IV, alpha 1"	"M000229_01"
1	5	1	"Ace_Angiotensin converting enzyme"	"M000233_01"
1	6	1	"Cd5_CD5 antigen"	"M000237_01"
1	7	1	"Psme1_Protease (prosome, macropain) 28 s"
</pre>

	<p>
	If the file is an Excel file the first sheet is automatically converted to a
	tab-separated text file by default. To use a different sheet call the
	{@link #setExcelSheet(String)} method with the name or index of the sheet.
	Initial parsing and regular expression matching is always done against the 
	text representation of the selected sheet. Note that empty cells on the top
	and left are usually cut away. When retrieving values via the {@link Data}
	class it will typically go directly to the mapped cell from the Excel sheet
	and get the value, which means that numeric and date values doesn't have to 
	be converted to and from strings if not needed. For example, if a date 
	value is requested and the mapped cell is date that will be used as it is, 
	but it the mapped cell is a string, the same parsers that are used for CSV 
	files are used to convert the string to a date.

	<p>
	<b>How to use</b><br>
	The parsing is controlled by regular expressions. Start by
	creating a new <code>FlatFileParser</code> object. Use the
	various <code>set</code> methods to provide regular expression
	used to match the data/headers.
	<p>
	Use the {@link #setInputStream(InputStream, String)}
	method to specify a file to parse, and {@link #parseHeaders()} to start the 
	parsing. Note! Even if you know that
	the file doesn't contain any headers, you should always call this
	method since the parser must initialize itself. If there are sections 
	in the file use {@link #nextSection()} first to control which section 
	you are parsing from. 
	<p>
	When the headers have been found use the {@link #hasMoreData()}
	and {@link #nextData()} methods in a loop to read all
	data from the section.
	<p>
	<b>Example</b><br>
<pre class="code">
FlatFileParser ffp = new FlatFileParser();
ffp.setHeaderRegexp(Pattern.compile("(.*)=(.*)"));
ffp.setDataHeaderRegexp(Pattern.compile("Block\\tColumn\\tRow\\tName\\tID"));
ffp.setDataSplitterRegexp(Pattern.compile("\\t"));
ffp.setIgnoreRegexp(Pattern.compile("#.*"));
ffp.setMinDataColumns(5);
ffp.setMaxDataColumns(5);
ffp.setInputStream(FileUtil.getInputStream(path_to_file), Config.getCharset());
ffp.parseHeaders();
for (int i = 0; i &lt; ffp.getLineCount(); i++)
{
   FlatFileParser.Line line = ffp.getLine(i);
   System.out.println(i+":"+line.type()+":"+line.line());
}
int i = 0;
while (ffp.hasMoreData())
{
   FlatFileParser.Data data = ffp.nextData();
   System.out.println(i+":"+data.columns()+":"+data.line());
}
</pre>

	<p>
	<b>Mapping column values</b><br>
	With the <code>FlatFileParser.Data</code> object you can only
	access the data by column index (0-based). Another approach is to 
	use {@link Mapper}:s. A mapper takes a string template and inserts the 
	values of the data columns where you specify. Here are some example:
	<pre class="code">
\1\
\row\
Row: \row\, Col:\col\
=2 * col('Radius')
</pre>
	
	The result can be retrieved either as a string, as a numeric value
	or as a date. It is even possible to create expressions that
	does a calculation on the value before it is returned.
	See the {@link #getMapper(String)} method for more information.

	@author Nicklas, Enell
	@version 2.0
	@base.modified $Date$
*/
public class FlatFileParser
{

	/**
		The default value for the number of unknown lines
		in a row that may be encountered by the {@link
		#parseHeaders() parseHeaders} method before it
		gives up.
		@see #setMaxUnknownLines(int) setMaxUnknownLines
	*/
	public static final int DEFAULT_MAX_UNKNOWN_LINES = 100;


	/**
		Pattern used to find column mappings in a string, ie. abc \col\ def
	*/
	private static final Pattern findColumn = Pattern.compile("\\\\([^\\\\]+)\\\\");

	/**
		Reads from the given input stream
	*/
	private BufferedReader reader;

	/**
		For keeping track of the number of bytes parsed.
	*/
	private InputStreamTracker tracker;
	
	/**
		Name of the Excel sheet to parse if the file is an Excel file.
	*/
	private String excelSheetName;
	
	/**
		Flag to indicate if only a single (=false) or all (=true) Excel sheets should
		be parsed in one go.
	*/
	private boolean parseAllExcelSheets;
	
	/**
		Excel workbook that has been loaded.
	*/
	private XlsxToCsvUtil excelWorkbook;
	
	/**
		Excel sheet that is currently being parsed.
	*/
	private XlsxToCsvUtil.SheetInfo excelSheet;
	
	/**
		The value of the parsedLines when the current Excel sheet is parsed.
		Used to find the correct row number in the current sheet (=parsedLines - excelParsedLinesOffset)
	 */
	private int excelParsedLinesOffset;
	
	/**
		The regular expression for matching the beginning-of-file marker
	*/
	private Pattern bofMarker;
	
	/**
		The regular expression for matching a header line.
	*/
	private Pattern header;

	/**
		The regular expression for matching the fist line of a section.
		The expression must have one capturing group.
	*/
	private Pattern section;
	
	/**
		The regular expression for matching the data header line.
		The expression must have two capturing groups.
	*/
	private Pattern dataHeader;

	/**
		The regular expression for splitting a data line.
	*/
	private Pattern dataSplitter;

	/**
	 	If quotes should be trimmed from data values or not. Default true.
	 	Quotes are double or single quotes.
	*/
	private boolean trimQuotes = true;
	
	/**
		If white space should be trimmed from data values or not.
		Default is false.
		@since 3.15.1
	*/
	private boolean trimWhiteSpace;
	
	/**
		The regular expression for matching the data footer line.
	*/
	private Pattern dataFooter;

	/**
		The minimun number of allowed data columns for a line to be considered
		a data line.
	*/
	private int minDataColumns;

	/**
		The maximun number of allowed data columns for a line to be considered
		a data line.
	*/
	private int maxDataColumns;

	/**
		The regular expression for matching a comment line.
	*/
	private Pattern ignore;

	/**
		The maximum number of unkown lines to parse before giving up.
	*/
	private int maxUnknownLines = DEFAULT_MAX_UNKNOWN_LINES;

	/**
		If <code>null</code> should be returned for empty columns (instead of an empty string).
	*/
	private boolean emptyIsNull = true;
	
	/**
		If <code>null</code> should be returned if a (numeric) value can't be parsed.
	*/
	private boolean useNullIfException = false;

	/**
		If non-existing columns should be ignored (true) or result in an
		exception (false) 
	*/
	private boolean ignoreNonExistingColumns = false;

	/**
		If <code>null</code> should be returned for the string NULL (ignoring case)
		or not. 
	*/
	private boolean nullIsNull = true;
	
	/**
		The default number formatter to use for creating mappers.
	*/
	private NumberFormat numberFormat = null;
	
	/**
		The default date formatter to use when creating mappers.
	*/
	private Formatter<Date> dateFormat = null;
	
	/**
		The default timestamp format to use when creating mappers.
	*/
	private Formatter<Date> timestampFormat = null;

	/**
		The value that was captured by the bofMarker pattern.
	*/
	private String bofType;
	
	/**
		List of lines parsed by the {@link #parseHeaders()} method.
	*/
	private List<Line> lines;

	/**
		The total number of lines parsed so far.
	*/
	private int parsedLines;

	/**
		The number of sections parsed so far.
	*/
	private int parsedSections;
	
	/**
		The total number of parsed characters so far.
	*/
	private long parsedCharacters;
	
	/**
		The number of data lines parsed in the current section so far.
		This value is reset at each new section.
	*/
	private int parsedDataLines;
	
	/**
		Map of header lines parsed by the {@link #parseHeaders()} method.
		The map contains name -&gt; value pairs
	*/
	private Map<String, String> headers;
	
	/**
		List of the column names found by splitting the data header using the
		data splitter regexp.
	*/
	private List<String> columnHeaders;
	
	/**
		The line that last matched the {@link #section}.
	*/
	private Line nextSection;

	/**
		The next available data line as parsed by the {@link #hasMoreData()}
		method.
	*/
	private Data nextData;

	/**
		Number of ignored lines in the {@link #nextData()} method.
	*/
	private int ignoredLines;
	/**
		Number of unknown lines in the {@link #nextData()} method.
	*/
	private int unknownLines;
	
	/**
		If unknown or ignored lines should be kept.
		@see #getSkippedLines()
	*/
	private boolean keepSkippedLines = false;
	
	/**
		List for keeping ignored and unknown lines in the {@link #nextData()} method.
	*/
	private List<Line> skippedLines;
	
	/**
		Create a new <code>FlatFileParser</code> object.
	*/
	public FlatFileParser()
	{
		parsedLines = 0;
		parsedCharacters = 0;
	}

	/**
		Set a regular expression that maches a beginning-of-file
		marker. This property should be set before starting to parse
		the file (otherwise it is ignored). The first method call that
		causes the parsing to be started will invoke {@link #parseToBof()} 
		(can also be invoked manually).
		<p>
		The regular expression may contain a single capturing group. The
		matched value is returned by {@link #getBofType()}.
		
		@param regexp A regular expression
		@since 2.15
	*/
	public void setBofMarkerRegexp(Pattern regexp)
	{
		this.bofMarker = regexp;
	}
	
	/**
		Set a regular expression that can be matched against a header.
		The regular expression must contain two capturing
		groups, the first should capture the name and the second the value
		of the header. For example, the file contains headers like:
<pre class="code">
"Type=GenePix Results 1.3"
"DateTime=2002/09/04 13:59:48"
</pre>
		To match this we can use the following regular expression:
		<code>"(.*)=(.*)"</code>.

		@param regexp A regular expression
	*/
	public void setHeaderRegexp(Pattern regexp)
	{
		this.header = regexp;
	}

	/**
		Set a regular expression that can be matched against the section
		line. For example, the file contains a section like:
<pre class="code">
[FileInformation]
</pre>
		To match this we can use the following regular expression:
		<code>section (.*)</code>. This will match to anything
		that starts with "section ". The section name will be in the
		capturing group.

		@param regexp A regular expression
	*/
	public void setSectionRegexp(Pattern regexp)
	{
		this.section = regexp;
	}
	
	/**
		Set a regular expression that can be matched against the data
		header. For example, the file contains a data header like:
<pre class="code">
"Block"{tab}"Column"{tab}"Row"{tab}"Name"{tab}"ID" ...and so on
</pre>
		To match this we can use the following regular expression:
		<code>"(.*?)"(\t"(.*?)")</code>. This will match to anything
		that has at least two columns. We could also be more specific and use:
		<code>"Block"\t"Column"\t"Row"\t"Name"\t"ID"...</code>

		@param regexp A regular expression
	*/
	public void setDataHeaderRegexp(Pattern regexp)
	{
		this.dataHeader = regexp;
	}

	/**
		Set a regular expression that is used to split a data line into
		columns. To split on tabs we use: <code>\t</code>. This regular expression is 
		also used to split the data header line into column names, which can then be used
		in the {@link #getMapper(String)} method.

		@param regexp A regular expression
		@see #setMinDataColumns(int) setMinDataColumns
		@see #setMaxDataColumns(int) setMaxDataColumns
	*/
	public void setDataSplitterRegexp(Pattern regexp)
	{
		this.dataSplitter = regexp;
	}
	
	/**
		Set if quotes around each data value should be removed or not. A quote is
		either a double quote (") or a single quote ('). The default setting of
		this option is true.
		@param trimQuotes TRUE to remove quotes, FALSE to keep them
	 */
	public void setTrimQuotes(boolean trimQuotes)
	{
		this.trimQuotes = trimQuotes;
	}
	
	/**
		Set a flag indicating if white-space should be trimmed from
		start and end of data values. The default setting is false.
		@param trimWhiteSpace TRUE to remove white-space, FALSE to keep them
		@since 3.15.1
	 */
	public void setTrimWhiteSpace(boolean trimWhiteSpace)
	{
		this.trimWhiteSpace = trimWhiteSpace;
	}

	/**
		Set the minimum number of columns a data line must contain in
		order for it to be counted as a data line.

		@param columns The minimum number of columns
	*/
	public void setMinDataColumns(int columns)
	{
		this.minDataColumns = columns;
	}
	/**
		Set the maximum number of columns a data line can contain in
		order for it to be counted as a data line. 

		@param columns The maximum number of columns, or 0 for an
			unlimited number, or -1 to disable counting the number of columns
	*/
	public void setMaxDataColumns(int columns)
	{
		this.maxDataColumns = columns;
	}

	/**
		Set a regular expression that can be matched against a
		data footer. If a line matching this pattern is found while
		looking for data with the {@link #hasMoreData() hasMoreData}
		method it will exit and no more data will be returned.

		@param regexp A regular expression
	*/
	public void setDataFooterRegexp(Pattern regexp)
	{
		this.dataFooter = regexp;
	}

	/**
		Set a regular expression that is used to match a line that should
		be ignored. For example, the file may contain comments starting
		with a #: <code>\#.*</code>

		@param regexp A regular expression
	*/
	public void setIgnoreRegexp(Pattern regexp)
	{
		this.ignore = regexp;
	}

	/**
		The number of unknown lines in a row that can be parsed by
		the {@link #parseHeaders() parseHeaders} method before it gives
		up. The default value is specified by {#link #DEFAULT_MAX_UNKNOWN_LINES}.
		This value is ignored while parsing data.

		@param lines The number of lines
	*/
	public void setMaxUnknownLines(int lines)
	{
		this.maxUnknownLines = lines;
	}

	/**
		Specify if <code>null</code> values should be returned instead of empty strings
		for columns that doesn't contain any value.
		@param emptyIsNull TRUE to return null, FALSE to return an empty string
	*/
	public void setUseNullIfEmpty(boolean emptyIsNull)
	{
		this.emptyIsNull = emptyIsNull;
	}
	
	/**
		Specify if <code>null</code> values should be returned for strings
		having the value "NULL" (ignoring case).
		@param nullIsNull TRUE to return null, FALSE to return the original string value
	*/
	public void setUseNullIfNull(boolean nullIsNull)
	{
		this.nullIsNull = nullIsNull;
	}

	/**
		If the {@link #nextData()} and {@link #hasMoreData()} methods should
		keep information of lines that was skipped because they matched the
		ignore pattern or could be interpreted as data lines. The default is
		FALSE. The number of lines that was skipped is always available regardless
		of this setting.
		
		@param keep TRUE to keep line information, FALSE to not
		@see #getSkippedLines()
		@see #getIgnoredLines()
		@see #getUnknownLines()
		@see #getNumSkippedLines()
	*/
	public void setKeepSkippedLines(boolean keep)
	{
		this.keepSkippedLines = keep;
	}
	
	/**
		If this flag is set and the source file is an Excel file,
		then all sheets will be parsed unless a named sheet
		specified. Each sheet is handled like a section with the sheet
		name inside brackets ([name]). The regular expression for 
		detecting a section is automatically updated to match this
		pattern.
		@since 3.15
	*/
	public void setParseAllExcelSheets(boolean parseAllExcelSheets)
	{
		this.parseAllExcelSheets = parseAllExcelSheets;
	}
	
	/**
		@see #setParseAllExcelSheets(boolean)
		@since 3.15
	*/
	public boolean getParseAllExcelSheets()
	{
		return parseAllExcelSheets;
	}
	
	/**
		Set the name of Excel worksheet to parse if the given file is an Excel
		file, otherwise this is ignored.
		@since 3.15
	*/
	public void setExcelSheet(String name)
	{
		this.excelSheetName = name;
	}
	
	/**
		Get the name of the Excel sheet that should be or is parsed.
		@since 3.15
	*/
	public String getExcelSheet()
	{
		return excelSheetName;
	}
	
	/**
		If the input stream that is being parsed is an Excel
		document, this method returns information about it.
		@return An XlsxToCsvUtil object or null if the stream is not an Excel document
		@since 3.15.1
	*/
	public XlsxToCsvUtil getCurrentExcelWorkbook()
	{
		return excelWorkbook;
	}
	
	/**
		If the input stream that is being parsed is an Excel
		document, this method returns information about the current
		worksheet.
		@return An XlsxToCsvUtil.SheetInfo object or null if the stream is not an Excel document
		@since 3.15.1
	*/
	public XlsxToCsvUtil.SheetInfo getCurrentSheet()
	{
		return excelSheet;
	}
	
	/**
		Set the input stream that will be parsed. The stream can be either a
		text CSV-like stream or an Excel workbook (xlsx). 
		
		If the stream is an Excel workbook the following apply:
		
		* If no date format has been specified, yyyy-MM-dd is used
		* If no timestamp format has been specified, yyyy-MM-dd hh:mm:ss is used
		* If no number format has been specified, 'dot' is used
		* The data splitter regular expression is changed to \\t
		* The section regular expression is changed to [.*] (if the {@link #getParseAllExcelSheets()} flag is set
		
		@param in The <code>InputStream</code>
		@param charsetOrSheetName If CSV, the name of the character set to 
			use when parsing the file, or null to use the default charset specified by
			{@link Config#getCharset()}
			If Excel, the name or index of the worksheet in the workbook, the default is
			to parse the first sheet (index=0) or the whole workbook if the {@link #getParseAllExcelSheets()} flag is set
		@since 2.1.1
	*/
	public void setInputStream(InputStream in, String charsetOrSheetName)
	{
		in = XlsxToCsvUtil.prepareForCheck(in);
		String charsetName = charsetOrSheetName;
		if (XlsxToCsvUtil.isAnExcelFile(in))
		{
			excelWorkbook = new XlsxToCsvUtil();
			if (dateFormat == null) dateFormat = new DateFormatter("yyyy-MM-dd");
			if (timestampFormat == null) timestampFormat = new DateFormatter("yyyy-MM-dd hh:mm:ss");
			if (numberFormat == null) numberFormat = NumberFormatUtil.getNumberFormat('.', (char)0);
			excelWorkbook.setDateFormat(dateFormat);
			excelWorkbook.setTimestampFormat(timestampFormat);
			excelWorkbook.setNumberFormat(new NumberFormatFormatter(numberFormat));
			excelWorkbook.setEvaluateFormulas(true);
			setDataSplitterRegexp(Pattern.compile("\\t"));
			charsetName = "UTF-8";
			try
			{
				excelWorkbook.readWorkbook(in);
			}
			catch (IOException ex)
			{
				throw new RuntimeException(ex);
			}
			
			if (parseAllExcelSheets)
			{
				setSectionRegexp(Pattern.compile("\\[(.*)\\]"));
				in = excelWorkbook.getWorkbookAsCsv();
			}
			else
			{
				List<String> sheets = excelWorkbook.getSheetNames();
				if (sheets.contains(charsetOrSheetName))
				{
					excelSheet = excelWorkbook.getSheetAsCsv(charsetOrSheetName);
				}
				else if (sheets.contains(excelSheetName))
				{
					excelSheet = excelWorkbook.getSheetAsCsv(excelSheetName);
				}
				else
				{
					// Or try to convert to index with 0 as default
					excelSheet = excelWorkbook.getSheetAsCsv(Values.getInt(charsetOrSheetName));
				}
				excelSheetName = excelSheet.getName();
				in = excelSheet.parseToCsv();
			}
		}
		Charset cs = CharsetUtil.getCharset(charsetName);
		this.tracker = new InputStreamTracker(in);
		this.reader = new BufferedReader(new InputStreamReader(tracker, cs));
	}

	/**
		Parse the file until the beginning-of-file marker is found. If no 
		regular expression has been set with {@link #setBofMarkerRegexp(Pattern)}
		or if the parsing of the file has already started, this method call is 
		ignored. 
		
		@return TRUE if this call resulted in parsing and the BOF marker was found,
			FALSE otherwise
		@since 2.15
	*/
	public boolean parseToBof()
		throws IOException
	{
		if (bofMarker == null || parsedLines > 0) return false;
		
		boolean done = false;
		boolean matched = false;
		while (!done)
		{
			ThreadSignalHandler.checkInterrupted();
			String line = reader.readLine();
			parsedLines++;
			if (line == null)
			{
				done = true;
			}
			else
			{
				parsedCharacters += line.length(); 
				Matcher m = bofMarker.matcher(line);
				if (m.matches())
				{
					bofType = m.groupCount() > 0 ? m.group(1) : m.group();
					matched = true;
					done = true;
				}
			}
		}
		return matched;
	}
	
	/**
		Get the value captured by the BOF marker regular expression. If
		no capturing groups was specified in the pattern this value is the
		string that matched the entire pattern.
		
		@return The matched value, or null if BOF matching has not been done
		@since 2.15
	*/
	public String getBofType()
	{
		return bofType;
	}
	
	/**
		Start parsing the input stream. The parser will read a single
		line at a time. Each line is checked in the following order:
		<ol>
		<li>Does it match the {@link #setSectionRegexp(Pattern) section}
			regular expression?
		<li>Does it match the {@link #setHeaderRegexp(Pattern) header}
			regular expression?
		<li>Does it match the {@link #setDataHeaderRegexp(Pattern) data header}
			regular expression?
		<li>Does it match the {@link #setIgnoreRegexp(Pattern) comment}
			regular expression?
		<li>Can it be split by the {@link #setDataSplitterRegexp(Pattern) data}
			regular expression into the appropriate number of columns?
		</ol>

		The first expression that matches stops the processing of that line.
		If the line matched a header or comment the parser continues with
		the next line. If the line matched the data header or data, the
		method returns. If none of the above is true the line is recorded
		as {@link LineType#UNKNOWN} and processing is continued
		with the next line. If too many unkown lines in a row has been found
		the method also returns. This should be considered as a failure to
		parse the specified file.
		<p>
		The method returns the type of the last line that was parsed as
		follows:
		<ul>
		<li>{@link LineType#SECTION}: The last line
			was a section. Header, data header or data may 
			follow this line.
		<li>{@link LineType#DATA_HEADER}: The last line
			was the data header. It is expected that data should
			follow.
		<li>{@link LineType#DATA}: The last line
			was a data line. More data may follow.
		<li>{@link LineType#UNKNOWN}: The last line
			was of unknown format. The file could not be parsed.
		</ul>

		@return The {@link LineType} of the last parsed line
	 	@throws IOException If reading the file fails.
	*/
	public LineType parseHeaders()
		throws IOException
	{
		if (parsedLines == 0) parseToBof();
		lines = new ArrayList<Line>(maxUnknownLines);
		headers = new HashMap<String,String>();

		nextData = null;
		nextSection = null;
		int unknownLines = 0;
		boolean done = false;
		while (!done)
		{
			ThreadSignalHandler.checkInterrupted();
			String line = reader.readLine();
			parsedLines++;
			if (line == null)
			{
				done = true;
			}
			else
			{
				parsedCharacters += line.length(); 
				boolean matched = false;
				
				// Check if the line is a section
				if (section != null)
				{
					Matcher m = section.matcher(line);
					if (m.matches() && (m.groupCount() == 1))
					{
						matched = true;
						done = true;
						nextSection = new Line(parsedLines, line, m.group(1), LineType.SECTION);
						lines.add(nextSection);
						if (parseAllExcelSheets)
						{
							excelSheet = excelWorkbook.getSheetAsCsv(parsedSections);
							excelSheetName = excelSheet.getName();
							excelParsedLinesOffset = parsedLines;
						}
						parsedSections++;
					}
				}
				
				// Check if the line is a header
				if (header != null && !matched)
				{
					Matcher m = header.matcher(line);
					if (m.matches() && (m.groupCount() == 2))
					{
						matched = true;
						String name = m.group(1);
						String value = convertToNull(m.group(2));
						headers.put(name, value);
						lines.add(new Line(parsedLines, line, name, value));
					}
				}

				// Check if the line is a comment
				if (ignore != null && !matched)
				{
					Matcher m = ignore.matcher(line);
					if (m.matches())
					{
						matched = true;
						lines.add(new Line(parsedLines, line, LineType.IGNORED));
					}
				}
	
				// Check if the line matches the data header
				if (dataHeader != null && !matched)
				{
					Matcher m = dataHeader.matcher(line);
					if (m.matches())
					{
						matched = true;
						done = true;
						lines.add(new Line(parsedLines, line, LineType.DATA_HEADER));

						if (dataSplitter != null)
						{
							String[] columns = trimQuotes(dataSplitter.split(line, -1));
							columnHeaders = Arrays.asList(columns);
						}
					}
				}
			
				// Try the data splitter and check the number of columns
				if (dataHeader == null && dataSplitter != null && !matched && maxDataColumns >= 0)
				{
					String[] columns = trimQuotes(dataSplitter.split(line, -1));
					if ((columns.length >= minDataColumns) &&
						((columns.length <= maxDataColumns) || (maxDataColumns == 0)))
					{
						matched = true;
						done = true;
						lines.add(new Line(parsedLines, line, LineType.DATA));
						parsedDataLines = 1;
						nextData = new Data(parsedLines, parsedDataLines, line, columns, emptyIsNull, nullIsNull, trimWhiteSpace);
					}
				}

				// The line was unknown
				if (!matched)
				{
					unknownLines++;
					if (unknownLines > maxUnknownLines) done = true;
					lines.add(new Line(parsedLines, line, LineType.UNKNOWN));
				}
				else
				{
					unknownLines = 0;
				}
			}
		}
		return getLineCount() == 0 ? LineType.UNKNOWN : getLine(getLineCount()-1).type();
	}
	
	private String convertToNull(String value)
	{
		return emptyIsNull && value.length() == 0 || 
			nullIsNull && "NULL".equalsIgnoreCase(value) ? null : value;
	}
	
	/**
		Get the names of all headers found by the {@link #parseHeaders()}
		method. To get the value of a header, use the {@link #getHeader(String)}
		method.
	*/
	public Set<String> getHeaderNames()
	{
		return Collections.unmodifiableSet(headers.keySet());
	}
	
	/**
		Get the value of the header with the specified name. This method
		should only be used after {@link #parseHeaders()} has
		been completed.

		@param name The name of the header
		@return The value of the header, or null if it was not found
		@see #getLine(int)
	*/
	public String getHeader(String name)
	{
		return headers.get(name);
	}

	/**
		Get the number of lines that the {@link #parseHeaders()}
		method parsed.

		@return The number of lines parsed
	*/
	public int getLineCount()
	{
		return lines.size();
	}

	/**
		Get the line with the specified number. This method
		should only be used after {@link #parseHeaders()} has
		been completed.
		@param index The line number, starting at 0
		@return A <code>Line</code> object
		@see #getHeader(String)
	*/
	public Line getLine(int index)
	{
		return lines.get(index);
	}
	
	/**
		Get the lines read by {@link #parseHeaders()}.
		@return The lines in the order that they have been read.
	*/
	public List<Line> getLines()
	{
		return new ArrayList<Line>(lines);
	}

	/**
		Get all column headers that were found by splitting the line matching the
		{@link #setDataHeaderRegexp(Pattern)} pattern using the 
		{@link #setDataSplitterRegexp(Pattern)} pattern. This method should only be called 
		after {@link #parseHeaders()} has been called.
		
		@return A list containing the column headers, or null if no headers have been 
			found
	*/
	public List<String> getColumnHeaders()
	{
		if (columnHeaders != null)
		{
			return Collections.unmodifiableList(columnHeaders);
		}
		return null;
	}
	
	/**
		Get the index of a column header with a given name. This method should only be called 
		after {@link #parseHeaders()} has been called. If more than one header
		with the same name exists the index of the first is returned.
		
		@param name The name of the column header
		@return The index, or null if no header with that name exists
		@see #findColumnHeaderIndex(String)
	*/
	public Integer getColumnHeaderIndex(String name)
	{
		int index = columnHeaders.indexOf(name);
		return index >= 0 ? index : null;
	}
	
	/**
		Find the index of a column header using a regular expression for pattern
		matching. This method should only be called after {@link #parseHeaders()} 
		has been called. If more than one header matches the regular expression
		only the first one found is returned.
		
		@param regex The regular expression used to match the header names
		@return The index, or null if no header is matching the regular expression
			or if the string is not a valid regular expression
		@see #getColumnHeaderIndex(String)
		@since 2.5
	*/
	public Integer findColumnHeaderIndex(String regex)
	{
		try
		{
			Pattern p = Pattern.compile(regex);
			int index = 0;
			for (String column : columnHeaders)
			{
				if (p.matcher(column).matches()) return index;
				++index;
			}
		}
		catch (PatternSyntaxException ex)
		{}
		return null;
	}
	
	/**
		Set the default number format to use when creating mappers.
		@param numberFormat The number format to use, or null to parse
			numbers with Float.valueOf or Double.valueOf
		@since 2.2
		@see #getMapper(String)
		@see #getMapper(String, NumberFormat)
	 */
	public void setDefaultNumberFormat(NumberFormat numberFormat)
	{
		this.numberFormat = numberFormat;
	}
	
	/**
		Get the default number format.
		@return The number format, or null if none has been specified
		@since 2.2
	*/
	public NumberFormat getDefaultNumberFormat()
	{
		return numberFormat;
	}
	
	/**
		Set the default date format to use when creating mappers.
		If null, xxx is used.
		@since 3.15
	*/
	public void setDefaultDateFormat(Formatter<Date> dateFormat)
	{
		this.dateFormat = dateFormat;
	}
	
	/**
		Get the default date format.
		@since 3.15
	*/
	public Formatter<Date> getDefaultDateFormat()
	{
		return dateFormat;
	}
	
	/**
		Set the default timestamp format to use when creating mappers.
		If null, xxx is used.
		@since 3.15
	*/
	public void setDefaultTimestampFormat(Formatter<Date> timestampFormat)
	{
		this.timestampFormat = timestampFormat;
	}
	
	/**
		Get the default timestamp format.
		@since 3.15
	*/
	public Formatter<Date> getDefaultTimestampFormat()
	{
		return timestampFormat;
	}
	
	/**
		Specify if <code>null</code> should be returned if a (numeric) 
		value can't be parsed. If this setting is set to TRUE all mappers
		created by one of the {@link #getMapper(String)} methods are wrapped
		in a {@link NullIfExceptionMapper}. It is not possible to log
		or get information about the exception.
		@param useNullIfException TRUE to return null, FALSE to throw an exception
		@since 2.4
	*/
	public void setUseNullIfException(boolean useNullIfException)
	{
		this.useNullIfException = useNullIfException;
	}
	
	/**
		@since 3.15
	*/
	public boolean getUseNullIfException()
	{
		return useNullIfException;
	}
	
	
	/**
		Specify if trying to create a mapper with one of the {@link #getMapper(String)}
		methods for an expression which references a non-existing column should
		result in an exception or be ignored.
		@param ignoreNonExistingColumns TRUE to ignore, or FALSE to throw an exception
		@since 2.6
	*/
	public void setIgnoreNonExistingColumns(boolean ignoreNonExistingColumns)
	{
		this.ignoreNonExistingColumns = ignoreNonExistingColumns;
	}
	
	/**
		Get a mapper using the default number format. 
		@see #getMapper(String, NumberFormat, boolean)
	*/
	public Mapper getMapper(String expression)
	{
		return getMapper(expression, numberFormat, useNullIfException);
	}
	
	public Mapper getMapper(String expression, JepFunction... functions)
	{
		return getMapper(expression, numberFormat, useNullIfException, functions);
	}
	
	/**
		Get a mapper using the default number format.
		@see #getMapper(String, NumberFormat, boolean)
		@since 2.4
	*/
	public Mapper getMapper(String expression, boolean nullIfException)
	{
		return getMapper(expression, numberFormat, nullIfException);
	}
	
	/**
		Get a mapper using a specific number format.
		@see #getMapper(String, NumberFormat, boolean)
		@since 2.2
	*/
	public Mapper getMapper(String expression, NumberFormat numberFormat)
	{
		return getMapper(expression, numberFormat, useNullIfException);
	}
	
	/**
		@see #getMapper(String, NumberFormat, boolean, JepFunction...)
		@since 2.4
	*/
	public Mapper getMapper(String expression, NumberFormat numberFormat, boolean nullIfException)
	{
		return getMapper(expression, numberFormat, nullIfException, (JepFunction[])null);
	}
	
	/**
		Get a mapper using the default date format.
		@since 3.15
	*/
	public Mapper getDateMapper(String expression)
	{
		return getMapper(expression, dateFormat, useNullIfException);
	}
	
	/**
		Get a mapper using the default timestamp format.
		@since 3.15
	*/
	public Mapper getTimestampMapper(String expression)
	{
		return getMapper(expression, timestampFormat, useNullIfException);
	}

	/**
		Get a mapper using the specified date format.
		@since 3.15
	*/
	public Mapper getMapper(String expression, Formatter<Date> dateFormat, boolean nullIfException)
	{
		return getMapper(expression, numberFormat, dateFormat, nullIfException, (JepFunction[])null);
	}
	
	public Mapper getMapper(String expression, NumberFormat numberFormat, boolean nullIfException, JepFunction... functions)
	{
		return getMapper(expression, numberFormat, dateFormat, nullIfException, functions);
	}
	
	/**
		Create a mapper object that maps an expression string to a value.
		An expression string is a regular string which contains placeholders
		where the data column values will be inserted. For example:
		<pre class="code">
\1\
\row\
Row: \row\, Col:\col\
</pre>

		It is also possible to use expressions that are evaluated dynamically.
		<pre class="code">
=2 * col('Radius')
</pre>
		If no column that is matching the exact name is found the placeholder
		is interpreted as a regular expression which is checked against each of
		the column headers. In all cases, the first column header found is used
		if there are multiple matches.
		<p>
		If the expression is null, a mapper returning en empty string is returned, 
		unless the {@link #setUseNullIfEmpty(boolean)} has been activated. In that
		case the mapper returns null.

		@param expression The string containing the mapping expression
		@param numberFormat The number format the mapper should use for
			parsing numbers, or null to use Float.valueOf or Double.valueOf
		@param dateFormat The date format the mapper should use for
			parsing dates, or null to use Type.DATE.parseString()
		@param nullIfException TRUE to return a null value instead of throwing
			an exception when a value can't be parsed.
		@param functions Optional array with Jep functions that should be
			included in the parser
		@return A mapper object
		@since 3.15
	*/
	public Mapper getMapper(String expression, NumberFormat numberFormat, Formatter<Date> dateFormat, boolean nullIfException, JepFunction... functions)
	{
		Mapper mapper = null;
		if (expression == null || expression.length() == 0)
		{
			mapper = new ConstantMapper(emptyIsNull ? null : "", numberFormat, dateFormat, nullIfException);
		}
		else if (expression.startsWith("="))
		{
			mapper = new JepMapper(expression.substring(1), columnHeaders, numberFormat, dateFormat, ignoreNonExistingColumns, functions);
			if (nullIfException) mapper = new NullIfExceptionMapper(mapper);
		}
		else
		{
			List<Mapper> mappers = new LinkedList<Mapper>();
			Matcher m = findColumn.matcher(expression);
			int nextStart = 0;
			while (m.find())
			{
				int matchStart = m.start();
				if (matchStart > nextStart)
				{
					mappers.add(new ConstantMapper(expression.substring(nextStart, matchStart), numberFormat, dateFormat, nullIfException));
				}
				try
				{
					int column = Integer.parseInt(m.group(1));
					mappers.add(new ColumnMapper(column, null, numberFormat, dateFormat, nullIfException));
				}
				catch (NumberFormatException ex)
				{
					if (columnHeaders != null)
					{
						String name = m.group(1);
						Integer column = getColumnHeaderIndex(name);
						if (column == null) column = findColumnHeaderIndex(name);
						if (column == null)
						{
							if (!ignoreNonExistingColumns)
							{
								throw new BaseException("Column '" + name + "' not found in column headers.");
							}
						}
						else
						{
							mappers.add(new ColumnMapper(column, name, numberFormat, dateFormat, nullIfException));
						}
					}
				}
				nextStart = m.end();
			}
			if (nextStart < expression.length())
			{
				mappers.add(new ConstantMapper(expression.substring(nextStart), numberFormat, dateFormat, nullIfException));
			}
			if (mappers.size() == 0)
			{
				mapper = new ConstantMapper(emptyIsNull ? null : "", numberFormat, dateFormat, nullIfException);
			}
			else if (mappers.size() == 1)
			{
				mapper = mappers.get(0);
			}
			else
			{
				mapper = new CompoundMapper(mappers, numberFormat, dateFormat, nullIfException);
			}
		}
		return mapper;
	}
	
	/**
		Check if the input stream contains more data. If it is unknown
		if there is more data or not, this method will start reading more
		lines from the stream. Each line is checked in the following order:
		<ol>
		<li>Does it match the {@link #setDataFooterRegexp(Pattern) data footer}
			regular expression?</li>
		<li>Does it match the {@link #setSectionRegexp(Pattern) section}
			regular expression?</li>
		<li>Does it match the {@link #setIgnoreRegexp(Pattern) ignore} regular
			expression?</li>
		<li>Can it be split by the {@link #setDataSplitterRegexp(Pattern) data}
			regular expression into the appropriate number of columns?</li>
		</ol>
		If the first or second check is true, FALSE is returned no more data may be
		retrieved, but a section may be retrived with the {@link #nextSection() nextSection} 
		method. If the third check is true, the line is ignored and the processing
		continues with the next line. If the fourth check is true, TRUE is 
		returned and the data may be retrieved with the {@link #nextData() nextData} method.

		@return TRUE if there is more data, FALSE otherwise
		@see #nextData() nextData
		@throws IOException If there is an error reading from the input stream
	*/
	public boolean hasMoreData()
		throws IOException
	{
		if (dataSplitter == null) return false;
		if (nextData == null && nextSection == null)
		{
			if (parsedLines == 0) parseToBof();
			ignoredLines = 0;
			unknownLines = 0;
			if (keepSkippedLines) skippedLines = new LinkedList<Line>();
			boolean done = false;
			while (!done)
			{
				ThreadSignalHandler.checkInterrupted();
				String line = reader.readLine();

				// End of file?
				if (line == null)
				{
					done = true;
				}
				else
				{
					parsedLines++;
					parsedDataLines++;
					parsedCharacters += line.length(); 

					// Check if it matches the data footer regexp
					if (!done && dataFooter != null)
					{
						Matcher m = dataFooter.matcher(line);
						done = m.matches();
					}
					
					// Check if it matches the section regexp
					if (!done && section != null)
					{
						Matcher m = section.matcher(line);
						if (m.matches() && (m.groupCount() == 1))
						{
							done = true;
							parsedDataLines = 0;
							nextSection = new Line(parsedLines, line, m.group(1), LineType.SECTION);
							lines.add(nextSection);
							if (parseAllExcelSheets)
							{
								excelSheet = excelWorkbook.getSheetAsCsv(parsedSections);
								excelSheetName = excelSheet.getName();
								excelParsedLinesOffset = parsedLines;
							}
							parsedSections++;
						}
					}
				
					// Check if it is a comment line
					if (!done && ignore != null)
					{
						Matcher m = ignore.matcher(line);
						if (m.matches()) 
						{
							ignoredLines++;
							if (keepSkippedLines) skippedLines.add(new Line(parsedLines, line, LineType.IGNORED));
							
							// Skip to next line
							/* #### CONTINUE-STATEMENT #### */
							continue;
						}
					}

					if (!done)
					{
						// Check if it is a data line
						String[] columns = trimQuotes(dataSplitter.split(line, -1));
						if ((columns.length >= minDataColumns) &&
							((columns.length <= maxDataColumns) || (maxDataColumns <= 0)))
						{
							done = true;
							if (excelSheet != null)
							{
								nextData = new ExcelData(excelSheet, excelParsedLinesOffset, parsedLines, parsedDataLines, line, columns, emptyIsNull, nullIsNull, trimWhiteSpace);
							}
							else
							{
								nextData = new Data(parsedLines, parsedDataLines, line, columns, emptyIsNull, nullIsNull, trimWhiteSpace);
							}
						}
					}
					
					if (!done)
					{
						unknownLines++;
						if (keepSkippedLines) skippedLines.add(new Line(parsedLines, line, LineType.UNKNOWN));
					}
				}
			}
		}
		return nextData != null;
	}

	/**
		Remove enclosing quotes (" or ') around all columns.
		@param columns The columns
		@return The trimmed columns
	*/
	public String[] trimQuotes(String[] columns)
	{
		if (trimQuotes)
		{
			for (int i = 0; i < columns.length; ++i)
			{
				String c = columns[i];
				if (c.length() >= 2)
				{
					// If start and end with " or ' --> remove them
					if ((c.startsWith("\"") && c.endsWith("\"")) || (c.startsWith("'") && c.endsWith("'")))
					{
						columns[i] = c.substring(1, c.length() - 1);
					}
				}
			}
		}
		return columns;
	}
	
	/**
		Get the number of parsed lines so far.
	*/
	public int getParsedLines()
	{
		return parsedLines;
	}
	
	/**
		Get the number of found sections so far.
		@since 3.15
	*/
	public int getParsedSections()
	{
		return parsedSections;
	}
	
	/**
		Get the number of parsed data lines so far in the current section.
		This value is reset for each new section.
	*/
	public int getParsedDataLines()
	{
		return parsedDataLines;
	}
	
	/**
		Get the number of parsed characters so far. This value may or may not
		correspond to the number of parsed bytes depending on the character
		set of the file.
		@see #getParsedBytes()
	*/
	public long getParsedCharacters()
	{
		return parsedCharacters;
	}
	
	/**
		Get the number of parsed bytes so far. This value may or may not
		correspond to the number of parsed characters depending on the character
		set of the file.
		@since 2.5.1
		@see #getParsedCharacters()
	*/
	public long getParsedBytes()
	{
		return tracker == null ? 0 : tracker.getNumRead();
	}
	
	/**
		Get the next available data.
		@return A <code>Data</code> object, or null if there is no more data
	 	@throws IOException If the is an error reading from the input stream. 
		@see #hasMoreData() hasMoreData
	*/
	public Data nextData()
		throws IOException
	{
		Data currentData = null;
		if (hasMoreData())
		{
			currentData = nextData;
			nextData = null;
		}
		return currentData;
	}
	
	/**
		Get the number of lines that the last call to {@link #nextData()} or
		{@link #hasMoreData()} ignored because they matched the ignore regular
		expression.
		@return The number of ignored lines
		@see #setIgnoreRegexp(Pattern)
		@see #setKeepSkippedLines(boolean)
	*/
	public int getIgnoredLines()
	{
		return ignoredLines;
	}
	
	/**
		Get the number of lines that the last call to {@link #nextData()} or
		{@link #hasMoreData()} ignored because they couldn't be interpreted as 
		data lines.
		@return The number of unknown lines
		@see #setKeepSkippedLines(boolean)
	*/
	public int getUnknownLines()
	{
		return unknownLines;
	}
	
	/**
		Get the number of lines that the last call to {@link #nextData()} or
		{@link #hasMoreData()} ignored because they matched the ignore regular
		expression or couldn't be interpreted as data lines.
		@return The number of ignored or unknown lines
		@see #getIgnoredLines()
		@see #getUnknownLines()
		@see #getSkippedLines()
	*/
	public int getNumSkippedLines()
	{
		return ignoredLines + unknownLines;
	}
	
	/**
		Get lines that was skipped during the last call to {@link #nextData()} or
		{@link #hasMoreData()}. The list is only available if the {@link #setKeepSkippedLines(boolean)}
		has been set to true (default is false).
		@return A list with the skipped lines
		@see #setKeepSkippedLines(boolean)
	*/
	public List<Line> getSkippedLines()
	{
		return skippedLines;
	}
	
	/**
		Check if the input stream contains more sections. If it is unknown
		if there is more sections or not, this method will start reading more
		lines from the stream. Each line is checked if it matches the
		{@link #setSectionRegexp(Pattern) section} regular expression. The
		parser will continue util a section line is found or end of file 
		is reached. If the metod return TRUE the section may be retrived with 
		the {@link #nextSection()} method. If the {@link 
		#setSectionRegexp(Pattern) section} regular expression isn't specified
		the method returns FALSE and won't parse any line.
		
		@return TRUE if there is more data, FALSE otherwise
		@see #nextData()
		@throws IOException If there is an error reading from the input stream
	*/
	public boolean hasMoreSections()
		throws IOException
	{
		if (section == null)
		{
			return false;
		}
		if (nextSection == null)
		{
			if (parsedLines == 0) parseToBof();
			boolean done = false;
			while (!done)
			{
				ThreadSignalHandler.checkInterrupted();
				String line = reader.readLine();

				// End of file
				if (line == null)
				{
					done = true;
				}
				else
				{
					parsedLines++;
					parsedCharacters += line.length();
					Matcher m = section.matcher(line);
					if (m.matches() && m.groupCount() == 1)
					{
						done = true;
						nextSection = new Line(parsedLines, line, m.group(1), LineType.SECTION);
						parsedDataLines = 0;
						if (parseAllExcelSheets)
						{
							excelSheet = excelWorkbook.getSheetAsCsv(parsedSections);
							excelSheetName = excelSheet.getName();
							excelParsedLinesOffset = parsedLines;
						}
						parsedSections++;
					}
				}
			}
		}
		return nextSection != null;
	}
	
	/**
		Get the next line that matches the {@link #setSectionRegexp(Pattern) section} 
		regular expression.
		@return The line that matched the regular expression
		@see #hasMoreSections()
	*/
	public Line nextSection()
		throws IOException
	{
		if (section == null)
		{
			return null;
		}
		Line currentSection = null;
		if (hasMoreSections())
		{
			currentSection = nextSection;
			nextSection = null;
		}
		return currentSection;
	}

	/**
		Represents the type of a line matched or unmatched by the parser.
	*/
	public enum LineType
	{
		/**
			The parsed line is a header line, matching the
			regular expression specified by {@link #setHeaderRegexp(Pattern)}.
		*/
		HEADER("Header"),
		
		/**
			The parsed line is the data header line, matching the
			regular expression specified by {@link #setDataHeaderRegexp(Pattern)}.
		*/
		DATA_HEADER("Data header"),
		
		/**
			The parsed line is a data line, which could be split
			into the specified number of columns by {@link #setDataSplitterRegexp(Pattern)}.
			@see #setMinDataColumns(int)
			@see #setMaxDataColumns(int)
		*/
		DATA("Data"),
		
		/**
			The parsed line is the line starting a section, matching the
			regular expression specified by {@link #setSectionRegexp(Pattern)}.
		*/
		SECTION("Section start"),

		/**
			The parsed line is the data footer line, matching the
			regular expression specified by {@link #setDataFooterRegexp(Pattern)}.
		*/
		DATA_FOOTER("Data footer"),

		/**
			The parsed line was ignored since it matched the
			regular expression specified by {@link #setIgnoreRegexp(Pattern)}.
		*/
		IGNORED("Ignored"),
		
		
		/**
			The parsed line is of an unknown type.
		*/
		UNKNOWN("Unknown");

		private final String title;
		private LineType(String title)
		{
			this.title = title;
		}
		
		@Override
		public String toString()
		{
			return title;
		}
	}
	
	/**
		This class holds data about a line parsed by the {@link
		FlatFileParser#parseHeaders()} method.

		@author Nicklas
		@version 2.0
		@see FlatFileParser#getLine(int)
	*/
	public static class Line
	{
		private final int lineNo;
		private final String line;
		private final LineType type;
		private final String name;
		private final String value;

		/**
			Create a new line object.
		*/
		private Line(int lineNo, String line, LineType type)
		{
			this.lineNo = lineNo;
			this.line = line;
			this.type = type;
			this.name = null;
			this.value = null;
		}
		/**
			Create a new section line object.
		*/
		private Line(int lineNo, String line, String name, LineType type)
		{
			this.lineNo = lineNo;
			this.line = line;
			this.type = type;
			this.name = name;
			this.value = null;
		}
		/**
			Create a new header line object.
		*/
		private Line(int lineNo, String line, String name, String value)
		{
			this.lineNo = lineNo;
			this.line = line;
			this.type = LineType.HEADER;
			this.name = name;
			this.value = value;
		}

		@Override
		public String toString()
		{
			return "Line " + lineNo + " [" + type + "]: " + line;
		}
		
		/**
			The line number of this line in the input stream.
			@return The line number
		*/
		public int lineNo()
		{
			return lineNo;
		}

		/**
			The entire line as it was read from the input stream.
			@return A string
		*/
		public String line()
		{
			return line;
		}

		/**
			The type of the line as determined by the {@link
			FlatFileParser#parseHeaders()} method. The value is
			one of the values in the {@link LineType} enumeration.
			@return A code indicating the type of the line
		*/
		public LineType type()
		{
			return type;
		}
		/**
			The name of the header if the line is a header line.
			@return The name of the header, or null if it isn't a header line
		*/
		public String name()
		{
			return name;
		}
		/**
			The value of the header if the line is a header line.
			@return The value of the header, or null if it isn't a header line
		*/
		public String value()
		{
			return value;
		}
	}

	/**
		This class holds data about a line parsed by the {@link
		FlatFileParser#hasMoreData()} method.

		@author Nicklas
		@version 2.0
		@see FlatFileParser#nextData()
	*/
	public static class Data
	{
		private final int lineNo;
		private final int dataLineNo;
		private final String[] result;
		private final String line;
		private final boolean emptyIsNull;
		private final boolean nullIsNull;
		private final boolean trimWhiteSpace;

		/**
			Create a new data object.
		*/
		private Data(int lineNo, int dataLineNo, String line, String[] columns, boolean emptyIsNull, boolean nullIsNull, boolean trimWhiteSpace)
		{
			this.lineNo = lineNo;
			this.dataLineNo = dataLineNo;
			this.line = line;
			this.result = columns;
			this.emptyIsNull = emptyIsNull;
			this.nullIsNull = nullIsNull;
			this.trimWhiteSpace = trimWhiteSpace;
		}
		
		@Override
		public String toString()
		{
			return "Line " + lineNo +" [Data]: " + line;
		}
		
		/**
			The line number of the data in the input stream.
			@return The line number
		*/
		public int lineNo()
		{
			return lineNo;
		}
		/**
			The line number of the data in the current section.
			@return The line number
		*/
		public int dataLineNo()
		{
			return dataLineNo;
		}
		/**
			The entire line as it was read from the input stream.
			@return A string
		*/
		public String line()
		{
			return line;
		}
		/**
			The number of columns as determined by the
			{@link FlatFileParser#setDataSplitterRegexp(Pattern)} 
			regular expression.

			@return The number of columns in the data line
		*/
		public int columns()
		{
			return result.length;
		}
		/**
			Get the value in the specified column.
			@param index The index for the column, starting at 0
			@return The value in the column
			@deprecated In 3.15, use {@link #getString(int)} instead or one of the
				other getXxx() method to return a given data type directly
		*/
		@Deprecated
		public String get(int index)
		{
			return getString(index);
		}
		
		/**
			The data line as an array of strings.
			@since 2.15
		*/
		public String[] data()
		{
			return result;
		}
		
		/**
			Get the value in the specified column.
			@param index The index for the column, starting at 0
			@return The value in the column
		*/
		public String getString(int index)
		{
			return fixString(result[index]);
		}
		
		/**
			Fix the value string according to options
			for 'emptyIsNull', 'nullIsNull', etc.
			@since 3.15.1
		*/
		protected String fixString(String value)
		{
			// TODO - In Java 11 we can use String.strip() which also handles other white-space than "space"
			if (value != null && trimWhiteSpace) value = value.trim();			
			
			return value == null || 
				emptyIsNull && value.length() == 0 || 
				nullIsNull && "NULL".equalsIgnoreCase(value) 
				? null : value;
		}
		
		/**
			Shortcut for getInt(index, null, true)
		*/
		public Integer getInt(int index)
		{
			return getInt(index, null, true);
		}
		
		/**
			Get the value as an integer. The default implementation parses
			the string value with the given parser. If the source file is
			an Excel file, the numeric value may be retrieved directly.
			@since 3.15
		*/
		public Integer getInt(int index, NumberFormat parser, boolean nullIfException)
		{
			return (Integer)Type.INT.parseString(getString(index), parser, nullIfException);
		}

		/**
			Get the value as a long. The default implementation parses
			the string value with the given parser. If the source file is
			an Excel file, the numeric value may be retrieved directly.
			@since 3.15
		*/
		public Long getLong(int index, NumberFormat parser, boolean nullIfException)
		{
			return (Long)Type.LONG.parseString(getString(index), parser, nullIfException);
		}
		
		/**
			Shortcut for getFloat(index, null, true)
		*/
		public Float getFloat(int index)
		{
			return getFloat(index, null, true);
		}
		
		/**
			Get the value as a float. The default implementation parses
			the string value with the given parser. If the source file is
			an Excel file, the numeric value may be retrieved directly.
			@since 3.15
		*/
		public Float getFloat(int index, NumberFormat parser, boolean nullIfException)
		{
			return (Float)Type.FLOAT.parseString(getString(index), parser, nullIfException);
		}

		/**
			Get the value as a double. The default implementation parses
			the string value with the given parser. If the source file is
			an Excel file, the numeric value may be retrieved directly.
			@since 3.15
		*/
		public Double getDouble(int index, NumberFormat parser, boolean nullIfException)
		{
			return (Double)Type.DOUBLE.parseString(getString(index), parser, nullIfException);
		}

		/**
			Get the value as a date. The default implementation parses
			the string value with the given parser. If the source file is
			an Excel file, the date value may be retrieved directly.
			@since 3.15
		*/
		public Date getDate(int index, Formatter<Date> parser, boolean nullIfException)
		{
			return (Date)Type.DATE.parseString(getString(index), parser, nullIfException);
		}
	}
	
	/**
		Subclass that is used to return data when the source file is an Excel
		file. It will forward all get() methods to retrieve the data directly
		from the Cell item in Excel.
	*/
	static class ExcelData
		extends Data
	{
		
		private final SheetInfo sheet;
		private final int rowOffset;
		private final int colOffset;
		
		/**
			Create a new data object.
		*/
		ExcelData(SheetInfo sheet, int lineOffset, int lineNo, int dataLineNo, String line, String[] columns, boolean emptyIsNull, boolean nullIsNull, boolean trimWhiteSpace)
		{
			super(lineNo, dataLineNo, line, columns, emptyIsNull, nullIsNull, trimWhiteSpace);
			this.sheet = sheet;
			this.rowOffset = sheet.getFirstRow()-1-lineOffset;
			this.colOffset = sheet.getFirstColumn();
		}
		
		@Override
		public String getString(int index)
		{
			Cell c = sheet.getCell(lineNo()+rowOffset, index+colOffset);
			return fixString(sheet.getCellValueAsString(c));
		}
		
		@Override
		public Integer getInt(int index, NumberFormat parser, boolean nullIfException)
		{
			Double d = getDouble(index, parser, nullIfException);
			return d == null ? null : d.intValue();
		}
		
		@Override
		public Long getLong(int index, NumberFormat parser, boolean nullIfException)
		{
			Double d = getDouble(index, parser, nullIfException);
			return d == null ? null : d.longValue();
		}
	
		@Override
		public Float getFloat(int index, NumberFormat parser, boolean nullIfException)
		{
			Double d = getDouble(index, parser, nullIfException);
			return d == null ? null : d.floatValue();
		}
	
		@Override
		public Double getDouble(int index, NumberFormat parser, boolean nullIfException)
		{
			Cell c = sheet.getCell(lineNo()+rowOffset, index+colOffset);
			return sheet.getCellValueAsNumber(c, parser, nullIfException);
		}
	
		@Override
		public Date getDate(int index, Formatter<Date> parser, boolean nullIfException)
		{
			Cell c = sheet.getCell(lineNo()+rowOffset, index+colOffset);
			return sheet.getCellValueAsDate(c, parser, nullIfException);
		}

	}
}
