/**
	$Id: Utf8PropertiesResourceBundleControl.java 7714 2019-05-21 11:37:09Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.resources;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Locale;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;

import net.sf.basedb.util.FileUtil;

/**
	ResourceBundle.Control implementation that open property files in UTF-8 format
	instead of ISO-8859-1 format. Use {@link #INSTANCE} as the control parameter
	when calling {@link ResourceBundle#getBundle(String, java.util.ResourceBundle.Control)}.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2019-05-21 13:37:09 +0200 (tis, 21 maj 2019) $
*/
public class Utf8PropertiesResourceBundleControl
	extends ResourceBundle.Control
{
	/**
		The one and only instance.
	*/
	public static final Utf8PropertiesResourceBundleControl INSTANCE = new Utf8PropertiesResourceBundleControl();
	
	private Utf8PropertiesResourceBundleControl()
	{}
	
	/*
		From the ResourceBundle.Control class
		-------------------------------------
	*/
	/**
		We don't need a fallback locale.
		@return Always null
	*/
	@Override
	public Locale getFallbackLocale(String baseName, Locale locale)
	{
		return null;
	}

	/**
		This control only handle property files.
		@return Always {@link java.util.ResourceBundle.Control#FORMAT_PROPERTIES}.
	*/
	@Override
	public List<String> getFormats(String baseName)
	{
		return FORMAT_PROPERTIES;
	}

	/**
		Load property files using UTF-8.
	*/
	@Override
	public ResourceBundle newBundle(String baseName, Locale locale,
			String format, ClassLoader loader, boolean reload)
		throws IllegalAccessException, InstantiationException, IOException
	{
		// Let the superclass handle non-property bundles, but this
		// *should* never happen since getFormats() only return "java.properties"
		if (!format.equals("java.properties"))
		{
			return super.newBundle(baseName, locale, format, loader, reload);
		}
		
	   String bundleName = toBundleName(baseName, locale);
		String resourceName = toResourceName(bundleName, "properties");
	   ResourceBundle bundle = null;
		InputStream in = null;
		try 
		{
			in = loader.getResourceAsStream(resourceName);
			if (in != null)
			{
				bundle = new PropertyResourceBundle(new InputStreamReader(in, StandardCharsets.UTF_8));
			}
		}
		finally
		{
			FileUtil.close(in);
		}
 	   return bundle;
	}
	// ---------------------------------------------------

}
