/**
	$Id: ThreadTimerTask.java 7786 2020-03-02 08:10:33Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.timer;

import java.util.TimerTask;

/**
	Class for forcing the execution of a <code>TimerTask</code> in
	a new thread. Instead of adding the original <code>TimerTask</code>
	to a <code>Timer</code> an instance of this class is added, which
	creates a new <code>Thread</code> where the original <code>TimerTask</code>:s
	run method is executed.
	<p>
	Note that for repeating tasks the use of this class may lead to that
	the same task is executed again before the previous execution has 
	finished. This may lead to complications if the task is not thread-safe.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2020-03-02 09:10:33 +0100 (mån, 02 mars 2020) $
*/
public class ThreadTimerTask
	extends TimerTask
{

	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(ThreadTimerTask.class);

	private final TimerTask task;
	private final boolean allowMultiple;
	
	private boolean isExecuting;
	
	/**
		Create a new threaded timer task. The task is executed in a new thread.
		@param task The task to execute when the timer fires
		@param allowMultiple If TRUE, it is allowed to start the same task
			again before the previous execution has returned, if FALSE 
			the timer event is ignored
	*/
	public ThreadTimerTask(TimerTask task, boolean allowMultiple)
	{
		this.task = task;
		this.allowMultiple = allowMultiple;
		this.isExecuting = false;
	}

	/*
		From the TimerTask class
		-------------------------------------------
	*/
	@Override
	public final synchronized void run()
	{
		try
		{
			Thread t = null;
			if (allowMultiple) 
			{
				// We just create another thread
				t = new Thread(task);
			}
			else if (!isExecuting)
			{
				// Create a new thread that also manages the isExecuting flag
				isExecuting = true;
				t = new Thread(
					new Runnable()
					{
						@Override
						public void run()
						{
							try
							{
								task.run();
							}
							finally
							{
								isExecuting = false;
							}
						}
					}
				);
			}
			else
			{
				// Do not allow multiple thread to execute the task
			}
			// Start the thread
			if (t != null) t.start();
		}
		catch (Throwable t)
		{
			log.error("Failed to start task: " + task, t);
		}
	}
	@Override
	public boolean cancel()
	{
		task.cancel();
		return super.cancel();
	}
	// -------------------------------------------

}
