/**
	$Id: FileUnpacker.java 5758 2011-09-23 07:05:23Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.zip;

import java.io.IOException;
import java.io.InputStream;
import java.util.Set;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.File;
import net.sf.basedb.core.plugin.Plugin;

/**
	This interface should be implemented by plugins that can unpack a single compressed
	file into the BASE file system. The plugin may extend the 
	{@link net.sf.basedb.util.zip.AbstractFileUnpacker} class which implements most of the
	plugin specified parts, leaving only the actual unpacking to the subclass.
	<p>
	The file upload functionality checks all plugins that implement this interface to
	provide automatic unpacking of uploaded files without having to store them in a
	temporary place first.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2011-09-23 09:05:23 +0200 (fre, 23 sep. 2011) $
	@see net.sf.basedb.plugins.ZipFileUnpacker
	@see net.sf.basedb.util.zip.AbstractFileUnpacker
*/
public interface FileUnpacker
	extends Plugin
{
	
	/**
		Get the name of the compressed file format this unpacker supports.
		For example: ZIP files
	*/
	public String getFormatName();
	
	/**
		Get the file extensions that this unpacker supports. For example
		[ zip, jar ].
		@return A set containing file extensions (do not inlcude the dot)
	*/
	public Set<String> getExtensions();
	
	
	/**
		Get the MIME type for the file formats this unpacker supports. For example
		[ application/zip, application/java-archive ].
		@return A set containing MIME types
	*/
	public Set<String> getMimeTypes();
	
	/**
		Unpack the given input stream into a BASE directory. If a source
		file is given it can be used as a template for setting certain
		properties on the unpcked files. See {@link 
		PackUtil#copyProperties(File, File)}.
		
		@param dc The DbControl to use for database access
		@param dir The directory to unpack the files to
		@param in The input stream containing the packed data
		@param sourceFile The source packed file that the input stream is
			coming from or null if not known
		@param overwrite If existing files should be overwritten or ignored
		@param progress A optional progress reporter which expects the number
			of compressed bytes that has been uncompressed
		@return The number of unpacked files
		@throws IOException If there is an error reading the input stream
		@throws BaseException If there is another error
		@since 3.0
	*/
	public int unpack(DbControl dc, Directory dir, InputStream in, File sourceFile, boolean overwrite, 
		AbsoluteProgressReporter progress)
		throws IOException, BaseException;
	
}
