/**
	$Id: AnnotationFlatFileImporter.java 7703 2019-04-11 07:31:26Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import java.io.InputStream;
import java.lang.reflect.Method;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.regex.Pattern;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationBatcher;
import net.sf.basedb.core.AnnotationBatcher.Change;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.DateFormatException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DisallowedValueException;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerUtil;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemList;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.NumberOutOfRangeException;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.StringTooLongException;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.ParameterValues;
import net.sf.basedb.core.plugin.ParameterValuesWrapper;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.plugins.batchimport.AnnotationIdMethod;
import net.sf.basedb.plugins.batchimport.IdMethod;
import net.sf.basedb.plugins.batchimport.InternalIdMethod;
import net.sf.basedb.plugins.batchimport.PropertyIdMethod;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.error.ClassMapErrorHandler;
import net.sf.basedb.util.error.ErrorHandler;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.parser.ColumnMapper;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.CropStringMapper;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.units.UnitCache;
import net.sf.basedb.util.units.UnitUtil;

/**
	Plug-in for importing annotations from simple text files. The plug-in supports
	all files that can be parsed with the {@link FlatFileParser} class. This plug-in
	works without a configuration but can use a configuration to store regular
	expressions and other settings for the flat file parser. In both cases,
	the job configuration is a three-step process:
	
	<ol>
	<li>Setup regular expressions and other options for the flat file parser. If a
		configuration is used all values should already be filled in. In this step
		a file to import from must also be selected.
	<li>Map file columns to annotation types. Annotation types that support multiple
		values may be mapped to more than one column.
	<li>Setup error handling options and other settings for the plug-in (for example
		if existing annotations should be replaced or not).
	</ol>

	@author nicklas
	@version 2.4
	@base.modified $Date: 2019-04-11 09:31:26 +0200 (tors, 11 apr. 2019) $
*/
public class AnnotationFlatFileImporter
	extends AbstractFlatFileImporter
	implements InteractivePlugin, ConfigureByExample
{

	private static volatile Set<GuiContext> guiContexts;
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	// second step in the wizard
	private static final String CONFIGURE_MAPPING = "configure_mapping";
	// third step in the wizard
	private static final String CONFIGURE_IMPORT = "configure_import";
	
	private static final StringParameterType optionalColumnMapping = new StringParameterType(255, null, false);
	private static final BooleanParameterType optionalBooleanType = new BooleanParameterType(null, false);

	private static final PluginParameter<String> nameColumnMapping = new PluginParameter<String>(
			"nameColumnMapping",
			"Name",
			"Mapping that picks the name of the item from the data columns. This " +
			"parameter is optional but must be given if a mapping for external or internal ID isn't given. " +
			"Example: \\Name\\",
			optionalColumnMapping
			);
	
	private static final PluginParameter<String> externalIdColumnMapping = new PluginParameter<String>(
			"externalIdColumnMapping",
			"External ID",
			"Mapping that picks the external ID of the item from the data columns. This " +
			"parameter is optional but must be given if a mapping for name or internal ID isn't given. " +
			"Example: \\External ID\\",
			optionalColumnMapping
			);

	private static final PluginParameter<String> internalIdColumnMapping = new PluginParameter<String>(
			"internalIdColumnMapping",
			"Internal ID",
			"Mapping that picks the BASE internal ID of the item from the data columns. This " +
			"parameter is optional but must be given if a mapping for name or external ID isn't given. " +
			"Example: \\BASE ID\\",
			optionalColumnMapping
			);
	
	/**
		Column mapping parameter for the annotation column used for item identification.
		@since 3.15
	*/
	private static final PluginParameter<String> idAnnotationColumnMapping = new PluginParameter<String>(
			"idAnnotationColumnMapping",
			"Annotation ID",
			"Mapping that picks the annotation value to use for item identification from the data columns. This " +
			"parameter is required if the selected identification method is based on an annotation, " +
			"otherwise it is ignored. Example: \\Sample-ID\\",
			optionalColumnMapping
			);

	/**
		Section definition for grouping all mappings of columns to annotation types.
		Parameters in this section will be generated on the fly based on the
		column headers from the file to import from.
	*/
	private static final PluginParameter<String> annotationTypeSection = new PluginParameter<String>(
		"annotationTypeSection",
		"Map columns to annotation types",
		"For each column found in the file, you can select the annotation type " +
		"it contains annotations for. It is possible to select the same annotation type " +
		"for more than one column.", 
		null
	);
	
	private static final PluginParameter<String> identificationSection = new PluginParameter<String>(
			"idSection",
			"Item identification",
			"Options related to how find existing items.", 
			null
		);

	private static final PluginParameter<String> addReplaceSection = new PluginParameter<String>(
		"addReplaceSection",
		"Existing annotations",
		"Options that affect what to do if an item already has annotations.",
		null
	);

	private static final PluginParameter<Boolean> addToUnlimitedParameter = new PluginParameter<Boolean>(
		"addToUnlimited",
		"Append to unlimited",
		"For annotation types that allows an unlimited number of values, " +
		"this setting specifies if new values should be appended to the existing ones. If false " +
		"is selected the 'Replace existing' setting decides if the old values should be kept " +
		"or replaced.",
		true,
		optionalBooleanType
	);

	private static final PluginParameter<Boolean> replaceExistingParameter = new PluginParameter<Boolean>(
		"replaceExisting",
		"Replace existing",
		"If existing annotations should be replaced with the values from this file.",
		true,
		optionalBooleanType
	);
	
	private static final PluginParameter<Boolean> removeExistingParameter = 
		new PluginParameter<Boolean>(
			"removeExisting",
			"Remove existing",
			"If existing annotations should be removed if this file only contains empty values " +
			"for an annotation.",
			null,
			optionalBooleanType
		);
	
	private static final PluginParameter<String> defaultErrorParameter = new PluginParameter<String>(
		"defaultError",
		"Default error handling",
		"The default way to handle an error when no other error handling options have been " +
		"specified to take care of specific exceptions.\n\n"+
		"skip = Skip the current annotation and continue\n"+
		"fail = Stop with an error message (default)",
		new StringParameterType(255, "skip", true, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);
		
	private static final PluginParameter<String> itemNotFoundErrorParameter = new PluginParameter<String>(
		"itemNotFoundError",
		"Item not found",
		"What to do if the plug-in can't find an item with the name or external ID from the file." +
		"This also include the case when name or external ID is unspecified in the file.\n\n"+
		"skip = Skip the current data line and continue\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);

	private static final PluginParameter<String> multipleItemsFoundErrorParameter = new PluginParameter<String>(
		"multipleItemsFoundError",
		"Multiple items found",
		"What to do if the plug-in finds more than one item with the same name or external ID.\n\n"+
		"all = Annotate all items\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, "all", false, 1, 0, 0, 
			Arrays.asList( new String[] { "all", "fail"} ))
	);

	private static final PluginParameter<String> stringTooLongErrorParameter = Parameters.stringTooLongError(
		"String too long",
		"How to handle errors that are caused by too long string values. If no value is specified the " +
		"default error handling is used.\n\n"+
		"crop = Crop the string to the maximum allowed length\n"+
		"skip = Skip the current annotation and continue\n"+
		"fail = Stop with an error message",
		"crop", "crop", "skip", "fail");

	protected static final PluginParameter<String> numberOutOfRangeErrorParameter = Parameters.numberOutOfRangeError(
		"Numeric value out of range",
		"How to handle errors that are caused by numbers that are either too small or too big. If no value is specified the " +
		"default error handling is used.\n\n"+
		"skip = Skip the current annotation and continue\n"+
		"fail = Stop with an error message",
		null, "skip", "fail");

	protected static final PluginParameter<String> numberFormatErrorParameter = Parameters.numberFormatError(
		"Invalid numeric value",
		"How to handle errors that are caused by strings that can't be converted to a numeric value. " +
		"If no value is specified the default error handling is used.\n\n"+
		"skip = Skip the current annotation and continue\n"+
		"fail = Stop with an error message",
		null, "skip", "fail");

	protected static final PluginParameter<String> dateFormatErrorParameter = Parameters.dateFormatError(
		"Invalid date",
		"How to handle errors that are caused by strings that can't be converted to a date/timestamp value. " +
		"If no value is specified the default error handling is used.\n\n"+
		"skip = Skip the current annotation and continue\n"+
		"fail = Stop with an error message",
		null, "skip", "fail");
	
	protected static final PluginParameter<String> invalidEnumErrorParameter = new PluginParameter<String>(
		"invalidEnumError",
		"Invalid enumerated value",
		"How to handle errors that are caused by values that doesn't match any value of " +
		"an enumerated annotation type. " +
		"If no value is specified the default error handling is used.\n\n"+
		"skip = Skip the current annotation and continue\n"+
		"fail = Stop with an error message",
		new StringParameterType(255, null, false, 1, 0, 0, 
			Arrays.asList( new String[] { "skip", "fail"} ))
	);	
	
	protected static final PluginParameter<String> tooManyValuesErrorParameter = new PluginParameter<String>(
		"tooManyValuesError",
		"Too many values",
		"How to handle errors that are caused by trying to set more values than " +
		"the annotation type allows. "+
		"If no value is specified the default error handling is used.\n\n"+
		"skip = Skip the current annotation and continue\n"+
		"crop = Use as many values as the annotation type allows and discard the rest\n" + 
		"fail = Stop with an error message",
		new StringParameterType(255, "crop", false, 1, 0, 0, 
				Arrays.asList( new String[] { "skip", "crop", "fail"} ))
		);	

	
	public AnnotationFlatFileImporter()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	/**
	 	Request read access to File:s, read access to annotation types
	 	and write access to all annotatable items.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.ANNOTATIONTYPE, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.UNIT, EnumSet.of(Permission.READ), null));
			permissions.add(new Permissions(Item.QUANTITY, EnumSet.of(Permission.READ), null));
			for (GuiContext ctx : getGuiContexts())
			{
				permissions.add(new Permissions(ctx.getItem(), null, EnumSet.of(Permission.WRITE)));
			}
		}
		return permissions;
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	/**
		This plug-in works in list context of all {@link Annotatable} items,
		except bioassay sets, bioassays and wells because they are not standalone
		items. We use {@link Metadata#getAnnotatableItems()} to create the contexts.
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		if (guiContexts == null)
		{
			synchronized (AnnotationFlatFileImporter.class)
			{
				Set<GuiContext> tmp = new HashSet<GuiContext>();
				for (Item item : Metadata.getAnnotatableItems())
				{
					if (item.getDefinedPermissions() != null)
					{
						tmp.add(new GuiContext(item, GuiContext.Type.LIST));
					}
				}
				guiContexts = Collections.unmodifiableSet(tmp);
			}
		}
		return guiContexts;
	}
	
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		List<String> itemTypes = configuration.getValues("itemTypes");
		Item itemType = context.getItem();

		if (itemTypes == null || itemTypes.isEmpty() || itemTypes.contains(itemType.name()))
		{
			return null;
		}
		
		return "Not used for " + itemType.toString().toLowerCase() + " items";
	}	
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigureParserParameters(context, false);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureParserParameters(context, true);
		}
		else if (command.equals(CONFIGURE_MAPPING))
		{
			requestInformation = getConfigureMappingParameters(context);
		}
		else if (command.equals(CONFIGURE_IMPORT))
		{
			requestInformation = getConfigureImportParameters(context);
		}
		return requestInformation;
	}
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		DbControl dc = null;
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN) || command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				boolean forJob = command.equals(Request.COMMAND_CONFIGURE_JOB);
				RequestInformation ri = getConfigureParserParameters(context, true);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				ParameterValues wrapper = new ParameterValuesWrapper(request, job, configuration, forJob);
				FlatFileParser parser = null;
				if (forJob)
				{
					// Check that file can be parsed with given regular expressions
					parser = getInitializedFlatFileParser(wrapper);
					File f = (File)request.getParameterValue("file");
					InputStream in = f.getDownloadStream(0);
					parser.setInputStream(in, getCharset(request));
					parser.setDefaultNumberFormat(getNumberFormat());
					FlatFileParser.LineType result = parser.parseHeaders();
					in.close();
					if (result == FlatFileParser.LineType.UNKNOWN)
					{
						response.setError("The file could not be parsed with the given settings, " +
							"no data found after " + parser.getLineCount() + " lines.", null);
						return;
					}
					// Store column headers for use in the next configuration step
					columnHeaders = parser.getColumnHeaders();
				}
				else
				{
					storeValues(configuration, request, ri.getParameter("itemTypes"));
				}

				dc = sc.newDbControl();
				// Check that a mapping for name, internal/external ID or annotation is given
				IdMethod idMethod = getIdMethod(dc, context == null ? null : context.getItem(), (String)request.getParameterValue("idMethod"));
				PluginParameter<?> idMappingParameter = ri.getParameter(idMethod.getColumnMappingParameterName());
				String idMapping = (String)request.getParameterValue(idMappingParameter.getName());
				if (idMapping == null)
				{
					response.setError("No colum mapping provided for item identification: " + 
							idMappingParameter.getLabel(), null);
					return;
				}
				
				// Check the mapping expressions
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				checkColumnMapping(parser, idMapping, allowComplex, idMappingParameter.getLabel());

				// Everything is ok, save values
				if (forJob) storeValue(wrapper, request, fileParameter);
				storeValue(wrapper, request, excelSheetParameter);
				storeValue(wrapper, request, ri.getParameter("idMethod"));
				// Include options
				if (forJob)
				{
					storeValue(job, request, ri.getParameter("memberOfList"));
					storeValue(job, request, ri.getParameter("includeMine"));
					storeValue(job, request, ri.getParameter("includeShared"));
					storeValue(job, request, ri.getParameter("includeInProject"));
					storeValue(job, request, ri.getParameter("includeOthers"));
				}

				storeValue(wrapper, request, dataHeaderRegexpParameter);
				storeValue(wrapper, request, dataSplitterRegexpParameter);
				storeValue(wrapper, request, trimQuotesParameter);
				storeValue(wrapper, request, ignoreRegexpParameter);
				storeValue(wrapper, request, dataFooterRegexpParameter);
				storeValue(wrapper, request, minDataColumnsParameter);
				storeValue(wrapper, request, maxDataColumnsParameter);
				storeValue(wrapper, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(wrapper, request, ri.getParameter(Parameters.DECIMAL_SEPARATOR_PARAMETER));
				storeValue(wrapper, request, ri.getParameter(Parameters.DATE_FORMAT_PARAMETER));
				storeValue(wrapper, request, ri.getParameter(Parameters.TIMESTAMP_FORMAT_PARAMETER));
				
				// Column mappings
				storeValue(wrapper, request, complexMappings);
				storeValue(wrapper, request, nameColumnMapping);
				storeValue(wrapper, request, externalIdColumnMapping);
				storeValue(wrapper, request, internalIdColumnMapping);
				storeValue(wrapper, request, idAnnotationColumnMapping);
				
				if (forJob)
				{
					response.setContinue(CONFIGURE_MAPPING);
				}
				else
				{
					response.setDone("Plug-in configuration completed");
				}
			}
			else if (command.equals(CONFIGURE_MAPPING))
			{
				RequestInformation ri = getConfigureMappingParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}

				// Save mapping parameters
				for (int i = 0; i < columnHeaders.size(); ++i)
				{
					storeValue(job, request, ri.getParameter("columnMapping." + i));
				}
				
				response.setContinue(CONFIGURE_IMPORT);
			}
			
			else if (command.equals(CONFIGURE_IMPORT))
			{
				RequestInformation ri = getConfigureImportParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Add/Replace options
				storeValue(job, request, replaceExistingParameter);
				storeValue(job, request, removeExistingParameter);
				storeValue(job, request, addToUnlimitedParameter);
				
				// Error handling
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, itemNotFoundErrorParameter);
				storeValue(job, request, multipleItemsFoundErrorParameter);
				storeValue(job, request, stringTooLongErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				storeValue(job, request, dateFormatErrorParameter);
				storeValue(job, request, numberOutOfRangeErrorParameter);
				storeValue(job, request, invalidEnumErrorParameter);
				storeValue(job, request, tooManyValuesErrorParameter);
				
				// Save context info
				job.setValue("itemType", new StringParameterType(), context.getItem().name());
				File file = (File)job.getValue("file");
				response.setSuggestedJobName("Import " + context.getItem().toString().toLowerCase() + 
						" annotations from file '" + file.getName() + "'");
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);		

			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------

	/*
		From the AbstractFlatFileImporter class
		-------------------------------------------
	*/
	private List<String> columnHeaders;
	private DbControl dc;
	private AnnotationBatcher batcher;
	
	private UnitCache unitCache;
	private FlatFileParser ffp;
	private NumberFormat numberFormat;
	private Formatter<Date> dateFormatter;
	private Formatter<Date> timestampFormatter;
	private Item itemType;
	private IdMethod idMethod;
	private String idMapping;
	private Mapper itemMapper;
	private Map<Mapper, AnnotationType> mappers;
	private Map<String, Set<NewAnnotations>> itemCache;
	private ItemQuery<?> itemQuery;
	private ItemList itemList;
	private boolean addToUnlimited;
	private boolean replaceExisting;
	private boolean removeExisting;
	
	// Error handling
	private ClassMapErrorHandler errorHandler;
	private boolean cropStrings;
	private boolean ignoreNotFoundItems;
	private boolean failIfMultipleFoundItems;
	private boolean failIfTooManyValues;
	private boolean cropTooManyValues;
	
	// Messaging
	private int numItems;
	private int numItemNotFound;
	private int numAnnotations;
	private int numReplaced;
	private int numError;
	private int numRemoved;
	private ProgressReporter progress;
	// Percentages will be adjusted based on number of annotations
	private int progressParser = 50; // Progress percentage for file-parsing
	private int progressDatabase = 48; // Progress percentage for database update
	
	@Override
	public void doImport(InputStream in, ProgressReporter progress)
			throws BaseException
	{
		this.progress = progress;
		super.doImport(in, progress);
	}
	
	/**
		Don't use AbstractFlatFileImporter to handle errors.
	*/
	@Override
	protected void setUpErrorHandling()
	{}
	
	/**
		Setup error handling and pre-load some of the configuration options.
	*/
	@Override
	protected void begin(FlatFileParser ffp)
	{
		this.dc = sc.newDbControl();
		this.ffp = ffp;
		// Formatting options
		this.numberFormat = ffp.getDefaultNumberFormat();
		this.dateFormatter = ffp.getDefaultDateFormat();
		this.timestampFormatter = ffp.getDefaultTimestampFormat();
		
		this.itemType = Item.valueOf((String)job.getValue("itemType"));
		this.idMethod = getIdMethod(dc, itemType, (String)job.getValue("idMethod"));
		this.idMapping = (String)job.getValue(idMethod.getColumnMappingParameterName());
		this.itemList = (ItemList)job.getValue("memberOfList");
		
		this.addToUnlimited = Boolean.TRUE.equals(job.getValue("addToUnlimited"));
		this.replaceExisting = Boolean.TRUE.equals(job.getValue("replaceExisting"));
		this.removeExisting = Boolean.TRUE.equals(job.getValue("removeExisting"));

		// Error handling, we don't use the error handling in AbstractFlatFileImporter
		this.ignoreNotFoundItems = "skip".equals(getErrorOption("itemNotFoundError"));
		this.failIfMultipleFoundItems = "fail".equals(getErrorOption("multipleItemsFoundError"));
		this.cropStrings = "crop".equals(job.getValue("stringTooLongError"));
		this.failIfTooManyValues = "fail".equals(getErrorOption("tooManyValuesError"));
		this.cropTooManyValues = "crop".equals(job.getValue("tooManyValuesError"));
		
		ErrorHandler defaultErrorHandler = new SimpleErrorHandler("skip".equals(job.getValue("defaultError")));
		errorHandler = new ClassMapErrorHandler(defaultErrorHandler);
		errorHandler.addErrorHandler(StringTooLongException.class, 
				getErrorHandler((String)job.getValue("stringTooLongError"), defaultErrorHandler));
		errorHandler.addErrorHandler(NumberOutOfRangeException.class, 
				getErrorHandler((String)job.getValue("numberOutOfRangeError"), defaultErrorHandler));
		errorHandler.addErrorHandler(NumberFormatException.class,
				getErrorHandler((String)job.getValue("numberFormatError"), defaultErrorHandler));
		errorHandler.addErrorHandler(DateFormatException.class, 
				getErrorHandler((String)job.getValue(Parameters.DATE_FORMAT_ERROR), defaultErrorHandler));
		errorHandler.addErrorHandler(DisallowedValueException.class, 
				getErrorHandler((String)job.getValue("invalidEnumError"), defaultErrorHandler));
	}
	
	/**
		Setup column mapping. Creates DbControl and query to find items.
	*/
	@Override
	protected void beginData()
	{
		this.unitCache = new UnitCache(dc);
		
		// Mapper to get name or external ID
		this.itemMapper = getMapper(ffp, idMapping, null, null);

		// Annotation type mappers
		this.mappers = new HashMap<Mapper, AnnotationType>();
		this.itemCache = new HashMap<String, Set<NewAnnotations>>();
		int i = 0;
		for (String column : ffp.getColumnHeaders())
		{
			AnnotationType at = (AnnotationType)job.getValue("columnMapping." + i);
			if (at != null)
			{
				at = AnnotationType.getById(dc, at.getId());
				Type valueType = at.getValueType();
				Formatter<Date> df = null;
				NumberFormat nf = null;
				if (valueType.isTemporal()) df = valueType == Type.TIMESTAMP ? timestampFormatter : dateFormatter;
				if (valueType.isNumerical()) nf = numberFormat;
				Mapper mapper = new ColumnMapper(i, column, nf, df, false);
				if (cropStrings && at.getValueType() == Type.STRING && at.getMaxLength() != null)
				{
					mapper = new CropStringMapper(mapper, at.getMaxLength());
				}
				mappers.put(mapper, at);
			}
			++i;
		}
		
		// Recalculate progress parts so that the file parsing is between 10-50% depedning on number of annotations
		int dbPart = (1 + mappers.size() % 20);
		progressParser = Math.max(100 / (1+dbPart), 10);
		progressDatabase = 98 - progressParser;
		
		batcher = new AnnotationBatcher(dc, itemType);
		batcher.addAnnotationTypes(mappers.values());
		
		// Include options
		Set<Include> includes = EnumSet.noneOf(Include.class);
		if (Boolean.TRUE.equals(job.getValue("includeMine"))) includes.add(Include.MINE);
		if (Boolean.TRUE.equals(job.getValue("includeShared"))) includes.add(Include.SHARED);
		if (Boolean.TRUE.equals(job.getValue("includeInProject"))) includes.add(Include.IN_PROJECT);
		if (Boolean.TRUE.equals(job.getValue("includeOthers"))) includes.add(Include.OTHERS);
		
		itemQuery = idMethod.prepareQuery(dc, createQuery(itemType, itemList));
		if (includes.size() > 0) itemQuery.setIncludes(includes);
	}
	
	/**
		Read annotations from a single data line. Errors are handled internally. 
		Errors thrown from this method should be reported back to client (ie. error
		handling in AbstractFlatFileParser) must be disabled. This method will
		load items and annotation values and put them in an internal cache. No
		items are annotated until the entire file has been parsed.
	*/
	@Override
	protected void handleData(Data data) 
		throws BaseException
	{

		// Find the item(s) to annotate. Mapper gives us name OR external ID
		String nameOrId = itemMapper.getString(data);
		// 1. check the cache
		Set<NewAnnotations> items = itemCache.get(nameOrId);
		if (items == null)
		{
			items = new HashSet<NewAnnotations>();
			
			if (nameOrId != null)
			{
				List<?> result = idMethod.find(dc, itemQuery, nameOrId);
				if (result.isEmpty())
				{
					// No item with specified name/external ID was found
					if (!ignoreNotFoundItems)
					{
						throw new ItemNotFoundException(itemType + 
							"[" + idMethod + "=" + nameOrId + "]");
					}
					else
					{
						numItemNotFound++;
					}
				}
				if (result.size() > 1 && failIfMultipleFoundItems)
				{
					throw new BaseException("Found " + result.size() + " " + itemType.toString().toLowerCase() + 
						" with " + idMethod + "=" + nameOrId);
				}
				for (Object item : result)
				{
					// Add each item that was found to the cache
					items.add(new NewAnnotations((Annotatable)item));
				}
				itemCache.put(nameOrId, items);
			}
			else if (!ignoreNotFoundItems)
			{
				throw new ItemNotFoundException(itemType + "[" + idMethod + " has no value in file]");
			}
			else
			{
				numItemNotFound++;
			}
		}
		
		if (!items.isEmpty())
		{
			// Load annotation values from column mappers
			for (Map.Entry<Mapper, AnnotationType> entry : mappers.entrySet())
			{
				Mapper mapper = entry.getKey();
				AnnotationType at = entry.getValue();
				Type valueType = at.getValueType();

				Object annotationValue = null;
				Unit unit = null;
				Unit defaultUnit = at.getDefaultUnit();
				try
				{
					if (at.isEnumeration() && valueType == Type.STRING)
					{
						// Case and leading/trailing whitespace are ignore for enumerated string annotations
						String sValue = mapper.getString(data);
						String enumValue = at.findValue(sValue, true, true);
						if (enumValue != null) annotationValue = enumValue;
					}
					else if (at.supportUnits())
					{
						// If a unit is used data is a string so we need to check that
						String sValue = mapper.getString(data);
						String[] valueAndUnit = UnitUtil.splitValueWithUnit(sValue);
						if (valueAndUnit != null && valueAndUnit[1] != null)
						{
							String unitSymbol = valueAndUnit[1];
							// Yes.. there is a unit, look it up
							unit = unitCache.findUnit(at.getQuantity(), unitSymbol);
							if (unit == null)
							{
								// The unit doesn't exists
								throw new NumberFormatException("Unknown unit for value: " + sValue);
							}
							annotationValue = Type.DOUBLE.parseString(valueAndUnit[0], numberFormat, false);
						}
						else
						{
							// No unit specified, use the default unit and get the type-specific value from the mapper
							unit = at.getDefaultUnit();
							annotationValue = mapper.getByType(data, valueType);
						}
					}
					else
					{
						annotationValue = mapper.getByType(data, valueType);
					}
					if (annotationValue != null)
					{
						at.validateAnnotationValue(annotationValue, unit);
					}
					for (NewAnnotations item : items)
					{
						// Add annotation value to the cache
						item.addValue(at, annotationValue, unit, cropTooManyValues);
					}
				}
				catch (Throwable t)
				{
					// In case the annotation value is not valid...
					numError++;
					try
					{
						errorHandler.handleError(t);
						// The error should be ignored if we get passed the above line
					}
					catch (Throwable t2)
					{
						// handleData doesn't allow us to throw any Throwable
						if (t2 instanceof RuntimeException)
						{
							throw (RuntimeException)t2;
						}
						else
						{
							throw new BaseException(t);
						}
					}
				}
			}
		}		
	}
	
	/**
		Now it's time to update the items in the cache with the new annotation
		values.
	*/
	@Override
	protected void end(boolean success)
	{
		try
		{
			if (success)
			{
				int total = itemCache.size();
				int current = 0;
				
				for (Set<NewAnnotations> n : itemCache.values())
				{
					// progressParser should be between 10-50 and is a good interval for progress reporting
					if (progress != null && current % progressParser == 0) 
					{
						progress.display(progressParser + (progressDatabase * current) / total, "Saving to database (" + current + " of  " + total + " items done)...");
					}
					current++;
					numItems += n.size();
					for (NewAnnotations na : n)
					{
						na.setNewAnnotations(batcher, addToUnlimited, replaceExisting, failIfTooManyValues, removeExisting);
						numAnnotations += na.getNumSet();
						numError += na.getNumError();
						numReplaced += na.getNumReplaced();
						numRemoved += na.getNumRemoved();
						ThreadSignalHandler.checkInterrupted();
					}
				}
				batcher.close();
				if (progress != null) progress.display(99, "Committing transaction (" + numItems + " items)");
				dc.commit();
				if (progress != null) progress.display(100, "Done");
			}
		}
		catch (BaseException ex)
		{
			success = false;
			throw ex;
		}
		finally
		{
			if (dc != null) dc.close();
			super.end(success);
			ffp = null;
			if (mappers != null) mappers.clear();
			if (itemCache != null) itemCache.clear();
		}
	}
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String msg = numItems + " item(s) annotated with " + numAnnotations + " annotation(s)";
		if (numReplaced > 0) msg += "; " + numReplaced + " was replaced)";
		if (numRemoved > 0) msg += "; " + numRemoved + " annotations removed";
		if (numItemNotFound > 0) msg += "; " + numItemNotFound + " row(s) skipped because no item was found";
		if (numError > 0) msg += "; " + numError + " annotation(s) skipped because of invalid values";
		return msg;
	}
	/**
		Overriding default implementation since we want file parsing to 
		only use 'progressParser'% and the remaining 'progressDatabase'% 
		for storing to the database.
	 */
	@Override
	protected int getProgress(FlatFileParser ffp)
	{
		if (ffp.getParsedDataLines() == 0) return 0;
		return super.getProgress(ffp) * progressParser / 100;
	}
	// -------------------------------------------

	private RequestInformation configureParser;
	private RequestInformation configureMapping;
	private RequestInformation configureImport;
	
	private Pattern getPattern(ParameterValues parameters, String name)
		throws BaseException
	{
		String regexp = (String)parameters.getValue(name);
		return regexp != null ? Pattern.compile(regexp) : null;
	}
	
	protected FlatFileParser getInitializedFlatFileParser(ParameterValues parameters)
		throws BaseException
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setDataHeaderRegexp(getPattern(parameters, dataHeaderRegexpParameter.getName()));
		ffp.setDataSplitterRegexp(getPattern(parameters, dataSplitterRegexpParameter.getName()));
		Boolean trimQuotes = (Boolean)parameters.getValue(trimQuotesParameter.getName());
		if (trimQuotes != null) ffp.setTrimQuotes(trimQuotes);
		ffp.setIgnoreRegexp(getPattern(parameters, ignoreRegexpParameter.getName()));
		ffp.setDataFooterRegexp(getPattern(parameters, dataFooterRegexpParameter.getName()));
		ffp.setMinDataColumns(IntegerUtil.getInt((Integer)parameters.getValue(minDataColumnsParameter.getName())));
		ffp.setMaxDataColumns(IntegerUtil.getInt((Integer)parameters.getValue(maxDataColumnsParameter.getName())));
		ffp.setExcelSheet(parameters.getValue(excelSheetParameter.getName()));
		return ffp;
	}	
	
	/**
		If the current context is the "Members" tab for an item list we load
		the item list and will later require that all items to be annotated
		are members of that list.
	*/
	private ItemList getItemListFromContext(DbControl dc, GuiContext context)
	{
		if (context == null) return null;
		ItemList list = null;
		String subcontext = context.getSubContext();
		if (subcontext != null && subcontext.startsWith("listmembers"))
		{
			// If subcontext is: "listmembers.nnnn" nnn is the id of the item list
			int listId = Values.getInt(subcontext.split("\\.")[1]);
			list = ItemList.getById(dc, listId);
			// Reject lists with different type of member item (should not happen)
			if (list.getMemberType() != context.getItem()) list = null;
		}
		return list;
	}
	
	private RequestInformation getConfigureParserParameters(GuiContext context, boolean forJob)
	{
		if (configureParser == null)
		{
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
	
				if (forJob)
				{
					// The file to import from
					parameters.add(fileParameter);
					parameters.add(excelSheetParameter);
					parameters.add(identificationSection);
				}
				else
				{
					Enumeration<String, String> items = new Enumeration<String, String>();
					for (GuiContext ctx : getGuiContexts())
					{
						Item i = ctx.getItem();
						ctx.getItem().name();
						items.add(i.name(), i.toString());
					}
					items.sortValues();
					
					PluginParameter<String> itemTypesParameter =
						new PluginParameter<String>("itemTypes", "Item types", 
							"Specify which types of items this file format is used for. If " +
							"not specified it will be used for all types.",
						new StringParameterType(255, null, false, 0, 0, 0, items));
					
					parameters.add(itemTypesParameter);
				}
				
				Enumeration<String, String> idMethods = new Enumeration<String, String>();
				boolean hasAnnotationIdMethod = false;
				for (IdMethod idm : getIdMethods(dc, context == null ? null : context.getItem()))
				{
					if (idm instanceof AnnotationIdMethod) hasAnnotationIdMethod = true;
					idMethods.add(idm.getMethod(), idm.getTitle());
				}
				PluginParameter<String> idMethodParameter = new PluginParameter<String>(
					"idMethod",
					"Identification method",
					"The method to use for finding existing items. You can use either the " +
					"internal ID, the name, external ID or an annotation. If not specified, a " +
					"method is automatically selected based on the available column mappings. " +
					"If a method is selected, you need to makes sure that the corresponding " +
					"column mapping is defined.",
					new StringParameterType(null, idMethods.getKey(0), true, 1, 0, 0, idMethods)
				);
				parameters.add(idMethodParameter);

				if (forJob)
				{
					ItemList list = getItemListFromContext(dc, context);
					if (list != null)
					{
						parameters.add(new PluginParameter<ItemList>(
							"memberOfList", "Member of list",
							"The plug-in will only update annotations for items if they are members of this list.",
							new ItemParameterType<>(ItemList.class, list, true, 1, Arrays.asList(list))
						));
					}

					ItemContext cc = sc.getCurrentContext(context.getItem());
					Set<Include> include = cc.getInclude();
					parameters.add(new PluginParameter<Boolean>(
							"includeMine",
							"Owned by me",
							"Search items that are owned by yourself.",
							include.contains(Include.MINE) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
					parameters.add(new PluginParameter<Boolean>(
							"includeShared",
							"Shared to me",
							"Search items that have been shared to you by other users.",
							include.contains(Include.SHARED) ? Boolean.TRUE : null,
							optionalBooleanType)
						);
						
					if (sc.getActiveProjectId() != 0)
					{
						parameters.add(new PluginParameter<Boolean>(
								"includeInProject",
								"In current project",
								"Search items that are shared to the current project.",
								include.contains(Include.IN_PROJECT) ? Boolean.TRUE : null,
								optionalBooleanType)
							);
					}
					if (sc.hasPermission(Permission.READ, context.getItem()))
					{
						parameters.add(new PluginParameter<Boolean>(
								"includeOthers",
								"Owned by others",
								"Search items that are owned by other users.",
								include.contains(Include.OTHERS) ? Boolean.TRUE : null,
								optionalBooleanType)
							);
					}
				}
				
				// Parser regular expressions
				parameters.add(parserSection);
				if (!forJob) parameters.add(excelSheetParameter);
				parameters.add(dataHeaderRegexpParameter);
				parameters.add(dataSplitterRegexpParameter);
				parameters.add(trimQuotesParameter);
				parameters.add(ignoreRegexpParameter);
				parameters.add(dataFooterRegexpParameter);
				parameters.add(minDataColumnsParameter);
				parameters.add(maxDataColumnsParameter);
				parameters.add(Parameters.charsetParameter(null, null, getCharset(), true));
				parameters.add(Parameters.decimalSeparatorParameter(null, null, getDecimalSeparator()));
				parameters.add(Parameters.dateFormatParameter(null, null, sc.getUserClientSetting("formatter.date.formatstring")));
				parameters.add(Parameters.timestampFormatParameter(null, null, sc.getUserClientSetting("formatter.datetime.formatstring")));
				
				// Mappings for Name and External ID
				parameters.add(mappingSection);
				parameters.add(complexMappings);
				parameters.add(nameColumnMapping);
				parameters.add(internalIdColumnMapping);
				if (context == null || hasExternalId(context.getItem()))
				{
					parameters.add(externalIdColumnMapping);
				}
				if (hasAnnotationIdMethod)
				{
					parameters.add(idAnnotationColumnMapping);
				}
				
				if (forJob)
				{
					configureParser = new RequestInformation
					(
						Request.COMMAND_CONFIGURE_JOB,
						"Select a file to import annotations from",
						"Here you select which file to import annotations from, and the " +
						"regular expressions and other settings used to parse it. Step 1 of 3",
						parameters
					);
				}
				else
				{
					configureParser = new RequestInformation
					(
						Request.COMMAND_CONFIGURE_PLUGIN,
						"Specify parser regular expressions and settings",
						"This plug-in can be configured with the regular expressions and " +
						"settings used to parse files. Mapping file columns to annotation types must " +
						"be done for each import.",
						parameters
					);				
				}
			}
			finally
			{
				if (dc != null) dc.close();
			}
		}
		return configureParser;
	}	
	
	/**
		Step 2 of 3. Parameters for configuring mapping between columns and annotation types
	*/
	private RequestInformation getConfigureMappingParameters(GuiContext context)
	{
		if (configureMapping == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			ItemQuery<AnnotationType> query = AnnotationType.getQuery(context.getItem());
			query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			query.setItemPermission(Permission.USE);
			query.order(Orders.asc(Hql.property("name")));
			List<AnnotationType> annotationTypes;
			Map<String, AnnotationType> names = 
				new TreeMap<String, AnnotationType>(String.CASE_INSENSITIVE_ORDER);
			DbControl dc = sc.newDbControl();
			try
			{
				annotationTypes = new ArrayList<AnnotationType>(query.list(dc));
				for (AnnotationType at : annotationTypes)
				{
					names.put(at.getName(), at);
				}
			}
			finally
			{
				if (dc != null) dc.close();
			}
			
			ParameterType<AnnotationType> annotationParameterType = 
				new ItemParameterType<AnnotationType>(AnnotationType.class, annotationTypes);
			
			parameters.add(annotationTypeSection);
			int i = 0;
			for (String columnHeader : columnHeaders)
			{
				PluginParameter<AnnotationType> pp = 
					new PluginParameter<AnnotationType>("columnMapping." + i, columnHeader, 
					"Select the annotation type that corresponds to the annotations in this column.",
					names.get(columnHeader),
					annotationParameterType);
				parameters.add(pp);
				++i;
			}
			
			configureMapping = new RequestInformation
			(
				CONFIGURE_MAPPING,
				"Map file columns to annotation types",
				"Here you select specify how the columns in the file map with annotation types. " +
				"Step 2 of 3.",
				parameters
			);
			
		}
		return configureMapping;
	}
	
	/**
		Step 3 of 3. Other options and error handling parameters
	*/
	private RequestInformation getConfigureImportParameters(GuiContext context)
	{
		if (configureImport == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
	
			// Update/add/replace options
			parameters.add(addReplaceSection);
			parameters.add(replaceExistingParameter);
			parameters.add(removeExistingParameter);
			parameters.add(addToUnlimitedParameter);
			
			// Error handling options
			parameters.add(errorSection);
			parameters.add(defaultErrorParameter);
			parameters.add(itemNotFoundErrorParameter);
			parameters.add(multipleItemsFoundErrorParameter);
			parameters.add(stringTooLongErrorParameter);
			parameters.add(numberFormatErrorParameter);
			parameters.add(dateFormatErrorParameter);
			parameters.add(numberOutOfRangeErrorParameter);
			parameters.add(invalidEnumErrorParameter);
			parameters.add(tooManyValuesErrorParameter);
			
			configureImport = new RequestInformation
			(
				CONFIGURE_IMPORT,
				"Other settings and error handling options",
				"Specify other settings and error handling options. Step 3 of 3.",
				parameters
			);
			
		}
		return configureImport;
	}
	
	/**
		Check if the current item has an 'externalId' property. We use reflection to
		look for the 'getExternalId' method in the item's data class.
	*/
	private boolean hasExternalId(Item item)
	{
		Class<?> dataClass = item.getDataClass();
		Method method = null;
		try
		{
			method = dataClass.getMethod("getExternalId");
		}
		catch (NoSuchMethodException ex)
		{}
		return method != null;
	}
	
	private List<IdMethod> getIdMethods(DbControl dc, Item item)
	{
		List<IdMethod> methods = new ArrayList<>();
		methods.add(PropertyIdMethod.NAME);
		methods.add(InternalIdMethod.INTERNAL_ID);
		if (item == null || hasExternalId(item)) methods.add(PropertyIdMethod.EXTERNAL_ID);
		methods.addAll(AnnotationIdMethod.getIdMethods(dc, item));
		return methods;
	}
	
	/**

	*/
	protected IdMethod getIdMethod(DbControl dc, Item item, String method)
	{
		for (IdMethod im : getIdMethods(dc, item))
		{
			if (im.getMethod().equals(method))
			{
				return im;
			}
		}
		return PropertyIdMethod.NAME;
	}

	/**
		Create a query that return items of the specified type. We use reflection
		to call the static method 'getQuery' on the item's item class, for
		example {@link net.sf.basedb.core.Sample#getQuery()}.
		
		@param itemType The type of items to search
		@return A query that searches items of the specified type.
	*/
	private ItemQuery<?> createQuery(Item itemType, ItemList list)
	{
		return list != null && list.getMemberType() == itemType ? list.getMembers() : itemType.getQuery();
	}
	
	private static class AnnotationValue
	{
		final Object value;
		final Unit unit;
		AnnotationValue(Object value, Unit unit)
		{
			this.value = value;
			this.unit = unit;
		}
	}
	
	/**
		Internal cache for storing annotation values from the file 
		for a single item. The items are not annotated until the entire
		file has been parsed.
	*/
	private static class NewAnnotations
	{
		private final Annotatable item;
		private final Map<AnnotationType, List<AnnotationValue>> values;
		private int numAnnotations;
		private int numSet;
		private int numError;
		private int numReplaced;
		private int numRemoved;
		
		/**
			Create a new cache for the specified item.
		*/
		NewAnnotations(Annotatable item)
		{
			this.item = item;
			this.values = new HashMap<AnnotationType, List<AnnotationValue>>();
			this.numAnnotations = 0;
		}
		
		/**
			Add an annotation value. The value should have been checked for
			error before calling this method.
		*/
		void addValue(AnnotationType at, Object value, Unit unit, boolean limitToMultiplicty)
		{
			List<AnnotationValue> listOfValues = values.get(at);
			if (listOfValues == null)
			{
				listOfValues = new LinkedList<AnnotationValue>();
				values.put(at, listOfValues);
			}
			// If "Too many values"=crop we check how many is allowed by the annotation type
			// and don't allow the list to grow larger
			if (limitToMultiplicty)
			{
				int multiplicity = at.getMultiplicity();
				if (multiplicity > 0 && listOfValues.size() >= multiplicity) value = null;
			}
			if (value != null) 
			{
				if (at.supportUnits() && unit == null) unit = at.getDefaultUnit();
				listOfValues.add(new AnnotationValue(value, unit));
				numAnnotations++;
			}
		}
		
		/**
			The total numer of annotation values that has been added to this 
			cache.
		*/
		int getNumAnnotations()
		{
			return numAnnotations;
		}
		
		/**
			Annotate the item with the annotation values from the file.
			@param addToUnlimited TRUE to append to annotation types with multiplicity==0, FALSE
				to replace
			@param replaceExisting TRUE to replace existing annotation values, FALSE to skip
			@param failIfTooManyValues FALSE to skip annotations with more values than
				the multiplicity settings allows, TRUE to throw an exception
		*/
		void setNewAnnotations(AnnotationBatcher batcher, boolean addToUnlimited, boolean replaceExisting, 
			boolean failIfTooManyValues, boolean removeAnnotations)
		{
			if (values.size() == 0) return;
			int numAnnotations = 0;
			
			batcher.setCurrentItem(item);
			
			for (Map.Entry<AnnotationType, List<AnnotationValue>> entry : values.entrySet())
			{
				AnnotationType at = entry.getKey();

				List<AnnotationValue> newValues = entry.getValue();
				
				int size = newValues.size();
				int multiplicity = at.getMultiplicity();
				boolean merge = addToUnlimited && multiplicity == 0;
				boolean tooManyValues = multiplicity > 0 && size > multiplicity;
				
				if (size == 0 && !removeAnnotations) continue; // Skip removing
				if (tooManyValues && !failIfTooManyValues) 
				{
					numError += size;
					continue; // Do not try to set too many values
				}
				
				List<Object> theValues = new ArrayList<Object>(newValues.size());
				// Make sure all annotation values are converted to the same unit
				Unit unitToUse = size > 0 ? newValues.get(0).unit : at.getDefaultUnit();
				for (AnnotationValue value : newValues)
				{
					Object theValue = value.value;
					if (unitToUse != null && value.unit != null && !unitToUse.equals(value.unit)) 
					{
						theValue = unitToUse.fromOther(((Number)value.value).doubleValue(), value.unit);
					}
					theValues.add(theValue);
				}
				Change change = batcher.setValues(at, theValues, unitToUse, merge);
				if (change == Change.DELETED)
				{
					numRemoved++;
				}
				else if (change == Change.ADDED)
				{
					numSet += theValues.size();
				}
				else if (change == Change.UPDATED)
				{
					numSet += theValues.size();
					if (!merge) numReplaced += theValues.size();
				}
			}
		}
		
		/**
			Number of annotations that was set (=added or replaced) by the call to the 
			{@link #setNewAnnotations(AnnotationBatcher, boolean, boolean, boolean, boolean)} method.
		*/
		int getNumSet()
		{
			return numSet;
		}

		/**
			Number of annotations that was replaced by the call to the
			{@link #setNewAnnotations(AnnotationBatcher, boolean, boolean, boolean, boolean)} method.
		*/
		int getNumReplaced()
		{
			return numReplaced;
		}			
		
		/**
			Number of annotations that was removed by the call to the
			{@link #setNewAnnotations(AnnotationBatcher, boolean, boolean, boolean, boolean)} method.
			@since 2.8
		*/
		int getNumRemoved()
		{
			return numRemoved;
		}
		
		/**
			Number of annotations that was couldn't be set by the call to the 
			{@link #setNewAnnotations(AnnotationBatcher, boolean, boolean, boolean, boolean)} method because of an
			error (too many values).
		*/
		int getNumError()
		{
			return numError;
		}
		
	}
}
