/**
	$Id: BaseFileExporterPlugin.java 7605 2019-02-26 10:10:15Z nicklas $

	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Jari Häkkinen, Peter Johansson, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2008 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileSetMember;
import net.sf.basedb.core.FileStoreUtil;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.ExtraValue.CoordinateType;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.EqualsHelper;
import net.sf.basedb.util.basefile.BaseFileWriter;
import net.sf.basedb.util.export.spotdata.BaseFileExporter;
import net.sf.basedb.util.export.spotdata.DynamicField;
import net.sf.basedb.util.export.spotdata.ExportableFieldFactory;
import net.sf.basedb.util.export.spotdata.MatrixBaseFileExporter;
import net.sf.basedb.util.export.spotdata.SerialBaseFileExporter;
import net.sf.basedb.util.formatter.Formatter;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
	 Exports bioassay set data as serial or matrix BASEfile:s.

	 @author Nicklas
	 @version 2.12
	 @base.modified $Date: 2019-02-26 11:10:15 +0100 (tis, 26 feb. 2019) $
*/
public class BaseFileExporterPlugin
	extends AbstractExporterPlugin
	implements InteractivePlugin, SignalTarget
{

	private static final Set<GuiContext> guiContexts =
		Collections.singleton(GuiContext.item(Item.BIOASSAYSET));

	private static final Set<Permissions> permissions = new HashSet<Permissions>();

 	private static final String FORMAT_BASEFILE_SERIAL = SerialBaseFileExporter.SUBTYPE;
 	private static final String FORMAT_BASEFILE_MATRIX = MatrixBaseFileExporter.SUBTYPE;

	
	private ThreadSignalHandler signalHandler;
	private RequestInformation configureJob;

	public BaseFileExporterPlugin()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return false;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return false;
	}

	/**
	 	Request create and write access to File:s and Directory:s and read access to
	 	Experiment:s and User:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.ANNOTATIONTYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.WRITE)));
			permissions.add(new Permissions(Item.FORMULA, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.RAWBIOASSAY, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.USER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.EXTRAVALUETYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DATAFILETYPE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// --------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = null;	
		if (item == null)
		{
			message = "The object is null";
		}
		else if (!(item instanceof BioAssaySet))
		{
			message = "The object is not a BioAssaySet: " + item;
		}
		else
		{
			BioAssaySet bas = (BioAssaySet)item;
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
		}
		return message;
	}

	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context, null);
		}
		return requestInformation;
	}
	
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		try
		{
			String command = request.getCommand();
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters (context, !request.isAllowedImmediateExecution());
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found " +
														"in the request", errors);
					return;
				}
				String fileName = (String)request.getParameterValue(SAVE_AS);
				if (fileName == null)
				{
					if (!request.isAllowedImmediateExecution())
					{
						response.setError("Immediate download is not allowed. Please specify a filename", null);
						return;
					}
				}
				else
				{
					boolean overwrite = Boolean.TRUE.equals(request.getParameterValue(OVERWRITE));
					if (!pathCanBeUsed(fileName, overwrite))
					{
						response.setError("File exists: " + fileName + 
								"; Please select a different file name.", null);
						return;
					}
				}

				// Source and file options
				storeValue(job, request, ri.getParameter("source"));
				storeValue(job, request, ri.getParameter(SAVE_AS));
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				storeValue(job, request, ri.getParameter(OVERWRITE));
				storeValue(job, request, ri.getParameter("attachToBioAssaySet"));
				
				// Export options
				storeValue(job, request, ri.getParameter("fileformat"));
				storeValue(job, request, ri.getParameter("averageOnReporter"));
				storeValues(job, request, ri.getParameter("spotFields"));
				storeValues(job, request, ri.getParameter("reporterFields"));

				if (job.getValue(SAVE_AS) == null)
				{
					// No filename --> immediate download
					BioAssaySet source = (BioAssaySet)request.getParameterValue("source");
					response.setSuggestedJobName("Download bioassay set '" +source.getName() + "' as " + 
							job.getValue("fileformat") + " BASEfile");
					
					response.setDownloadImmediately("Export bioassay set " + 
							source.getName(), Job.ExecutionTime.SHORT, true);
				}
				else
				{
					BioAssaySet source = (BioAssaySet)job.getValue("source");
					response.setSuggestedJobName("Export bioassay set '" +source.getName() + "' to " + 
							job.getValue("fileformat") + " BASEfile: " + job.getValue(SAVE_AS));
					
					response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
				}
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// --------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	@Override
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	/*
		From AbstractExporterPlugin
		------------------------------------------.
	*/
	private DbControl dc;
	private BioAssaySet bas;
	private Map<String, ExtendedDynamicField> exportableFields;

	@Override
	protected void begin(DbControl dc)
	{
		super.begin(dc);
		this.dc = dc;
		bas = (BioAssaySet)job.getValue("source");
	}	

	@Override
	protected void performExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);

		// Get job configuration parameters
		BioAssaySet source = job.getValue("source");
		source = BioAssaySet.getById(dc, source.getId());
		boolean attachToBioAssaySet = Boolean.TRUE.equals(job.getValue("attachToBioAssaySet"));
		String characterSet = job.getValue(Parameters.CHARSET_PARAMETER);
		if (characterSet == null) characterSet = Config.getCharset();
		String format = job.getValue("fileformat");
		List<String> reporterFields = job.getValues("reporterFields");
		List<String> spotFields = job.getValues("spotFields");
		if (reporterFields == null) reporterFields = Collections.emptyList();
		if (spotFields == null) spotFields = Collections.emptyList();
		boolean averageOnReporter = Boolean.TRUE.equals(job.getValue("averageOnReporter"));

		if (progress != null) progress.display(0, "Exporting ...");
		
		// Make settings on the exported file
		out.setMimeType("text/plain");
		String filename = source.getName() + "-" + source.getId() + "-"+ format + ".txt";
		out.setFilename(Path.makeSafeFilename(filename, ""));
		out.setCharacterSet(characterSet);
		File exportFile = out.getFile();
		String dataFileTypeId = null;
		
		// Prepare for export
		BaseFileExporter exporter = null;
		if (FORMAT_BASEFILE_SERIAL.equals(format))
		{
			exporter = new SerialBaseFileExporter();
			dataFileTypeId = DataFileType.BASEFILE_SERIAL;
		}
		else if (FORMAT_BASEFILE_MATRIX.equals(format))
		{
			exporter = new MatrixBaseFileExporter();
			dataFileTypeId = DataFileType.BASEFILE_MATRIX;
		}
		else
		{
			throw new BaseException("Unknown format. "+format);
		}
		exporter.setDbControl(dc);
		exporter.setProgressReporter(progress);
		exporter.setAverageOnReporter(averageOnReporter);
		exporter.setSource(source);
		exporter.setWriter(new BaseFileWriter(new OutputStreamWriter(out, characterSet)));
		Map<String, ExtendedDynamicField> all = getExportableFields(dc, source);
		for (String field : reporterFields)
		{
			ExtendedDynamicField ef = all.get(field);
			exporter.addReporterField(ef);
		}
		for (String field : spotFields)
		{
			ExtendedDynamicField ef = all.get(field);
			exporter.addSpotField(ef);
		}
		exporter.doExport();
		
		if (attachToBioAssaySet && exportFile != null && dataFileTypeId != null)
		{
			FileSetMember file = FileStoreUtil.setDataFile(dc, source, dataFileTypeId, exportFile);
			file.setValid(true, null);
			source.setNumFileSpots(source.getNumSpots());
			source.setNumFileReporters(source.getNumReporters());
			for (BioAssay ba : source.getBioAssays().list(dc))
			{
				ba.setNumFileSpots(ba.getNumSpots());
			}
		}
	}
	
	@Override
	protected void end(boolean success)
	{
		this.dc = null;
	}
	
	@Override
	protected String getSuccessMessage()
	{
		return "Bioassay set '" + bas.getName() + "' exported successfully";
	}
	// -------------------------------------------

	public final static String PARAMETER_FILEFORMAT = "fileformat";
	
	/**
		Create a plug-in parameter that asks for the BASEFile subtype to use in
		the export.
	*/
	public static PluginParameter<String> getFileFormatParameter()
	{
		Enumeration<String, String> formats = new Enumeration<String, String>();
		formats.add(FORMAT_BASEFILE_SERIAL, "Serial BASEfile");
		formats.add(FORMAT_BASEFILE_MATRIX, "Matrix BASEfile");
		return new PluginParameter<String>(
				PARAMETER_FILEFORMAT,
				"File format",
				"Please specify if you want to export a serial or matrix BASEfile. " +
				"A serial file has one section for each assay, and a matrix file " +
				"has all data in one big section",
				new StringParameterType(255, FORMAT_BASEFILE_SERIAL, true, 1, 0, 0, formats) 
			);
	}

	
	private RequestInformation getConfigureJobParameters(GuiContext context, Boolean requireFile)
	{
		if (configureJob != null) return configureJob;

		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			int currentBioAssaySetId = sc.getCurrentContext(Item.BIOASSAYSET).getId();			
			BioAssaySet source = BioAssaySet.getById(dc, currentBioAssaySetId);
			
			// Generate a default path if a file is required
			String defaultPath = null;
			if (requireFile == null)
			{
				PluginDefinition pd = job.getPluginDefinition();
				requireFile = pd == null ? false : !pd.getAllowImmediateExecution();
			}
			if (requireFile)
			{
				defaultPath = "~/" + Path.makeSafeFilename(
						source.getName() + "-" + source.getId(), "").toLowerCase() + ".txt";
			}
			
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			// Source bioassay set parameter
			parameters.add(new PluginParameter<BioAssaySet>(
					"source",
					"Bioassay set",
					"The bioassay set to export",
					new ItemParameterType<BioAssaySet>(BioAssaySet.class, source, true, 1, null) 
				));

			// Export File format
			parameters.add(getFileFormatParameter());

			// Spot and reporter fields
			Enumeration<String, String> spotFields = new Enumeration<String, String>();
			Enumeration<String, String> reporterFields = new Enumeration<String, String>();
			
			for (ExtendedDynamicField f : getExportableFields(dc, source).values())
			{
				if (f.isReporterField())
				{
					reporterFields.add(f.getId(), f.getGuiTitle());
				}
				else
				{
					spotFields.add(f.getId(), f.getGuiTitle());
				}
			}

			parameters.add(new PluginParameter<String>(
					"spotFields",
					"Spot fields",
					"Select the spot data that should be exported.",
					new StringParameterType(null, null, true, 0, 0, 0, spotFields)
				));

			parameters.add(new PluginParameter<String>(
					"reporterFields",
					"Reporter fields",
					"Select the reporter annotations that should be exported.",
					new StringParameterType(null, null, false, 0, 0, 0, reporterFields)
				));

			// Average on reporter
			parameters.add(new PluginParameter<Boolean>("averageOnReporter",
					"Average on reporters",
					"Average on reporters within each assay. If false or not specified " +
					"no averaging is done",
					new BooleanParameterType(null, false)
				));
			
			// Save to file options
			parameters.add(getSaveAsParameter(null, null, defaultPath, requireFile));
			parameters.add(Parameters.charsetParameter(
					"Character set", 
					"The character set to use when writing the BASEfile. If this parameter " +
					"is not set, the system default is used (" + Config.getCharset() + ").", 
					Config.getCharset()));
			parameters.add(getOverwriteParameter(null, null));
			parameters.add(new PluginParameter<Boolean>(
					"attachToBioAssaySet",
					"Attach file to bioassay set",
					"Deselect this option if you don't want the generated file to " +
					"become attached to the current bioassay set. This option is ignored " +
					"if the file isn't saved in the BASE file system.",
					new BooleanParameterType(true, false)
				));

			configureJob = new RequestInformation(
					Request.COMMAND_CONFIGURE_JOB,
					"Select export options.",
					"Select a file name (if required) and other options such as " +
					"which fields to export.",
					parameters
				);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}
	
	public Map<String, ExtendedDynamicField> getExportableFields(DbControl dc, BioAssaySet source)
	{
		if (exportableFields != null) return exportableFields;
		exportableFields = new LinkedHashMap<String, ExtendedDynamicField>();
		
		List<ExtendedDynamicField> all = new ArrayList<ExtendedDynamicField>();
		all.addAll(getSpotFields(source));
		all.addAll(getReporterFields());
		all.addAll(getFormulaFields(dc, source));
		all.addAll(getRawdataFields(source));
		all.addAll(getExtraValueFields(dc, source));
		
		for (ExtendedDynamicField ef : all)
		{
			exportableFields.put(ef.getId(), ef);
		}
		return exportableFields;
	}

	private List<ExtendedDynamicField> getFormulaFields(DbControl dc, BioAssaySet source)
	{
		List<ExtendedDynamicField> items = new ArrayList<ExtendedDynamicField>();
		ItemQuery<Formula> queryFormula = Formula.getQuery(Formula.Type.COLUMN_EXPRESSION, 
				source.getRawDataType(), source.getIntensityTransform());
		queryFormula.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
		queryFormula.order(Orders.asc(Hql.property("name")));
		for (Formula f : queryFormula.list(dc))
		{
			DynamicField fld = ExportableFieldFactory.formula(f, f.getName(), null);
			items.add(new ExtendedDynamicField("frm." + f.getId(), "[Formula] "+f.getName(), false, fld));
		}
		return items;
	}
	
	private List<ExtendedDynamicField> getRawdataFields(BioAssaySet source)
	{
		Formula.AverageMethod NONE = Formula.AverageMethod.NONE;
		List<ExtendedDynamicField> items = new ArrayList<ExtendedDynamicField>();
		items.add(rawField("position", "[Raw] Position", Type.INT, NONE));
		items.add(rawField("block", "[Raw] Array block", Type.INT, NONE));
		items.add(rawField("metaGridX", "[Raw] Meta grid X", Type.INT, NONE));
		items.add(rawField("metaGridY", "[Raw] Meta grid Y", Type.INT, NONE));
		items.add(rawField("row", "[Raw] Row in block", Type.INT, NONE));
		items.add(rawField("column", "[Raw] Column in block", Type.INT, NONE));
		items.add(rawField("x", "[Raw] X", Type.FLOAT, NONE));
		items.add(rawField("y", "[Raw] Y", Type.FLOAT, NONE));
		for (RawDataProperty rdp : source.getRawDataType().getProperties())
		{
			items.add(rawField(rdp.getName(), "[Raw] "+rdp.getTitle(), rdp.getType(), rdp.getAverageMethod()));
		}
		return items;
	}

	private ExtendedDynamicField rawField(String property, String guiTitle, Type type, Formula.AverageMethod averageMethod)
	{
		DynamicField df = ExportableFieldFactory.rawData(property, guiTitle, type, averageMethod, null);
		return new ExtendedDynamicField("raw." + property, guiTitle, false, df);
	}

	private Collection<ExtendedDynamicField> getReporterFields()
	{
		Formula.AverageMethod NONE = Formula.AverageMethod.NONE;
		Formula.AverageMethod MAX = Formula.AverageMethod.MAX;
		List<ExtendedDynamicField> items = new ArrayList<ExtendedDynamicField>();
		items.add(reporterField("id", "Internal id", Type.INT, NONE));
		items.add(new ExtendedDynamicField("rep.position", true, 
				ExportableFieldFactory.jep("pos()", "Position", Type.INT, NONE, null)));
		items.add(reporterField("name", "Name", Type.STRING, MAX));
		items.add(reporterField("externalId", "External id", Type.STRING, MAX));
		items.add(reporterField("symbol", "Gene symbol", Type.STRING, MAX));
		items.add(reporterField("description", "Description", Type.TEXT, MAX));
		items.add(reporterField("lastUpdate", "Last updated", Type.TIMESTAMP, MAX));
		
		for (ExtendedProperty ep : ExtendedProperties.getProperties("ReporterData"))
		{
			items.add(reporterField(ep.getName(), ep.getTitle(), ep.getType(), MAX));
		}
		return items;
	}

	
	private ExtendedDynamicField reporterField(String property, String guiTitle, Type type, Formula.AverageMethod averageMethod)
	{
		DynamicField df = ExportableFieldFactory.reporter(property, guiTitle, type, averageMethod, null);
		return new ExtendedDynamicField("rep." + property, guiTitle, true, df);
	}
	
	private Collection<ExtendedDynamicField> getExtraValueFields(DbControl dc, BioAssaySet source)
	{
		List<ExtendedDynamicField> items = new ArrayList<ExtendedDynamicField>();
		ItemQuery<ExtraValue> query = source.getExtraValues();
		query.order(Orders.asc(Hql.property("extraValueType.name")));
		for (ExtraValue ev : query.list(dc))
		{
			ExtraValueType evt = ev.getExtraValueType();
			DynamicField df = ExportableFieldFactory.extraValue(evt, evt.getName(), null);
			boolean positionValue = ev.getCoordinateType() == CoordinateType.POSITION;
			items.add(new ExtendedDynamicField("xtra." + evt.getId(), "[Extra] " + evt.getName(), positionValue, df));
		}
		return items;
	}
	
	private Collection<ExtendedDynamicField> getSpotFields(BioAssaySet source)
	{
		List<ExtendedDynamicField> items = new ArrayList<ExtendedDynamicField>();
		int channels = source.getRawDataType().getChannels();
		for (int channel = 1; channel <= channels; ++channel)
		{
			DynamicField df = ExportableFieldFactory.channel(channel, "Ch " + channel, null);
			items.add(new ExtendedDynamicField("ch" + channel, false, df));
		}
		return items;
	}
	
 	
 	public static class ExtendedDynamicField
 		implements DynamicField
 	{

 		private final String id;
 		private final DynamicField parent;
 		private final String guiTitle;
 		private final boolean reporterField;
 		
 		public ExtendedDynamicField(String id, boolean reporterField, DynamicField parent)
 		{
 			this(id, parent.getTitle(), reporterField, parent);
 		}
 		
 		public ExtendedDynamicField(String id, String guiTitle, boolean reporterField, DynamicField parent)
 		{
 			this.id = id;
 			this.guiTitle = guiTitle;
 			this.reporterField = reporterField;
 			this.parent = parent;
 		}
		
 		/*
 			From the DynamicField interface
 			-------------------------------
 		*/
		@Override
		public Expression getExpression(DbControl dc, DynamicQuery query, BioAssaySet source, boolean forAverage)
		{
			return parent.getExpression(dc, query, source, forAverage);
		}

		@Override
		public Formatter<?> getFormatter()
		{
			return parent.getFormatter();
		}

		@Override
		public String getTitle()
		{
			return parent.getTitle();
		}
		
		@Override
		public Type getType()
		{
			return parent.getType();
		}
		// ---------------------------------
		/**
			An extended dynamic field is equal to another field if their parent fields
			equal.
			@since 2.15
		*/
		@Override
		public boolean equals(Object obj)
		{
			if (this == obj) return true;
			if (!(obj instanceof ExtendedDynamicField)) return false;
			ExtendedDynamicField other = (ExtendedDynamicField)obj;
			return EqualsHelper.equals(this.parent, other.parent);
		}
		
		/**
			@since 2.15
		*/
		@Override
		public int hashCode()
		{
			return EqualsHelper.hashCode(parent);
		}

		public String getId()
		{
			return id;
		}
		
		public String getGuiTitle()
		{
			return guiTitle;
		}
		
		public boolean isReporterField()
		{
			return reporterField;
		}
 	}
 	
}
