/*
	$Id: PlateFlatFileImporter.java 7666 2019-03-20 15:21:26Z nicklas $
	

	Copyright (C) 2005 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.NumberOutOfRangeException;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Plate;
import net.sf.basedb.core.PlateType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Well;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.plugins.util.Parameters;
import net.sf.basedb.util.Coordinate;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.error.SimpleErrorHandler;
import net.sf.basedb.util.parser.ConfigureByExample;
import net.sf.basedb.util.parser.ConstantMapper;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 
 @base.modified $Date: 2019-03-20 16:21:26 +0100 (ons, 20 mars 2019) $
 @author Enell
 @version 2.0
 */
public class PlateFlatFileImporter 
	extends AbstractFlatFileImporter 
	implements InteractivePlugin, ConfigureByExample
{

	private static final Set<GuiContext> guiContexts = 
		Collections.singleton(new GuiContext(Item.PLATE, GuiContext.Type.LIST));
	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();

	private static final PluginParameter<String> columnColumnMapping = new PluginParameter<String>
	(
		"columnColumnMapping",
		"Column",
		"Mapping that picks the well's column in the plate from the data columns. " +
		"For example: \\Column\\",
		new StringParameterType(255, null, true)
	);

	private static final PluginParameter<String> rowColumnMapping = new PluginParameter<String>
	(
		"rowColumnMapping",
		"Row",
		"Mapping that picks the well's row in the plate from the data columns. " +
		"For example: \\Row\\",
		new StringParameterType(255, null, true)
	);
	
	private static final PluginParameter<String> reporterColumnMapping = new PluginParameter<String>
	(
		"reporterColumnMapping",
		"Reporter ID",
		"Mapping that picks the 'External ID' of the well's reporter from the data columns. " +
		"For example: \\ID\\",
		new StringParameterType(255, null, true)
	);

	private static final PluginParameter<String> nameColumnMapping = new PluginParameter<String>
	(
		"nameColumnMapping",
		"Plate number/name",
		"Mapping that picks the plate name from the data columns. " +
		"For example: \\Plate\\",
		new StringParameterType(255, null, true)
	);
	
	private static final PluginParameter<String> barcodeColumnMapping = new PluginParameter<String>
	(
		"barcodeColumnMapping",
		"Plate barcode",
		"Mapping that picks the plate barcode from the data columns. " +
		"For example: \\Barcode\\",
		new StringParameterType(255, null, false)
	);
	
	private PluginParameter<String> plateNamePrefixParameter = new PluginParameter<String>
	(
		"plateNamePrefix",
		"Plate name prefix",
		"A prefix that is appended to the plate name.",
		new StringParameterType(255, null, false)
	);
	
	private PluginParameter<String> reporterPrefixParameter = new PluginParameter<String>
	(
		"reporterPrefix",
		"Reporter ID prefix",
		"This prefix is prepended to the external reporter IDs in the input file before " +
		"a reporter lookup is performed.",
		new StringParameterType(255, null, false)
	);
	
	private PluginParameter<Integer> namePaddingParameter = new PluginParameter<Integer>
	(
		"namePadding",
		"Plate name padding",
		"The plate name will be left-padded with spaces to this length.",
		new IntegerParameterType(0, 255, 4, false)
	);
	
	protected static final PluginParameter<String> missingReporterErrorParameter = new PluginParameter<String>(
			"missingReporterError",
			"Reporter not found",
			"How to handle errors that are caused by a reporter not beeing present in the datbase. If not specified the " +
				"default error handling is used.\n\n"+
				"null = Import the data but set the reporter to null\n"+
				"skip = Skip the current data line and continue\n"+
				"fail = Stop with an error message",
				new StringParameterType(255, null, false, 1, 0, 0, 
						Arrays.asList( new String[] { "null", "skip", "fail"} ))
			);

	
	private ItemParameterType<PlateType> plateTypeType;
	private PluginParameter<PlateType> plateTypeParameter;
	
	private List<PluginParameter<String>> allColumnMappings;
	private RequestInformation configurePlugin;
	private RequestInformation configureJob;
	
	private DbControl dc;
	private FlatFileParser ffp;
	private int numPlates;
	private int numWells;
	private PlateType plateType;
	private String plateNamePrefix;
	private Integer plateNamePadding;
	private String reporterPrefix;
	
	private Mapper reporterMapper;
	private Mapper nameMapper;
	private Mapper barcodeMapper;
	private Mapper rowMapper;
	private Mapper columnMapper;
	
	private ReporterBatcher reporterBatcher;
	private boolean errorIfNotFound;
	private HashMap<String, Plate> plates;
		
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	@Override
	public boolean requiresConfiguration()
	{
		return true;
	}
	/**
	 	Request create access to Plate:s, use access to PlateType:s and
	 	read access to Reporter:s, PlateGeometry:s and File:s.
	*/
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.PLATEGEOMETRY, EnumSet.of(Permission.READ), null));
			permissions.add(new Permissions(Item.PLATETYPE, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.PLATE, null, EnumSet.of(Permission.CREATE, Permission.READ)));
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	// -------------------------------------------
	
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Always null, since plugin doesn't operate on individual items.
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		return null;
	}

	/**
		The {@link Request#COMMAND_CONFIGURE_PLUGIN} command will ask for
		parser regular expressions and column mappings.
		The {@link Request#COMMAND_CONFIGURE_JOB} command will ask for
		a file and what {@link PlateType} the plates in the file has.
	
	*/
	@Override
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters(context);
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters(context);
			ItemContext fileContext = sc.getCurrentContext(Item.FILE);
			fileContext.removeAllTemporaryFilters();
		}
		return requestInformation;
	}
	
	/**
		Store configuration settings for {@link Request#COMMAND_CONFIGURE_PLUGIN} 
		and {@link Request#COMMAND_CONFIGURE_JOB}.
	*/
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				RequestInformation ri = getConfigurePluginParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Parser settings
				storeValue(configuration, request, headerRegexpParameter);
				storeValue(configuration, request, dataHeaderRegexpParameter);
				storeValue(configuration, request, dataSplitterRegexpParameter);
				storeValue(configuration, request, trimQuotesParameter);
				storeValue(configuration, request, ignoreRegexpParameter);
				storeValue(configuration, request, dataFooterRegexpParameter);
				storeValue(configuration, request, minDataColumnsParameter);
				storeValue(configuration, request, maxDataColumnsParameter);
				storeValue(configuration, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				
				// Column mappings
				storeValue(configuration, request, complexMappings);
				boolean allowComplex = "allow".equals(request.getParameterValue(complexMappings.getName()));
				for (PluginParameter<String> pp : getAllColumnMappings())
				{
					String mapExpression = (String)request.getParameterValue(pp.getName());
					checkColumnMapping(mapExpression, allowComplex, pp.getLabel());
					storeValue(configuration, pp, mapExpression);
				}
				response.setDone("Plugin configuration complete");
			}
			else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size()+" invalid parameter(s) were found in the request", errors);
					return;
				}
				
				storeValue(job, request, namePaddingParameter);
				storeValue(job, request, reporterPrefixParameter);
				storeValue(job, request, plateNamePrefixParameter);
				storeValue(job, request, plateTypeParameter);
				storeValue(job, request, fileParameter);
				storeValue(job, request, ri.getParameter(Parameters.CHARSET_PARAMETER));
				
				// Error handling parameters
				storeValue(job, request, defaultErrorParameter);
				storeValue(job, request, missingReporterErrorParameter);
				storeValue(job, request, stringTooLongErrorParameter);
				storeValue(job, request, invalidUseOfNullErrorParameter);
				storeValue(job, request, numberFormatErrorParameter);
				
				File file = (File)job.getValue("file");
				response.setSuggestedJobName("Import plates from file '" + file.getName() + "'");
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}

	// -------------------------------------------

	/*
		From the AbstractFlatFileReporter class
		-------------------------------------------
	*/
	/**
		Create a {@link DbControl} and a {@link net.sf.basedb.core.RawDataBatcher}. 
		Load the {@link net.sf.basedb.core.RawBioAssay} that has been specified.
	*/
	@Override
	protected void begin(FlatFileParser ffp) 
		throws BaseException
	{
		super.begin(ffp);
		dc = sc.newDbControl();
		plates = new HashMap<String, Plate>();
		plateType = dc.reattachItem((PlateType)job.getValue("plateType"), false);
		plateNamePrefix = (String)job.getValue("plateNamePrefix");
		if (plateNamePrefix == null) plateNamePrefix = "";
		reporterPrefix = (String)job.getValue("reporterPrefix");
		if (reporterPrefix == null) reporterPrefix = "";
		reporterBatcher = ReporterBatcher.getNew(dc);
		plateNamePadding = (Integer)job.getValue("namePadding");
		String method = (String)job.getValue("missingReporterError");
		errorIfNotFound = !"null".equals(method);
		if (method != null)
		{
			addErrorHandler(ItemNotFoundException.class, new SimpleErrorHandler("skip".equals(method)));
		}
		numPlates = 0;
		numWells = 0;
		this.ffp = ffp;
	}

	/**
		Initialise column <code>Mapper</code>:s.
	*/
	@Override
	protected void beginData()
	{
		Mapper nullMapper = new ConstantMapper((String)null);
		boolean cropStrings = ("crop".equals(job.getValue("stringTooLongError")));
		reporterMapper = getMapper(ffp, (String)configuration.getValue("reporterColumnMapping"), 
			cropStrings ? ReporterData.MAX_EXTERNAL_ID_LENGTH : null, nullMapper);

		nameMapper = getMapper(ffp, (String)configuration.getValue("nameColumnMapping"), 
			cropStrings ? Plate.MAX_NAME_LENGTH : null, nullMapper);
		barcodeMapper = getMapper(ffp, (String)configuration.getValue("barcodeColumnMapping"), 
			cropStrings ? Plate.MAX_BARCODE_LENGTH : null, nullMapper);
		rowMapper = getMapper(ffp, (String)configuration.getValue("rowColumnMapping"), null, nullMapper);
		columnMapper = getMapper(ffp, (String)configuration.getValue("columnColumnMapping"), null, nullMapper);
	}
	
	/**
		Close and commit/rollback and DbControl.
	*/
	@Override
	protected void end(boolean success) 
		throws BaseException
	{
		try
		{
			if (success)
			{
				dc.commit();
			}
			else
			{
				dc.close();
			}
		}
		catch (BaseException ex)
		{
			dc.close();
			throw ex;
		}
		finally
		{
			super.end(success);
		}
	}

	@Override
	protected void handleData(Data data) 
		throws BaseException
	{
		String plateName = nameMapper.getString(data);
		if (plateName != null)
		{
			Plate plate = plates.get(plateName);
			if (plate == null)
			{
				plate = Plate.getNew(dc, plateType);
				plates.put(plateName, plate);
				if (plateNamePadding != null)
				{
					plateName = MD5.leftPad(plateName, '0', plateNamePadding);
				}
				plate.setName(plateNamePrefix + plateName);
				plate.setBarcode(barcodeMapper.getString(data));
				dc.saveItem(plate);
				numPlates++;
			}

			// Find well from coordinates
			String row = rowMapper.getString(data);
			String column = columnMapper.getString(data);
			if (row == null) throw new InvalidUseOfNullException("row");
			if (column == null) throw new InvalidUseOfNullException("column");
			int r = Coordinate.alphaToNumeric(row);
			int c = Coordinate.alphaToNumeric(column);
			
			Well well = plate.getWell(r-1, c-1);
			if (well == null)
			{
				throw new NumberOutOfRangeException("Well coordinates [" + r + ", " + c + "] is outside the range defined by the geometry, "+plate.getPlateType().getPlateGeometry().getName()+".");
			}

			// Find reporter
			String reporterId = reporterMapper.getString(data);
			if (reporterId != null)
			{
				well.setReporter(reporterBatcher.getByExternalId(reporterPrefix + reporterId, errorIfNotFound));
			}
			numWells++;
		}
	}
	/**
		Return <code>x plates inserted</code>.
	*/
	@Override
	protected String getSuccessMessage(int skippedLines)
	{
		String skipped = skippedLines > 0 ? "; " + skippedLines + " items skipped due to errors" : "";
		return numPlates + " plate(s) inserted; " +
			numWells + " well(s) inserted"+
			skipped;
	}
	// -------------------------------------------

	
	private RequestInformation getConfigurePluginParameters(GuiContext context)
	{
		if (configurePlugin == null)
		{
			// Parameters for CONFIGURE_PLUGIN
			ArrayList<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			
			// Parser regular expressions
			parameters.add(parserSection);
			parameters.add(headerRegexpParameter);
			parameters.add(dataHeaderRegexpParameter);
			parameters.add(dataSplitterRegexpParameter);
			parameters.add(trimQuotesParameter);
			parameters.add(ignoreRegexpParameter);
			parameters.add(dataFooterRegexpParameter);
			parameters.add(minDataColumnsParameter);
			parameters.add(maxDataColumnsParameter);
			parameters.add(Parameters.charsetParameter(null, null, null));
			
			// Column mappings
			parameters.add(mappingSection);
			parameters.add(complexMappings);
			parameters.addAll(getAllColumnMappings());
		
			configurePlugin = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Parser settings",
				"Enter the regular expressions used to parse the text file " +
				"and the column mappings used to find matching properties for the new plates.",
				parameters
			);
		}
		return configurePlugin;
	}
	
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			// Parameters for CONFIGURE_JOB
			DbControl dc = sc.newDbControl();
			ItemQuery<PlateType> query = PlateType.getQuery();
			query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			query.order(Orders.asc(Hql.property("name")));
			List<PlateType> plateTypes = new ArrayList<PlateType>(query.list(dc));
			dc.close();
			
			plateTypeType = 
				new ItemParameterType<PlateType>(PlateType.class, null, true, 1, plateTypes);
			plateTypeParameter = new PluginParameter<PlateType>(
				"plateType",
				"Plate type",
				"Select which plate type you are uploading since the plate file itself "+
				"does not necessarily contain information about the plate type.",
				plateTypeType);

			ArrayList<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(plateTypeParameter);
			parameters.add(fileParameter);
			parameters.add(Parameters.charsetParameter(null, null, 
					(String)configuration.getValue(Parameters.CHARSET_PARAMETER)));
			parameters.add(plateNamePrefixParameter);
			parameters.add(reporterPrefixParameter);
			parameters.add(namePaddingParameter);
			
			// Error handling parameters
			parameters.add(errorSection);
			parameters.add(defaultErrorParameter);
			parameters.add(missingReporterErrorParameter);
			parameters.add(stringTooLongErrorParameter);
			parameters.add(invalidUseOfNullErrorParameter);
			parameters.add(numberFormatErrorParameter);
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Select a file to import plates from",
				"Here you specify which file to import from and some other settings having to "+
				"do with the naming of the new plates and reporters.",
				parameters
			);
		}
		return configureJob;
	}
	
	private List<PluginParameter<String>> getAllColumnMappings()
	{
		if (allColumnMappings == null)
		{
			// Column mappings
			allColumnMappings = new ArrayList<PluginParameter<String>>();
			allColumnMappings.add(nameColumnMapping);
			allColumnMappings.add(barcodeColumnMapping);
			allColumnMappings.add(rowColumnMapping);
			allColumnMappings.add(columnColumnMapping);
			allColumnMappings.add(reporterColumnMapping);
		}
		return allColumnMappings;
	}
	
}
