/**
	$Id: TestBioPlate.java 6100 2012-09-04 08:43:21Z nicklas $

	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.BioWell;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemProxy;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.MeasuredBioMaterial;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PlateGeometry;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Tag;
import net.sf.basedb.core.Trashcan;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;
import net.sf.basedb.util.formatter.WellCoordinateFormatter;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Set;

/**
	 @author martin
	 @version 2.10
	 @base.modified $Date: 2012-09-04 10:43:21 +0200 (tis, 04 sep. 2012) $
 */

public class TestBioPlate 
{
	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}
	
	static boolean test_all()
	{
		write("++Testing bioplates");
		write_header();
		
		int sampleId = TestSample.test_create(0, "Sample A.1", 0, true);
		int sample2Id = TestSample.test_create(0, "Sample A.2", 0, true);
		int sample3Id = TestSample.test_create(0, "Sample A.3", 0, true);
		int extractId = TestExtract.test_create(sampleId, "Extract A.1", 0, 0, true);
		int labelId = TestTag.test_create("Test label", SystemItems.getId(Tag.LABEL), true);
		int labeledExtractId = TestExtract.test_create(extractId, "Labeled extract A.1", 
				SystemItems.getId(Extract.LABELED), labelId, true);
		int freezerId = TestHardware.test_create(SystemItems.getId(Hardware.FREEZER), null, false);
		int protocolId = TestProtocol.test_create(SystemItems.getId(Protocol.SAMPLING), 
			"A plate event protocol", false);
		int hybProtocolId = TestProtocol.test_create(SystemItems.getId(Protocol.HYBRIDIZATION), 
				"A hyb event protocol", false);
		
		// Standard tests: create, load, list
		int plateGeometryId = TestPlateGeometry.test_create(8, 12, false);
		int bioPlateTypeId = TestBioPlateType.test_create("Test for bioplate", null, BioWell.LockMode.UNLOCKED);
		int addOnlyTypeId = TestBioPlateType.test_create("Add-only plate type", null, BioWell.LockMode.LOCKED_AFTER_ADD);
		int moveTypeId = TestBioPlateType.test_create("Move plate type", null, BioWell.LockMode.LOCKED_AFTER_MOVE);
		int id = test_create("Test BioPlate", plateGeometryId, bioPlateTypeId, freezerId, true);
		int id2 = test_create("Add-only plate", plateGeometryId, addOnlyTypeId, 0, false);
		int id3 = test_create("Move plate", plateGeometryId, moveTypeId, 0, false);
		test_load(id);
		test_list(-1);
		
		// Wells -- unlocked plate
		boolean FAIL = true;
		test_set_well(id, 0, 0, Item.SAMPLE, sampleId, !FAIL);
		test_set_well(id, 0, 1, Item.EXTRACT, extractId, !FAIL);
		test_list_wells(id, 2);
		test_clear_well(id, 0, 0, !FAIL);
		test_clear_well(id, 0, 1, !FAIL);
		test_list_wells(id, 0);
		
		// Wells -- add-only plate
		test_set_well(id2, 0, 0, Item.SAMPLE, sample2Id, !FAIL);
		test_clear_well(id2, 0, 0, FAIL);
		test_set_well(id2, 0, 0, Item.SAMPLE, sample3Id, FAIL);
		
		// Wells -- locked-after-move plate
		test_set_well(id3, 0, 0, Item.SAMPLE, sample3Id, !FAIL);
		test_clear_well(id3, 0, 0, !FAIL);
		test_list_wells(id3, 0);
		test_list_original_biomaterial(id3, 1);
		test_set_well(id3, 0, 0, Item.SAMPLE, sample3Id, FAIL);
		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		// Standard test: Delete
		TestExtract.test_delete(labeledExtractId);
		TestTag.test_delete(labelId);
		TestExtract.test_delete(extractId);
		TestSample.test_delete(sampleId);
		TestSample.test_delete(sample2Id);
		TestSample.test_delete(sample3Id);
		test_delete(id);
		test_delete(id2);
		test_delete(id3);
		
		TestPlateGeometry.test_delete(plateGeometryId);
		TestBioPlateType.test_delete(bioPlateTypeId);
		TestBioPlateType.test_delete(addOnlyTypeId);
		TestBioPlateType.test_delete(moveTypeId);
		TestHardware.test_delete(freezerId);
		TestProtocol.test_delete(protocolId);
		TestProtocol.test_delete(hybProtocolId);
		
		write("++Testing plates "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();		
			BioPlate bp = BioPlate.getById(dc, id);
			dc.deleteItem(bp);
			Set<ItemProxy> using = bp.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(bp + " is used by " + using);
			}
			dc.commit();
			write("--Delete bioplate OK");
		}
		catch (Throwable ex)
		{
			write("--Delete bioplate FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	static boolean test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<BioPlate> plates = BioPlate.getQuery().list(dc);
			for (int i = 0; i<plates.size(); i++)
			{
				write_item(i, plates.get(i));
			}
			if (expectedResults >=0 && expectedResults != plates.size())
			{
				throw new BaseException("Expected " + expectedResults + " results, " +
						"not " + plates.size());
			}
			write ("--List bioplates OK (" + plates.size() +")");
		}
		catch (Throwable ex)
		{
			write("--List bioplates FAILED");
			ex.printStackTrace();
			ok = false;
			return false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return ok;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioPlate bp = BioPlate.getById(dc, id);
			write("--Load bioplate OK");
		}
		catch (Throwable ex)
		{
			write("--Load bioplate FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static int test_create(String name, int plateGeometryId, int plateTypeId, int freezerId, boolean setAll)
	{
		if (plateGeometryId == 0 || plateTypeId == 0 || !TestUtil.hasPermission(Permission.CREATE, Item.BIOPLATE))
		{
			return 0;
		}
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PlateGeometry pg = PlateGeometry.getById(dc, plateGeometryId);	
			BioPlateType pt = BioPlateType.getById(dc, plateTypeId);
			BioPlate bp = BioPlate.getNew(dc, pg, pt);
			Item bioMaterialType = pt.getBioMaterialType();
			bp.setName(name);
			if (setAll)
			{
				bp.setDescription("Added at " + new Date());
				bp.setExternalId("test.bioplate.plate1");
				bp.setBarcode("995522");
				bp.setSection("1");
				bp.setTray("B4");
				bp.setPosition("3");
				if (freezerId > 0) bp.setFreezer(Hardware.getById(dc, freezerId));
			}
			dc.saveItem(bp);
			dc.commit();
			id = bp.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bp, false);
			write_item(0, bp);
			write("--Create bioplate OK");
		}
		catch (Throwable ex)
		{
			write("--Create bioplate FAIILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<BioPlate> l = BioPlate.getQuery().list(dc);
			int numDeleted = Trashcan.delete(dc.getSessionControl(), l, true, null);
			if (numDeleted != l.size()) throw new BaseException("Could not delete all items: " + numDeleted + " of " + l.size());
			dc.commit();
			write("--Delete all bioplates OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all bioplates FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			System.out.println("   \tID \tExternal id\tName      \tPlate type\tDescription\tBarcode\tGeometry\tFreezer\tSection\tTray\tPosition");
			System.out.println("-- \t-- \t-----------\t----------\t----------\t-----------\t-------\t--------\t-------\t-------\t----\t--------");
		}
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void write_item(int i, BioPlate bioPlate)
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+bioPlate.getId()+"\t"+bioPlate.getExternalId()+
				"\t"+bioPlate.getName()+"\t"+bioPlate.getBioPlateType()+"\t"+bioPlate.getDescription()+
				"\t"+bioPlate.getBarcode()+"\t"+bioPlate.getPlateGeometry()+"\t"+bioPlate.getFreezer()+
				"\t"+bioPlate.getSection()+"\t"+bioPlate.getTray()+"\t"+bioPlate.getPosition());
	}
	
	static void write_item(int i, BioWell well)
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+well.getId()+"\t["+well.getRow()+
				","+well.getColumn()+"]\t"+well.getBioMaterial()+"\t"+well.getOriginalBioMaterial());
	}

	
	static int test_set_well(int bioPlateId, int row, int column, Item bioMaterialType, int bioMaterialId, boolean shouldFail)
	{
		if (bioPlateId == 0 || bioMaterialId == 0)
		{
			return 0;
		}
		int wellId = 0;
		DbControl dc = null;
		try
		{
			Throwable t = null;
			BioPlate plate = null;
			try
			{
				dc = TestUtil.getDbControl();
				plate = BioPlate.getById(dc, bioPlateId);
				MeasuredBioMaterial bm = (MeasuredBioMaterial)bioMaterialType.getById(dc, bioMaterialId);
				BioWell well = plate.getBioWell(row, column);
				bm.setBioWell(well);
				dc.commit();
				wellId = well.getId();
			}
			catch (Throwable t1)
			{
				t = t1;
			}
			if (shouldFail)
			{
				if (t == null) throw new RuntimeException("Well shuld be locked");
				write("--Set well " + plate + "[" + row + "," + column + "] OK (well was locked)");
			}
			else
			{
				if (t != null) throw t;
				write("--Set well " + plate + "[" + row + "," + column + "] OK");
			}
		}
		catch (Throwable ex)
		{
			write("--Set well BioPlate[id=" + bioPlateId + "][" + row + "," + column + "] FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return wellId;
	}
	
	
	static void test_clear_well(int bioPlateId, int row, int column, boolean shouldFail)
	{
		if (bioPlateId == 0) return;
		int wellId = 0;
		DbControl dc = null;
		try
		{
			Throwable t = null;
			BioPlate plate = null;
			try
			{
				dc = TestUtil.getDbControl();
				plate = BioPlate.getById(dc, bioPlateId);
				BioWell well = plate.getBioWell(row, column);
				well.getBioMaterial().setBioWell(null);
				dc.commit();
			}
			catch (Throwable t1)
			{
				t = t1;
			}
			if (shouldFail)
			{
				if (t == null) throw new RuntimeException("Well shuld be locked");
				write("--Clear well " + plate + "[" + row + "," + column + "] OK (well was locked)");
			}
			else
			{
				if (t != null) throw t;
				write("--Clear well " + plate + "[" + row + "," + column + "] OK");
			}
		}
		catch (Throwable ex)
		{
			write("--Clear well BioPlate[id=" + bioPlateId + "][" + row + "," + column + "] FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	static void test_list_wells(int bioPlateId, int expectedResults)
	{
		if (bioPlateId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioPlate plate = BioPlate.getById(dc, bioPlateId);
			ItemResultList<BioWell> wells = plate.getBioWells().list(dc);
			int numWells = 0;
			for (int i = 0; i<wells.size(); i++)
			{
				BioWell well = wells.get(i);
				if (well.getBioMaterial() != null)
				{
					write_item(i, well);
					numWells++;
				}
			}
			if (expectedResults >=0 && expectedResults != numWells)
			{
				throw new BaseException("Expected " + expectedResults + " results, " +
						"not " + numWells);
			}
			write ("--List wells OK (" + numWells +")");
		}
		catch (Throwable ex)
		{
			write("--List wells FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_original_biomaterial(int bioPlateId, int expectedResults)
	{
		if (bioPlateId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioPlate plate = BioPlate.getById(dc, bioPlateId);
			ItemResultList<BioWell> wells = plate.getBioWells().list(dc);
			int numWells = 0;
			for (int i = 0; i<wells.size(); i++)
			{
				BioWell well = wells.get(i);
				if (well.getOriginalBioMaterial() != null)
				{
					write_item(i, well);
					numWells++;
				}
			}
			if (expectedResults >=0 && expectedResults != numWells)
			{
				throw new BaseException("Expected " + expectedResults + " results, " +
						"not " + numWells);
			}
			write ("--List wells with original biomaterial OK (" + numWells +")");
		}
		catch (Throwable ex)
		{
			write("--List wells with original biomaterial FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	static List<Integer> test_fill(int bioPlateId, Item itemType, String rootName, int tagId)
	{
		if (bioPlateId == 0) return null;
		List<Integer> ids = null;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			BioPlate plate = BioPlate.getById(dc, bioPlateId);
			Tag label = tagId == 0 ? null : Tag.getById(dc, tagId);
			PlateGeometry geometry = plate.getPlateGeometry();
			int rows = geometry.getRows();
			int cols = geometry.getColumns();
			List<MeasuredBioMaterial> items = new ArrayList<MeasuredBioMaterial>(rows * cols);
			WellCoordinateFormatter rowFormat = new WellCoordinateFormatter(true);
			WellCoordinateFormatter colFormat = new WellCoordinateFormatter(false);
			for (int row = 0; row < rows; ++row)
			{
				for (int col = 0; col < cols; ++col)
				{
					BioWell well = plate.getBioWell(row, col);
					MeasuredBioMaterial item = null;
					if (itemType == Item.SAMPLE)
					{
						item = Sample.getNew(dc);
					}
					else if (itemType == Item.EXTRACT)
					{
						item = Extract.getNew(dc);
					}
					item.setName(rootName + " [" + rowFormat.format(row) + colFormat.format(col) + "]");
					dc.saveItem(item);
					item.setBioWell(well);
					items.add(item);
				}
			}
			dc.commit();
			ids = new ArrayList<Integer>(items.size());
			new BasicItemToIdTransformer().transform(items, ids);
			write("--Fill bioplate OK");
		}
		catch (Throwable ex)
		{
			write("--Fill bioplate FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return ids;
	}
	
	
	static void test_delete_contents(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();		
			BioPlate bp = BioPlate.getById(dc, id);
			List<BioWell> wells = bp.getBioWells().list(dc);
			for (BioWell well : wells)
			{
				MeasuredBioMaterial item = well.getBioMaterial();
				if (item != null) dc.deleteItem(item);
			}
			dc.commit();
			write("--Delete bioplate contents OK");
		}
		catch (Throwable ex)
		{
			write("--Delete bioplate contents FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	
}